#ifndef PSD_UTILS_H
#define PSD_UTILS_H

#include <qlabel.h>
#include "qar.h"
#include "Qdos.h"
#include "Xrefs.h"
#include "triangulation.h"


#define MOM_DENSITY 0x01
#define MOM_VELOCITY 0x02
#define MOM_PRESSURE 0x04
#define MOM_HEATFLUX 0x08

// This class holds all the data and functions to compute the moments
// from Alban Rochel's MomentParameters

class PSD_Parameters
{
  
    protected:

    
    public:
        // Constructor
        // obj is the PSD object, with all the required metadata
        // direction_in is the direction of the flow (-1 if the flow comes opposite of the probe direction -angle is the probe direction, +1 if the data angle is the flow direction)

        PSD_Parameters(QdObject_var& obj, int direction_in);

        ~PSD_Parameters(){
			if(_x) delete _x;
			if(_y) delete _y;
			if(_z) delete _z;
			
			if(_vel) delete _vel; 
			
			if(_phi_centre) delete _phi_centre; 
			if(_phi_plus) delete _phi_plus; 
			if(_phi_minus) delete _phi_minus; 
			
			if(_theta_centre) delete _theta_centre; 
			if(_theta_plus) delete _theta_plus; 
			if(_theta_minus) delete _theta_minus; 
		}

        // Tries to deduce all the required computation parameters
        // from the data PSD object
        // Return false if anything fails
        bool init_data(QdObject_var & data);

        // Initialises the Frame objects, used to convert the frame to Target
        bool init_rotn(QdObject_var & rotnMtx);

        // Initialises the potential object
        // Converts into volts
        // Converts into time series if only scalar
        bool init_potential(QdObject_var & pot);

        // Initialises energy threshold object
        bool init_threshold(QdObject_var & threshold);
 
		// Initialises space for quantities computed often (in rotated frame)
        bool init_cartesian();
		
		// Stores quantities computed often (in rotated frame)
        bool fill_cartesian(int time);
		
		// Stores quantities computed often (in rotated frame)
        bool init_PSDcentre(QdObject_var & centreOfPSD);
		
		// Stores quantities computed often (in rotated frame)
        bool init_sliceNormal(QdObject_var & normal);
		
        
        double massUnit;
        double chargeUnit;
        double mass;
        double charge;
		double _PI;
		bool stop; // for interrupt

        
        enum data_type {MATRIX_SEQ, MATRIX, SCALAR_SEQ, SCALAR, UNKNOWN};

        // True if and only if the initialisation process went fine
        bool init_ok;

        // The following variables are initialised by the protected initialisation functions called by the constructor
        QdRMatrixSeq_var data; // the phase-space distribution
        double dataFillVal; // holds the fill value for data

        QdRMatrix_var e_const; // Energy values, if constant
        QdRMatrixSeq_var e_seq; // Energy values, if sequence
        QdRMatrix_var e_minus_const; // Lower boundaries of energy values, if constant
        QdRMatrixSeq_var e_minus_seq; // ditto, if sequence
        QdRMatrix_var e_plus_const; // Upper boundaries, constants
        QdRMatrixSeq_var e_plus_seq; // Upper boundaries, sequence
			
		// rotated quantities
		double * _x; // cartesians for each bin
		double * _y; // cartesians for each bin
		double * _z; // cartesians for each bin
			
		// invariant under rotn quantities
		double * _vel; // magnitude of vel for bin (depends on potential and cutoff)


		
        double eFillVal; // Fill values for energies
//         bool logE;

        // phi bins
        double *_phi_centre;
		double *_phi_plus;
		double *_phi_minus;
        double phiFillVal;
        double phi2rad; // Conversion to radians
        

        // Same thing for theta
		double *_theta_centre;
		double *_theta_plus;
		double *_theta_minus;
        double thetaFillVal;
        double theta2rad;
        
        QdRScalarSeq_var potential; // Potential in Volts
        QdRScalarSeq_var threshold; // Threshold in Coulombs
		
		QdRMatrixSeq_var centre_PSD; // Centre of PSD in velocity
		QdRMatrixSeq_var sliceNormal; // Centre of PSD in velocity
		int resolution;               // slice resolution
        
        QdRMatrix_var rotn_const; // GSE conversion matrix, constant
        QdRMatrixSeq_var rotn_seq; // GSE conversion matrix, sequence
        QdObject_var rotn_obj; // GSE conversion matrix, ptr
        data_type type_rotn;
        
        list<Element3D> tetra;
        
        int indexE, indexPhi, indexTheta;
        int maxTime, maxE, maxTheta, maxPhi;
        
        double v_sign;
		QString message;

//         double getIntegratedValue(int time);

       // Returns the PSD value at the input indexes 
        double getData(int time, int r, int phi, int theta);

        // ********************************************************
        // In the following functions, the "_raw" suffix indicates that the energies are not corrected using thresholds and potential
        // ********************************************************
        
        // Initialises the mass and charge
        inline void init_species(double m, int c)
		{
			mass = massUnit * m;
			charge = chargeUnit * c;
		}

        // Returns the volume of the element in the phase space
        inline double getElementVolume(int time, int v, int phi, int theta)
        {
            double v1(getVMinus(time, v)), v2(getVPlus(time, v)), theta1(getThetaMinus(time, theta)), theta2(getThetaPlus(time, theta)), phi1(getPhiMinus(time, phi)), phi2(getPhiPlus(time, phi));

            if(v1<0. || v2<0.|| isFill(phi1, phiFillVal) || isFill(phi2, phiFillVal) || isFill(theta1, thetaFillVal) || isFill(theta2, thetaFillVal) )
                return -1.;
            else
            {
                return fabs((v2*v2*v2-v1*v1*v1)*(cos(theta1)-cos(theta2))*(phi2-phi1))/3.;
            }
        }

        inline double getElementVolume_raw(int time, int v, int phi, int theta)
        {
            double v1(getVMinus_raw(time, v)), v2(getVPlus_raw(time, v)), theta1(getThetaMinus(time, theta)), theta2(getThetaPlus(time, theta)), phi1(getPhiMinus(time, phi)), phi2(getPhiPlus(time, phi));

            if(v1<0. || v2<0. || isFill(phi1, phiFillVal) || isFill(phi2, phiFillVal) || isFill(theta1, thetaFillVal) || isFill(theta2, thetaFillVal) )
                return -1.;
            else
            {
                return fabs((v2*v2*v2-v1*v1*v1)*(cos(theta1)-cos(theta2))*(phi2-phi1))/3.;
            }
        }

        // In velocity space, rather than phase space
        inline double getElementVolumeVelocitySpace(double m, int time, int r, int phi, int theta)
        {
            double mcube=m*m*m;
            double r1(getEMinus(time, r)), r2(getEPlus(time, r)), theta1(getThetaMinus(time, theta)), theta2(getThetaPlus(time, theta)), phi1(getPhiMinus(time, phi)), phi2(getPhiPlus(time, phi));
    
            if(isFill(r1, eFillVal) || isFill(r2, eFillVal) || isFill(phi1, phiFillVal) || isFill(phi2, phiFillVal) || isFill(theta1, thetaFillVal) || isFill(theta2, thetaFillVal) )
                return -1.;
            else
                return sqrt(8./mcube)*(sqrt(r2*r2*r2)-sqrt(r1*r1*r1))*(cos(theta1)-cos(theta2))*(phi2-phi1)/3.;
        }

        // Get energy, in joules
        inline double getE(int time, int e)
        {
            double energy=(e_const.is_nil() ? e_seq[time][e] : (*e_const)[e]);
			if(energy == eFillVal) return -1.;
            if (energy < threshold->elt(time)) return -1.;
            if((double)potential->elt(time) > 0.)
            	energy = energy - charge*(double)potential->elt(time);
            else
            	energy = energy + charge*(double)potential->elt(time);
            return energy;
        }

        // Get energy, in Joules, without prior correction
        inline double getE_raw(int time, int e)
        {
            double energy=(e_const.is_nil() ? e_seq[time][e] : (*e_const)[e]);
            return energy;
        }
        
        inline double getEPlus(int time, int e)
        {
            double _e=getE(time, e);
            if(_e<0.)
                return _e;
            else
                return _e+(double)(e_plus_const.is_nil() ? e_plus_seq[time][e] : (*e_plus_const)[e]);
        }
        
        inline double getEPlus_raw(int time, int e)
        {
            double _e=getE_raw(time, e);
            if(_e==eFillVal)
                return eFillVal;
            else
                return _e+(double)(e_plus_const.is_nil() ? e_plus_seq[time][e] : (*e_plus_const)[e]);
        }

        inline double getEMinus(int time, int e)
        {
            double _e=getE(time, e);
            if(_e<0.)
                return _e;
            else
                return _e-(double)(e_minus_const.is_nil() ? e_minus_seq[time][e] : (*e_minus_const)[e]);
        }
        
        inline double getEMinus_raw(int time, int e)
        {
            double _e=getE_raw(time, e);
            if(_e==eFillVal)
                return eFillVal;
            else
                return _e-(double)(e_minus_const.is_nil() ? e_minus_seq[time][e] : (*e_minus_const)[e]);
        }

        inline double getTheta(int time, int theta)
        {
			if(_theta_centre == 0) return thetaFillVal;
            return _theta_centre[time*(maxTheta+1) + theta];
        }
        
        inline double getThetaPlus(int time, int theta)
        {	
			if(_theta_plus == 0) return thetaFillVal;
            return _theta_plus[time*(maxTheta+1) + theta];
		}

        inline double getThetaMinus(int time, int theta)
        {
			if(_theta_minus == 0) return thetaFillVal;
            return _theta_minus[time*(maxTheta+1) + theta];
		}

        inline double getPhi(int time, int phi)
        {
			if(_phi_centre == 0) return phiFillVal;
            return _phi_centre[time*(maxPhi+1) + phi];
        }

        inline double getPhiPlus(int time, int phi)
        {
			if(_phi_plus == 0) return phiFillVal;
            return _phi_plus[time*(maxPhi+1) + phi];
        }

        inline double getPhiMinus(int time, int phi)
        {
			if(_phi_minus == 0) return phiFillVal;
            return _phi_minus[time*(maxPhi+1) + phi];
        }

        // Energy to velocity
        inline double E2V(double e)
        {
            if(e>0)
            {
                return sqrt(e*2./mass);
            }
            else
            {
                return -1.;
            }
        }
        
        inline double getV(int time, int v)
        {
            //return E2V((e_const.is_nil() ? e_seq[time][v] : (*e_const)[v]));
            return E2V(getE(time, v));
        }
        
        inline double getV_raw(int time, int v)
        {
            return E2V(getE_raw(time, v));
        }
        
        inline double getVPlus(int time, int v)
        {
            return E2V(getEPlus(time, v));
        }
        
        inline double getVPlus_raw(int time, int v)
        {
            return E2V(getEPlus_raw(time, v));
        }

        inline double getVMinus(int time, int v)
        {
            return E2V(getEMinus(time, v));
        }
        
        inline double getVMinus_raw(int time, int v)
        {
            return E2V(getEMinus_raw(time, v));
        }
                
        QdRMatrix_var computeValuesAlongVector(int time, QdRMatrix_var& CoM, QdRMatrix_var& direction, double v_min, double v_max, int nb_values);
        
        QdRMatrix_var computePAD(int time, QdRMatrix_var& CoM, QdRMatrix_var& direction, double v_min, double v_max, int nb_v, int nb_theta);
        
        void buildTetra(int time); // TODO use triangulation2 where possible
        
		void getRotationMatrix(double direction[], double rotation[][3]);
        
        void getMomentsSummed(int flags, QdRScalarSeq_var& density, QdRMatrixSeq_var& velocity, QdRMatrixSeq_var& pressure, QdRMatrixSeq_var& heat, QdRScalarSeq_var& temperature, QLabel *percLabl);
        
        void getMomentsIntegrated(int flags, QdRScalarSeq_var& density, QdRMatrixSeq_var& velocity, QdRMatrixSeq_var& pressure, QdRMatrixSeq_var& heat, QdRScalarSeq_var& temperature, QLabel *percLabl);
        
		void getSlice(QdRMatrixSeq_var& PSD_slice, QLabel *percLabl);
};


inline double tetra_volume(double xa, double ya, double za, double xb, double yb, double zb, double xc, double yc, double zc, double xd, double yd, double zd)
{
    double ax(xa-xd), ay(ya-yd), az(za-zd),
    bx(xb-xd), by(yb-yd), bz(zb-zd),
    cx(xc-xd), cy(yc-yd), cz(zc-zd);

    return fabs(ax*(by*cz-bz*cy)+ay*(bz*cx-bx*cz)+az*(bx*cy-by*cx))/6.;
}

#endif
