#include "qcalc.h"
#include "qdutil.h"
#include "qwui.h"

#include <qapplication.h>
#include <qprogressdialog.h>
#include <qmessagebox.h>



using namespace QSAS;


bool QcalcWindow::CheckSelfLinkedVirtualObjs()
{
	// Check if the Virtual objects are linked to themselves (directly or indirectly)
	bool                  bReturn    = true;

	QList<QcalcItem*> items = _join_listview->listItems();
	foreach(QcalcItem *item, items)
	{
		if ( item->m_eObjType == VIRTUAL_OBJECT )
		{
			QString qstrAlias = QString("");
			bReturn = item->CheckOperatorSelfLink( qstrAlias );
			if ( !bReturn )
			{
				QString qstrError = QString( "Error: Item [ %1 ] is linked to itself (directly or indirectly)" ).arg(item->_alias);
				QcalcAppendTextDisplay( qstrError.toLatin1().data() );
				DisplayEndMessage();
				QApplication::restoreOverrideCursor();
				return ( true );
			}
		}
	}

	return ( false );
}

bool QcalcWindow::JoinFirst()
{
    QdObjectSeq_var          qdJoinedList = new QdObjectSeq();
	if ( qdJoinedList.is_nil() )
	{
		QcalcAppendTextDisplay( "Error: Cannot start calculation - Memory cannot be allocated for intermediary objects" );
		DisplayEndMessage();
		return ( false );
	}

	// Set up the target time tags here
	// these are used for all joins:- join first, join for operations and join after
	
	this->_join_listview->targetD0 = getTimeTags();  // this is member of QcalcListView as it applies to all items
	if(this->_join_listview->targetD0.is_nil()) this->_join_listview->targetD0 = getScalarTags();
	
	if ( this->_join_listview->targetD0.is_nil() )
	{
		QcalcAppendTextDisplay( "No target tags found for the object list" );
		if ( multiJn_check1->isChecked() ) return (false); // no tt so cannot join first
		                                   return (true); // no tt but join may be unnecessary
	}

	vector <string> nameList;

	if ( multiJn_check1->isChecked() )
	{
		QcalcAppendTextDisplay( "All sequence objects will be multi-joined before calculation" );
		
		// do Multi-join
		
		if ( !joinObjects( qdJoinedList, nameList, TRUE ) ) // TRUE skips virtul objects
		{
			QApplication::restoreOverrideCursor();
			QcalcAppendTextDisplay( "Multi-join: Error joining time-tagged objects before calculation" );
			DisplayEndMessage();
			return ( false );
		}

		// ensure each WL item has a joined return value
		QList<QcalcItem*> items = _join_listview->listItems();
		foreach(QcalcItem *item, items)
		{
			if ( item->m_eObjType == WORKING_LIST_OBJECT  )
			{		
				// go through joined list and attach each joined object back to item m_qdObjJoinedFirst pointer
				for ( int nIndex = 0; nIndex < qdJoinedList->size(); nIndex++ )
				{
					QdObject_var qObj = qdJoinedList[ nIndex ];
					if ( qObj.is_nil() ) continue;

					// Get the name of the objects being joined
					if ( item->_alias == QString( nameList[ nIndex ].c_str() ) )
					{
						item->m_qdObjJoinedFirst = qObj;
					}
				}
				// put constants onto m_qdObjJoinedFirst pointer
				if(item->m_qdObjJoinedFirst.is_nil())
				{
					item->m_qdObjJoinedFirst = item->m_pcalcTab->operandStack[0]->slot->getSlotObject();
					if( item->m_qdObjJoinedFirst->xref_exists(DEPEND_0) )
					{
						QApplication::restoreOverrideCursor();
						QcalcAppendTextDisplay( " time-tagged object failed to join before calculation" );
						DisplayEndMessage();
						return ( false );
					}
					
				}
			}
		}
	}

	return ( true );

}


bool QcalcWindow::JoinLast()
{
	if ( multiJn_check2->isChecked() )
	{
		QdObjectSeq_var          qdJoinedList = new QdObjectSeq();
		if ( qdJoinedList.is_nil() )
		{
			QcalcAppendTextDisplay( "Error: Cannot start calculation - Memory cannot be allocated for intermediary objects" );
			DisplayEndMessage();
			return ( false );
		}
	
		vector <string> nameList;

		QcalcAppendTextDisplay( "All time-tagged objects will be multi-joined after calculation" );
		
		// do Multi-join
		
		if ( !joinObjects( qdJoinedList, nameList, FALSE ) ) // FALSE does not skip virtul objects - they should exist
		{
			QApplication::restoreOverrideCursor();
			QcalcAppendTextDisplay( "Multi-join: Error joining time-tagged objects after calculation" );
			DisplayEndMessage();
			return ( false );
		}

		// ensure each WL item has a joined return value
		QList<QcalcItem*> items = _join_listview->listItems();
		foreach(QcalcItem *item, items)
		{
		
			// go through joined list and attach each joined object back to item m_qdObjJoinedFirst pointer
			for ( int nIndex = 0; nIndex < qdJoinedList->size(); nIndex++ )
			{
				QdObject_var qObj = qdJoinedList[ nIndex ];
				if ( qObj.is_nil() ) continue;

				// Get the name of the objects being joined

				if ( item->_alias == QString( nameList[ nIndex ].c_str() ) )
				{
					item->m_qdObjResult = qObj;
				}
			}
			// put constants onto m_qdObjJoinedFirst pointer
			if(item->m_qdObjResult.is_nil())
			{
				item->m_qdObjResult = item->m_pcalcTab->operandStack[0]->slot->getSlotObject();
				if( item->m_qdObjResult->xref_exists(DEPEND_0) )
				{
					QApplication::restoreOverrideCursor();
					QcalcAppendTextDisplay( " time-tagged object failed to join before calculation" );
					DisplayEndMessage();
					return ( false );
				}
			}
		}
	}
	
	return ( true );
}


void QcalcWindow::DoCalculation()
{
	QdObjectSeq_var          qdJoinedList = new QdObjectSeq();
	QdObject_var nilObj;

	char   bufLog[ 1024 ];
	QDate  qdate = QDate::currentDate();
	QTime  qTime = QTime::currentTime();

	newCalcBtn->setEnabled( false );
	// Display start message log
	
	sprintf( bufLog, "%s Calculator Started", qTime.toString( "hh:mm:ss.zzz" ).toLatin1().data() );


	QcalcAppendTextDisplay( bufLog );

	if ( qdJoinedList.is_nil() )
	{
		QcalcAppendTextDisplay( "Error: Cannot start calculation - Memory cannot be allocated for intermediary objects" );
		DisplayEndMessage();
		newCalcBtn->setEnabled( true );
		return;
	}

	QApplication::setOverrideCursor( QCursor( Qt::WaitCursor ) );
	
	qApp->processEvents();
	if ( CheckSelfLinkedVirtualObjs() )
	{
		newCalcBtn->setEnabled( true );
		QApplication::restoreOverrideCursor();
		return;
	}


	// Reset the claculation done flag and check that SOMETHING gets saved to WL
	bool noResultSaved = true;
	QList<QcalcItem*> items = _join_listview->listItems();
	foreach(QcalcItem *item, items)
	{
		if (item->m_pcalcTab && item->m_pcalcTab->saveWL_check->isChecked() ) noResultSaved = false;
		
		item->m_bCalcDone = false;
		item->m_qdObjJoinedFirst = nilObj;
		item->m_qdObjResult = nilObj;
	}

	if(noResultSaved) {
		QcalcAppendTextDisplay( "No item has 'Save to WL' ticked" );
		QcalcAppendTextDisplay( "Calculations ignored" );
		DisplayEndMessage();
		messageHandler( "All operations virtual: No results would be saved", true );
		newCalcBtn->setEnabled( true );
		QApplication::restoreOverrideCursor();
		return;
	}
	
	qApp->processEvents();

	// Check if joining first is required
	if ( !JoinFirst() )
	{
		// JoinFirst returned an error
		QcalcAppendTextDisplay( "Problem in joining first" );
		newCalcBtn->setEnabled( true );
		QApplication::restoreOverrideCursor();
		return;
	}
	
	// Do the actual calculation
	int nStep = items.size();
	if(nStep == 0) {
		QApplication::restoreOverrideCursor();
		QcalcAppendTextDisplay( "Calculation: Null pointer encountered" );
		DisplayEndMessage();
		newCalcBtn->setEnabled( true );
		QApplication::restoreOverrideCursor();
		return;
	}
	QProgressDialog dlgProg( "Calculating...", "Abort", 0,  nStep, this, Qt::Dialog );

	dlgProg.setWindowTitle( "Calculator" );
	dlgProg.show();
	qApp->processEvents();

	int nCounter = 0;

	foreach(QcalcItem *item, items)
	{

		dlgProg.setValue( nCounter );

	 	qApp->processEvents();

		if ( dlgProg.wasCanceled() )
		{
			switch	(	QMessageBox::warning(
			                                  0,
			                                  "Calculator",
			                                  "Abort?",
					                          QMessageBox::No | QMessageBox::Default,
					                          QMessageBox::Yes,
											  QMessageBox::Cancel | QMessageBox::Escape
											)	 )
			{
				case QMessageBox::Yes:
						QApplication::restoreOverrideCursor();
						newCalcBtn->setEnabled( true );
						QcalcAppendTextDisplay( "Calculation: Calculator aborted" );
						DisplayEndMessage();
						//break;
						return;

				case QMessageBox::No:
				case QMessageBox::Cancel:
				default:
						dlgProg.reset();
						dlgProg.setLabelText("Calculating...");

                                                dlgProg.setValue( nCounter );

						dlgProg.show();
						qApp->processEvents();
						break;
	   		}
		}

		item->DoOperation();
		nCounter++;

	}

	dlgProg.setValue( nStep );

	qApp->processEvents();
	// Check if joinig last is required
	if ( !JoinLast() )
    {
		// Debugging info are within the calling function
		newCalcBtn->setEnabled( true );
		QApplication::restoreOverrideCursor();
		return;
	}


	// Write the objects into the working list
	foreach(QcalcItem *item, items)
	{
		qApp->processEvents();
		QcalcTab *qTab = item->m_pcalcTab;
		if(qTab == NULL) continue;
		
		if ( qTab->saveWL_check->isChecked() )
		{
			if (  !( item->m_qdObjResult.is_nil() )  )
			{
				try
				{
					QString newObjectName = qTab->_nameWL_field->text();
					if( newObjectName.isEmpty() ) newObjectName = QString(item->_alias);
					
					messageHandler( "Calculator: Placing object " + newObjectName + " on the working list", false );
					QcalcAppendTextDisplay( QString("Calculator: Placing object " + newObjectName + " on the working list" ).toLatin1().data());
					
					// set FIELDNAM to be object name
					QuSetTxtAttr( "FIELDNAM", newObjectName.toLatin1().data(), item->m_qdObjResult );

					QwuiPutOnWL( item->m_qdObjResult, newObjectName.toLatin1().data(), false);
				}
				catch( Exception &e )
				{
					messageHandler( QString( e.msg().c_str() ), true );
					QcalcAppendTextDisplay( e.msg().c_str() );
					DisplayEndMessage();
					QApplication::restoreOverrideCursor();
					newCalcBtn->setEnabled( true );
					return;
				}
			}
		}
	}


	QApplication::restoreOverrideCursor();
	QcalcWindow::DisplayEndMessage();
	QcalcAppendTextDisplay( " " );
	newCalcBtn->setEnabled( true );
}

void QcalcWindow::DisplayEndMessage()
{
	QDate qdate = QDate::currentDate();
	QTime qTime = QTime::currentTime();
	char  bufLog[ 1024 ];

	sprintf( bufLog, "%s %s Calculator Ended",  qdate.toString().toLatin1().data(), qTime.toString( "hh:mm:ss.zzz" ).toLatin1().data() );

	QcalcAppendTextDisplay( bufLog );
}

