#include "qcalc.h"
#include "qwui.h"
#include <qprogressbar.h>

extern qwuiWindow* qsasWindow;

using namespace QSAS;

// CALC Window
//---------------


/*!
    \class QcalcWindow
    \brief The class that encapsulates the calculator main window. It is inherited from the QSAS QguWindow class.

	The main widgets included in this class are the widgets for the join tab page. In addition, it includes the QTabWidget
	that is used when adding dynamically the operation tab pages.
*/

QcalcWindow::QcalcWindow( QWidget *parent, const char *name )
  : QguWindow( parent,name ),
   _join_listview(  NULL )
{
	this->setPalette(qApp->palette());
	this->setFont(qApp->font());
  
	setAcceptDrops( true );
	setFocusPolicy( Qt::WheelFocus );

	verbose    = false;

   _top_widget = new QWidget( this );
   setCentralWidget( _top_widget );

   _top_gl = new QGridLayout( _top_widget );
   _top_gl->setSpacing(1);
   _top_gl->setMargin(0);

  	// splitter to manage database list view and data selection widget

   _top_splitter = new QSplitter( _top_widget );
   _top_gl->addWidget( _top_splitter, 0,0 );

  	//  list view of objects to perform join and calculation on
 	 _join_listview = new QcalcListView( _top_splitter, this, "Object List" );

  _join_listview->setAllColumnsShowFocus( TRUE );
  _join_listview->setRootIsDecorated( false );
  
  // Create the tab widget (Joining / Calculator )
  tabWidget      = new QTabWidget( _top_splitter );
  connect( tabWidget, SIGNAL( currentChanged( QWidget * ) ), _join_listview, SLOT( tabChanged( QWidget * ) ) );
  tabWidget->setContentsMargins(0,0,0,0);

  // Create the Feed-back group box
  CreateFeedBackGroupBox();

  // Button group box

  buttons_groupbox = new QGroupBox( _top_widget);
  QHBoxLayout * buttons_layout=new QHBoxLayout;
  _top_gl->addWidget( buttons_groupbox, 3,0 );
  buttons_layout->setMargin(0);
  buttons_layout->setSpacing(0);
  buttons_groupbox->setLayout(buttons_layout);


  // Log Area

  QGroupBox* logGroupbox = new QGroupBox( _top_widget );
  QHBoxLayout * logLayout=new QHBoxLayout;
  _top_gl->addWidget( logGroupbox, 4,0 );
  logLayout->setMargin(0);

  logArea = new QTextEdit( logGroupbox );
  logArea->setMinimumHeight(125);
  logLayout->addWidget(logArea);
  logGroupbox->setLayout(logLayout);


  //  buttons added in JOIN or EXPORT create windows

  // Create menus
  CreatePopUpMenus();

  messageHandler(" ", false);
  this->setMinimumSize( this->sizeHint() );

  this->fromSession = false;

  // Load the operators map
  m_mapOps.LoadAll();
  
  this->setDefaults(); // initialise values used for new items

  // insert a blank operation
  this->_join_listview->insertVirtualItem();

  _top_splitter->setMinimumSize(_top_splitter->sizeHint());
  this->setMinimumSize(this->sizeHint());
  // set this item to use defaults
  this->defaultAll();
  setMinimumHeight(800);
}



void QcalcWindow::keyPressEvent(QKeyEvent * e )
{  
  if(e->key() == Qt::Key_Escape){
	  if(   JoinBtn->isDefault()) {
		 JoinBtn->setDefault(false);
		 newCalcBtn->setDefault(true);
	  }
	  else if(newCalcBtn->isDefault()) {
		 newCalcBtn->setDefault(false);
		 dismissBtn->setDefault(true);
	  }

	  else if(dismissBtn->isDefault())  {
		 dismissBtn->setDefault(false);
		 JoinBtn->setDefault(true);
	  }
	  return;
  }
  if(e->key() == Qt::Key_Return ||  e->key() == Qt::Key_Enter) {
     if(   JoinBtn->isDefault()) DoCalculation();
	 else if(newCalcBtn->isDefault()) _join_listview->insertVirtualItem();
	 else if(dismissBtn->isDefault()) close(); 
     return;
  }

  // else pass on to list view 
  e->ignore(); // let selection change

}


void QcalcWindow::CreatePopUpMenus()
{

  _menu_bar->clear();

  QMenu* file_menu = new QMenu( this);
  edit_menu = new QMenu( this);

  file_menu->clear();
  file_menu->addAction("&Save", this, SLOT(save()), Qt::CTRL+Qt::Key_S);
  file_menu->addAction("Save &As...", this, SLOT(saveAs()), Qt::CTRL+Qt::Key_A);
  file_menu->addSeparator();
  file_menu->addAction("&Include...", this, SLOT(include()), Qt::CTRL+Qt::Key_I);
  file_menu->addAction("&Restore...", this, SLOT(restore()), Qt::CTRL+Qt::Key_R);
  file_menu->addSeparator();
  file_menu->addAction("Use Default Settings", this, SLOT(defaultAll()));
  _menu_bar->addMenu(file_menu );
  file_menu->setTitle("File");

  edit_menu->clear();

  edit_menu->addAction("&Copy",       _join_listview, SLOT( copy() ),      Qt::CTRL+Qt::Key_C );
  edit_menu->addAction("Paste",       _join_listview, SLOT(paste()) );
  edit_menu->addAction("select &All", _join_listview, SLOT( selectAll() ), Qt::CTRL+Qt::Key_A );
  edit_menu->addSeparator();

  edit_menu->addAction("Copy Join Settings",  this,  SLOT(keepChoices()),  Qt::CTRL+Qt::Key_Left);
  edit_menu->addAction("Apply Join Settings", this,  SLOT(applyChoices()), Qt::CTRL+Qt::Key_Right);
  edit_menu->addSeparator();

  edit_menu->addAction("Delete Item", _join_listview, SLOT(deleteListItems()), Qt::CTRL+Qt::Key_X);
  edit_menu->addAction("Clear List",  _join_listview, SLOT(deleteAllItems()));
  edit_menu->addSeparator();
	
  edit_menu->addAction( "Move Down",      _join_listview, SLOT( moveItemDown() ) ,                Qt::CTRL+Qt::Key_Down);
  edit_menu->addAction( "Move Up",        _join_listview, SLOT( moveItemUp() ) ,                  Qt::CTRL+Qt::Key_Up);

  
  _menu_bar->addMenu(edit_menu );
  edit_menu->setTitle("Edit");

  options_menu = new QMenu(this);

  options_menu->clear();

  toggleAction = options_menu->addAction("Verbose", this, SLOT(verboseToggle()));
  toggleAction->setCheckable(true);
  
  _menu_bar->addMenu(options_menu );
  options_menu->setTitle("Options");

  _menu_bar->addSeparator();   

  QMenu * help = new QMenu( this );

  _menu_bar->addMenu(help );
  help->setTitle( " Help");

  help->addAction("QSAS &Manual",   this, SLOT( launchManual() ), Qt::CTRL+Qt::Key_M );
  help->addSeparator();

  //help->insertItem("...Data Export", this, SLOT( launchManExport() ));
  help->addAction("...Calculator",        this, SLOT( launchManJoin() ));
  help->addSeparator();

  help->addAction("Module Version", this, SLOT( launchVersion() ));
}


void QcalcWindow::CreateFeedBackGroupBox()
{
	QGroupBox* feedback_groupbox=new QGroupBox(_top_widget);
	QVBoxLayout * feedback_layout=new QVBoxLayout;
	feedback_layout->setMargin(QSAS_profiles->current()->GUImargins/2);
	feedback_layout->setSpacing(QSAS_profiles->current()->GUImargins);
	feedback_groupbox->setTitle( "  Sequence Dependency (Time or Scalar Sequence)  ");
	
	// Checks for join options
	
	QGridLayout* multiJn_layout = new QGridLayout;
	multiJn_layout->setSpacing(1);
	
	multiJn_check1 = new QCheckBox("Multi-Join Before calculation");
	multiJn_layout->addWidget( multiJn_check1, 0, 0 );
	multiJn_check1->setChecked( false );

	multiJn_check2 = new QCheckBox("Multi-Join After calculation");
	multiJn_layout->addWidget( multiJn_check2, 0, 1 );
	multiJn_check2->setChecked( false );
	
	feedback_layout->addLayout(multiJn_layout);
	
	// Time source
	
	QGridLayout * time_layout=new QGridLayout;
	time_layout->setSpacing( 1 );
	
	time_layout->addWidget(new QLabel("Interval "), 0, 0, Qt::AlignRight);
	
	_TimeIvlSelect = new QComboBox;
	_TimeIvlSelect->insertItem(0, "from Data");
	_TimeIvlSelect->insertItem(1, "from Object");
	connect( _TimeIvlSelect, SIGNAL( activated( int ) ), this, SLOT( setTTivl( int ) ) );
	time_layout->addWidget( _TimeIvlSelect, 0, 1, Qt::AlignLeft );
	
	IvlSource_label = new QLabel("  ");
	time_layout->addWidget( IvlSource_label, 1, 0, Qt::AlignRight );
	
	_IvlSource = new QguWLinput(_top_widget, "IvlSource" );
	QFontMetrics calc_font_mt(qApp->font());
	int char_width = calc_font_mt.width('m');
	_IvlSource->setMinimumWidth( 35*char_width );
	_IvlSource->setMaxLength( 100 );
	_IvlSource->active = false;
	_IvlSource->setToolTip("not active");
	time_layout->addWidget( _IvlSource, 1, 1, Qt::AlignLeft );
	
	time_layout->addWidget( new QLabel("Target Tags "), 2, 0, Qt::AlignRight );
	
	_TimeTagSelect = new QComboBox;
	_TimeTagSelect->insertItem(0, "Create Regular");
	_TimeTagSelect->insertItem(1, "from Object" );
	_TimeTagSelect->setCurrentIndex(1);
	connect( _TimeTagSelect, SIGNAL(activated(int)), this, SLOT( setTTsource( int ) ) );
	time_layout->addWidget( _TimeTagSelect, 2, 1, Qt::AlignLeft );
	
	TagSource_label = new QLabel("Tag Object ");
	time_layout->addWidget( TagSource_label, 3, 0, Qt::AlignRight );
	
	_TagSource = new QguWLinput(_top_widget, "TagSource" );
	_TagSource->setMinimumWidth( 35*char_width );
	_TagSource->setMaxLength( 100 );
	setTTsource(1);
	connect( _TagSource,     SIGNAL( edited_OK()  ),  this,       SLOT( setTimeFeedback()  ));
	_TagSource->setToolTip( "drag and drop any object with DEPEND_0 (or DEPEND_0 object itself)");
	time_layout->addWidget( _TagSource, 3, 1, Qt::AlignLeft );
	connect( _IvlSource,     SIGNAL( edited_OK()  ),   this,          SLOT( setTimeFeedback() ) );
	
	time_layout->addWidget(new QLabel("Start:"), 0, 2, Qt::AlignRight );
	
	Start_field = new QLabel("<Not yet set>");
	time_layout->addWidget( Start_field, 0, 3, Qt::AlignLeft );

	QLabel* End_label = new QLabel("End: ");
	time_layout->addWidget( End_label, 1, 2, Qt::AlignRight );

	End_field = new QLabel("<Not yet set>");
	time_layout->addWidget( End_field, 1, 3, Qt::AlignLeft );

	QLabel* Ntags_label = new QLabel("     Approx number of Tags in Ivl: ");
	time_layout->addWidget( Ntags_label, 2, 2, Qt::AlignRight );

	Ntags_field = new QLabel("<Not yet set>");
	time_layout->addWidget( Ntags_field, 2, 3, Qt::AlignLeft );

	spacing_label = new QLabel("Tag Spacing: ");
	time_layout->addWidget( spacing_label, 3, 2, Qt::AlignRight );

	spacing_field = new QLabel("<Not yet set>");
	time_layout->addWidget( spacing_field, 3, 3, Qt::AlignLeft );
	
	
	feedback_layout->addLayout(time_layout);
	
	// Done
	
	
	feedback_groupbox->setLayout(feedback_layout);
	_top_gl->addWidget(feedback_groupbox);
}


QcalcWindow::~QcalcWindow()
{
 	delete _IvlSource;
 	delete _TagSource;
 	m_mapOps.clear();
}

