#include "qcalcEvaluator.h"


void QEvaluatorMultiply::evaluate(QVector<VInSlot*>& in, QVector<VOutSlot*>& out, bool& ok, QString& message)
{
	ok=true;
	QdObject_var result;
	
	QEvaluator::evaluate(in, out, ok, message);
	if(!ok)
		return;
	
	QdObject_var val1=in[0]->connectedSlot()->getValue(ok);
	if(!ok)
	{
		message=QString("Input ")+in[0]->getName()+QString(" invalid");
		ok=false;
		return;
	}
	QdObject_var val2=in[1]->connectedSlot()->getValue(ok);
	if(!ok)
	{
		message=QString("Input ")+in[1]->getName()+QString(" invalid");
		ok=false;
		return;
	}
	
	bool isVector1=QuIsThreeVector(val1);
	bool isVector2=QuIsThreeVector(val2);
	
	if( isVector1 &&  isVector2)
	{
		message=QString("Use vector operations for multiplying vectors");
		ok=false;
		return; 
	}
	
	// Test validity of inputs...
	int nQarStatus=QarTestConformal(val1, val2);
	if ( nQarStatus == QAR_ERR )
	{
		message=QString("Non conformal inputs");
		ok=false;
		return;
	}
	
	// Test Frames for vectors
	
	string  attrPtr;
	string  stringBuf;
	string flag1 = QuGetOrder(val1);
	string flag2 = QuGetOrder(val2);
	bool isV2 = false;
	bool isR1 = false;
	bool isR2 = false;

	if(QuToUpper(flag2).find("VECTOR") != string::npos) isV2 = true;
	if(QuToUpper(flag1).find("ROT") != string::npos) isR1 = true;
	if(QuToUpper(flag2).find("ROT") != string::npos) isR2 = true;
	
	if(isR1 && isV2)
	{
		string fr1 = QuGetFrame(val1);
		string fr2 = QuGetFrame(val2);
		if(fr1 != fr2)
		{
			message=QString("Frames for Rotation Matrix and Vector differ");
			ok=false;
			return;
		}
	}
	if(isR1 && isR2)
	{
		string fr1 = QuGetFrame(val1);
		string fr2 = QuGetToFrame(val2);
		if(fr1 != fr2)
		{
			message=QString("Frames for Rotation Matrices differ...\nR1.Frame != R2.ToFrame");
			ok=false;
			return;
		}
	}
	
	// do arithmetic...
	try
	{
		result = val1 * val2;
	}
	catch( Exception &e )
	{
		message=QString(e.msg().c_str());
		message+=QString(" The operation Multiplication is not yet supported between these input types");
		ok=false;
		return;
	}

	if ( result.is_nil() )
	{
		message=QString("Arithmetic>> This Operation is not yet supported between these input types\n" );
		ok=false;
		return;
	}
	
	// get xrefs from inputs if requested
	result->use_xrefs_from(val1, val2);

	// Add timeline (or DEPEND_0) : Either timeline is safe if it exists
	QdObject_var qd0 = get_Depend_0( val1 );
	if ( qd0.is_nil() ) qd0 = get_Depend_0( val2 );
	if ( !qd0.is_nil() ) result->change_xref(DEPEND_0 , qd0 );

	
	// Set QSAS essential attriutes
	// SI_conversion
	attrPtr = QarSIConv_product( val1, val2 );
	if ( !attrPtr.empty() )
	{
		QuSetTxtAttr( "SI_conversion", attrPtr.c_str(), result );
	}

	// Frame
	attrPtr = QarFrame_product( val1, val2 );
	if ( !attrPtr.empty() )
	{
		QuSetFrame(result, attrPtr.c_str() );
	}
	
	if(isR1 && isR2)
	{
	  // new ToFrame from R1 
	  // Note Frame is from ToFrame of R2 and is handled by QarFrame_producet
		string toFr = QuGetAttrText("ToFrame", val1);
		QuSetTxtAttr( "ToFrame", toFr.c_str(), result );
	}
	

	// UNITS
	attrPtr = QarUNITS_product( val1, val2 );
	if ( !attrPtr.empty() )
	{
		QuSetTxtAttr("UNITS", attrPtr.c_str(), result);
	}
	
	// LABLAXIS
	QString labl_s = QuGetAttrText( "LABLAXIS", val1 ).c_str();
	if(val2->xref_exists("LABLAXIS")){
		labl_s += " x ";
		labl_s += QuGetAttrText( "LABLAXIS", val2 ).c_str();
	}
	else if(val2->sequence_size() == 1){
		QdRScalarSeq_var ss = QdRScalarSeq_var::narrow(val2);
		QdRScalar_var s = QdRScalar_var::narrow(val2);
		if(!ss.is_nil()) labl_s += QString(" x %1").arg(ss[0]);
		else if(!s.is_nil()) labl_s += QString(" x %1").arg(*s);
	}
	QuSetTxtAttr( "LABLAXIS", labl_s.toLatin1().data(), result );


	// DEPEND_i
	
	// special case of one operand is a scalar
	QdRScalarSeq_var sObj2 = QdRScalarSeq_var::narrow(val2);
	QdRScalarSeq_var sObj1 = QdRScalarSeq_var::narrow(val1);
	QdRScalar_var Obj2 = QdRScalar_var::narrow(val2);
	QdRScalar_var Obj1 = QdRScalar_var::narrow(val1);
	if ( !sObj2.is_nil() || !Obj2.is_nil() )
	{
		// 2nd obj is scalar use depends from first
		if (val1->xref_exists(DEPEND_1)) 
		{
			QdObject_var d1_xref = val1->get_xref(DEPEND_1);
			result->change_xref( DEPEND_1, d1_xref  );
		}
   
		if (val1->xref_exists(DEPEND_2)) 
		{
			QdObject_var d1_xref = val1->get_xref(DEPEND_2);
			result->change_xref( DEPEND_2, d1_xref  );
		}

		if (val1->xref_exists(DEPEND_3)) 
		{
			QdObject_var d1_xref = val1->get_xref(DEPEND_3);
			result->change_xref( DEPEND_3, d1_xref  );
		}
	}
	else if ( !sObj1.is_nil() ||  !Obj1.is_nil() )
	{
		// 1st obj is scalar use depends from second
		if (val2->xref_exists(DEPEND_1)) 
		{
			QdObject_var d1_xref = val2->get_xref(DEPEND_1);
			result->change_xref( DEPEND_1, d1_xref  );
		}
   
		if (val2->xref_exists(DEPEND_2)) 
		{
			QdObject_var d1_xref = val2->get_xref(DEPEND_2);
			result->change_xref( DEPEND_2, d1_xref  );
		}

		if (val2->xref_exists(DEPEND_3)) 
		{
			QdObject_var d1_xref = val2->get_xref(DEPEND_3);
			result->change_xref( DEPEND_3, d1_xref  );
		}
	}
	else{
 	   // handle matrix
		QdRMatrixSeq_var Mobj2s = QdRMatrixSeq_var::narrow(val2);
		QdRMatrixSeq_var Mobj1s = QdRMatrixSeq_var::narrow(val1);
		QdRMatrix_var    Mobj2 = QdRMatrix_var::narrow(val2);
		QdRMatrix_var    Mobj1 = QdRMatrix_var::narrow(val1);
		if( (!Mobj2s.is_nil() || !Mobj2.is_nil()) && (!Mobj1s.is_nil() || !Mobj1.is_nil())){
	     // both objects are arrays or array sequences
			setProductDepends(result, val1, val2);
		}
	}
	
	out[0]->setValue(result, true);
}

void QEvaluatorMultiply::setProductDepends(QdObject_var resultObj, QdObject_var obj1, QdObject_var obj2)
{
	int dims1 = QuArrayDims(obj1);
	int dims2 = QuArrayDims(obj2);

	int thisDim=1;
	while( thisDim < dims1 ){ // this ignores last dim = first dim of obj2
		QString depName = QString("DEPEND_%1").arg(thisDim);
		if (obj1->xref_exists(depName.toLatin1().data()))
		{
			QdObject_var d_xref = obj1->get_xref(depName.toLatin1().data());
#if QT_VERSION < 0x040000
			resultObj->change_xref( depName, d_xref  );
#else
			resultObj->change_xref( depName.toStdString(), d_xref  );
#endif
		}
		thisDim++;
	}
   
	int i=2;
	while( i <= dims2 ){ // this ignores first dim of obj2
		QString depNamei = QString("DEPEND_%1").arg(i);
		QString depName = QString("DEPEND_%1").arg(thisDim);
		if (obj2->xref_exists(depNamei.toLatin1().data()))
		{
			QdObject_var d_xref = obj2->get_xref(depNamei.toLatin1().data());
#if QT_VERSION < 0x040000
			resultObj->change_xref( depName, d_xref ); // shuffles dependencies along as needed
#else
			resultObj->change_xref( depName.toStdString(), d_xref);
#endif
		}
		thisDim++;
		i++;
	}
}
