#include "qcalcEvaluator.h"


void QEvaluatorVectProd::evaluate(QVector<VInSlot*>& in, QVector<VOutSlot*>& out, bool& ok, QString& message)
{
	ok=true;
	QdObject_var result;
	string attrPtr;
	
	QEvaluator::evaluate(in, out, ok, message);
	if(!ok)
		return;
	
	QdObject_var val1=in[0]->connectedSlot()->getValue(ok);
	if(!ok)
	{
		message=QString("Input ")+in[0]->getName()+QString(" invalid");
		ok=false;
		return;
	}
	QdObject_var val2=in[1]->connectedSlot()->getValue(ok);
	if(!ok)
	{
		message=QString("Input ")+in[1]->getName()+QString(" invalid");
		ok=false;
		return;
	}
	
	bool isVector1=QuIsThreeVector(val1);
	bool isVector2=QuIsThreeVector(val2);
	
	if( !isVector1 ||  !isVector2)
	{
		message=QString("Operation is only valid between vectors or tensors");
		ok=false;
		return; 
	}
	
	// Test validity of inputs...
	int nQarStatus =  QarTestConformalForVecProds( val1, val2 );
	if ( nQarStatus == QAR_ERR )
	{
		message=QString("Inner and Outer product only valid between vectors or tensors" );
		ok=false;
		return; 
	}
	
	int frameOK = QarAreFramesSame(val1, val2 );
	if(frameOK == QAR_WARN)
	{
		message=QString("Warning: Vect Prod: Cannot confirm vexctor frame");
		ok=false;
		return; //TODO warning rather than error
	}
	else if(frameOK == QAR_ERR)
	{
		message=QString("Error: Vect Prod: Frames differ");
		ok=false;
		return;
	}

	// do arithmetic...
	try
	{
		result = outer(val1, val2);
	}
	catch( Exception &e )
	{
		message=QString(e.msg().c_str());
		message+=QString(" The operation Vect Prod is not yet supported between these input types");
		ok=false;
		return;
	}

	if(result.is_nil())
	{
		message=QString("Arithmetic>> Not valid for these input types");
		ok=false;
		return;
	}

	// Add timeline (or DEPEND_0) : Either timeline is safe if it exists
	QdObject_var qd0 = get_Depend_0( val1 );
	if ( qd0.is_nil() ) qd0 = get_Depend_0( val2 );
	if ( !qd0.is_nil() ) result->change_xref(DEPEND_0 , qd0 );


	// Set QSAS essential attriutes
	// SI_conversion
	attrPtr = QarSIConv_product(val1, val2);
	if ( !attrPtr.empty() )
	{
		QuSetTxtAttr( "SI_conversion", attrPtr.c_str(), result );
	}


	// Frame
	QuSetFrame( result, QuGetFrameAttr(val1, val2).c_str() );

	// UNITS
	attrPtr = QarUNITS_product(val1, val2);
	if ( !attrPtr.empty() )
	{
		QuSetTxtAttr("UNITS", attrPtr.c_str(), result);
	}
	
	out[0]->setValue(result, true);
}

// ---------------------

void QEvaluatorDiff3::evaluate(QVector<VInSlot*>& in, QVector<VOutSlot*>& out, bool& ok, QString& message)
{
	ok=true;
	QdObject_var result;
	string attrPtr;
	
	QEvaluator::evaluate(in, out, ok, message);
	if(!ok)
		return;
	
	QdObject_var val1=in[0]->connectedSlot()->getValue(ok);
	if(!ok)
	{
		message=QString("Input ")+in[0]->getName()+QString(" invalid");
		ok=false;
		return;
	}
	QdObject_var val2=in[1]->connectedSlot()->getValue(ok);
	if(!ok)
	{
		message=QString("Input ")+in[1]->getName()+QString(" invalid");
		ok=false;
		return;
	}
	
	QdRScalar_var gap = QdRScalar_var::narrow(val2);
	if( gap.is_nil() )
	{
		message=QString("Gap input is not a scalar value");
		ok=false;
		return; 
	}
	double gapSize = *gap;	
	
	if(val1->sequence_size() < 2)
	{
		message=QString("Input sequence is too short" );
		ok=false;
		return; 
	}
	
	QdTimeSeq_var tt;
	QdRScalarSeq_var ss;
	QdObject_var d0;
	QdRScalarSeq_var h = new QdRScalarSeq(val1->sequence_size());
	
	if(val1->xref_exists(DEPEND_0) ){
		d0 = val1->get_xref(DEPEND_0);
		tt = QdTimeSeq_var::narrow(d0);
		ss = QdRScalarSeq_var::narrow(d0);
	}
	if(tt.is_nil() && ss.is_nil()){
		message=QString("Input sequence has no useable DEPEND_0" );
		ok=false;
		return;
	}
	
	if(!tt.is_nil()){
		h[0] = 0.0;
		for(int i=1; i<tt->sequence_size(); i++){
			h[i] = tt[i] - tt[i-1];
		}
	}
	else if(!ss.is_nil()){
		h[0] = 0.0;
		for(int i=1; i<ss->sequence_size(); i++){
			h[i] = ss[i] - ss[i-1];
		}
	}
	
	QdRMatrixSeq_var msobj = QdRMatrixSeq_var:: narrow(val1);
	QdRScalarSeq_var ssobj = QdRScalarSeq_var:: narrow(val1);

	// do arithmetic...
	try
	{
		// 3 point estimate	

		if (!ssobj.is_nil()) 
		{

			QdRScalarSeq_var retS = new QdRScalarSeq(ssobj->sequence_size());
			// do first entry as 2-point estimate
			retS[0] = (ssobj[1] - ssobj[0])	/ h[1];
			for(int i=1; i<ssobj->sequence_size()-1; i++){
				// use 2-point measure if gap
				if(h[i+1] > gapSize) {
					retS[i] = (ssobj[i] - ssobj[i-1]) / h[i];
					i++;
					// and set start of next block with 2-point estimate
					if(i<ssobj->sequence_size()-1) retS[i] = (ssobj[i+1] - ssobj[i]) / h[i+1]; 

				}
				else retS[i] = (ssobj[i+1] - ssobj[i-1]) / (h[i] + h[i+1]);
			}
			// do last entry as 2-point estimate
			retS[ssobj->sequence_size()-1] = (ssobj[ssobj->sequence_size()-1] - ssobj[ssobj->sequence_size()-2]) / h[ssobj->sequence_size()-1];
			
			result = retS;
			result->copy_xrefs_from(ssobj);
			
		}
  		else if (!msobj.is_nil()) 
		{

			QdRMatrixSeq_var retM = Factory::clone(msobj);
			
			// do first entry as 2-point estimate
			retM[0] = (msobj[1] - msobj[0])	/ h[1];
			for(int i=1; i<msobj->sequence_size()-1; i++){
				// ensure no gaps 
				if(h[i+1] > gapSize) {
					retM[i] = (msobj[i] - msobj[i-1]) / h[i];
					i++;
					// set start of next block with 2-point estimate
					if(i<msobj->sequence_size()-1) retM[i] = (msobj[i+1] - msobj[i]) / h[i+1]; 

				}
				else retM[i] = (msobj[i+1] - msobj[i-1]) / (h[i] + h[i+1]); // 3-point estimate
			}
			// do last entry as 2-point estimate
			retM[msobj->sequence_size()-1] = (msobj[msobj->sequence_size()-1] - msobj[msobj->sequence_size()-2]) / h[msobj->sequence_size()-1];
			
			result = retM;
			result->copy_xrefs_from(msobj);

		}
		else
		{
			message=QString("Input not a valid numeric sequence type (scalar or matrix)" );
			ok=false;
			return;
		}

	}
	catch( Exception &e )
	{
		message=QString(e.msg().c_str());
		message+=QString(" Differentiation is not supported between these input types");
		ok=false;
		return;
	}

	if(result.is_nil())
	{
		message=QString("Differentiation>> Not valid for these input types");
		ok=false;
		return;
	}

	// Add timeline (or DEPEND_0)
	QdObject_var qd0 = get_Depend_0( val1 );
	if ( !qd0.is_nil() ) result->change_xref(DEPEND_0 , qd0 );


	// Fix QSAS essential attriutes
	
	// Units
	
	attrPtr = QuGetAttrText(UNITS, d0);
	if( !attrPtr.empty() ){
		string string_buf = " (";
     	if(tt.is_nil() ) string_buf += attrPtr;
		else string_buf += "s"; // QSAS does time arithmetic in seconds whatever the metadata says
     	string_buf += ")^-1";
 	
	
		attrPtr = QuGetAttrText(UNITS, val1);
		if( !attrPtr.empty() ){
	
			QdString_var unitS = new QdString( attrPtr.c_str() );
			*unitS += QdString(string_buf.c_str());
			result->change_xref(UNITS, unitS);
		}
	} 
	
	// SI_conversion
    attrPtr = QuSIConv_inverse(d0);
    QdRScalar_var inv_ptr = new QdRScalar(0.0); // this is irrelevant as we only need metadata
   
    if(!attrPtr.empty() ){
		if(tt.is_nil() ) QuSetTxtAttr(SI_CONVERSION, attrPtr.c_str(), inv_ptr);
		else QuSetTxtAttr(SI_CONVERSION, "1.0 > s^-1", inv_ptr);

		attrPtr = QuSIConv_product(val1, inv_ptr);
		if ( !attrPtr.empty() ) {
			QuSetTxtAttr(SI_CONVERSION, attrPtr.c_str(), result);
 		}
	}
	
	if(val1->xref_exists("FIELDNAM") ){
		QdObject_var xref = val1->get_xref("FIELDNAM");
		QdString_var sxref = QdString_var::narrow(xref);
		if(!sxref.is_nil()){
			QdString_var fName = new QdString("d ");
			*fName +=  *sxref;
			if(!tt.is_nil() ) *fName += "/ dt";
			else *fName += "/ ds";
			result->change_xref("FIELDNAM", fName);
		}
	}

	QuSetTxtAttr("Method", "Derivative (3-point estimate)", result);

	
	out[0]->setValue(result, true);
}

// ----------------------------------


void QEvaluatorDiff5::evaluate(QVector<VInSlot*>& in, QVector<VOutSlot*>& out, bool& ok, QString& message)
{
	ok=true;
	QdObject_var result;
	string attrPtr;
	
	QEvaluator::evaluate(in, out, ok, message);
	if(!ok)
		return;
	
	QdObject_var val1=in[0]->connectedSlot()->getValue(ok);
	if(!ok)
	{
		message=QString("Input ")+in[0]->getName()+QString(" invalid");
		ok=false;
		return;
	}
	
	if(val1->sequence_size() < 2){
		message=QString("Input sequence is too short" );
		ok=false;
		return; 
	}
	
	QdTimeSeq_var tt;
	QdRScalarSeq_var ss;
	QdObject_var d0;
	QdRScalarSeq_var h = new QdRScalarSeq(val1->sequence_size());
	
	if(val1->xref_exists(DEPEND_0) ){
		d0 = val1->get_xref(DEPEND_0);
		tt = QdTimeSeq_var::narrow(d0);
		ss = QdRScalarSeq_var::narrow(d0);
	}
	if(tt.is_nil() && ss.is_nil()){
		message=QString("Input sequence has no useable DEPEND_0" );
		ok=false;
		return;
	}
	
	if(!tt.is_nil()){
		h[0] = 0.0;
		for(int i=1; i<tt->sequence_size(); i++){
			h[i] = tt[i] - tt[i-1];
		}
	}
	else if(!ss.is_nil()){
		h[0] = 0.0;
		for(int i=1; i<ss->sequence_size(); i++){
			h[i] = ss[i] - ss[i-1];
		}
	}
	
	QdRMatrixSeq_var msobj = QdRMatrixSeq_var:: narrow(val1);
	QdRScalarSeq_var ssobj = QdRScalarSeq_var:: narrow(val1);

	// do arithmetic...
	try
	{
		// 5 point estimate

		// First test for regular data
		double hReg;
		bool isReg;
		if(!tt.is_nil()) isReg = QuIsTSRegular(tt, &hReg);
		else  isReg = QuIsSSRegular(ss, &hReg);
		
		double h12 = 12.* hReg;
		double h2 = 2.* hReg;
		double gapSize = 1.5 * hReg;
		
		if(!isReg){
		
			QplugAppendTextDisplay("WARNING: 5pt Derivative requires regular time tags, making new regular series");
			
			// make tt or ss regular and join
			if(tt.is_nil() ){
				// scalar sequence
				double range = ss[ss->sequence_size()-1] - ss[0];
				int Ntags = (int) (range/hReg + 1.0);
    			QdRScalarSeq_var stags = new QdRScalarSeq();
    			stags->assign_regular(Ntags, ss[0], hReg );
				
				if(!ssobj.is_nil()){
					QdRScalarSeq_var newSS = new QdRScalarSeq(Ntags);
					newSS[0] = ssobj[0];
					int j = 0;
					for(int i=1; i<Ntags; i++){
						while(ss[j] < stags[i] ) j++;
						if(j >= ssobj->sequence_size() ) newSS[i] = ssobj[ssobj->sequence_size() - 1]; // catch last entry
						else newSS[i] = (ssobj[j-1] * (ss[j] - stags[i]) + ssobj[j] * (stags[i] - ss[j-1]) ) / (ss[j] - ss[j-1]); 
					}
					newSS->copy_xrefs_from(ssobj);
					ssobj = newSS;
					ssobj->change_xref(DEPEND_0, stags);
				}
				else if(!msobj.is_nil()){
					QdRMatrixSeq_var newMS = new QdRMatrixSeq(Ntags, msobj[0]);
					newMS[0] = msobj[0];
					int j = 0;
					for(int i=1; i<Ntags; i++){
						while(ss[j] < stags[i] ) j++;
						if(j >= msobj->sequence_size() ) newMS[i] = msobj[msobj->sequence_size() - 1]; // catch last entry
						else newMS[i] = (msobj[j-1] * (ss[j] - stags[i]) + msobj[j] * (stags[i] - ss[j-1]) ) / (ss[j] - ss[j-1]); 
					}
					newMS->copy_xrefs_from(msobj);
					msobj = newMS;
					msobj->change_xref(DEPEND_0, stags);
				}
				
			}
			else{
				// time sequence
				double Tinterval = tt->get_interval_bounds().duration_abs();
				int Ntags = (int) (Tinterval/hReg + 1.0);

    			QdTimeSeq_var ttags = new QdTimeSeq();
    			ttags->assign_regular(Ntags, tt[0], hReg );
			
  				if(!ssobj.is_nil()){
					QdRScalarSeq_var newSS = new QdRScalarSeq(Ntags);
					newSS[0] = ssobj[0];
					int j = 0;
					for(int i=1; i<Ntags; i++){
						while(tt[j] < ttags[i] && j < ssobj->sequence_size()) j++;
						if(j >= ssobj->sequence_size() ) newSS[i] = ssobj[ssobj->sequence_size() - 1]; // catch last entry
						else newSS[i] = (ssobj[j-1] * (tt[j] - ttags[i]) + ssobj[j] * (ttags[i] - tt[j-1]) ) / (tt[j] - tt[j-1]); 
					}
					newSS->copy_xrefs_from(ssobj);
					ssobj = newSS;
					ssobj->change_xref(DEPEND_0, ttags);
				}
				else if(!msobj.is_nil()){
					QdRMatrixSeq_var newMS = new QdRMatrixSeq(Ntags, msobj[0]);
					newMS[0] = msobj[0];
					int j = 0;

					for(int i=1; i<Ntags; i++){

						while(tt[j] < ttags[i] && j < msobj->sequence_size()) j++;
						if(j >= msobj->sequence_size() ) newMS[i] = msobj[msobj->sequence_size() - 1]; // catch last entry
						else newMS[i] = (msobj[j-1] * (tt[j] - ttags[i]) + msobj[j] * (ttags[i] - tt[j-1]) ) / (tt[j] - tt[j-1]); 
					}
					newMS->copy_xrefs_from(msobj);
					msobj = newMS;
					msobj->change_xref(DEPEND_0, ttags);
				}

			}
		}

		// Do derivative
		
		if (!ssobj.is_nil()) 
		{

			QdRScalarSeq_var retS = new QdRScalarSeq(ssobj->sequence_size());
			
			// do first entry as 2-point estimate
			retS[0] = (ssobj[1] - ssobj[0])	/ hReg;
			// do second entry as 3-point estimate
			retS[1] = (ssobj[2] - ssobj[0])	/ h2;
			
			for(int i=2; i<ssobj->sequence_size()-2; i++){
				retS[i] = (-ssobj[i+2] + 8.*ssobj[i+1] - 8.*ssobj[i-1] + ssobj[i-2]) / h12; // 5-point estimate
			}
			// do last entries as 2-point and 3-point estimate
			retS[ssobj->sequence_size()-2] = (ssobj[ssobj->sequence_size()-1] - ssobj[ssobj->sequence_size()-3]) / h2;
			retS[ssobj->sequence_size()-1] = (ssobj[ssobj->sequence_size()-1] - ssobj[ssobj->sequence_size()-2]) / hReg;
			
			result = retS;
			result->copy_xrefs_from(ssobj);
			
		}
  		else if (!msobj.is_nil()) 
		{

			QdRMatrixSeq_var retM = new QdRMatrixSeq(msobj->sequence_size(), msobj[0]);
			
			// do first entry as 2-point estimate
			retM[0] = (msobj[1] - msobj[0])	/ hReg;
			// do second entry as 3-point estimate
			retM[1] = (msobj[2] - msobj[0])	/ h2;
			
			for(int i=2; i<msobj->sequence_size()-2; i++){
				retM[i] = (-msobj[i+2] + 8.*msobj[i+1] - 8.*msobj[i-1] + msobj[i-2]) / h12; // 5-point estimate
			}
			// do last entries as 2-point and 3-point estimate
			retM[msobj->sequence_size()-2] = (msobj[msobj->sequence_size()-1] - msobj[msobj->sequence_size()-3]) / h2;
			retM[msobj->sequence_size()-1] = (msobj[msobj->sequence_size()-1] - msobj[msobj->sequence_size()-2]) / hReg;
			
			result = retM;
			result->copy_xrefs_from(msobj);

		}
		else
		{
			message=QString("Input not a valid numeric sequence type (scalar or matrix)" );
			ok=false;
			return;
		}

	}
	catch( Exception &e )
	{
		message=QString(e.msg().c_str());
		message+=QString(" Differentiation is not supported between these input types");
		ok=false;
		return;
	}

	if(result.is_nil())
	{
		message=QString("Differentiation>> Not valid for these input types");
		ok=false;
		return;
	}


	// Fix QSAS essential attriutes
	
	// Units
	
	attrPtr = QuGetAttrText(UNITS, d0);
	if( !attrPtr.empty() ){
     	string string_buf = " (";
     	if(tt.is_nil() ) string_buf += attrPtr;
		else string_buf += "s"; // QSAS does time arithmetic in seconds whatever the metadata says
     	string_buf += ")^-1";
 	
	
		attrPtr = QuGetAttrText(UNITS, val1);
		if( !attrPtr.empty() ){
	
			QdString_var unitS = new QdString(attrPtr.c_str());
			*unitS += QdString(string_buf.c_str());
			result->change_xref(UNITS, unitS);
		}
	} 
	
	// SI_conversion
    attrPtr = QuSIConv_inverse(d0);
    QdRScalar_var inv_ptr = new QdRScalar(0.0); // this is irrelevant as we only need metadata
   
    if( !attrPtr.empty() ){
		if(tt.is_nil() ) QuSetTxtAttr(SI_CONVERSION, attrPtr.c_str(), inv_ptr);
		else QuSetTxtAttr(SI_CONVERSION, "1.0 > s^-1", inv_ptr);

		attrPtr = QuSIConv_product(val1, inv_ptr);
		if ( !attrPtr.empty() ) {
			QuSetTxtAttr(SI_CONVERSION, attrPtr.c_str(), result);
 		}
	}
	
	if(val1->xref_exists("FIELDNAM") ){
		QdObject_var xref = val1->get_xref("FIELDNAM");
		QdString_var sxref = QdString_var::narrow(xref);
		if(!sxref.is_nil()){
			QdString_var fName = new QdString("d ");
			*fName +=  *sxref;
			if(!tt.is_nil() ) *fName += "/ dt";
			else *fName += "/ ds";
			result->change_xref("FIELDNAM", fName);
		}
	}

	QuSetTxtAttr("Method", "Derivative (5-point estimate)", result);

	
	out[0]->setValue(result, true);
}

// ---------------------

void QEvaluatorIntDef::evaluate(QVector<VInSlot*>& in, QVector<VOutSlot*>& out, bool& ok, QString& message)
{
	ok=true;
	QdObject_var retObj;
	string attrPtr;
	
	QEvaluator::evaluate(in, out, ok, message);
	if(!ok)
		return;
	
	QdObject_var dobj=in[0]->connectedSlot()->getValue(ok);
	if(!ok)
	{
		message=QString("Input ")+in[0]->getName()+QString(" invalid");
		ok=false;
		return;
	}
	QdObject_var val2=in[1]->connectedSlot()->getValue(ok);
	if(!ok)
	{
		message=QString("Input ")+in[1]->getName()+QString(" invalid");
		ok=false;
		return;
	}
	
	QdRScalar_var opt = QdRScalar_var::narrow(val2);
	if( opt.is_nil() )
	{
		message=QString("Method input is not a scalar value");
		ok=false;
		return; 
	}
	double gapMethod = *opt;	
	
	if(dobj->sequence_size() < 2)
	{
		message=QString("Input sequence is too short" );
		ok=false;
		return; 
	}
	
	QdTimeSeq_var tt;
	QdRScalarSeq_var ss;
	QdObject_var d0;
	QdRScalarSeq_var x = new QdRScalarSeq(dobj->sequence_size(), 0.0);
	QdRScalarSeq_var xm = new QdRScalarSeq(dobj->sequence_size(), 0.0);
	QdRScalarSeq_var xp = new QdRScalarSeq(dobj->sequence_size(), 0.0);
	double deltaTop = 0.0;
	
	if(dobj->xref_exists(DEPEND_0) ){
		d0 = dobj->get_xref(DEPEND_0);
		if(!d0.is_nil()) {
			tt = QdTimeSeq_var::narrow(d0);
			ss = QdRScalarSeq_var::narrow(d0);
		}
	}
	if(tt.is_nil() && ss.is_nil()){
		message=QString("Input sequence has no useable DEPEND_0 ");
		ok=false;
		return;
	}
	
	bool deltasFound=false;
	if(d0->xref_exists(DELTA_PLUS) && d0->xref_exists(DELTA_MINUS)){

		// deltas exist and gaps ignored (so use bin width from deltas)
		QdObject_var dp = tt->get_xref(DELTA_PLUS);
		QdObject_var dm = tt->get_xref(DELTA_MINUS);
		QdRScalar_var dps = QdRScalar_var::narrow(dp);
		QdRScalarSeq_var dpss;
		QdRScalar_var dms = QdRScalar_var::narrow(dm);
		QdRScalarSeq_var dmss;
		
		if(!dps.is_nil()){
			dpss = new QdRScalarSeq(dobj->sequence_size(), *dps);
			dpss->copy_xrefs_from(dps);
		}
		else dpss = QdRScalarSeq_var::narrow(dp);
		
		if(!dms.is_nil()){
			dmss = new QdRScalarSeq(dobj->sequence_size(), *dms);
			dmss->copy_xrefs_from(dms);
		}
		else dmss = QdRScalarSeq_var::narrow(dm);
		
		if(!dpss.is_nil() && !dmss.is_nil()){
			deltasFound=true;
			// we now have deltas as sequences
			if(!tt.is_nil()){
				QdRScalarSeq_var dmsec;
				QdRScalarSeq_var dpsec;
				if(dmss->xref_exists(SI_CONVERSION) && dpss->xref_exists(SI_CONVERSION) ){
					dmsec = QdRScalarSeq_var::narrow(QarConvertUnitsTo(dmss, "1.0>s"));
					dpsec = QdRScalarSeq_var::narrow(QarConvertUnitsTo(dpss, "1.0>s"));
				}
				else{
					dmsec = dmss;
					dpsec = dpss;
					message=QString("Timetag deltas have no SI_CONVERSION, result magnitude assumed sec");
					
				}
				xm[0] = 0.0;
				x[0] = dmsec[0];
				xp[0] = x[0] + dpsec[0];
				for(int i=1; i<dobj->sequence_size(); i++){
					// done this way to get scalar from time
					x[i] = tt[i] - tt[i-1] + x[i-1];
					xm[i] = x[i] - dmsec[i];
					xp[i] = x[i] + dpsec[i];
				}
				deltaTop = dpsec[dobj->sequence_size()-1];
			}
			else {
				xm[0] = 0.0;
				x[0] = dmss[0];
				xp[0] = x[0] + dpss[0];
				for(int i=1; i<dobj->sequence_size(); i++){
					// done this way to get scalar from time
					x[i] = ss[i] - ss[i-1] + x[i-1];
					xm[i] = x[i] - dmss[i];
					xp[i] = x[i] + dpss[i];
				}
				deltaTop = dpss[dobj->sequence_size()-1];

			} // end if on tt vs ss
		} // end if on deltas found ok
	}

	if(gapMethod != 2 || !deltasFound){
	
		// not using deltas except at ends if possible (fill gaps)
		if(!tt.is_nil()){
			// replacing x, xm and xp, so use ends before overwritten
			if(!deltasFound) x[0] = 0.0;
			// else is already incremented by delta_minus
			
			xm[0] = 0.0;
			for(int i=1; i<dobj->sequence_size(); i++){
				// done this way to get scalar from time
				xm[i] = x[i-1] + 0.5*(tt[i] - tt[i-1]);
				x[i] = x[i-1] + (double)(tt[i] - tt[i-1]) ;
				xp[i-1] = xm[i];
			}
			xp[dobj->sequence_size()-1] = x[dobj->sequence_size()-1] + deltaTop;
		}
		else {
			if(deltasFound) x[0] = xm[0];
			else  x[0] = 0.0;
			xm[0] = 0.0;
			for(int i=1; i<dobj->sequence_size(); i++){
				// done this way to get scalar from time
				xm[i] = x[i-1] + 0.5*(ss[i] - ss[i-1]);
				x[i] = x[i-1] + ss[i] - ss[i-1];
				xp[i-1] = xm[i];
			}
			xp[dobj->sequence_size()-1] = x[dobj->sequence_size()-1] + deltaTop;
	
		}
	}
	
	QdRMatrixSeq_var msobj = QdRMatrixSeq_var:: narrow(dobj);
	QdRScalarSeq_var ssobj = QdRScalarSeq_var:: narrow(dobj);
	
	// Note step lengths may not be constant, so Trapezium, Simpson's rule etc don't work.
	
	if(gapMethod == 0 || gapMethod == 2){
		// histogram
		// same method for ignore gaps or histogram as difference is only in xm and xp bin boundaries
	
		// definite integral
		if(!ssobj.is_nil()){
			QdRScalar_var Iobj = new QdRScalar(0.);
			for(int i=0; i<dobj->sequence_size(); i++){
				*Iobj += ssobj[i]*(xp[i] - xm[i]);
			}
			retObj = Iobj;
		}
		if(!msobj.is_nil()){
			QdRMatrix_var Iobj = new QdRMatrix(msobj[0].dims(), 0.0);
			for(int i=0; i<dobj->sequence_size(); i++){
				*Iobj += msobj[i]*(xp[i] - xm[i]);
			}
			retObj = Iobj;
		}
			
	}
	else if(gapMethod == 1){
		// linear
	

		// definite integral
		if(!ssobj.is_nil()){
			QdRScalar_var Iobj = new QdRScalar(0.);
			double h = xp[0] - xm[0];
			*Iobj = (double)ssobj[0] * h;
			for(int i=1; i<dobj->sequence_size()-1; i++){
				// does gap filling linearly
				double h = xp[i] - xm[i];
				double hm = 0.5 * h * (xm[i] - x[i]) / (x[i-1] - x[i]);
				double hp = 0.5 * h * (xp[i] - x[i]) / (x[i+1] - x[i]) ;
				h = h - hp - hm;
				*Iobj += ssobj[i]*h + ssobj[i+1]*hp + ssobj[i-1]*hm  ;
			}
			h = xp[dobj->sequence_size()-1] - xm[dobj->sequence_size()-1];
			*Iobj = (double) ssobj[dobj->sequence_size()-1] * h;

			retObj = Iobj;
		}
		if(!msobj.is_nil()){
			QdRMatrix_var Iobj = new QdRMatrix(msobj[0].dims(), 0.0);
			*Iobj = (xp[0] - xm[0]) * msobj[0];
			for(int i=1; i<dobj->sequence_size()-1; i++){
				// does gap filling linearly
				double h = xp[i] - xm[i];
				double hm = 0.5 * h * (xm[i] - x[i]) / (x[i-1] - x[i]);
				double hp = 0.5 * h * (xp[i] - x[i]) / (x[i+1] - x[i]) ;
				h = h - hp - hm;
				*Iobj += msobj[i]*h + msobj[i+1]*hp + msobj[i-1]*hm  ;
			}
			*Iobj = (xp[dobj->sequence_size()-1] - xm[dobj->sequence_size()-1]) * msobj[dobj->sequence_size()-1];

			retObj = Iobj;
		}
			
	}

	if( retObj.is_nil() ) {
		message=QString("NULL object from integral");
		ok=false;
		return;
	}


  	// set xrefs
	
	retObj->copy_xrefs_from(dobj);

	retObj->delete_xref(DEPEND_0);
		
	// Units
	
	string attr_ptr = QuGetAttrText(UNITS, d0);
	string string_buf=" ";
	if( !attr_ptr.empty() ){
     	if(tt.is_nil() ) string_buf += attr_ptr;
		else string_buf += "s"; // QSAS does time arithmetic in seconds whatever the metadata says
	
		attr_ptr = QuGetAttrText(UNITS, dobj);
		if( !attr_ptr.empty() ){
	
			QdString_var unitS = new QdString(attr_ptr.c_str());
			*unitS += QdString(string_buf.c_str());
			retObj->change_xref(UNITS, unitS);
		}
	} 
	
	// SI_conversion
    attr_ptr = QuGetAttrText(SI_CONVERSION, d0);
    QdRScalar_var dummy_ptr = new QdRScalar(0.0); // this is irrelevant as we only need metadata
   
    if( !attr_ptr.empty() ){
		if(tt.is_nil() ) QuSetTxtAttr(SI_CONVERSION, attr_ptr.c_str(), dummy_ptr);
		else QuSetTxtAttr(SI_CONVERSION, "1.0 > s", dummy_ptr);// QSAS does time arithmetic in seconds
		
		attr_ptr = QuSIConv_product(dobj, dummy_ptr);
		if ( !attr_ptr.empty() ) {
			QuSetTxtAttr(SI_CONVERSION, attr_ptr.c_str(), retObj);
 		}
	}
	
	if(dobj->xref_exists("FIELDNAM") ){
		QdObject_var xref = dobj->get_xref("FIELDNAM");
		QdString_var sxref = QdString_var::narrow(xref);
		if(!sxref.is_nil()){
			QdString_var fName = new QdString("Integral(");
			*fName +=  *sxref;
			if(!tt.is_nil() ) *fName += ")dt";
			else *fName += ") ds";
			retObj->change_xref("FIELDNAM", fName);
		}
	}

	if(gapMethod == 0) QuSetTxtAttr("Method", "Histogram sum (gaps filled)", retObj);
	else if(gapMethod == 1) QuSetTxtAttr("Method", "Linear fit (gaps filled)", retObj);
	else QuSetTxtAttr("Method", "Histogram sum (gaps excluded)", retObj);

	
	out[0]->setValue(retObj, true);
}

// ---------------------

void QEvaluatorIntIndef::evaluate(QVector<VInSlot*>& in, QVector<VOutSlot*>& out, bool& ok, QString& message)
{
	ok=true;
	QdObject_var retObj;
	string attrPtr;
	
	QEvaluator::evaluate(in, out, ok, message);
	if(!ok)
		return;
	
	QdObject_var dobj=in[0]->connectedSlot()->getValue(ok);
	if(!ok)
	{
		message=QString("Input ")+in[0]->getName()+QString(" invalid");
		ok=false;
		return;
	}
	QdObject_var val2=in[1]->connectedSlot()->getValue(ok);
	if(!ok)
	{
		message=QString("Input ")+in[1]->getName()+QString(" invalid");
		ok=false;
		return;
	}
	
	QdRScalar_var opt = QdRScalar_var::narrow(val2);
	if( opt.is_nil() )
	{
		message=QString("Method input is not a scalar value");
		ok=false;
		return; 
	}
	double gapMethod = *opt;	
	
	if(dobj->sequence_size() < 2)
	{
		message=QString("Input sequence is too short" );
		ok=false;
		return; 
	}
	
	QdTimeSeq_var tt;
	QdRScalarSeq_var ss;
	QdObject_var d0;
	QdRScalarSeq_var x = new QdRScalarSeq(dobj->sequence_size(), 0.0);
	QdRScalarSeq_var xm = new QdRScalarSeq(dobj->sequence_size(), 0.0);
	QdRScalarSeq_var xp = new QdRScalarSeq(dobj->sequence_size(), 0.0);
	double deltaTop = 0.0;
	
	if(dobj->xref_exists(DEPEND_0) ){
		d0 = dobj->get_xref(DEPEND_0);
		if(!d0.is_nil()) {
			tt = QdTimeSeq_var::narrow(d0);
			ss = QdRScalarSeq_var::narrow(d0);
		}
	}
	if(tt.is_nil() && ss.is_nil()){
		message=QString("Input sequence has no useable DEPEND_0 ");
		ok=false;
		return;
	}
	
	bool deltasFound=false;
	if(d0->xref_exists(DELTA_PLUS) && d0->xref_exists(DELTA_MINUS)){

		// deltas exist and gaps ignored (so use bin width from deltas)
		QdObject_var dp = tt->get_xref(DELTA_PLUS);
		QdObject_var dm = tt->get_xref(DELTA_MINUS);
		QdRScalar_var dps = QdRScalar_var::narrow(dp);
		QdRScalarSeq_var dpss;
		QdRScalar_var dms = QdRScalar_var::narrow(dm);
		QdRScalarSeq_var dmss;
		
		if(!dps.is_nil()){
			dpss = new QdRScalarSeq(dobj->sequence_size(), *dps);
			dpss->copy_xrefs_from(dps);
		}
		else dpss = QdRScalarSeq_var::narrow(dp);
		
		if(!dms.is_nil()){
			dmss = new QdRScalarSeq(dobj->sequence_size(), *dms);
			dmss->copy_xrefs_from(dms);
		}
		else dmss = QdRScalarSeq_var::narrow(dm);
		
		if(!dpss.is_nil() && !dmss.is_nil()){
			deltasFound=true;
			// we now have deltas as sequences
			if(!tt.is_nil()){
				QdRScalarSeq_var dmsec;
				QdRScalarSeq_var dpsec;
				if(dmss->xref_exists(SI_CONVERSION) && dpss->xref_exists(SI_CONVERSION) ){
					dmsec = QdRScalarSeq_var::narrow(QarConvertUnitsTo(dmss, "1.0>s"));
					dpsec = QdRScalarSeq_var::narrow(QarConvertUnitsTo(dpss, "1.0>s"));
				}
				else{
					dmsec = dmss;
					dpsec = dpss;
					message=QString("Timetag deltas have no SI_CONVERSION, result magnitude assumed sec");
					
				}
				xm[0] = 0.0;
				x[0] = dmsec[0];
				xp[0] = x[0] + dpsec[0];
				for(int i=1; i<dobj->sequence_size(); i++){
					// done this way to get scalar from time
					x[i] = tt[i] - tt[i-1] + x[i-1];
					xm[i] = x[i] - dmsec[i];
					xp[i] = x[i] + dpsec[i];
				}
				deltaTop = dpsec[dobj->sequence_size()-1];
			}
			else {
				xm[0] = 0.0;
				x[0] = dmss[0];
				xp[0] = x[0] + dpss[0];
				for(int i=1; i<dobj->sequence_size(); i++){
					// done this way to get scalar from time
					x[i] = ss[i] - ss[i-1] + x[i-1];
					xm[i] = x[i] - dmss[i];
					xp[i] = x[i] + dpss[i];
				}
				deltaTop = dpss[dobj->sequence_size()-1];

			} // end if on tt vs ss
		} // end if on deltas found ok
	}

	if(gapMethod != 2 || !deltasFound){
	
		// not using deltas except at ends if possible (fill gaps)
		if(!tt.is_nil()){
			// replacing x, xm and xp, so use ends before overwritten
			if(!deltasFound) x[0] = 0.0;
			// else is already incremented by delta_minus
			
			xm[0] = 0.0;
			for(int i=1; i<dobj->sequence_size(); i++){
				// done this way to get scalar from time
				xm[i] = x[i-1] + 0.5*(tt[i] - tt[i-1]);
				x[i] = x[i-1] + (double)(tt[i] - tt[i-1]) ;
				xp[i-1] = xm[i];
			}
			xp[dobj->sequence_size()-1] = x[dobj->sequence_size()-1] + deltaTop;
		}
		else {
			if(deltasFound) x[0] = xm[0];
			else  x[0] = 0.0;
			xm[0] = 0.0;
			for(int i=1; i<dobj->sequence_size(); i++){
				// done this way to get scalar from time
				xm[i] = x[i-1] + 0.5*(ss[i] - ss[i-1]);
				x[i] = x[i-1] + ss[i] - ss[i-1];
				xp[i-1] = xm[i];
			}
			xp[dobj->sequence_size()-1] = x[dobj->sequence_size()-1] + deltaTop;
	
		}
	}
	
	QdRMatrixSeq_var msobj = QdRMatrixSeq_var:: narrow(dobj);
	QdRScalarSeq_var ssobj = QdRScalarSeq_var:: narrow(dobj);
	
	// Note step lengths may not be constant, so Trapezium, Simpson's rule etc don't work.
	
	if(gapMethod == 0 || gapMethod == 2){
		// histogram
		// same method for ignore gaps or histogram as difference is only in xm and xp bin boundaries
		
		// indefinite integral
		if(!ssobj.is_nil()){
			QdRScalarSeq_var Iobj = new QdRScalarSeq(dobj->sequence_size(), 0.0);
			Iobj[0] = (xp[0] - xm[0]) * ssobj[0];
			for(int i=1; i<dobj->sequence_size(); i++){
				Iobj[i] = Iobj[i-1] + ssobj[i]*(xp[i] - xm[i]);
			}
			retObj = Iobj;
		}
		if(!msobj.is_nil()){
			QdRMatrixSeq_var Iobj = new QdRMatrixSeq(dobj->sequence_size(), msobj[0].dims());
			Iobj[0] = msobj[0]*(xp[0] - xm[0]);
			for(int i=1; i<dobj->sequence_size(); i++){
				Iobj[i] = Iobj[i-1] + msobj[i]*(xp[i] - xm[i]);
			}
			retObj = Iobj;
		}
		
	}
	else if(gapMethod == 1){
		// linear
	
		// indefinite integral
		if(!ssobj.is_nil()){
			QdRScalarSeq_var Iobj = new QdRScalarSeq(dobj->sequence_size(), 0.0);
			double h = xp[0] - xm[0];
			Iobj[0] = ssobj[0] * h;
			for(int i=1; i<dobj->sequence_size()-1; i++){
				double h = xp[i] - xm[i];
				double hm = 0.5 * h * (xm[i] - x[i]) / (x[i-1] - x[i]);
				double hp = 0.5 * h * (xp[i] - x[i]) / (x[i+1] - x[i]) ;
				h = h - hp - hm;
				Iobj[i] = Iobj[i-1] + ssobj[i]*h + ssobj[i+1]*hp + ssobj[i-1]*hm  ;
			}
			Iobj[dobj->sequence_size()-1] = Iobj[dobj->sequence_size()-2] + (xp[dobj->sequence_size()-1] - xm[dobj->sequence_size()-1]) * ssobj[dobj->sequence_size()-1];
			retObj = Iobj;
		}
		if(!msobj.is_nil()){
			QdRMatrixSeq_var Iobj = new QdRMatrixSeq(dobj->sequence_size(), msobj[0].dims());
			Iobj[0] = (xp[0] - xm[0]) * msobj[0];
			for(int i=1; i<dobj->sequence_size()-1; i++){
				double h = xp[i] - xm[i];
				double hm = 0.5 * h * (xm[i] - x[i]) / (x[i-1] - x[i]);
				double hp = 0.5 * h * (xp[i] - x[i]) / (x[i+1] - x[i]) ;
				h = h - hp - hm;
				Iobj[i] = Iobj[i-1] + msobj[i]*h + msobj[i+1]*hp + msobj[i-1]*hm  ;
			}
			Iobj[dobj->sequence_size()-1] = Iobj[dobj->sequence_size()-2] + (xp[dobj->sequence_size()-1] - xm[dobj->sequence_size()-1]) * msobj[dobj->sequence_size()-1];
			retObj = Iobj;
		}
		

	}

	if( retObj.is_nil() ) {
		message=QString("NULL object from integral");
		ok=false;
		return;
	}


  	// set xrefs
	
	retObj->copy_xrefs_from(dobj);
		
	// Units
	
	string attr_ptr = QuGetAttrText(UNITS, d0);
	string string_buf=" ";
	if( !attr_ptr.empty() ){
     	if(tt.is_nil() ) string_buf += attr_ptr;
		else string_buf += "s"; // QSAS does time arithmetic in seconds whatever the metadata says
	
		attr_ptr = QuGetAttrText(UNITS, dobj);
		if( !attr_ptr.empty() ){
	
			QdString_var unitS = new QdString(attr_ptr.c_str());
			*unitS += QdString(string_buf.c_str());
			retObj->change_xref(UNITS, unitS);
		}
	} 
	
	// SI_conversion
    attr_ptr = QuGetAttrText(SI_CONVERSION, d0);
    QdRScalar_var dummy_ptr = new QdRScalar(0.0); // this is irrelevant as we only need metadata
   
    if( !attr_ptr.empty() ){
		if(tt.is_nil() ) QuSetTxtAttr(SI_CONVERSION, attr_ptr.c_str(), dummy_ptr);
		else QuSetTxtAttr(SI_CONVERSION, "1.0 > s", dummy_ptr);// QSAS does time arithmetic in seconds
		
		attr_ptr = QuSIConv_product(dobj, dummy_ptr);
		if ( !attr_ptr.empty() ) {
			QuSetTxtAttr(SI_CONVERSION, attr_ptr.c_str(), retObj);
 		}
	}
	
	if(dobj->xref_exists("FIELDNAM") ){
		QdObject_var xref = dobj->get_xref("FIELDNAM");
		QdString_var sxref = QdString_var::narrow(xref);
		if(!sxref.is_nil()){
			QdString_var fName = new QdString("Integral(");
			*fName +=  *sxref;
			if(!tt.is_nil() ) *fName += ")dt";
			else *fName += ") ds";
			retObj->change_xref("FIELDNAM", fName);
		}
	}

	if(gapMethod == 0) QuSetTxtAttr("Method", "Histogram sum (gaps filled)", retObj);
	else if(gapMethod == 1) QuSetTxtAttr("Method", "Linear fit (gaps filled)", retObj);
	else QuSetTxtAttr("Method", "Histogram sum (gaps excluded)", retObj);


	out[0]->setValue(retObj, true);
}



