#include "qcalcInputItem.h"
#include "qcalcSlots.h"
#include "qcalcScene.h"
#include <QGraphicsProxyWidget>
#include <QGraphicsSceneDragDropEvent>
#include "qdutil.h"

using namespace QSAS;

QVector<VInputItem*> InputItems;

VInputItem::VInputItem():
		VCalcItem()
{
	VOutSlot* slot=new VOutSlot(this);
	outputs.push_back(slot);
	joinSettings=new JoinSettings;
    not_joined_value=QdObject_var();
	value=QdObject_var();
	gui=NULL;
	input_id=counter++;
	name=QString("Input #%1").arg(input_id);
#ifdef GRAPH_DEBUG
	std::cout << "Creating VInputItem " << this << std::endl;
#endif
    joinSettings->noJoin=false;
	joinSettings->algorithm=0;
	joinSettings->width=60.;
	joinSettings->gap=0;
	joinSettings->gapSize=6.;
    valueText="";
	
	_keepXrefs = false;
	InputItems.append(this);
}

VInputItem::~VInputItem()
{
#ifdef GRAPH_DEBUG
	std::cout << "Deleting VInputItem " << this << std::endl;
#endif
	if(gui!=NULL)
	{
		QGraphicsScene * scene=gui->scene();
		if(scene!=NULL)
		{
			scene->removeItem(gui);
		}
		gui->parentItem=NULL; // thus gui knows it is being deleted
		
		delete gui;
	}
	
	int index=InputItems.indexOf(this);
	if(index>=0)
		InputItems.remove(index);
	
	delete joinSettings;
}

void VInputItem::setKeepXref(bool keep)
{
	_keepXrefs = keep;
	flagOutSlotsForReevaluation();
}


QDomElement VInputItem::toXMLElement(QDomDocument& document) const
{
	QDomElement tag=document.createElement("Input");
    if(gui!=NULL)
    {
		tag.appendChild(gui->toXMLElement(document));
    }
	tag.appendChild(createTextXMLNode(document, "XPosnInBlackBox", positionInBlackBox.x()));
	tag.appendChild(createTextXMLNode(document, "YPosnInBlackBox", positionInBlackBox.y()));
	tag.appendChild(createTextXMLNode(document, "ID", calc_id));

    tag.appendChild(createTextXMLNode(document, "Data", valueText));
	tag.appendChild(createTextXMLNode(document, "Page", page));
	tag.appendChild(createTextXMLNode(document, "Join_No_Join", joinSettings->noJoin));
	tag.appendChild(createTextXMLNode(document, "Join_Algorithm", joinSettings->algorithm));
	tag.appendChild(createTextXMLNode(document, "Join_Width", joinSettings->width));
	tag.appendChild(createTextXMLNode(document, "Join_Gap", joinSettings->gap));
	tag.appendChild(createTextXMLNode(document, "Join_Gap_Size", joinSettings->gapSize));
	
    return tag;
}


void VInputItem::reevaluateOutSlots()
{
 	if(not_joined_value.is_nil()){
		outputs[0]->setValue(QdObject_var(), false);
		return;
	}
	
    if(joinSettings->noJoin)
    {	
        value=not_joined_value;
    }
    else
    {
       if(!not_joined_value->is_sequence() || joinSettings->noJoin)
        {
            value=not_joined_value;
        }
        else
        {
			
			string gap_opt_ptr;
			switch(joinSettings->gap)
			{
				case 1:
					gap_opt_ptr = String(ZERO_FILL_GAP);
					break;
				case 2:
					gap_opt_ptr = String(FILL_GAP);
					break;
				case 3:
					gap_opt_ptr = String(NEAREST_NEIGHBOUR_GAP);
					break;
				case 4:
					gap_opt_ptr = String(REMOVE_GAP);
					break;
				case 0:
				default:
					gap_opt_ptr = String(LINEAR_INTERP_GAP);
					break;
			}
				
			if(joinSettings->algorithm == 0){
				value = QuLinearJoin(not_joined_value, tags, joinSettings->gapSize, gap_opt_ptr);
			}
			else {
				value = QuBoxcarJoin(not_joined_value, tags, joinSettings->width, gap_opt_ptr);					 
			}	
         
            if(value.is_nil()){
                outputs[0]->setValue(value, false);
                return;
            }
			if(value->sequence_size() == 0)  QplugAppendTextDisplay( "Input failed on join: Check Timeline is valid");
        }
    }

	if(_keepXrefs){
		QdString_var attr = new QdString("QSAS performed arithmetic, but some attributes unchanged");
		value->change_xref("QSAS_MODS", attr);
	}
	else{
		if(value->xref_exists("QSAS_MODS") ) value->delete_xref("QSAS_MODS");
	}
	
	outputs[0]->setValue(value, true);
}

InputItem* VInputItem::getGUI()
{
	if(gui==NULL)
	{
		gui=new InputItem(this, outputs[0]);
	}
	return gui;
}

void VInputItem::removeGUI()
{
	if(gui!=NULL)
	{
		gui->scene()->removeItem(gui);
		delete gui;
		gui=NULL;
	}
}


void VInputItem::setJoinTags(QdObject_var tags)
{
	this->tags=tags;
	flagOutSlotsForReevaluation();
}


void VInputItem::fromXMLElement(QDomElement& el, QGraphicsScene* scene)
{
	VCalcItem::fromXMLElement(el, scene);


	QDomElement guiElement=el.firstChildElement("GUI");
	QString data="";
	QString val;

	if(getXMLNodeText(el, "XPosnInBlackBox", val))
	{
		positionInBlackBox.setX(val.toDouble());
	}
	if(getXMLNodeText(el, "YPosnInBlackBox", val))
	{
		positionInBlackBox.setY(val.toDouble());
	}

	if(getXMLNodeText(el, "Data", val))
	{
		data=val;
	}
	if(getXMLNodeText(el, "Page", val))
	{
		page=val.toInt();
	}
	if(getXMLNodeText(el, "Join_No_Join", val))
	{
		joinSettings->noJoin=val.toInt();
	}
	if(getXMLNodeText(el, "Join_Algorithm", val))
	{
		joinSettings->algorithm=val.toInt();
	}
	if(getXMLNodeText(el, "Join_Width", val))
	{
		joinSettings->width=val.toDouble();
	}
	if(getXMLNodeText(el, "Join_Gap", val))
	{
		joinSettings->gap=val.toInt();
	}
	if(getXMLNodeText(el, "Join_Gap_Size", val))
	{
		joinSettings->gapSize=val.toDouble();
	}
	if(!guiElement.isNull())
	{
		InputItem* item=this->getGUI();
        scene->addItem(item);
        item->fromXMLElement(guiElement);
		item->blockSig();
		item->setValue(data);
		item->allowSig();
	}
	else
	{
		QdObject_var obj=QuGetObjByNameAndComp(data);
        if(!obj.is_nil())
        {
            setValue(obj, data);
        }
        else
        {
            bool ok;
            QdRScalar_var scalar_var=new QdRScalar(data.toDouble(&ok));
            if(ok)
            {
                setValue(scalar_var, data);
            }
            else
            {
                setValue(new QdString(data.toLatin1().data()), data);
            }
        }
	}

}


InputItem::InputItem(VInputItem* parentItem, VOutSlot* output, QGraphicsItem* parent):
		CalcItem(parent)
{
	this->parentItem=parentItem;
	
	outlineColor=QColor(100, 75, 00);
	color1=QColor(255, 255, 100);
	color2=QColor(230, 190, 80);
	
	title=parentItem->getName();
	
	width=BASE_WIDTH;
	height=BASE_HEIGHT;
	setRect(-width/2., -height/2., width, height);
	
	edit=new QguWLinput(NULL, "");
	edit->setMaximumWidth((int)(BASE_WIDTH-6));
	edit->setMinimumWidth((int)(BASE_WIDTH-6));
	edit->setMaximumHeight((int)(HEADER_HEIGHT-2));
	edit->setMinimumHeight((int)(HEADER_HEIGHT-2));
    edit->setAcceptable(Qcondition());
    edit->addAcceptable(Qcondition(QS_NARROW, QS_INT, QS_0D, QS_NOT_SEQ, QS_ANY));
    edit->addAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_0D, QS_NOT_SEQ, QS_ANY));
    edit->addAcceptable(Qcondition(QS_NARROW, QS_TEXT, QS_0D, QS_NOT_SEQ, QS_ANY));
	
	QGraphicsProxyWidget * editProxy=new QGraphicsProxyWidget(this);
	editProxy->setWidget(edit);
	editProxy->moveBy(-edit->width()/2., 0.);
	
	OutSlot* slot=new OutSlot(this, output);
	slot->setPos(0., height/2.);
	outSlots.push_back(slot);
	
	rzButton->setPos(-width/2.+1.5*HEADER_HEIGHT, -height/2.+HEADER_HEIGHT/2.);
	closeButton->setPos(width/2.-HEADER_HEIGHT/2., -height/2.+HEADER_HEIGHT/2.);
	errButton->setPos(width/2.-1.5*HEADER_HEIGHT, -height/2.+HEADER_HEIGHT/2.);
	
	connect(edit, SIGNAL(edited_OK()), this, SLOT(setValue()));
	
	setAcceptDrops(false);
	edit->setAcceptDrops(false);
	
//	setValue(parentItem->valueText);

#ifdef GRAPH_DEBUG
	std::cout << "Creating InputItem " << this << " for VInputItem " << parentItem << std::endl;
#endif

}

InputItem::~InputItem()
{
#ifdef GRAPH_DEBUG
	std::cout << "Deleting InputItem " << this << std::endl;
#endif
	foreach(InSlot* slot, inSlots)
	{
		delete slot;
	}
	foreach(OutSlot* slot, outSlots)
	{
		delete slot;
	}
}


QVariant InputItem::itemChange(GraphicsItemChange change, const QVariant& value)
{
	if (change == ItemPositionChange && scene()) {
		QPointF newPos = value.toPointF();
		
		updateConnectorsPositions();
		
		return newPos;
	}
	else if(change==ItemSelectedChange && scene())
	{
		if(value.toBool())
		{
			setZValue(1.);
		}
		else setZValue(0.);
		parentItem->setSelected(value.toBool());
	}
	return QGraphicsItem::itemChange(change, value);
}

void InputItem::dragEnterEvent(QGraphicsSceneDragDropEvent * e)
{
	e->accept();
	QGraphicsItem::dragEnterEvent(e);
}

void InputItem::dragLeaveEvent(QGraphicsSceneDragDropEvent * e)
{
	e->accept();
	QGraphicsItem::dragLeaveEvent(e);
}

void InputItem::dropEvent(QGraphicsSceneDragDropEvent* e)
{
	e->accept();
	QGraphicsItem::dropEvent(e);
	setValue(e->mimeData()->text());
} 

void InputItem::paint(QPainter* painter, const QStyleOptionGraphicsItem* option, QWidget* widget)
{
	CalcItem::paint(painter, option, widget);
}

void InputItem::setValue()
{
    setValue(edit->text());
}

void InputItem::setValue(const QString& text)
{
	QdObject_var obj=QuGetObjByNameAndComp(text);
	if(!obj.is_nil())
	{
		QdTimeIntervalSeq_var TI = QdTimeIntervalSeq_var::narrow(obj);
		if( !TI.is_nil()) emit newInput(QString("scalar"));
		else if(obj->sequence_size() > 1) emit newInput(text);
		else emit newInput(QString("scalar"));
		
		parentItem->setValue(obj, text);
	}
	else
	{
		bool ok;
		QdRScalar_var scalar_var=new QdRScalar(text.toDouble(&ok));
		if(ok)
		{
			emit newInput(QString("scalar"));
			parentItem->setValue(scalar_var, text);
		}
		else
		{
			emit newInput(QString("scalar"));
			parentItem->setValue(new QdString(text.toLatin1().data()), text);
		}
	}
	
	if(edit->text()!=text)
	{
		edit->setText(text);
		edit->reValidate();
	}
}

QDomElement InputItem::toXMLElement(QDomDocument& doc) const
{
	return CalcItem::toXMLElement(doc);
}

void InputItem::fromXMLElement(QDomElement& el)
{
	CalcItem::fromXMLElement(el);
}
