#include "qcalcJoinItem.h"
#include "qcalcSlots.h"
#include "qcalcInputItem.h"
#include "qcalcOutputItem.h"

using namespace QSAS;


VJoinItem::VJoinItem():
		VCalcItem()
{
	join_id=counter++;
	gui=NULL;
	VInSlot* slot=new VInSlot(this);
	inputs.push_back(slot);
	VOutSlot* oslot=new VOutSlot(this);
	outputs.push_back(oslot);
#ifdef GRAPH_DEBUG
	std::cout << "Creating VJoinItem " << this << std::endl;
#endif

	joinSettings=new JoinSettings;
    joinSettings->noJoin=true;
	joinSettings->algorithm=0;
	joinSettings->width=60.;
	joinSettings->gap=0;
	joinSettings->gapSize=6.;
	
	name="Join";
}

VJoinItem::~VJoinItem()
{
#ifdef GRAPH_DEBUG
	std::cout << "Deleting VJoinItem " << this << std::endl;
#endif
	
	if(gui!=NULL)
	{
		QGraphicsScene * scene=gui->scene();
		if(scene!=NULL)
		{
			scene->removeItem(gui);
		}
		gui->parentItem=NULL; // thus gui knows it is being deleted

		delete gui;
	}
	
	delete joinSettings;
}

QDomElement VJoinItem::toXMLElement(QDomDocument& document) const
{
	QDomElement tag=document.createElement("Join");
    if(gui!=NULL)
    {
		tag.appendChild(gui->toXMLElement(document));
    }

	tag.appendChild(createTextXMLNode(document, "XPosnInBlackBox", positionInBlackBox.x()));
	tag.appendChild(createTextXMLNode(document, "YPosnInBlackBox", positionInBlackBox.y()));
	tag.appendChild(createTextXMLNode(document, "ID", calc_id));
	tag.appendChild(createTextXMLNode(document, "Page", page));
	tag.appendChild(createTextXMLNode(document, "Join_No_Join", joinSettings->noJoin));
	tag.appendChild(createTextXMLNode(document, "Join_Algorithm", joinSettings->algorithm));
	tag.appendChild(createTextXMLNode(document, "Join_Width", joinSettings->width));
	tag.appendChild(createTextXMLNode(document, "Join_Gap", joinSettings->gap));
	tag.appendChild(createTextXMLNode(document, "Join_Gap_Size", joinSettings->gapSize));
	
		QDomElement inputSlots=document.createElement("Input_Slots");
	tag.appendChild(inputSlots);
	for(int i=0; i<inputs.size(); ++i)
	{
		QDomElement s=document.createElement("Inslot");
		s.setAttribute("Index", i);
		s.setAttribute("Name", inputs[i]->getName());
		s.setAttribute("Description", inputs[i]->getDescription());
		inputSlots.appendChild(s);
	}
	QDomElement outputSlots=document.createElement("Output_Slots");
	tag.appendChild(outputSlots);
	for(int i=0; i<outputs.size(); ++i)
	{
		QDomElement s=document.createElement("Outslot");
		s.setAttribute("Index", i);
		s.setAttribute("Name", outputs[i]->getName());
		s.setAttribute("Description", outputs[i]->getDescription());
		s.setAttribute("open", (int) outputs[i]->openSlot);
		outputSlots.appendChild(s);
	}

    return tag;
}

JoinItem* VJoinItem::getGUI()
{
	if(gui==NULL)
	{
		gui=new JoinItem(this, inputs[0], outputs[0]);
	}
	return gui;
}

void VJoinItem::removeGUI()
{
	if(gui!=NULL)
	{
		gui->scene()->removeItem(gui);
		delete gui;
		gui=NULL;
	}
}

void VJoinItem::setErrMessage(bool err, const QString& message)
{
	if(gui) gui->setErrMessage(err, message);
}

void VJoinItem::reevaluateOutSlots()
{
	bool ok=true;
	QdObject_var value;
	QdObject_var not_joined_value=inputs[0]->connectedSlot()->getValue(ok);
	
	if(!ok)
	{
		setErrMessage(true, "Input invalid");
		outputs[0]->setValue(QdObject_var(), false);
	}
	else setErrMessage(false, "");

	
	if(joinSettings->noJoin)
    {
        value=not_joined_value;
    }
    else
    {
        if(!not_joined_value->is_sequence() || joinSettings->noJoin)
        {
            value=not_joined_value;
        }
        else
        {
 
			string gap_opt_ptr;
			switch(joinSettings->gap)
			{
				case 1:
					gap_opt_ptr = String(ZERO_FILL_GAP);
					break;
				case 2:
					gap_opt_ptr = String(FILL_GAP);
					break;
				case 3:
					gap_opt_ptr = String(NEAREST_NEIGHBOUR_GAP);
					break;
				case 4:
					gap_opt_ptr = String(REMOVE_GAP);
					break;
				case 0:
				default:
					gap_opt_ptr = String(LINEAR_INTERP_GAP);
					break;
			}
				
			if(joinSettings->algorithm == 0){
				value = QuLinearJoin(not_joined_value, tags, joinSettings->gapSize, gap_opt_ptr);
			}
			else {
				value = QuBoxcarJoin(not_joined_value, tags, joinSettings->width, gap_opt_ptr);					 
			}				
								
            if(value.is_nil()) {
                outputs[0]->setValue(value, false);
                return;
            }
        }
    }
    
	outputs[0]->setValue(value, true);
	
}

bool VJoinItem::hasGui()
{
	return gui!=NULL;
}


void VJoinItem::fromXMLElement(QDomElement& el, QGraphicsScene* scene)
{
	VCalcItem::fromXMLElement(el, scene);

	QDomElement guiElement=el.firstChildElement("GUI");
	QString val;
	
	if(getXMLNodeText(el, "XPosnInBlackBox", val))
	{
		positionInBlackBox.setX(val.toDouble());
	}
	if(getXMLNodeText(el, "YPosnInBlackBox", val))
	{
		positionInBlackBox.setY(val.toDouble());
	}

	if(getXMLNodeText(el, "Page", val))
	{
		page=val.toInt();
	}
	if(getXMLNodeText(el, "Join_No_Join", val))
	{
		joinSettings->noJoin=val.toInt();
	}
	if(getXMLNodeText(el, "Join_Algorithm", val))
	{
		joinSettings->algorithm=val.toInt();
	}
	if(getXMLNodeText(el, "Join_Width", val))
	{
		joinSettings->width=val.toDouble();
	}
	if(getXMLNodeText(el, "Join_Gap", val))
	{
		joinSettings->gap=val.toInt();
	}
	if(getXMLNodeText(el, "Join_Gap_Size", val))
	{
		joinSettings->gapSize=val.toDouble();
	}
	if(!guiElement.isNull())
	{
		JoinItem* item=this->getGUI();
        scene->addItem(item);
        item->fromXMLElement(guiElement);
	}

	QDomElement inputsElement=el.firstChildElement("Input_Slots");
	if(!inputsElement.isNull())
	{
		QDomElement inputSlot=inputsElement.firstChildElement("Inslot");
		while(!inputSlot.isNull())
		{
			int index=inputSlot.attribute("Index", "-1").toInt();
			QString name=inputSlot.attribute("Name", "");
			QString description=inputSlot.attribute("Description", "");
			if(index>=0 && index< inputs.size())
			{
				inputs[index]->setName(name);
				inputs[index]->setDescription(description);
			}
			inputSlot=inputSlot.nextSiblingElement("Inslot");
		}
	}
	
	QDomElement outputsElement=el.firstChildElement("Output_Slots");
	if(!outputsElement.isNull())
	{
		QDomElement outputSlot=outputsElement.firstChildElement("Outslot");
		while(!outputSlot.isNull())
		{
			int index=outputSlot.attribute("Index", "-1").toInt();
			QString name=outputSlot.attribute("Name", "");
			QString description=outputSlot.attribute("Description", "");
			int isOpenSlot=outputSlot.attribute("open", "0").toInt();
			if(index>=0 && index< outputs.size())
			{
				outputs[index]->setName(name);
				outputs[index]->setDescription(description);
				outputs[index]->openSlot=(bool)isOpenSlot;
			}
			outputSlot=outputSlot.nextSiblingElement("Outslot");
		}
	}

}


JoinItem::JoinItem(VJoinItem* parentItem, VInSlot* input, VOutSlot* output, QGraphicsItem* parent):
		CalcItem(parent)
{
#ifdef GRAPH_DEBUG
	std::cout << "Creating JoinItem " << this << " for VJoinItem " << parentItem << std::endl;
#endif
	this->parentItem=parentItem;
	
	outlineColor=QColor(50, 50, 70);
	color1=QColor(220, 220, 255);
	color2=QColor(200, 200, 240);

	
	title=parentItem->getName();
	
	width=BASE_WIDTH;
	height=BASE_HEIGHT;
	setRect(-width/2., -height/2., width, height);
	
	InSlot* islot=new InSlot(this, input);
	input->setGUI(islot);
	input->resetToolTips();
	islot->setPos(0., -height/2.);
	inSlots.push_back(islot);

	OutSlot* oslot=new OutSlot(this, output);
	output->setGUI(oslot);
	output->resetToolTips();
	oslot->setPos(0., height/2.);
	outSlots.push_back(oslot);

	rzButton->setPos(-width/2.+1.5*HEADER_HEIGHT, -height/2.+HEADER_HEIGHT/2.);
	closeButton->setPos(width/2.-HEADER_HEIGHT/2., -height/2.+HEADER_HEIGHT/2.);
	errButton->setPos(width/2.-1.5*HEADER_HEIGHT, -height/2.+HEADER_HEIGHT/2.);
}

JoinItem::~JoinItem()
{
#ifdef GRAPH_DEBUG
	std::cout << "Deleting JoinItem " << this << std::endl;
#endif
	foreach(InSlot* slot, inSlots)
	{
		delete slot;
	}
	foreach(OutSlot* slot, outSlots)
	{
		delete slot;
	}
}

QVariant JoinItem::itemChange(GraphicsItemChange change, const QVariant& value)
{
	if (change == ItemPositionChange && scene()) {
		QPointF newPos = value.toPointF();
		
		updateConnectorsPositions();
		
		return newPos;
	}
	else if(change==ItemSelectedChange && scene())
	{
		if(value.toBool())
		{
			setZValue(1.);
		}
		else setZValue(0.);
		parentItem->setSelected(value.toBool());
	}
	return QGraphicsItem::itemChange(change, value);
}

void JoinItem::paint(QPainter* painter, const QStyleOptionGraphicsItem* option, QWidget* widget)
{
	CalcItem::paint(painter, option, widget);
}


QDomElement JoinItem::toXMLElement(QDomDocument& doc) const
{
	QDomElement el=CalcItem::toXMLElement(doc);
	return el;
}

void JoinItem::fromXMLElement(QDomElement& el)
{
	CalcItem::fromXMLElement(el);

}

