#include "qcalcScene.h"
#include "qcalcOutputItem.h"
#include "qcalcInputItem.h"
#include "qcalcJoinItem.h"
#include <QGraphicsProxyWidget>
#include "qcalcTimeLine.h"

extern TimeLineHandler timeLineHandler;

VOutputItem::VOutputItem():
		VCalcItem()
{
	output_id=counter++;
	VInSlot* slot=new VInSlot(this);
	inputs.push_back(slot);
	gui=NULL;
	name="Output";
	outputName=QString("result_%1").arg(output_id);
#ifdef GRAPH_DEBUG
	std::cout << "Creating VOutputItem" << this << std::endl;
#endif

	timeLineName="Default";
}

VOutputItem::~VOutputItem()
{
#ifdef GRAPH_DEBUG
	std::cout << "Deleting VOutputItem " << this << std::endl;
#endif
	if(gui!=NULL)
	{
		QGraphicsScene * scene=gui->scene();
		if(scene!=NULL)
		{
			scene->removeItem(gui);
		}
		gui->parentItem=NULL; // thus gui knows it is being deleted
		
		delete gui;
	}

}

QDomElement VOutputItem::toXMLElement(QDomDocument& document) const
{
	QDomElement tag=document.createElement("Output");
    if(gui!=NULL)
    {
		tag.appendChild(gui->toXMLElement(document));
    }

	tag.appendChild(createTextXMLNode(document, "XPosnInBlackBox", positionInBlackBox.x()));
	tag.appendChild(createTextXMLNode(document, "YPosnInBlackBox", positionInBlackBox.y()));

	tag.appendChild(createTextXMLNode(document, "ID", calc_id));
	tag.appendChild(createTextXMLNode(document, "Output_Name", outputName));
	tag.appendChild(createTextXMLNode(document, "Page", page));
	
	QDomElement tlElement=document.createElement("Time_Line");
	tag.appendChild(tlElement);
	TimeLine* tl=timeLineHandler.getTimeLine(timeLineName);
	if(tl==NULL)
	{
		tl=timeLineHandler.getDefaultTimeLine();
	}
	tlElement.appendChild(createTextXMLNode(document, "Name", timeLineName));
	tlElement.appendChild(createTextXMLNode(document, "From_Data", tl->fromData));
	tlElement.appendChild(createTextXMLNode(document, "Regular_Time_Tags", tl->regularTT));
	tlElement.appendChild(createTextXMLNode(document, "From_Obj", tl->fromObjString));
	tlElement.appendChild(createTextXMLNode(document, "Time_Tags", tl->TTString));
	
    
    return tag;
}

void VOutputItem::reevaluateOutSlots()
{
}

void VOutputItem::computeValue()
{
	VOutSlot* slot=inputs[0]->connectedSlot();
//	std::cout << QString(outputName+": ").toLatin1().data();
	if(slot==NULL)
	{
		if(gui!=NULL) gui->setErrMessage(true, "No connection");
		return;
	}

	// Now get the time line, and spread it to all the connected input and join items
    
    QdObject_var tags;

	tags=getDep0Tags();

	QLinkedList<VInputItem*> inputs=parentInputs();

    foreach(VInputItem* input, inputs)
    {
		input->setJoinTags(tags);
    }
	
	QLinkedList<VJoinItem*> joins=parentJoins();

    foreach(VJoinItem* join, joins)
    {
		join->setJoinTags(tags);
    }
  
	// compute
	bool validity;
	QdObject_var res;

	res=slot->getValue(validity);
	if(validity && !outputName.isEmpty())
	{
		QwuiPutOnWL(res, QString("Data/")+QString(outputName));
		if(gui!=NULL) gui->setErrMessage(false);
	}
	else
	{
		if(gui!=NULL) gui->setErrMessage(true, "Error when evaluating");
	}
}

OutputItem* VOutputItem::getGUI()
{
	if(gui==NULL)
	{
		gui=new OutputItem(this, inputs[0]);
	}
	return gui;
}

void VOutputItem::removeGUI()
{
	if(gui!=NULL)
	{
		gui->scene()->removeItem(gui);
		delete gui;
		gui=NULL;
	}
}

QdObject_var VOutputItem::getDep0Tags()
{
	TimeLine* tl=timeLineHandler.getTimeLine(timeLineName);

	if(tl==NULL)
		tl=timeLineHandler.getDefaultTimeLine();
	
    QdObject_var tags; 
	
    QdTimeInterval_var ivl;
    double Spacing;
    double Interval;

  // set interval

    ivl = getInterval();

    if ( ivl.is_nil() ) {
	
		// Try Scalar series for DEPEND_0
		
		double start, end;
		bool OK = getScalarInterval(start, end);
		if(!OK) return tags;   // will be .is_nil()
		
		Interval = end - start;
		QdRScalarSeq_var stags;
		
		if( tl->regularTT)
    	{
    		// create regular tags
        	Spacing = tl->TTString.toDouble(&OK);
        	if(!OK || fabs(Spacing) < 1.e-20) return tags;

			if(Interval < 0) Spacing = -fabs(Spacing); // handle decreasing sequence
			
        	int Ntags = (int) (fabs(Interval/Spacing) + 1.0);

			stags = new QdRScalarSeq();
        	stags->assign_regular(Ntags, start, Spacing );
			tags = stags;

    	}
    	else
    	{
    		// tags from object
        	QdObject_var tag_obj = QuGetObjByName(tl->TTString);
        	QdObject_var D0=get_Depend_0(tag_obj);
			QdRScalarSeq_var ss = QdRScalarSeq_var::narrow(D0);
        	if(ss.is_nil())  return tags;    // will be .is_nil()

    		// truncate to interval if necessary

    		if ( start < end && (ss[0] < start || ss[ss->sequence_size()-1] > end) ) {

      			SequenceMask msk(ss->get_mask(QdRScalarSeq::InRange(start,end)));
      			stags = new QdRScalarSeq();
      			ss->get_subsequence(*stags, msk);
    		}
    		else if ( start > end && (ss[0] > start || ss[ss->sequence_size()-1] < end) ) {

      			SequenceMask msk(ss->get_mask(QdRScalarSeq::InRange(start,end)));
      			stags = new QdRScalarSeq();
      			ss->get_subsequence(*stags, msk);
    		}
    		else stags = new QdRScalarSeq(*ss);
			tags = stags;
		}
	}
    else
	{
		//  DEPEND_0 is Time tags 
		
		Interval = ivl->duration_abs();

  		// set tags

		if( tl->regularTT)
    	{
    		// create regular tags
			bool OK;
        	Spacing = tl->TTString.toDouble(&OK);
        	if(!OK || Spacing < 1.e-20) return tags;

        	int Ntags = (int) (Interval/Spacing + 1.0);

        	QdTimeSeq_var ttags = new QdTimeSeq();
        	ttags->assign_regular(Ntags, ivl->start(), Spacing );
			tags = ttags;

    	}
    	else
    	{
    		// tags from object
        	QdObject_var tag_obj = QuGetObjByName(baseObjectText(tl->TTString));
        	QdTimeSeq_var ts=get_timetags(tag_obj);

        	if(ts.is_nil())  return tags;    // will be .is_nil()

    		// truncate to interval if necessary

        	if (( ivl->start() != ts->get_interval_bounds().start()) ||
                ( ivl->end()   != ts->get_interval_bounds().end())) {

            	SequenceMask msk(ts->get_mask(QdTimeSeq::InRange(ivl->start(),ivl->end())));
            	QdTimeSeq_var ttags = new QdTimeSeq();
            	ts->get_subsequence(*ttags, msk);
				tags = ttags;
            }
			else {
				QdTimeSeq_var ttags = new QdTimeSeq(*ts);
				tags = ttags;
			}
		}
    }

    return tags;
}

QdTimeInterval_var VOutputItem::getInterval()
{
    QdTimeInterval_var ivl;
	TimeLine* tl=timeLineHandler.getTimeLine(timeLineName);
	if(tl==NULL)
		tl=timeLineHandler.getDefaultTimeLine();
	
    if(tl->fromData)
    {
    // interval from data
        ivl = getIntervalFromData();
    }
    else
    {
    // interval from object
        QdObject_var ivl_obj = QuGetObjByName(tl->fromObjString);
        if ( !ivl_obj.is_nil() ) {
            ivl = QdTimeInterval_var::narrow(ivl_obj);
            if(ivl.is_nil()){
                QdTimeSeq_var ts=get_timetags(ivl_obj);
                if(!ts.is_nil())
                    ivl=new QdTimeInterval(ts->get_interval_bounds());
            }
        }

    }
    return ivl; 
}

bool VOutputItem::getScalarInterval(double &start, double &end)
{	
	TimeLine* tl=timeLineHandler.getTimeLine(timeLineName);
	if(tl==NULL)
		tl=timeLineHandler.getDefaultTimeLine();
	
    if(tl->fromData)
    {
    	// interval from data
        return getScalarIntervalFromData(start, end);
    }
    else
    {
    	// interval from object
        QdObject_var ivl_obj = QuGetObjByName(tl->fromObjString);
		return get_ScalarInterval(ivl_obj, start, end );
    }
	
    return false; 
}

QLinkedList<VInputItem*> VOutputItem::parentInputs()
{
    QLinkedList<VInputItem*> res;

    foreach(VCalcItem* item, CalcScene::currentItems)
    {

        VInputItem* in=dynamic_cast<VInputItem*>(item);
        if(in!=NULL)
        {
            if(in->isParentOf(this))
            {
                res.append(in);
            }
        }
    }

    return res;
}

QLinkedList<VJoinItem*> VOutputItem::parentJoins()
{
    QLinkedList<VJoinItem*> res;

    foreach(VCalcItem* item, CalcScene::currentItems)
    {

        VJoinItem* in=dynamic_cast<VJoinItem*>(item);
        if(in!=NULL)
        {
            if(in->isParentOf(this))
            {
                res.append(in);
            }
        }
    }

    return res;
}

QdTimeInterval_var VOutputItem::getIntervalFromData()
{
    QdTimeInterval_var ivl;
    QdTimeInterval_var ivlIntersect;

    QLinkedList<VInputItem*> inputs=parentInputs();
    foreach(VInputItem* input, inputs)
    {
        QdObject_var ivl_obj = input->getNotJoinedValue();

        if ( !ivl_obj.is_nil())
        {
            ivl = QdTimeInterval_var::narrow(ivl_obj);
            if(ivl.is_nil())
            {
                QdTimeSeq_var ts=get_timetags(ivl_obj);
                if(!ts.is_nil())
                    ivl=new QdTimeInterval(ts->get_interval_bounds());
            }
        }

        if(ivlIntersect.is_nil() ) ivlIntersect = ivl;
        else if ( !ivl.is_nil())
        {
            QdTimeInterval_var ivlNew = new QdTimeInterval(ivlIntersect->int_intersect(*ivl));
            ivlIntersect = ivlNew;
        }
    }
  
    return ivlIntersect;
}

bool VOutputItem::getScalarIntervalFromData(double &start, double &end)
{
	bool OK = false;
	double obj_start, obj_end;
	
    QLinkedList<VInputItem*> inputs=parentInputs();
    foreach(VInputItem* input, inputs)
    {
        QdObject_var ivl_obj = input->getNotJoinedValue();

		if( get_ScalarInterval(ivl_obj, obj_start, obj_end ) )
		{		
			if(OK){	
				if(start < end){
					if(start > obj_start) start = obj_start;
					if(end < obj_end) end = obj_end;
				}
				else{
					if(start < obj_start) start = obj_start;
					if(end > obj_end) end = obj_end;
				}	
			}
			else{
				// first valid data	
				start = obj_start;
				end = obj_end;
			}

			OK = true;
		}

    }
  
    return OK;
}

void VOutputItem::fromXMLElement(QDomElement& el, QGraphicsScene* scene)
{
	VCalcItem::fromXMLElement(el, scene);


	QString outString="";
	getXMLNodeText(el, "Output_Name", outString);
	QDomElement guiElement=el.firstChildElement("GUI");
	if(!guiElement.isNull())
	{
		OutputItem* item=this->getGUI();
        scene->addItem(item);
        item->fromXMLElement(guiElement);
		item->setValue(outString);
	}
	else
	{
		setOutputName(outString);
	}

	QString val;
	
	if(getXMLNodeText(el, "XPosnInBlackBox", val))
	{
		positionInBlackBox.setX(val.toDouble());
	}
	if(getXMLNodeText(el, "YPosnInBlackBox", val))
	{
		positionInBlackBox.setY(val.toDouble());
	}

	if(getXMLNodeText(el, "Page", val))
	{
		page=val.toInt();
	}
	
	QDomElement tlElement=el.firstChildElement("Time_Line");
	if(!tlElement.isNull())
	{
		bool fromData=true;
		bool regularTT=true;
		QString fromObjString="";
		QString TTString="";
		QString name="Default";
		if(getXMLNodeText(tlElement, "Name", val))
		{
			name=val;
		}
		if(getXMLNodeText(tlElement, "From_Data", val))
		{
			fromData=val.toInt();
		}
		if(getXMLNodeText(tlElement, "Regular_Time_Tags", val))
		{
			regularTT=val.toInt();
		}
		if(getXMLNodeText(tlElement, "From_Obj", val))
		{
			fromObjString=val;
		}
		if(getXMLNodeText(tlElement, "Time_Tags", val))
		{
			TTString=val;
		}
		
		QString existingName=timeLineHandler.getTimeLine(fromData, regularTT, fromObjString, TTString, name);
		if(!existingName.isEmpty())
		{
			timeLineName=existingName;
		}
		else
		{
			if(existingName==name) // no problem, the time line already exists
			{
				timeLineName=name;
			}
			else
			{
				QList<QString> names=timeLineHandler.getNames();
				if(names.contains(name)) // ouch, we already have a time line by this name, and it is not the same
				{
					while(names.contains(name))
						name.append("A");
				}

				timeLineName=name;
				timeLineHandler.addTimeLine(name, fromData, regularTT, fromObjString, TTString);
				qApp->processEvents();
				
			}
		} 
	}
	else
	{
		timeLineName="Default";
	}
}



OutputItem::OutputItem(VOutputItem* parentItem, VInSlot* input, QGraphicsItem* parent):
		CalcItem(parent)
{
	this->parentItem=parentItem;
	
	outlineColor=QColor(50, 75, 100);
	color1=QColor(150, 200, 255);
	color2=QColor(120, 170, 220);
	
	title=parentItem->getName();
	
	width=BASE_WIDTH;
	height=BASE_HEIGHT;
	setRect(-width/2., -height/2., width, height);
	
	edit=new QLineEdit;
	edit->setMaximumWidth((int)(BASE_WIDTH-6));
	edit->setMinimumWidth((int)(BASE_WIDTH-6));
	edit->setMaximumHeight((int)(HEADER_HEIGHT-2));
	edit->setMinimumHeight((int)(HEADER_HEIGHT-2));
	
	QGraphicsProxyWidget * editProxy=new QGraphicsProxyWidget(this);
	editProxy->setWidget(edit);
	editProxy->moveBy(-edit->width()/2., 0.);
	
	InSlot* slot=new InSlot(this, input);
	slot->setPos(0., -height/2.);
	inSlots.push_back(slot);
	
	
	rzButton->setPos(-width/2.+1.5*HEADER_HEIGHT, -height/2.+HEADER_HEIGHT/2.);
	closeButton->setPos(width/2.-HEADER_HEIGHT/2., -height/2.+HEADER_HEIGHT/2.);
	errButton->setPos(width/2.-1.5*HEADER_HEIGHT, -height/2.+HEADER_HEIGHT/2.);
	
	connect(edit, SIGNAL(textChanged(const QString&)), this, SLOT(setValue()));
	
	setAcceptDrops(false);
	edit->setAcceptDrops(false);
	
	edit->setText(parentItem->outputName);
	
#ifdef GRAPH_DEBUG
	std::cout << "Creating OutputItem " << this << " for VOutputItem "<< parentItem << std::endl;
#endif
}

OutputItem::~OutputItem()
{
#ifdef GRAPH_DEBUG
	std::cout << "Deleting OutputItem " << this << std::endl;
#endif
	foreach(InSlot* slot, inSlots)
	{
		delete slot;
	}
	foreach(OutSlot* slot, outSlots)
	{
		delete slot;
	}
}

QVariant OutputItem::itemChange(GraphicsItemChange change, const QVariant& value)
{
	if (change == ItemPositionChange && scene()) {
		QPointF newPos = value.toPointF();
		
		updateConnectorsPositions();
		
		return newPos;
	}
	else if(change==ItemSelectedChange && scene())
	{
		if(value.toBool())
		{
			setZValue(1.);
		}
		else setZValue(0.);
		parentItem->setSelected(value.toBool());
	}
	return QGraphicsItem::itemChange(change, value);
}

void OutputItem::paint(QPainter* painter, const QStyleOptionGraphicsItem* option, QWidget* widget)
{
	CalcItem::paint(painter, option, widget);
}

void OutputItem::setValue()
{
	parentItem->setOutputName(edit->text());
}

void OutputItem::setValue(const QString& text)
{
	edit->setText(text);
	// will emit text changed which calls setValue() above
}

void OutputItem::updateConnectorsPositions()
{
	CalcItem::updateConnectorsPositions();
}

QDomElement OutputItem::toXMLElement(QDomDocument& doc) const
{
	return CalcItem::toXMLElement(doc);
}

void OutputItem::fromXMLElement(QDomElement& el)
{
	CalcItem::fromXMLElement(el);
}




