/***********************************************************************
* SCCS id: @(#)qbowshock.h	1.1 18:23:45 04/29/99  
************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT:    Bow Shock Models header file
* 
* MODULE:       Qbowshock.h
*
* LANGUAGE:     C
*
* PURPOSE:
*
* DESCRIPTION: Header for Qbowshock.c
* Routines to calculate model bow shock (and magnetopause) surfaces, normals
* based on ISSI Methods for analysing data from multi-satellites book Ch10
* section 4.6
*
* FUNCTIONS:
*
* MODULES_USED:
*
* NOTES:
*
* COPYRIGHT:
* ---------------------------------------------------------------------
* Copyright (c) 1999 Steve Schwartz
*
* Astronomy Unit, Queen Mary and Westfield College, London, UK.
* This software is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
* If you are associated with the Cluster project, you may copy, or
*   modify this software freely, solely for purposes associated with
*   the Cluster project, and provided that you maintain a complete
*   copy of this copyright message with the software.
* If this software is distributed to you at the same time as a separate
*   file called COPYRIGHT, then the terms of copyright listed in that
*   file apply to this software.
* ---------------------------------------------------------------------
*
*
* HISTORY: SJS April 1999; up-dated July 2001
*                          also changed argument of QBSChooseDistanceRoot
*                          to XX to avoid X clash on solaris
*
* Jeremy Mitchell, Feb 2010
* Introduce QBSPosDistanceAlongVec, QBSCalcPosDistAlongVec and
* QBSChooseAlongPosVecRoot. Find the bowshock by following Vec in the
* positive direction only.
* Add NO_ROOT_IN_POS_VEC_DIRECTION flag
*
***********************************************************************/
#ifndef _QBS__  
#define _QBS__

/* c++ compliance... */

  #ifdef   __cplusplus
  extern "C" {
  #endif


/********************************************* BS and MP MODELS  ****/
/* properties of all BS and MP models in ISSI chapter. This is not the
*  most elegant way to hold this information, and modifications should
*  ensure that order, etc. is preserved */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>

#define N_MODELS 14

typedef enum QSBSMP_models{
  BS_PEREDO,
  BS_SLAVIN_HOLZER,
  BS_FAIRFIELD4DEG,
  BS_FAIRFIELDNO4DEG,
  BS_FORMISANO,
  BS_FARRIS,
  MP_ROELOF_SIBECK,
  MP_FAIRFIELD4DEG,
  MP_FAIRFIELDNO4DEG,
  MP_FARRIS,
  MP_PETRINECBZPOS,
  MP_PETRINECBZNEG,
  MP_FORMISANOUNNORM,
  MP_FORMISANONORM
} QBSMP_models;

static const char* BSMP_name[] = { 
 "BS: Peredo",
 "BS: Slavin & Holzer",
 "BS: Fairfield 4 deg",
 "BS: Fairfield no 4 deg",
 "BS: Formisano",
 "BS: Farris et al.",
 "MP: Roelof & Sibeck",
 "MP: Fairfield 4 deg",
 "MP: Fairfield no 4 deg",
 "MP: Farris et al.",
 "MP: Petrinec et al. Bz>0",
 "MP: Petrinec et al. Bz<0",
 "MP: Formisano Unnorm.",
 "MP: Formisano Norm."
};

static const float BSMP_eps[N_MODELS] = 
  { 0.98, 1.16, 1.02, 1.05, 0.97, 0.81,    /* BS epsilons */
    0.91, 0.79, 0.80, 0.43, 0.42, 0.50, 0.82, 0.69  /* MP epsilons */
  };
static const float BSMP_L[N_MODELS] =
  { 26.1, 23.3, 22.3, 20.5, 22.8, 24.8,
    11.2, 13.1, 12.8, 14.7, 14.6, 14.6, 12.5, 13.5
  };
static const float BSMP_xo[N_MODELS] =
  {  2.0,  3.0,  3.4,  4.6,  2.6,  0.0,
    4.82,  3.6,  3.9,  0.0,  0.0,  0.0,  4.1,  0.9
  };
static const float BSMP_yo[N_MODELS] =
  {  0.3,  0.0,  0.3,  0.4,  1.1,  0.0,
     0.0,  0.4,  0.6,  0.0,  0.0,  0.0,  0.1, -0.4
  }; 

/*************************************************** end BS and MP Models*/


#define V_EARTH   29.7854 /* Earth orbittal velocity in km/s  */
#define VSW_MIN    1.0    /* If |vsw| set less than VSW_MIN use VSW_DEFAULT */
#define VSW_DEFAULT 450.0
#define QUARTIC_EPS 1.e-6
#define N_MAX_ITER 500
#define ABSOLUTE_ROOT_ACCURACY 1.e-6

typedef enum QSBSMP_IN_OR_OUT{
   INSIDE,
   OUTSIDE,
   ON_SURFACE,        /* most unlikely a test for equality will succeed!*/
   CONFUSED
} QBSMP_IN_OR_OUT;


/* function prototypes */

void QBSAberrate( float* vec_gse,              /* input vec in gse xyz */
                  float vsw,                   /* sw speed in km/s     */
                  float *vec_ab) ;             /* vec in aberrated frame*/

void QBSUnAberrate( float* vec,              /* input vec in gse xyz */
                    float vsw,               /* sw speed in km/s     */
                    float *vec_unab) ;       /* vec in aberrated frame*/


void QBSBuildRotateXY( float angle, /* angle in RADIANS: + in RH sense wrt z*/
                       float rot[3][3] ); /* rotation matrix rot[row][col] */

void QBSRotate( float* vec,            /* input vector */
                float rot[3][3],            /* rotation matrix */
                float* vec_rot);        /* rotated vector */

void QBSDisplace( float* vec,
                  float* ro_vec,
                  float* rdispl_vec ) ;    /* rdispl_vec = vec - ro_vec */

float QBSVectorMagnitude (float *vec) ;

float QBSVectorScalarProduct(float *vec1,
                             float *vec2) ;

float QBSurface( float* r_vec,     /* position, already aberrated if req'd */
               QBSMP_models model, /* which model to use */
               float sigma );      /* scaling for L, ro; 1 means no scaling */

QBSMP_IN_OR_OUT QBSInside_Outside(  
           float*        r_vec,   /* pos (pre-aberrated) */
           QBSMP_models  model,   /* which model to use */
           float         sigma ); /* scaling for L, ro; 1 means no scaling */

int QBSScaleToXing (float* r_vec,         /* position, pre-aberrated if req'd */
                   QBSMP_models  model,   /* which model to use */
                   float* sigma );        /* returned scaling for L, ro; */
                         /* Multiply L,ro by sigma to make model fit point */
int QBSNormalAtSurface ( 
                   float* r_vec,         /* position, pre-aberrated if req'd */
                   QBSMP_models  model,  /* which model to use */
                   float* normal);       /* same system as r_vec */

int QBSNormalAtSurfaceGSE( float* r_vec,   /* Position in Re_GSE */
                           float  vsw,     /* Vsw (x) in km/s */
                           QBSMP_models nmodel, /* model to use */
                           float* normal_vec); /*returned outward GSE normal */

int QBSDistanceAlongNormal(float* r_vec,   /* Position in Re_GSE */
                           float  vsw,     /* Vsw (x) in km/s */
                           QBSMP_models model, /* model to use */
                           float  sigma,    /* mult scaling to apply to model*/
                           float* distance,/*returned signed distance along n */
                           float* r_at_shock);/* returned GSE pos at shock */

int QBSCalcNormalDist( float* R,          /* position (pre-aber and displ)*/
                       float eps,         /* model eps and L - pre-scaled */
                       float L, 
                       float* distance,   /*signed distance along normal in Re*/
                       float* r_at_shock); /* intersect n,shock; in frame of R*/

int QBSPositionAtShockRoot(float eps,  /* model parameters */
                    float L,    /* NB can pass L=1 and R normalised to L */
                    float *R,   /* position in question (pre-aber, displ)*/
                    float t,    /* root of quartic */
                    float* r);   /*returned pos vector; same units/frame as R*/

int QBSChooseDistanceRoot(float  XX,      /* On-axis position, normalised by L */
                          float  rho,    /* Dist to axis, normalised by L */
                          float  eps,    /* eccentricity */
                          float* roots,  /* roots t of r + tN = R vec eqn */
                                         /* (only real ones; others garbage */
                          int    nreal,  /* number of real roots */
                          float* chosen_root); /* chosen root */

int QBSSolveQuartic(double* c,     /* coefficients of quartic; c[0] const term*/
                    double* roots, /*sorted array of returned real roots */
                    int*   nreal); /* number of real roots returned */

int QBSBracketQuarticRoots(double* c,/*coefficients of quartic;c[0] const term*/
                    double* t_low, /*   low and high brackets*/
                    double* t_high );

void poly( double  t,      /* value of argument */
              int    degree, /* highest pwr of t inpolynomial */
              double* c,      /* coefficients, c[0] constant term */ 
              double* f,      /* evaluated polynomial */
              double* df );    /* evaluated derivative */

int rtsafe_poly( double*  c,     /* polynomial coeffs. c[0] constant term */
                 int  deg,   /* highest power of x */
                 double  x1,    /* lower bound for bracketed root */
                 double  x2,    /* upper bound for bracked root */
                 double  xacc,  /* absolute accuracy for root */
                 double* root );       /* found root */

int QBSSolveCubic ( double* c3,   /* coeffs of cubic; c[0] = constant term */
                    double* roots, /* returned real roots - sorted*/
                    int*   nreal);/* no of real roots; complex roots NOT retnd*/

int QBSDoubleCompare (const void* el1, const void* el2);/* double comp 4 qsort*/

int QBSDistanceAlongVec(float* r_vec,   /* Position in Re_GSE */
                           float  vsw,     /* Vsw (x) in km/s */
                           QBSMP_models model, /* model to use */
                           float  sigma,    /* mult scaling to apply to model*/
                           float* vec,      /* vector direction to follow */
                           float* distance, /*returned signed dist along vec */
                           float* r_at_shock);/* returned GSE pos at shock */

int QBSCalcDistAlongVec(float* R,          /* position (pre-aber and displ)*/
                       float eps,         /* model eps and L - pre-scaled */
                       float L,
                       float* vec,   /*vector direction to follow (pre-aber) */ 
                       float* distance, /*signed distance along vector in Re */
                       float* r_at_shock); /* pos at shock in frame of R*/

int QBSChooseAlongVecRoot(float* R,   /* pre-aber, displaced position */
                          float  eps,  /* model, scaled L */
                          float  L,
                          float* vec_norm,  /* unit direction vector */ 
                          float* roots, /* roots to quadratic for t */
                          float* chosen_root); /* returned selected root */

int QBSPosDistanceAlongVec(float* r_vec,   /* Position in Re_GSE */
                           float  vsw,     /* Vsw (x) in km/s */
                           QBSMP_models model, /* model to use */
                           float  sigma,    /* mult scaling to apply to model*/
                           float* vec,      /* vector direction to follow */
                           float* distance, /*returned signed dist along vec (always >= 0) */
                           float* r_at_shock);/* returned GSE pos at shock */

int QBSCalcPosDistAlongVec(float* R,          /* position (pre-aber and displ)*/
                       float eps,         /* model eps and L - pre-scaled */
                       float L,
                       float* vec,   /*vector direction to follow (pre-aber) */ 
                       float* distance, /*signed distance along vector in Re (always >= 0) */
                       float* r_at_shock); /* pos at shock in frame of R*/

int QBSChooseAlongPosVecRoot(float* R,   /* pre-aber, displaced position */
                          float  eps,  /* model, scaled L */
                          float  L,
                          float* vec_norm,  /* unit direction vector */ 
                          float* roots, /* roots to quadratic for t */
                          float* chosen_root); /* returned selected root */

void QBSListModels(FILE* fp);      /* output list of BS/MP models to fp */

int QBSCheckModelNumber(int n_model);  /* check model number in range */

void QBSErrorMessage(FILE *fp, int  errno); /* i/o stream, e.g., stdout, error code number */

/* Error codes */

#define QBS_OK                                      0
#define DISCRIM_NEGATIVE_IN_SCALE                   1
#define LARGEST_SCALING_ROOT_NOT_POSITIVE           2
#define NO_REAL_ROOTS_FOR_CLOSEST_DIST_ALONG_NORMAL 3
#define TOO_MANY_REAL_ROOTS_FOR_DIST_ALONG_NORMAL   4
#define NO_POSSIBLE_ROOTS_FOR_DIST_ALONG_NORMAL     5
#define DEGENERATE_1_ROOT_AT_TANGENCY               6
#define NO_REAL_ROOTS_1_MAX_MIN                     7
#define NO_REAL_ROOTS_3_MAX_MIN                     8
#define ERROR_FINDING_MAX_MIN_FOR_BRACKET           9
#define QUARTIC_BRACKETING_CONFUSED                 10
#define RTSAFE_INPUTS_DONT_BRACKET_ROOT             11
#define RTSAFE_POLY_N_MAX_ITER_EXCEEDED             12
#define CUBIC_BRACKET_FAILED                        13
#define NO_INTERSECTION_WITH_SURFACE                14
#define BOTH_INTERSECTIONS_ON_WRONG_BRANCH          15
#define MODEL_NUMBER_OUT_OF_RANGE                   16
#define NO_ROOT_IN_POS_VEC_DIRECTION                17

  #ifdef   __cplusplus
  }
  #endif

#endif
