/*****************************************************************
* PROJECT: Cluster UK CDHF
* COMPONENT: Plug In library
* MODULE: Dwt.cc
* LANGUAGE: C++
* FUNCTION: QSAS Dwt Plug In
* PURPOSE: Computes the Discrete Wavelet Transform of a scalar series.
* 	
* ARGUMENTS:  
*    IN: scalar time series.
*	 time interval
*	 transform DWT or MODWT
*	 wavelet filter and its width
*	 level J
*	 plotting device
*    OUT: 
*	 QSAS scalar sequence containing wavelet coefficients for
*	 j=1,...,J and the scaling coefficients at level J.
*
* RETURN: QplugReturnStatus
*
* History: version 1.0 Created Dec. 2001. 
* 	   version 1.1 June 2002.
* A. Serroukh, 
******************************************************************/

#include<cstdio>
#include<cmath>
#include<algorithm> 
#include<string.h>
#include<cstdlib>
#include<fstream>

#include "qplug_if.h"
#include "qdutil.h"
#include "Qdos.h"
#include "Xrefs.h"

// #include "cpgplot.h"
// #include "PlotWindow.h"
#include "Dwt_set_plot.h"

extern "C" QplugReturnStatus Dwt_plugin(QplugArgList *);

using namespace QSAS;

/*****************************************************************/
QplugReturnStatus Dwt_plugin(QplugArgList *dwt_list){

   QdObject_var  X_in = (* dwt_list)[0];
   QdObject_var  Int_in = (* dwt_list)[1]; 
   QdObject_var  dwt_yes_no = (* dwt_list)[2];
   QdObject_var  scal_filt_in = (* dwt_list)[3];
   QdObject_var  filt_len_in = (* dwt_list)[4];
   QdObject_var  level_in = (* dwt_list)[5];
   char* buf_out = new char[100];

   //narrow the series to a scalar seq.
   QdRScalarSeq_var xx = QdRScalarSeq_var::narrow(X_in);
   if(xx.is_nil() || is_empty(xx)){
     QplugAppendTextDisplay("DWT Error. Input data type not handled\n");
     return QPLUG_FAILURE;
   }

   //read the time interval
   QdTimeInterval_var timeInt = get_timeinterval(Int_in);
   if(timeInt.is_nil() || is_empty(timeInt) ){
     QplugAppendTextDisplay("DWT Error. Input time interval not handled\n");
     return QPLUG_FAILURE;
   }
   //get timetags from X_in
   QdTimeSeq_var in_tags = get_timetags(X_in);     
   if( in_tags.is_nil() ||  xx->sequence_size() < 2){
     QplugAppendTextDisplay("DWT Error. Object is not a time series of at least 2 records\n");
     return QPLUG_FAILURE;
   }    
   //check whether the inputed time tags in timeInt are valid in in_tags
   QdTimeSeq_var tmp_tags = get_timetags_inrange(in_tags, timeInt);
   if(tmp_tags.is_nil() || tmp_tags->sequence_size()<2){
      QplugAppendTextDisplay("DWT Error. Unvalid time interval.\n");
      return QPLUG_FAILURE;    
   }
   
   double* spacing = new double(0.0);
   //QdRScalar_var spacing = new QdRScalar(0.0);
   if(!QuIsTSRegular(in_tags, spacing)){
      QplugAppendTextDisplay("Dwt Error: Irregular time series. \n");
      return QPLUG_FAILURE;
   }   
   delete spacing;
   //read what to run DWT or MODWT
   Yes_or_No DWT_YES_NO = NO;
   Yes_or_No MODWT_YES_NO = YES;
   Yes_or_No PAD_DWT_YES_NO = NO;
   
   QdString_var dwt_str = QdString_var::narrow(dwt_yes_no);
   if(dwt_str.is_nil()){
      QplugAppendTextDisplay("DWT Error: Transform undefined\n");
      return QPLUG_FAILURE;
   } 
   const char* transf_str = dwt_str->c_str();
   if(strcmp(transf_str , "MODWT") == 0 || strcmp(transf_str , "") == 0)  
   	DWT_YES_NO = NO;
   else if(strcmp(transf_str , "DWT") == 0) {
   	DWT_YES_NO = YES;
	MODWT_YES_NO = NO;
   }
   else {     
     QplugAppendTextDisplay("DWT Error. Wavelet Transform undefined\n");
     return QPLUG_FAILURE;
   } 
   //get the filter name
   QdString_var filt_name = QdString_var::narrow(scal_filt_in);  
   if(filt_name.is_nil()){
     QplugAppendTextDisplay("DWT Error. Filter name not handled\n");
     return QPLUG_FAILURE;
   } 

   //get the filter length
   QdRScalar_var flen = QdRScalar_var::narrow(filt_len_in);
   //int filterLen = (int)(*flen); 
   if(flen.is_nil() || (int)(*flen) < 2){
     QplugAppendTextDisplay("DWT Error. Filter not handled\n");
     return QPLUG_FAILURE;
   }
   //check whether the filter length is even
   float nf = (*flen)/2.0;
   if((nf - floor(nf)) != 0.0){
      QplugAppendTextDisplay("DWT Error. Filter length is not even.\n");
      return QPLUG_FAILURE; 
   }
   
   QplugReturnStatus ret_status;  
   
   //read the filter from file /home/aserr/myqsas/Bank.flt
   const char* filterName = filt_name->c_str();
   QdRScalarSeq_var gscal = new QdRScalarSeq((int)(*flen));
   ret_status = ReadFilter(filterName,gscal);
   if(ret_status==QPLUG_FAILURE) return QPLUG_FAILURE;
   //gscal->diag_print();

   //rescale twice if MODWT
   if(MODWT_YES_NO == YES) *gscal /= sqrt(2.0);
    //gscal->diag_print();   
     
   //narrow the level to an integer
   QdRScalar_var J = QdRScalar_var::narrow(level_in);
   if(J.is_nil()){
     QplugAppendTextDisplay("Dwt Error: Level not handled\n");
     return QPLUG_FAILURE;
   }
   //test if the sample size can be processed
   if( xx->sequence_size() <= (int)(*flen) ){
      QplugAppendTextDisplay("DWT Error. Unvalid sample size.\n");
      return QPLUG_FAILURE; 
   }   
  //create a new subsequence of time tags
  SequenceMask   
     msk(in_tags->get_mask(QdTimeSeq::InRange(timeInt->start(),timeInt->end())));
   QdTimeSeq_var out_tags = new QdTimeSeq();
   //get the time tags from in_tags for which the mask msk was true
   in_tags->get_subsequence(*out_tags, msk);
 
  if( out_tags->sequence_size() < (int)(*flen) ){
      sprintf(buf_out, "Number of tags: %i \n", out_tags->sequence_size());
      QplugAppendTextDisplay (buf_out);
      QplugAppendTextDisplay("DWT Error. Unvalid time interval size.\n");
      return QPLUG_FAILURE; 
   }
   
   long Nsize = xx->sequence_size();
   long Nsubsize = out_tags->sequence_size(); 
   QdRScalarSeq_var xx_subseq = new QdRScalarSeq(Nsubsize);
   QdRScalarSeq_var pad_xx_subseq;
   
   xx->get_subsequence(*xx_subseq , msk);
   xx_subseq->copy_xrefs_from(X_in);//copy xref first
   set_timetags((QdObject_var)xx_subseq, out_tags);
   
   //display sample size on the plug-in window
   sprintf(buf_out,"input size: %ld processed sample size:%ld\n",Nsize,Nsubsize);
   QplugAppendTextDisplay (buf_out);
   delete buf_out;
   if(DWT_YES_NO ==  YES) {
      int jpower = 1;
      //if the series length is not a power of 2, padd by zeros without
      //changing the size of out_tags.    
     if(isPowerOf2(Nsubsize) == NO){
         PAD_DWT_YES_NO = YES;
	 pad_xx_subseq = new QdRScalarSeq(Nsubsize);
	 xx->get_subsequence(*pad_xx_subseq , msk);
     	 while( Nsubsize > pow(2.0,jpower)) ++jpower;
	 long pad_size = (long)(pow(2.0, jpower) - Nsubsize);
	 if(pad_size >0 ) Nsubsize = (long)pow(2.0,jpower);	 
	 for(int t=0; t<pad_size; t++) pad_xx_subseq->push_back(0.0);
      } else { //if Nsubsize is a power of 2.
	    pad_xx_subseq = new QdRScalarSeq(Nsubsize);
	    xx->get_subsequence(*pad_xx_subseq , msk);
      }
   }
   //work out the maximum level 
   int Jlevel = (int)*J;
   float Jmax = Nsubsize - 1.0; 
   Jmax = Jmax/((float)*flen-1.0);
   Jmax = log(Jmax + 1.0)/log(2.0);
   //run the full dwt if Jlevel =0 
   if(Jlevel == 0) Jlevel = (int)Jmax;   
   //check whether the required level transform is <= Jmax
   if( Jlevel <0 || Jlevel > (int)Jmax){
     QplugAppendTextDisplay("Dwt Error: DWT level is out of range.\n");
     return QPLUG_FAILURE;
   }
   
   //compute the wavelet filter
   QdRScalarSeq_var hwav = new QdRScalarSeq(*gscal);
   reverse(hwav->begin(), hwav->end());
   int lsign = -1;
   for(int l=0; l< (int)(*flen); l++){
	lsign = -lsign;
	hwav[l] = lsign*hwav[l]; //g[Flen-1-l];
   }       
   //define the output data 
   QdRScalarSeq_var wav_coef = new QdRScalarSeq();
   try{
      if(DWT_YES_NO == YES){
         ret_status = Run_Dwt(pad_xx_subseq, Jlevel, hwav, gscal, wav_coef);
      }
      else {
         ret_status = Run_MoDwt(xx_subseq, Jlevel, hwav, gscal, wav_coef); 
      }
   } catch(Exception &e){
	e.print_msg();
	return QPLUG_FAILURE;
   }   
   				
   //now apply the dot product to get |W_{j,t}|^2
   QdRScalarSeq_var wav_coef_cp = new QdRScalarSeq(*wav_coef);	
   wav_coef = QdRScalarSeq_var::narrow(
   	((QdObject_var)wav_coef_cp)*((QdObject_var)wav_coef_cp));

   if(wav_coef.is_nil()){
     QplugAppendTextDisplay("Dwt Error: mult operation error\n");
     return QPLUG_FAILURE;
   }
   
   //copy cross references  
   wav_coef->copy_xrefs_from(xx_subseq);
   wav_coef->delete_xref(DEPEND_0); 
   
   //setting and defining attributes
   //QdString_var timeInt_s_start = new QdString(timeInt->start().iso_srep());
   QdString_var timeIntrv = new QdString(timeInt->iso_srep());
   QdTime_var time_step = new QdTime(out_tags[1]-out_tags[0]);   
   QdString_var timeStep  = new QdString(time_step->iso_srep());
   
   char* timeIntrv_s = str_dup(*timeIntrv, timeIntrv->length());
   QuSetTxtAttr("TimeInterval", timeIntrv_s,(QdObject_var)wav_coef); 
   char* timeStep_s = str_dup(*timeStep, timeStep->length());
   QuSetTxtAttr("TimeStep", timeStep_s,(QdObject_var)wav_coef);      
   delete timeIntrv_s; delete timeStep_s;	


   
   //set the transform type
   QdString_var transf_name;
   if(DWT_YES_NO == YES) transf_name = new QdString("DWT_"); 
   else transf_name = new QdString("MODWT_"); 
      
   QdString_var transf_type = new QdString(*transf_name);   
   if(PAD_DWT_YES_NO == YES) *transf_type += "Padded with zeros"; 
   char* transf_type_s = str_dup(*transf_type, transf_type->length()); 
   QuSetTxtAttr("Discret_Wav_Transform", transf_type_s,(QdObject_var)wav_coef); 
        
   QdString_var oname = new QdString(QuGetAttrText("ObjectName",X_in));
   if(DWT_YES_NO == YES) *oname += "_DwtCoef";
   else *oname += "_MODwtCoef";
 
   char* new_obj_name = str_dup(*oname,oname->length());
   QuSetTxtAttr("DwtObjectName", new_obj_name, (QdObject_var)wav_coef);
   delete new_obj_name;
   
   //set the scaling filter name
   QdString_var scal_filter = new QdString(*filt_name); 
   *scal_filter += "(" + int_to_string((int)*flen) + ")";
   char* filt_s_rep = str_dup(*scal_filter, scal_filter->length());
   QuSetTxtAttr("Scaling_Filter", filt_s_rep,(QdObject_var)wav_coef);
   delete filt_s_rep;
   
   //set the DWT run level
   QdString_var dwt_level = new QdString("DWT run level: ");
   *dwt_level += int_to_string(Jlevel);
   char* level_s_rep = str_dup(*dwt_level, dwt_level->length());
   QuSetTxtAttr("DWT_Level", level_s_rep,(QdObject_var)wav_coef);
   delete level_s_rep;
   
   dwt_list->push_back((QdObject_var)wav_coef);
      
   /*---------------------------------------------------------*/
   /* prepare the plotting of X_t and wav. coef.             */

   PGSpectro_ret  ret_value;
   PGSpectro pgspectro, *pgspectrop;
//    PGXYtrace trace, *tracep;
//    tracep = &trace;
   pgspectrop = &pgspectro;
      
   QdString_var tmp_str;
   
   //get the subtitle of the time series plot
   const char* txtlegend = " ";
   tmp_str = new QdString(QuGetAttrText("FIELDNAM", X_in));
   char* subttle_s_rep = str_dup(*tmp_str, tmp_str->length()); 
   
   //FIELDNAM 
   tmp_str = new QdString(*transf_name); 
   *tmp_str += QuGetAttrText("FIELDNAM", X_in);   
   char* fldname_s_rep = str_dup(*tmp_str, tmp_str->length());
   QuSetTxtAttr("FIELDNAM", fldname_s_rep,(QdObject_var)wav_coef);
   delete fldname_s_rep;
   
   //LABLAXIS
   tmp_str = new QdString(*transf_name);
   *tmp_str += QuGetAttrText("LABLAXIS", X_in);
   char* labaxis_s_rep = str_dup(*tmp_str, tmp_str->length());
   QuSetTxtAttr("LABLAXIS",labaxis_s_rep,(QdObject_var)wav_coef);
   delete labaxis_s_rep;
   
   //SI_conversion   
   tmp_str = new QdString(QuGetAttrText("SI_conversion", X_in));
   *tmp_str += "^2";
   char* siconvr_s_rep = str_dup(*tmp_str, tmp_str->length());   
   QuSetTxtAttr("SI_conversion",siconvr_s_rep,(QdObject_var)wav_coef);
   delete siconvr_s_rep;
   
   //UNIT y_label and wedge_label
   string y_lab = QuGetAttrText("UNITS", X_in); //dont delete
   tmp_str = new QdString(y_lab.c_str());
   *tmp_str += "^2";
   char* units_s_rep = str_dup(*tmp_str, tmp_str->length());   
   QuSetTxtAttr("UNITS",units_s_rep,(QdObject_var)wav_coef);
      
   //widget label
   //tmp_str = new QdString("Log(");
   //*tmp_str +=  QdString(units_s_rep) + ")";
   //char* w_lab = str_dup(*tmp_str, tmp_str->length());
   char* w_lab  = units_s_rep;//dont delete 
   
   try{      
//        ret_status = XYtrace_set(xx_subseq, tracep, "Time", y_lab, 
// 			" ",subttle_s_rep, txtlegend);
//        tracep->subtitle = subttle_s_rep;
//    
//        ret_status = Time_set(out_tags, pgspectrop, tracep, DWT_YES_NO);

       ret_status = SPectro_set(wav_coef, out_tags, Jlevel, pgspectrop,
		gscal , DWT_YES_NO, (char*)"Time ", (char*)"Scale(Second)", w_lab);
	
   } catch(Exception &e){
	e.print_msg();
	return QPLUG_FAILURE;
   }		
//    QdString_var spec_title = new QdString("Wavelet Power Spectrum. ");
//    *spec_title += *transf_name;
//    *spec_title += *scal_filter;
//    char* spec_title_s_rep = str_dup(*spec_title, spec_title->length());
//    pgspectrop->title = spec_title_s_rep; //dont delete  
//    
//    QdRScalar_var pgdev = QdRScalar_var::narrow(pgdev_in);
//    if( pgdev.is_nil() ) {
//        QplugAppendTextDisplay("Device Error: plot device undefined.\n");
//        return QPLUG_FAILURE;
//    }
   //ALBAN
   std::cout << pgspectrop->nx << " " << pgspectrop->ny << std::endl;
   QdRMatrixSeq_var Spectro_return=new QdRMatrixSeq(pgspectrop->nx, Dim(pgspectrop->ny));
   for(int i=0; i<pgspectrop->nx; ++i)
   {
	   for(int j=0; j<pgspectrop->ny; ++j)
	   {
		   Spectro_return->elt(i)[j]=pgspectrop->z[i*pgspectrop->ny+j];
	   }
   }
   
   
   QdRMatrix_var spectro_depend_1=new QdRMatrix(Dim(pgspectrop->ny));
   QdRMatrix_var spectro_d1_delta_plus=new QdRMatrix(Dim(pgspectrop->ny));
   QdRMatrix_var spectro_d1_delta_minus=new QdRMatrix(Dim(pgspectrop->ny));
   for(int i=0; i<pgspectrop->ny; ++i)
   {
	   (*spectro_depend_1)[i]=(pgspectrop->ybinlow[i]+pgspectrop->ybinhigh[i])/2.;
	   (*spectro_d1_delta_plus)[i]=(pgspectrop->ybinhigh[i]-pgspectrop->ybinlow[i])/2.;
	   (*spectro_d1_delta_minus)[i]= (*spectro_d1_delta_plus)[i];
   }

   QuSetTxtAttr("UNITS", "s", (QdObject_var)spectro_depend_1);
   QuSetTxtAttr("LABLAXIS", "Wavelet scale", (QdObject_var)spectro_depend_1);
   QuSetTxtAttr("SI_CONVERSION", "1>s", (QdObject_var)spectro_depend_1);
   QuSetTxtAttr("SCALETYP", "Log", (QdObject_var)spectro_depend_1);
   QuSetTxtAttr("UNITS", "s", (QdObject_var)spectro_d1_delta_plus);
   QuSetTxtAttr("SI_CONVERSION", "1>s", (QdObject_var)spectro_d1_delta_plus);
   QuSetTxtAttr("UNITS", "s", (QdObject_var)spectro_d1_delta_minus);
   QuSetTxtAttr("SI_CONVERSION", "1>s", (QdObject_var)spectro_d1_delta_minus);
   Spectro_return->set_xref("UNITS", wav_coef->get_xref("UNITS"));
   Spectro_return->set_xref("SI_CONVERSION", wav_coef->get_xref("SI_CONVERSION"));
   
   Spectro_return->set_xref("DEPEND_0", out_tags);
   
   spectro_depend_1->set_xref("DELTA_PLUS", spectro_d1_delta_plus);
   spectro_depend_1->set_xref("DELTA_MINUS", spectro_d1_delta_minus);
   Spectro_return->set_xref("DEPEND_1", spectro_depend_1);
   QdRScalar_var sFill = new QdRScalar(-1e+31);
   Spectro_return->set_xref("FILLVAL", sFill);
   QuSetTxtAttr("SCALETYP","Log", Spectro_return);
   
   QdRScalar_var smin = new QdRScalar(pgspectrop->zscalemin);
   QdRScalar_var smax = new QdRScalar(pgspectrop->zscalemax);
   Spectro_return->set_xref("SCALEMIN", smin);
   Spectro_return->set_xref("SCALEMAX", smax);
   
   dwt_list->push_back((QdObject_var)Spectro_return);
	//FIN ALBAN
  
   clean_all(pgspectrop);
   
   return QPLUG_SUCCESS;
  }
  
