/****************************************************************************
							LICENSE

  Copyright (C) 2012 Eugen Sorbalo

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  A copy of the GNU General Public License is available on the internet at:

  http://www.gnu.org/copyleft/gpl.html

  or you can write to:

  The Free Software Foundation, Inc.
  675 Mass Ave
  Cambridge, MA 02139, USA
****************************************************************************/

/**
 * \file ClusterCommons.h
 * For more information, please contact e.sorbalo (AT) jacobs-university DOT de
 * 
 * \brief 
 * 
 * 
 */

#ifndef CLUSTERCOMMONS_H_
#define CLUSTERCOMMONS_H_

#include <vector>
#include <string.h>

namespace cluster {

const double kPi = 4.0*atan(1.0);

const short kScalarRank = 1;  ///compare nr_coordinates with this to decided if it is a scalar
const short kVectorRank = 3;  ///compare nr_coordinates with this to decided if it is a vector

const short kMaxNrSatellite = 4;

///Same as strcmp, but transforms strings to lowercase first.
int strcmp_ignorecase(const char* str1, const char* str2);

/**
 * The class Status has the purpose to print messages either to user or to the Console.
 */
class Status
{
public:
	static const bool kPrintToUserAndConsole = true;
	static const bool kPrintToConsoleOnly = false;
	static void PrintMessage(bool to_user, const char fmt[], ...);
public:
	enum ReturnStatus
	{
		kOk = 0, kUnknownError, kWrongInput, kNoInput, kVectorFieldRequired
	} code;
	std::string message() const;
	inline void set_message(const std::string message)
	{
		message_ = message;
	}
	;
	Status();
	virtual ~Status();
private:
	std::string message_;
};

/*
 *
 */
class ClusterAnalysis
{
public:
	class CEstimateNormalComponentFor3SpacecraftMethod {
	public:
		enum EOption {kNone, kGradientParallel, kGradientPerpendicular, kStationarity};
		QdRMatrix_var qvector;
		QdRMatrixSeq_var qvector_seq;

		CEstimateNormalComponentFor3SpacecraftMethod():
			option_(kNone),qvector(NULL),qvector_seq(NULL),slope_(NULL){};
		CEstimateNormalComponentFor3SpacecraftMethod(const unsigned short option);
		inline EOption get_option() const {return option_;}
		void set_option(const unsigned short option);
		inline double get_slope(const int t, const unsigned long coord) const;
		int computeSlopeOfTimeSeries(QdRMatrixSeq_var &g);
	private:
		EOption option_;  ///< Option for normal component estimation method.
		QdRMatrixSeq_var slope_;  ///< Is computed for stationarity assumption.
	} estimator;

public:
	//Inputs:
	QdRMatrixSeq_var field_TS_;  ///< The field. Assume that field_TS_ = new QdRMatrixSeq(nr_meas, RMatrix(Dim(nr_spacecrafts,nr_coordinates)));
	QdRMatrixSeq_var position_TS_;  ///< The position of satellites. Assume that field_TS_ = new QdRMatrixSeq(nr_meas, RMatrix(Dim(nr_spacecrafts,nr_coordinates)));
	double deltat_;  ///< Time interval between successive measurements. \todoCheck why do we need it ???

	//Outputs:
	QdRMatrixSeq_var grad_magn_TS_;
	QdRMatrixSeq_var grad_TS_;
	QdRMatrixSeq_var curl_TS_;
	QdRScalarSeq_var div_TS_;
	QdRScalarSeq_var vol_TS_;  ///< volume of the tetrahedron or the surface of the triangle.
	QdRMatrixSeq_var bary_vector_TS_;
	QdRScalarSeq_var bary_scalar_TS_;
	QdRMatrixSeq_var bary_coordinates_TS_;
	QdRMatrixSeq_var normal_TS_;   ///< normal to the surface of the triangle (formed by first 3 spacecrafts).
	QdRMatrixSeq_var curl_TS_MVA_;  ///< minimum variance analysis
	QdRMatrixSeq_var bary_TS_MVA_;
	QdRMatrixSeq_var velo_TS_MVA_;
	QdRMatrix_var rotation_matrix_;

	Status status;

	ClusterAnalysis();
	virtual ~ClusterAnalysis();

	/**
	 * Computes the gradient of the field.
	 *
	 * If nr_spacecrafts is left default, then the function selects automatically
	 * either Compute4PointsEstimate or Compute3PointsEstimate depending on how
	 * much data is available. Scalar or vector field is determined automatically.
	 */
	Status::ReturnStatus ComputeGradientEstimate(signed short nr_spacecrafts = -1);

	/**
	 * Computes curl and div estimates provided that the gradient was already estimated.
	 * \param curl_scale_factor A constant factor for the curl. Used mainly for the
	 * 		formula: J = curl B / mu_0 to compute the current.
	 */
	Status::ReturnStatus ComputeCurlAndDiv(double curl_scale_factor = 1);

	/**
	 * Computes barycenter of the field (the average of the field data from the measurements point).
	 */
	Status::ReturnStatus ComputeBarycenter();

	/**
	 * Computes MVA, normalizes curl and others.
	 * \note Available only if grad_TS_ and curl_TS_ are not null.
	 */
	Status::ReturnStatus Postprocess();

	inline unsigned short nr_coordinates() const {return nr_coordinates_;};

private:
	unsigned short nr_satellites_;
	unsigned short nr_coordinates_;  ///< If this is 1, then we work with scalars.
	int nr_measurements_;

	Status::ReturnStatus Compute4PointsEstimate();

	Status::ReturnStatus Compute3PointsEstimate();

	/**
	 \brief Computes the reciprocal vectors for the 4 spacecraft configuration.
		 Position is taken from the position_TS_ variable.
	 \param[in] t the index in the position_TS_ array for which the reciprocal will be computed.
	 \param[out] K matrix with reciprocal vectors. The user should pre-allocate memory for K.
	 \return Total volume of the tetrahedron formed by 4 spacecrafts.
	 */
	double ComputeReciprocal4Spacecrafts(int t, QdRMatrix_var &K);

	/**
	 \brief Computes the reciprocal vectors for the 3 spacecraft configuration.
	 \param[in] P matrix which contains positions of spacecrafts.
	 \param[out] K matrix with reciprocal vectors.
	 \param[out] normal with the normal vector to the plane of the spacecrafts.
	 \note The user should pre-allocate memory for \e K and \e normal.
	 */
	//void computeReciprocal3Spacecrafts(const QdRMatrixSeq_var &P, const int t, QdRMatrix_var &normal, const double surface, QdRMatrix_var &K);

	///Finds argmax(norm(V))
	static double maxVal(QdRMatrixSeq_var &V);

	/// sum_{i<nr_spacecraft} (V1_compV1 * V2_compV2)
	static double sum_SCs(QdRMatrix_var &V1, QdRMatrix_var &V2, int coordV1, int coordV2, unsigned short nr_spacecrafts);

	/**
	 Gradient of a scalar quantity as measured from 4 satellites.
	 */
	void computeGradScalar(QdRMatrix_var &Vs, QdRMatrix_var &R, QdRMatrix_var &Grad);

	/**
	 Gradient of the magnitude of a vector quantity as measured from 4 satellites.
	 */
	void computeGradMagnitude(QdRMatrix_var &V, QdRMatrix_var &R, QdRMatrix_var &Grad);

	/// DEPRECATED. This method is not used any more. Please see ComputeCurlAndDiv method.
	void CurlVec(QdRMatrix_var &V, QdRMatrix_var &R, const unsigned short nr_spacecrafts, QdRMatrix_var &Curl);

	/// DEPRECATED. This method is not used any more. Please see ComputeCurlAndDiv method.
	double DivergenceVec(QdRMatrix_var &V, QdRMatrix_var &R);

	/**
	 Gradient of a vector quantity as measured from 4 satellites.
	 */
	void computeGradVec(QdRMatrix_var &V, QdRMatrix_var &R, QdRMatrix_var &Grad);

	/**
	 Computes square of the norm of a vector V
	 */
	static double computeSquareVectorNorm(const QdRMatrix_var &V);

	static void normalizeVector(const QdRMatrix_var &vector, QdRMatrix_var &unit_vector);

	double computeEstimateDgDn(const QdRMatrix_var &grad_p, const QdRMatrix_var &normal,
														 const CEstimateNormalComponentFor3SpacecraftMethod &estimator,
														 const unsigned int t,
														 const unsigned short coord,
														 QdRScalar &dgdn);

	///
	void computeReciprocal3Spacecrafts(const int t, QdRMatrix_var &normal, const double surface, QdRMatrix_var &K);

	/**
	 * \brief Computes the normal vector to the plane and the surface area of the triangle
	 * 		 formed by the first three satellites.
	 * \param[in] nr_position_measurements is the number of position measurements for which
	 * 		it is needed to computed the surface and volume. If relative positions are almost
	 * 		constant, it makes sense to compute value for only one. The function takes minimum
	 * 		between this parameter and the actual sequence size of the position_TS_.
	 * \param[out] surface_TS area of the triangle. The function allocates memory and resizes the vector.
	 * \param[out] normal_TS normal to the plane. The function allocates memory and resizes the vector.
	 */
	void computeNormalToSurfaceOfFirst3Sat(int nr_position_measurements,
			QdRScalarSeq_var &surface_TS, QdRMatrixSeq_var &normal_TS);

	//=======================================================================
public:
	struct TCrossSpectralDensityParameters
	{
		short nr_points_subinterval;  ///WAVE_NSUB = 128  data points in each fft sub-interval
		short nr_points_overlap_between_subintervals;  ///WAVE_NOL = 64  overlap between fft sub-intervals
		enum {kHanning, kHamming, kBartlett, kRectang} window_type;  ///WAVE_WINDOW = 'hann'  window function for fft
		std::vector<double> window_function;  ///< will be filled by ComputeWindowFunction and used in CSD estimation.
		double window_factor;  ///normalization factor, computed by ComputeWindowFunction
		double frequency_sample;  /// basically, this is 1/delta_
		double csd_normalization_factor;  ///all csd elements are multiplied with this factor.
		int nr_segments;
		std::vector<double> frequencies;
	};

	typedef std::vector<std::complex<double> > MATXYZ4S_COMPLEX[kVectorRank][kMaxNrSatellite][kVectorRank][kMaxNrSatellite];

public:
	/** Remove the linear trend from a time series.
	 * \param u measurements;
	 */
	static void Detrend(std::vector<double> &u);

public:
	//Inputs:
	bool do_detrend_;  ///< Does the user want detrending?

	TCrossSpectralDensityParameters csd_parameters_;
	bool use_dopler_correction_;  //WAVE_DOPPLER_VECTOR = 0  define DOPPLER_VECTOR [km/s] for Doppler correction
	QdRMatrix_var doppler_vector_;  //DOPPLER_VECTOR = [0.,0.,100.]

	//Outputs:

	Status::ReturnStatus FindWaves(QDOS::Matrix<double> &power,
			QDOS::Matrix<double> &amplitude, QDOS::Matrix<double> &wave_vector,
			QDOS::Matrix<double> &slowness);

	Status::ReturnStatus ComputeCrossSpectralDensityMatrix4Satellites(MATXYZ4S_COMPLEX &csd_matrix);

	Status::ReturnStatus ComputeWindowFunction();

private:
	///Finds nr of satellites, nr of coordinates and nr measurements.
	Status::ReturnStatus Prelimenary();

	/**
	 * Seeks waves in the field.
	 * \param[OUT] power
	 * \param[OUT] amplitude
	 * \param[OUT] wave_vector
	 * \param[OUT] slowness = \vect{k} / \omega
	 */
	Status::ReturnStatus FindWaves4Satellites(QDOS::Matrix<double> &power,
			QDOS::Matrix<double> &amplitude, QDOS::Matrix<double> &wave_vector,
			QDOS::Matrix<double> &slowness);

	/// EIGvalues and eigenvectors of a HERmitian MATrix associated with array measurements at 4 Sensors.
	inline Status::ReturnStatus EigenValuesOf_MATXYZ4S_COMPLEX(MATXYZ4S_COMPLEX &matrix, size_t t,
			std::vector<std::complex<double> > &eigenvects, std::vector<double> &eigvalues,
			size_t &size_of_eigenvects);

	Status::ReturnStatus VogtWaveSurveyor(QDOS::Matrix<std::complex<double> > &eigenvects,
			QDOS::Matrix<double> &eigenvals, QDOS::Matrix<double> &noise, QDOS::Matrix<double> &wave_vector,
			QDOS::Matrix<double> &amplitude);
};


}  //namespace cluster

#endif /* CLUSTERCOMMONS_H_ */
