/**
 * \file MyMatrixEig.cpp
 * See the header file MyMatrixEig.h
 * 
 */

#include <cmath>
#include <iostream>
#include <complex>
#include <vector>
#include <string.h>

namespace QDOSS {

const float kEpsilon = 1E-10;  //used to see if a float variable is zero.

double house(double * x, int n, double * v)
{
	int i;
	double beta = 0., sigma = 0.;
	//double norm = 0.;

	for (i = 1; i < n; i++) {
		v[i] = x[i];
		sigma += x[i] * x[i];
	}
	v[0] = 1.;

	if (sigma != 0.) {
		double mu = std::sqrt(x[0] * x[0] + sigma);
		if (x[0] < 0.) {
			v[0] = x[0] - mu;
		}
		else {
			v[0] = -sigma / (x[0] + mu);
		}
		beta = 2 * v[0] * v[0] / (sigma + v[0] * v[0]);
		for (i = 1; i < n; i++) {
			v[i] = v[i] / v[0];
		}
		v[0] = 1.;
	}
	return beta;
}

void tridiag(double * a, double * q, int n)
{
	int i, j, k, sub_size;
	double beta, coeff, norm;
	double * v = new double[n - 1]; // Container for Householder vectors
	double * p = new double[n - 1]; // Container for Householder matrices
	double * w = new double[n - 1];
	double * beta_arr = new double[n - 1];
	double * q_temp = new double[n * n];

	for (k = 0; k < n; k++) {
		q[k * n + k] = 1.;
	}

	for (k = 0; k < n - 2; k++) {
		double * x = &a[k * n + k + 1];
		sub_size = n - k - 1;

		beta = house(x, sub_size, v);
		beta_arr[k] = beta;
		//p=b*a(k+1:n, k+1:n)*v;

		for (i = 0; i < sub_size; i++) {
			p[i] = 0.;
			for (j = 0; j < sub_size; j++) {
				p[i] += a[(i + k + 1) * n + j + k + 1] * v[j];
			}
			p[i] *= beta;
		}

		//w=p-(b*p'*v/2)*v;
		coeff = 0.;
		norm = 0.;
		for (i = 0; i < sub_size; i++) {
			coeff += p[i] * v[i];
			norm += a[(i + k + 1) * n + k] * a[(i + k + 1) * n + k];
		}
		coeff = coeff * beta / 2.;
		for (i = 0; i < sub_size; i++) {
			w[i] = p[i] - coeff * v[i];
		}

		//a(k+1, k)=norm(a(k+1:n, k)); a(k, k+1)=a(k+1, k);
		norm = std::sqrt(norm);
		a[(k + 1) * n + k] = a[k * n + k + 1] = norm;

		for (i = 0; i < sub_size; i++) {
			for (j = 0; j <= i; j++) {
				a[(i + k + 1) * n + j + k + 1] = a[(j + k + 1) * n + i + k + 1] = a[(i + k + 1) * n + j + k + 1] - v[i]
						* w[j] - w[i] * v[j];
			}
			// v storage for Q computation
			if (i != 0) {
				a[(i + k + 1) * n + k] = v[i];
			}
		}
	}

	for (k = n - 3; k >= 0; k--) {
		sub_size = n - k - 1;
		for (i = 0; i < sub_size; i++) {
			if (i == 0)
				v[i] = 1;
			else {
				v[i] = a[(i + k + 1) * n + k];
				a[(i + k + 1) * n + k] = a[k * n + i + k + 1] = 0.;
			}
		}
		// Copy of q in q_temp
		for (i = 0; i < sub_size; i++) {
			for (j = 0; j < sub_size; j++) {
				q_temp[(i + k + 1) * n + j + k + 1] = q[(i + k + 1) * n + j + k + 1];
			}
		}
		for (i = 0; i < sub_size; i++) {
			for (j = 0; j < sub_size; j++) {
				coeff = 0.;
				for (int l = 0; l < sub_size; l++) {
					coeff += v[l] * q_temp[(l + k + 1) * n + j + k + 1];
				}
				q[(i + k + 1) * n + j + k + 1] -= beta_arr[k] * coeff * v[i];
			}
		}
	}

	delete[] p;
	delete[] v;
	delete[] w;
	delete[] beta_arr;
	delete[] q_temp;
}

void givens(double a, double b, double * c, double * s)
{
	if (b == 0) {
		*c = 1.;
		*s = 0.;
	}
	else {
		double tau;
		if (fabs(b) > fabs(a)) {
			tau = -a / b;
			*s = 1. / std::sqrt(1 + tau * tau);
			*c = (*s) * tau;
		}
		else {
			tau = -b / a;
			*c = 1. / std::sqrt(1 + tau * tau);
			*s = (*c) * tau;
		}
	}
}

void apply_givens(double * a, int size, double c, double s, int i, int k, int p, int q)
{
	int j;
	double t1, t2;

	//one side of the diagonal
	for (j = p; j < q; j++) {
		t1 = a[i * size + j];
		t2 = a[k * size + j];
		a[i * size + j] = c * t1 - s * t2;
		a[k * size + j] = s * t1 + c * t2;
	}

	//same procedure to the other side of the diagonal.
	for (j = p; j < q; j++) {
		t1 = a[j * size + i];
		t2 = a[j * size + k];
		a[j * size + i] = c * t1 - s * t2;
		a[j * size + k] = s * t1 + c * t2;
	}
}

void symmqr_step(double * t, double * Q, int size, int p, int q)
{
	double d = (t[(q - 2) * size + q - 2] - t[(q - 1) * size + q - 1]) / 2.;
	double tq2 = t[(q - 1) * size + q - 2] * t[(q - 1) * size + q - 2];
	//we know that off-diagonal elements are not zero, so tq2<>0
	double sg = (d >= 0.0 ? 1.0 : -1.0);
	// d can be zero, so 'sg' should not be zero, otherwise we have Inf.
	double mu = t[(q - 1) * size + q - 1] - tq2 / (d + sg * std::sqrt(d * d + tq2));

	double x = t[p * size + p] - mu;
	double z = t[(p + 1) * size + p];
	int k, j;
	double c, s;
	double t1, t2;
	for (k = p; k < q - 1; k++) {
		givens(x, z, &c, &s);
		apply_givens(t, size, c, s, k, k + 1, p, q);

		for (j = 0; j < size; j++) {
			t1 = Q[j * size + k];
			t2 = Q[j * size + k + 1];
			Q[j * size + k] = c * t1 - s * t2;
			Q[j * size + k + 1] = s * t1 + c * t2;
		}

		if (k < q - 2) {
			x = t[(k + 1) * size + k];
			z = t[(k + 2) * size + k];
		}
	}
}

void display_mat(double *t, int size1, int size2)
{
#ifdef DIAGI
	std::cout << "CLAPACK: Matrix at this stage:" << std::endl;
	for (int i = 0; i < size1; ++i) {
		for (int j = 0; j < size2; ++j)
			std::cout << t[i * size2 + j] << " ";
		std::cout << std::endl;
	}
#endif
}

void symmqr(double* t, double * Q, int size)
{
	int p, q, i, end_loop, iter;
	p = 0;
	q = size - 1;
	end_loop = 0;

	//The hash is used to compute a Check Sum of the matrix.
	//If check sums after two iterations are equal, then the procedure must give up.
	double hash[20];
	hash[0] = 1;
	for (int i = 1; i < 20; i++)
		hash[i] = hash[i - 1] * 10;
	double prev_sum = 0, new_sum = 1;
	int iteration = 500;

	while ((q > p) && (new_sum != prev_sum) && (iteration > 0)) {
		//some code for checking if we don't loop too much
		iteration--;
		prev_sum = new_sum;
		new_sum = 0;
		for (int ii = 0; ii < size; ++ii)
			for (int j = 0; j < size; ++j)
				new_sum += t[ii * size + j] * hash[ii * size + j];
#ifdef DIAG
		if (new_sum == prev_sum)
			std::cout << "CLAPACK ERROR: symmqr: no progress" << std::endl;
#endif

		//if off-diagonal elements are almost zero, make them zero.
		for (i = 0; i < size - 1; i++) {
			if (fabs(t[i * size + i + 1]) <= kEpsilon) {
				t[i * size + i + 1] = t[(i + 1) * size + i] = 0.;
			}
		}

		//Find portions where we have non-diagonal elements:
		end_loop = 0;
		iter = q - 1;  //look from the up
		while (!end_loop && iter > 0) {
			if (fabs(t[(iter + 1) * size + iter]) < kEpsilon) {
				q = iter;
				iter--;
			}
			else {
				end_loop = 1;
			}
		}

		end_loop = 0;
		iter = p;  //look from bottom
		while (!end_loop && iter < size - 1) {
			if (fabs(t[(iter + 1) * size + iter]) < kEpsilon) {
				p = ++iter;
			}
			else {
				end_loop = 1;
			}
		}

		if (q > p) {
			symmqr_step(t, Q, size, p, q + 1);
			display_mat(t, size, size);
		}

	}
}

void sort(double * a, double * q, int size)
{
	double temp;
	for (int i = 0; i < size; i++) {
		for (int j = 0; j < size - i - 1; j++) {
			if (a[j * size + j] > a[(j + 1) * size + j + 1]) {
				temp = a[(j + 1) * size + j + 1];
				a[(j + 1) * size + j + 1] = a[j * size + j];
				a[j * size + j] = temp;

				for (int k = 0; k < size; k++) {
					temp = q[k * size + j + 1];
					q[k * size + j + 1] = q[k * size + j];
					q[k * size + j] = temp;
				}

			}
		}
	}
}

void eigen(double * a, double * v, int n)
{
	double * q = new double[n * n];
	memset(q, 0x0, n * n * sizeof(double));

#ifdef DIAG
	std::cout << "CLAPACK: tridiag" << std::endl;
#endif
	tridiag(a, q, n);

#ifdef DIAG
	display_mat(a, n, n);
	std::cout << "CLAPACK: symmqr" << std::endl;
#endif
	symmqr(a, q, n);

#ifdef DIAG
	std::cout << "CLAPACK: sort" << std::endl;
#endif
	sort(a, q, n);

#ifdef DIAG
	std::cout << "CLAPACK: Finished" << std::endl;
#endif
	for (int i = 0; i < n; i++)
		v[i] = a[i * n + i];
	memcpy(a, q, n * n * sizeof(double));
#ifdef DIAG
	display_mat(a, n, n);
#endif
	delete[] q;
}

void eigen(std::vector<std::complex<double> > &array, std::vector<double> &eigvalues, unsigned int size)
{
	using namespace std;
	if (size * size == array.size())
	{
		//if array = A + iB, then the new array is [[A, -B]; [B, A]]
		double *equiv_array = new double[(2*size) * (2*size)];
		for (unsigned int i = 0; i < size; ++i)
			for (unsigned int j = 0; j < size; ++j)
			{
				equiv_array[i * (2*size) + j] = array[i * size + j].real();  //A
				equiv_array[i * (2*size) + j + size] = -array[i * size + j].imag();  //-B
				equiv_array[(i + size) * (2*size) + j] = array[i * size + j].imag();  //B
				equiv_array[(i + size) * (2*size) + j + size] = array[i * size + j].real();  //A
			}
		double *evalues = new double[2 * size];
		eigen(equiv_array, evalues, 2*size);

		//Compound the solution
		eigvalues.empty();
		eigvalues.resize(size, 0.0);
		for (unsigned int col = 0; col < size; ++col)
		{
			//we assume here that the matrix was hermitian => eigenvalues are real!!!
			eigvalues[col] = evalues[2 * col];
			for (unsigned int row = 0; row < size; ++row)
				array[row*size + col] = complex<double>(equiv_array[row * (2*size) + 2*col],
														equiv_array[(row+size) * (2*size) + 2*col]);
		}
		delete[] equiv_array;
		delete[] evalues;
	}
}

} // end namespace QDOS

