/****************************************************************************
							LICENSE

  Copyright (C) 2012 Eugen Sorbalo

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  A copy of the GNU General Public License is available on the internet at:

  http://www.gnu.org/copyleft/gpl.html

  or you can write to:

  The Free Software Foundation, Inc.
  675 Mass Ave
  Cambridge, MA 02139, USA
****************************************************************************/

/**
 * \file eugen_gradient.cpp
 * For more information, please contact e.sorbalo (AT) jacobs-university DOT de
 *
 * \brief QSAS plug-in for the gradient estimation.
 *
 * This file is a transition between the ClusterAnalysis class and QSAS.
 *
 */

#include "qplug_if.h"
#include "qdutil.h"
#include "qar.h"
#include "Qdos.h"
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "ClusterCommons.h"

using namespace QSAS;
using namespace std;
using namespace cluster;

const int VERBOSE = 1;

bool getFieldFromAllSatellites(QplugArgList *call_list, QdObjectSeq_var &joined, 
							 const bool is_Bfield, const bool scalars, const unsigned short nr_spacecrafts,
							 cluster::ClusterAnalysis &nabla);

bool getPositionFromAllSatellites(QplugArgList *call_list, QdObjectSeq_var &joined, 
								 const unsigned short nr_spacecrafts, cluster::ClusterAnalysis &nabla);

std::string QplugNewSTR_2string(const char *in_str)
{
	char * str = QplugNewSTR(in_str);
	std::string out_string = str;
	free(str);
	return out_string;
}

extern "C" QplugReturnStatus  eugen_gradient  ( QplugArgList *call_list )
{
	QplugReturnStatus status = QPLUG_SUCCESS;
	cluster::ClusterAnalysis nabla;
	int  nseq = 8;
	
	//specifies whether we are working with an array of scalars:
	bool scalars = !QuIsThreeVector((QdObject_var)((* call_list)[0]));  
	if (scalars)
	{
		Status::PrintMessage(Status::kPrintToUserAndConsole, "Input is a field of scalars\n");
	}
	else
	{
		Status::PrintMessage(Status::kPrintToUserAndConsole, "Input is a field of vectors\n");
	}
	
	Status::PrintMessage(Status::kPrintToUserAndConsole, "1. Extracting data\n");
	
	//INPUT 10: number of satellites
	unsigned short nr_spacecrafts = 0;  //Indicates whether we should apply 3- or 4-satellites technique
	{
		QdRScalar_var qnr_satel = QdRScalar_var::narrow((*call_list)[10]);	
		if (!qnr_satel.is_nil()) {
			nr_spacecrafts = ((unsigned short) *qnr_satel);
			Status::PrintMessage(Status::kPrintToConsoleOnly, "Operating with %d satellites\n", nr_spacecrafts);
		}
		if ((qnr_satel.is_nil()) || ((nr_spacecrafts!=3) && (nr_spacecrafts !=4)))
		{
			Status::PrintMessage(Status::kPrintToUserAndConsole, "Currently %d-spacecraft methods is not supported yet", nr_spacecrafts);
			return QPLUG_FAILURE;
		}
	}
	
	//INPUT 11 and 12: options to estimate the normal component for 3-satellites data
	if (nr_spacecrafts == 3) {
		unsigned short option_for_estimate = 0;
		//Option choice:
		QdRScalar_var qoption = QdRScalar_var::narrow((*call_list)[11]);	
		if (!qoption.is_nil())
			option_for_estimate = ((unsigned short) *qoption);
		if ((qoption.is_nil()) || (option_for_estimate > 3)) {
			Status::PrintMessage(Status::kPrintToUserAndConsole, "Undefined option %u %s\n", option_for_estimate
					, "to estimate the normal component");
			return QPLUG_FAILURE;
		}
		nabla.estimator.set_option(option_for_estimate);
		if (VERBOSE) {Status::PrintMessage(Status::kPrintToConsoleOnly, "Option for estimate: %u\n", option_for_estimate); }
	}
	short additional_join = 0;
	// ---- INPUT 12: the vector for the normal estimate for 3-sat method
	if (nabla.estimator.get_option() != nabla.estimator.kNone) {
		QdRMatrixSeq_var temp = QdRMatrixSeq_var::narrow((*call_list)[12]);
		if (!temp.is_nil()) additional_join++;
	}
	if (VERBOSE) {Status::PrintMessage(Status::kPrintToConsoleOnly, "Additional join: %u\n", additional_join);}

	// ---- Get frame of vectors ------
	string frame_V = "";
	if (!scalars)
	{
		frame_V = QuGetFrameAttr((* call_list)[0]) ;
		string frame_i = "";
		for (int i=1; i<nr_spacecrafts; i++){
			frame_i = QuGetFrameAttr( (* call_list)[i]);
			if (frame_i != frame_V) {
				QplugAppendTextDisplay("Inputs not in same vector frame:");
				QplugAppendTextDisplay(frame_i.c_str());
				QplugAppendTextDisplay(" and ");
				QplugAppendTextDisplay(frame_V.c_str());
				return QPLUG_FAILURE;
			}
		}
		if ((nr_spacecrafts==3) && (nabla.estimator.get_option() != nabla.estimator.kNone) ){
			frame_i = QuGetFrameAttr( ((* call_list)[12]));
			if (frame_i != frame_V) {
				QplugAppendTextDisplay("Vector for the normal component estimation is not in same vector frame:");
				QplugAppendTextDisplay(frame_i.c_str());
				QplugAppendTextDisplay(" and ");
				QplugAppendTextDisplay(frame_V.c_str());
				return QPLUG_FAILURE;
			}
		}
	}
	
	//Frame for satellites position vectors:
	string frame_P = QuGetFrameAttr( ((* call_list)[4]) );  //first position sequence
	{
		string frame_i = "";
		for (int i=1; i<nr_spacecrafts; i++){
			frame_i = QuGetFrameAttr( ((* call_list)[i+4]));  //only position vectors
			if (frame_i != frame_P) {
				QplugAppendTextDisplay("Satellites' position not in same vector frame:");
				QplugAppendTextDisplay(frame_i.c_str());
				QplugAppendTextDisplay(" and ");
				QplugAppendTextDisplay(frame_P.c_str());
				return QPLUG_FAILURE;
			}
		}
	}

	bool is_Bfield = false;
	string SI_V = QuGetAttrText("SI_conversion", ((* call_list)[0]));
	if(QarSameBaseSI(SI_V.c_str(), "1.0e-9>T") ) is_Bfield = true;

	string SI_i = "";
	for (unsigned int i = 1; i < nr_spacecrafts; i++) {
		SI_i = QuGetAttrText("SI_conversion", ((*call_list)[i]));
		if (SI_i.empty())
			continue;
		if (!QarSameBaseSI(SI_V.c_str(), SI_i.c_str())) {
			QplugAppendTextDisplay("Inputs not in same base units:");
			QplugAppendTextDisplay(SI_i.c_str());
			QplugAppendTextDisplay(" and ");
			QplugAppendTextDisplay(SI_V.c_str());
			return QPLUG_FAILURE;
		}
	}

	string SI_R = QuGetAttrText("SI_conversion", ((*call_list)[4]));
	for (int i = 5; i < 8; i++) {
		SI_i = QuGetAttrText("SI_conversion", ((*call_list)[i]));
		if (SI_i.empty())
			continue;
		if (!QarSameBaseSI(SI_R.c_str(), SI_i.c_str())) {
			QplugAppendTextDisplay("Inputs not in same base units:");
			QplugAppendTextDisplay(SI_i.c_str());
			QplugAppendTextDisplay(" and ");
			QplugAppendTextDisplay(SI_R.c_str());
			return QPLUG_FAILURE;
		}
	}

	//	..first, get input vectors from each SC and join ..
	QdTimeSeq_var ttJoined = get_timetags((*call_list)[2]);    // tt of joined V and Pos form SC3 vec
	QuIsTSRegular(ttJoined, &(nabla.deltat_)); // AJA - don't care about regular, but need estimate of spacing
	double avg_width = 1.5 * nabla.deltat_;
	Status::PrintMessage(Status::kPrintToConsoleOnly, "delta t = %f  ->  gap_fill if t > %f\n", nabla.deltat_, avg_width);

	QdObjectSeq_var toJoin  = new QdObjectSeq(nseq + additional_join);
	QdObjectSeq_var joined  = new QdObjectSeq(nseq + additional_join);
	QdStringSeq_var joinMethod = new QdStringSeq(nseq + additional_join);
	vector <double> gapValue(nseq + additional_join);
	QdStringSeq_var gapHandler = new QdStringSeq(nseq + additional_join);
	
	int correspondence_array[] = {0, 1, 2, 3, 4, 5, 6, 7, 12};  //refers to the order of inputs
	for (int i=0; i<nseq + additional_join; i++) {
		if ((scalars) && (i<4))  //only the field can be scalar
			toJoin[i] = (*call_list)[correspondence_array[i]];
		else
			toJoin[i] = QuGetVectorXYZ ((*call_list)[correspondence_array[i]]);  // this is no slower then the test on representation as xyz vectors are returned untouched
		if ( (toJoin[i]).is_nil() ){
			QplugAppendTextDisplay("Failed to get vector in cartesian form");
		}
		
		joinMethod[i] = QdString(LINEAR);
		gapValue[i] = avg_width;
		gapHandler[i] = QdString(LINEAR);
	}
		

	// ---- INPUT 8: time interval
	QdTimeInterval_var start_end = get_timeinterval((*call_list)[8]);
	SequenceMask msk(ttJoined->get_mask(QdTimeSeq::InRange(start_end->start(), start_end->end())));
	QdTimeSeq_var tt_out = new QdTimeSeq(); // tt of masked V and Pos
	ttJoined->get_subsequence(*tt_out, msk);


	// subsequence is now done in join, no need to subsequence afterwards
	
	QuMultiJoin(toJoin, tt_out, joined, gapValue, gapHandler, joinMethod);
	

	// ---- INPUT 0-3 : get the measurements of the field as measured by all satellites -------------
	QdRMatrixSeq_var V = NULL;
	if (!getFieldFromAllSatellites(call_list, joined, is_Bfield, scalars, nr_spacecrafts, nabla))
		return QPLUG_FAILURE;

	// ---- INPUT 4-7 : get position for given time interval - store in P ---------------------------
	QdRMatrixSeq_var P = NULL;
	if (!getPositionFromAllSatellites(call_list, joined,  nr_spacecrafts, nabla))
		return QPLUG_FAILURE;

	// ---- INPUT 9: get scalefactor for Curl and check if this is a current (scale = mu_0 so that J = curl B / mu_0) ----
	char curlunit[63] = "";
	double scalefac = 1.0;
	bool iscurrent = false;
	if (status == QPLUG_SUCCESS) {
		QdRScalar_var sfac = QdRScalar_var::narrow((*call_list)[9]);

		char scaleunit[63] = "";
		sprintf(curlunit, "%s/%s", QuGetAttrText(UNITS, (*call_list)[2]).c_str(), QuGetAttrText(UNITS, (*call_list)[6]).c_str());

		if (!sfac.is_nil()) {
			scalefac = ((double) *sfac);
			sprintf(scaleunit, "%s", QuGetAttrText(UNITS, (*call_list)[9]).c_str());
			if (strlen(scaleunit) > 1)
				strcat(curlunit, "/");
			strcat(curlunit, scaleunit);
			printf("> Scalefactor supplied : %6.4g %s\n", scalefac, scaleunit);
			fflush(stdout);
			if (!strncasecmp(curlunit, "nT/km/H/m", 9)) {
				iscurrent = true;
			//	sprintf(curlunit, "A/m^2");
				//scalefac *= 1.0E12; // divide (10^-9 nT / 10^3 m) to get current in A/m2
			}
		}
		else {
			printf("> Curl unscaled\n");
			fflush(stdout);
		}
	}
	
	// ---- INPUT 12: the vector for the normal estimate for 3-sat method
	if (nabla.estimator.get_option() != nabla.estimator.kNone) {
		//Extract the vector for the estimate of the normal component:
		nabla.estimator.qvector = QdRMatrix_var::narrow((*call_list)[12]);
		if(!nabla.estimator.qvector.is_nil()) {
			if ((nabla.estimator.qvector->rank()==1) && ((nabla.estimator.qvector->dims())[0] == 3)){
				QplugAppendTextDisplay("Static vector for estimating the normal component.\n");
				printf("Static estimate vector: %f  %f  %f\n", (double)(*nabla.estimator.qvector)[0],
							 (double)(*nabla.estimator.qvector)[1], (double)(*nabla.estimator.qvector)[2]);
				fflush(stdout);
			}
			else{
				QplugAppendTextDisplay("Optional vector is not a vector.\n");
				return QPLUG_FAILURE;
			}
		}
		else {
			QdRMatrixSeq_var estimator_vector_seq = QdRMatrixSeq_var::narrow(joined[nseq]);
			if(!estimator_vector_seq.is_nil()) {
				QplugAppendTextDisplay("Dynamic vector for estimating the normal component.\n");
				nabla.estimator.qvector_seq = estimator_vector_seq;
			}
			else{
				QplugAppendTextDisplay("Optional vector is not a vector.\n");
				return QPLUG_FAILURE;
			}
		}
	}
	// ---------- now, calculate curl, div and grad for each time step ---------
	if (status == QPLUG_SUCCESS) {
		QplugAppendTextDisplay("2. Computing reciprocal vectors and gradient.\n");

		if (nabla.ComputeGradientEstimate(nr_spacecrafts) != cluster::Status::kOk)
				status = QPLUG_FAILURE;
		if (status == QPLUG_SUCCESS)
			cluster::Status::PrintMessage(false, "NablaEstimator: Computing curl and divergence.\n");
		if ((status == QPLUG_SUCCESS) && (nabla.nr_coordinates() == cluster::kVectorRank))
			if (nabla.ComputeCurlAndDiv(scalefac) != cluster::Status::kOk)
				status = QPLUG_FAILURE;
		if (status == QPLUG_SUCCESS) {
			cluster::Status::PrintMessage(false, "NablaEstimator: Computing barycenter.\n");
			if (nabla.ComputeBarycenter() != cluster::Status::kOk)
				status = QPLUG_FAILURE;
		}
		if ((status == QPLUG_SUCCESS) && (nabla.nr_coordinates() == cluster::kVectorRank)) {
			cluster::Status::PrintMessage(false, "NablaEstimator: Computing MVA.\n");
			if (nabla.Postprocess() != cluster::Status::kOk)
				status = QPLUG_FAILURE;
		}
		if (status != QPLUG_SUCCESS)
			cluster::Status::PrintMessage(false, "NablaEstimator: Finished processing with errors.\n");
	}

	// -------- send some outputs ---------------------------
	if (status == QPLUG_SUCCESS) {
		cluster::Status::PrintMessage(true, "NablaEstimator: Preparing the output.\n");
		char divunit[63] = "";
		char reciprocunit[63] = "";
		sprintf(divunit, "%s/%s", QuGetAttrText(UNITS, (*call_list)[2]).c_str(), QuGetAttrText(UNITS, (*call_list)[6]).c_str());
		sprintf(reciprocunit, "1/%s", QuGetAttrText(UNITS, (*call_list)[6]).c_str());

		cluster::Status::PrintMessage(false, "NablaEstimator: find inverse of SI_conversion\n");
		// SI_conversion: find inverses first
		string attr_inv_ptr = QarSIConv_inverse((*call_list)[4]);
		QdObject_var km_inv_ptr = (*call_list)[4]->create(); // this is empty as we only need metadata
		if( !attr_inv_ptr.empty() ){
			QuSetTxtAttr(SI_CONVERSION, attr_inv_ptr.c_str(), km_inv_ptr);
		}
		
		//In the following 6 lines we read the (char *) from functions and convert them to safe strings.
		string si_convers = QarSIConv_product((*call_list)[0], km_inv_ptr);  // T/m

		// ---- correct the SI conversion of the curl by considering the scale factor. --------------
		attr_inv_ptr = QarSIConv_inverse((*call_list)[9]);
		QdObject_var scalefactor_inv_ptr = (*call_list)[9]->create(); // this is empty as we only need metadata
		if( !attr_inv_ptr.empty() ){
			QuSetTxtAttr(SI_CONVERSION, attr_inv_ptr.c_str(), scalefactor_inv_ptr);
		}
		QdObject_var curl_temp_obj_ptr = (*call_list)[4]->create(); // this is empty as we only need metadata
		QuSetTxtAttr(SI_CONVERSION, si_convers.c_str(), curl_temp_obj_ptr);
		string si_convers_curl = QarSIConv_product(curl_temp_obj_ptr, scalefactor_inv_ptr);

		//cluster::Status::PrintMessage(false, "NablaEstimator curl unit: %s\n unit: %s\n"
		//		, si_convers_curl.c_str(), curlunit);
		// ------ analyze if we have current --------------------------------------------------------
		//1e-12 > m^-2 A
		size_t pos_delim = si_convers_curl.find('>');
		if (pos_delim != std::string::npos) {
			std::string analyze_si_curl = si_convers_curl.substr(pos_delim+1);
			while (analyze_si_curl[0] == ' ') analyze_si_curl.erase(0, 1);  //remove the trailing space
			while (analyze_si_curl[analyze_si_curl.length()-1] == ' ')
				analyze_si_curl.erase(analyze_si_curl.length()-1);  //trim trailing spaces
			if (!analyze_si_curl.compare("m^-2 A") || !analyze_si_curl.compare("A m^-2") ||
					!analyze_si_curl.compare("T H^-1") || !analyze_si_curl.compare("H^-1 T"))
				iscurrent = true;
		}
		// ------------------------------------------------------------------------------------------

		std::string frame = (scalars)? QplugNewSTR_2string(frame_P.c_str()) : QplugNewSTR_2string(frame_V.c_str());
		std::string curl_lablaxis = (iscurrent)? QplugNewSTR_2string("J") : QplugNewSTR_2string("Curl V");
				
		// ------------------------------------------------------------------------------------------
		cluster::Status::PrintMessage(false, "NablaEstimator: preparing output of gradient\n");
		if (!nabla.grad_TS_.is_nil())
		{
			QuSetFrame((QdObject_var) nabla.grad_TS_, frame.c_str());
			QuSetTxtAttr(UNITS,divunit,(QdObject_var) nabla.grad_TS_);
			QuSetTxtAttr(SI_CONVERSION, si_convers.c_str(), nabla.grad_TS_);
			QuSetTxtAttr("LABLAXIS","Grad V",(QdObject_var) nabla.grad_TS_);
			set_timetags((QdObject_var) nabla.grad_TS_, tt_out);
		}

		cluster::Status::PrintMessage(false, "NablaEstimator: preparing output of gradient's magnitude\n");
		if (!nabla.grad_magn_TS_.is_nil())
		{
			QuSetFrame((QdObject_var) nabla.grad_magn_TS_, frame.c_str());
			QuSetTxtAttr(UNITS,divunit,(QdObject_var) nabla.grad_magn_TS_);
			QuSetTxtAttr(SI_CONVERSION, si_convers.c_str(), nabla.grad_magn_TS_);
			QuSetTxtAttr("LABLAXIS","Grad magnitude V",(QdObject_var) nabla.grad_magn_TS_);
			set_timetags((QdObject_var) nabla.grad_magn_TS_, tt_out);
		}
		
		cluster::Status::PrintMessage(false, "NablaEstimator: preparing output of curl\n");
		if (!nabla.curl_TS_.is_nil())
		{
			QuSetFrame((QdObject_var) nabla.curl_TS_, frame.c_str());
			QuSetTxtAttr(UNITS,curlunit,(QdObject_var) nabla.curl_TS_);
			QuSetTxtAttr(SI_CONVERSION, si_convers_curl.c_str(), nabla.curl_TS_);
			QuSetTxtAttr("LABLAXIS", curl_lablaxis.c_str(), (QdObject_var) nabla.curl_TS_);
			set_timetags((QdObject_var) nabla.curl_TS_, tt_out);
		}
		
		cluster::Status::PrintMessage(false, "NablaEstimator: preparing output of curl in MVA\n");
		if (!nabla.curl_TS_MVA_.is_nil())
		{
			QuSetFrame((QdObject_var) nabla.curl_TS_MVA_, frame.c_str());
			QuSetTxtAttr(UNITS,curlunit,(QdObject_var) nabla.curl_TS_MVA_);
			QuSetTxtAttr("LABLAXIS", curl_lablaxis.c_str(), (QdObject_var) nabla.curl_TS_MVA_);
			QuSetTxtAttr("SI_conversion", si_convers_curl.c_str(), nabla.curl_TS_MVA_);
			set_timetags((QdObject_var) nabla.curl_TS_MVA_, tt_out);
		}
		
		cluster::Status::PrintMessage(false, "NablaEstimator: preparing output of barycenter\n");
		if (!nabla.bary_vector_TS_.is_nil())
		{
			QuSetTxtAttr(UNITS, QuGetAttrText(UNITS, (*call_list)[0]).c_str(),
					(QdObject_var) nabla.bary_vector_TS_);
			QuSetTxtAttr(SI_CONVERSION, QuGetAttrText(SI_CONVERSION, (*call_list)[0]).c_str(),
					(QdObject_var) nabla.bary_vector_TS_);
			QuSetTxtAttr(LABLAXIS, "Field in barycenter", (QdObject_var) nabla.bary_vector_TS_);
			QuSetFrame((QdObject_var) nabla.bary_vector_TS_, frame.c_str());
			set_timetags((QdObject_var) nabla.bary_vector_TS_, tt_out);
		}
		if (!nabla.bary_scalar_TS_.is_nil())
		{
			QuSetTxtAttr(UNITS, QuGetAttrText(UNITS, (*call_list)[0]).c_str(),
					(QdObject_var) nabla.bary_scalar_TS_);
			QuSetTxtAttr(SI_CONVERSION, QuGetAttrText(SI_CONVERSION, (*call_list)[0]).c_str(),
					(QdObject_var) nabla.bary_scalar_TS_);
			QuSetTxtAttr(LABLAXIS, "Field in barycenter", (QdObject_var) nabla.bary_scalar_TS_);
			set_timetags((QdObject_var) nabla.bary_scalar_TS_, tt_out);
		}

		cluster::Status::PrintMessage(false, "NablaEstimator: preparing output of barycenter position\n");
		if (!nabla.bary_coordinates_TS_.is_nil())
		{
			QuSetTxtAttr(UNITS, QuGetAttrText(UNITS, (*call_list)[4]).c_str(),
					(QdObject_var) nabla.bary_coordinates_TS_);
			QuSetTxtAttr(SI_CONVERSION, QuGetAttrText(SI_CONVERSION, (*call_list)[4]).c_str(),
					(QdObject_var) nabla.bary_coordinates_TS_);
			QuSetTxtAttr(LABLAXIS, "Barycenter position", (QdObject_var) nabla.bary_coordinates_TS_);
			QuSetFrame((QdObject_var) nabla.bary_coordinates_TS_, frame_P.c_str());
			set_timetags((QdObject_var) nabla.bary_coordinates_TS_, tt_out);
		}
		
		cluster::Status::PrintMessage(false, "NablaEstimator: preparing output of divergence\n");
		if (!nabla.div_TS_.is_nil())
		{
			QuSetFrame((QdObject_var) nabla.div_TS_, "scalar>na");
			QuSetTxtAttr(UNITS,divunit,(QdObject_var) nabla.div_TS_);
			QuSetTxtAttr("LABLAXIS","Div V",(QdObject_var) nabla.div_TS_);
			QuSetTxtAttr(SI_CONVERSION, si_convers.c_str(), nabla.div_TS_);
			set_timetags((QdObject_var) nabla.div_TS_, tt_out);
		}
		
		cluster::Status::PrintMessage(false, "NablaEstimator: preparing output of volume/surface \n");
		if (!nabla.vol_TS_.is_nil())
		{
			string strunit;
			string str_si_conv;
			if (nr_spacecrafts == 3){
				QuSetTxtAttr(LABLAXIS,"TriangleSurface",(QdObject_var) nabla.vol_TS_);
				strunit = QarUNITS_product((*call_list)[6], (*call_list)[6]);
				str_si_conv = QarSIConv_product((*call_list)[6], (*call_list)[6]);
			}
			else{
				QuSetTxtAttr(LABLAXIS,"TetraVolume",(QdObject_var) nabla.vol_TS_);
				strunit = QarUNITS_product((*call_list)[6], (*call_list)[6]);
				str_si_conv = QarSIConv_product((*call_list)[6], (*call_list)[6]);
				QdObject_var km2_ptr = (*call_list)[4]->create(); // this is empty as we only need metadata
				QuSetTxtAttr(UNITS,strunit.c_str(), km2_ptr);
				QuSetTxtAttr(SI_CONVERSION, str_si_conv.c_str(), km2_ptr);
				strunit = QarUNITS_product((*call_list)[6], km2_ptr);  //km^3
				str_si_conv = QarSIConv_product((*call_list)[6], km2_ptr);  //km^3
			}			
			QuSetFrame((QdObject_var) nabla.vol_TS_, "scalar>na");
			QuSetTxtAttr(UNITS,strunit.c_str(),(QdObject_var) nabla.vol_TS_);
			QuSetTxtAttr(SI_CONVERSION, str_si_conv.c_str(), (QdObject_var) nabla.vol_TS_);
			set_timetags((QdObject_var) nabla.vol_TS_, tt_out);
			cluster::Status::PrintMessage(false, "NablaEstimator: cleaning after output of volume/surface \n");

		}

		cluster::Status::PrintMessage(false, "NablaEstimator: preparing output of normals \n");
		if (!nabla.normal_TS_.is_nil())
		{
			QuSetTxtAttr(UNITS, QuGetAttrText(UNITS, (*call_list)[6]).c_str(),
					(QdObject_var) nabla.normal_TS_);
			QuSetTxtAttr(SI_CONVERSION, QuGetAttrText(SI_CONVERSION, (*call_list)[6]).c_str(),
					(QdObject_var) nabla.normal_TS_);
			QuSetTxtAttr(LABLAXIS, "NormalToPlane", (QdObject_var) nabla.normal_TS_);
			set_timetags((QdObject_var) nabla.normal_TS_, tt_out);
			QuSetFrame((QdObject_var) nabla.normal_TS_, frame_P.c_str());
		}
		
		cluster::Status::PrintMessage(false, "NablaEstimator: preparing output of velocity in MVA \n");
		if (!nabla.velo_TS_MVA_.is_nil())
		{
			///\todo Add SI_CONVERSION
			if(iscurrent) QuSetTxtAttr(UNITS,"km/s",(QdObject_var) nabla.velo_TS_MVA_);
			QuSetTxtAttr(LABLAXIS,"J",(QdObject_var) nabla.velo_TS_MVA_);
			set_timetags((QdObject_var) nabla.velo_TS_MVA_, tt_out);
			QuSetFrame((QdObject_var) nabla.velo_TS_MVA_, frame.c_str());
		}
	}

	if (status == QPLUG_SUCCESS) {
		cluster::Status::PrintMessage(false, "NablaEstimator: output to QSAS.\n");
		call_list->push_back((QdObject_var) nabla.grad_TS_);
		call_list->push_back((QdObject_var) nabla.grad_magn_TS_);
		call_list->push_back((QdObject_var) nabla.curl_TS_);
		call_list->push_back((QdObject_var) nabla.curl_TS_MVA_);
		call_list->push_back((QdObject_var) nabla.rotation_matrix_);
		if (!nabla.bary_vector_TS_.is_nil())
			call_list->push_back((QdObject_var) nabla.bary_vector_TS_);
		else
			call_list->push_back((QdObject_var) nabla.bary_scalar_TS_);
		call_list->push_back((QdObject_var) nabla.bary_coordinates_TS_);
		call_list->push_back((QdObject_var) nabla.div_TS_);
		call_list->push_back((QdObject_var) nabla.vol_TS_);
		call_list->push_back((QdObject_var) nabla.normal_TS_);
	}

	return status;
}

// -------------------------------------------------------------

int getVectorFieldOneSatellite(QplugArgList *call_list, QdObjectSeq_var &joined, const bool is_Bfield,
															  const unsigned short index, QdRMatrixSeq_var &V1)
{
	bool result = 0; //success

	QdRMatrixSeq_var V1j = QdRMatrixSeq_var::narrow(joined[index]);     // V*j = joined vector
	if(V1j.is_nil()) {
		QplugAppendTextDisplay("Input is not vector sequence");
		return 1;
	}
	
	if(is_Bfield){
		V1 = QdRMatrixSeq_var::narrow(QarConvertUnitsTo(V1j, "1.0e-9>T")); // ensure units are in Cluster units assumed
	}
	else{
		V1 = (QdRMatrixSeq_var) V1j;                                    // V* = vector TS to use
	}

	return result;
}

int getScalarFieldOneSatellite(QplugArgList *call_list, QdObjectSeq_var &joined, const bool is_Bfield,
															  const unsigned short index, QdRScalarSeq_var &V1)
{
	bool result = 0; //success

	QdRScalarSeq_var V1j = QdRScalarSeq_var::narrow(joined[index]);     // V*j = joined vector
	if(V1j.is_nil()) {
		QplugAppendTextDisplay("Input is not a scalar sequence");
		return 1;
	}
	if(is_Bfield){
		V1 = QdRScalarSeq_var::narrow(QarConvertUnitsTo(V1j, "1.0e-9>T")); // ensure units are in Cluster units assumed
	}
	else{
		V1 = (QdRScalarSeq_var) V1j;                                    // V* = vector TS to use
	}
	return result;
}

bool getFieldFromAllSatellites(QplugArgList *call_list, QdObjectSeq_var &joined, 
							 const bool is_Bfield, const bool scalars, const unsigned short nr_spacecrafts,
							 cluster::ClusterAnalysis &nabla)
{
	bool result = true;

	// --- SEH 2 May 2005 : set this to 1.0 to remove offsets in e.g. E-fields
	double offsremove = 0.00;


	// ---- get Vector quantity for given time interval - store in nabla.field_TS_ ----
	int nestsize;
	if (scalars)
	{
		QdRScalarSeq_var V1;
		for (unsigned short satellite=0; satellite<nr_spacecrafts; satellite++)
		{
			if (getScalarFieldOneSatellite(call_list, joined, is_Bfield, satellite, V1)>0)
				return false;
			if (satellite==0)
			{
				nestsize = V1->sequence_size();
				nabla.field_TS_ = new QdRMatrixSeq(nestsize, RMatrix(Dim(nr_spacecrafts,1)));
			}

			for (int t=0; t<nestsize; t++) {
				(*nabla.field_TS_)[t][satellite][0] = (double)(*V1)[t] - ( offsremove * (double)(*V1)[0]);
			}
		}
		nabla.field_TS_->copy_xrefs_from(V1);

	}
	else
	{
		QdRMatrixSeq_var V1;
		for (unsigned short satellite=0; satellite<nr_spacecrafts; satellite++)
		{
			if (getVectorFieldOneSatellite(call_list, joined, is_Bfield, satellite, V1)>0)
				return false;
			if (satellite==0)
			{
				nestsize = V1->sequence_size();
				nabla.field_TS_ = new QdRMatrixSeq(nestsize, RMatrix(Dim(nr_spacecrafts,3)));
			}

			for (int t=0; t<nestsize; t++) {
				for (int i=0; i<3; i++) {
					(*nabla.field_TS_)[t][satellite][i] = (double)(*V1)[t][i] - ( offsremove * (double)(*V1)[0][i]) ;
				}
			}
		}
		nabla.field_TS_->copy_xrefs_from(V1);
	}
	return result;
}

int getPositionOneSatellite(QplugArgList *call_list, QdObjectSeq_var &joined, const unsigned short index, QdRMatrixSeq_var &Pos1)
{
	bool result = 0; //success
	unsigned short index_cor = index+4;

	QdRMatrixSeq_var P1j = QdRMatrixSeq_var::narrow(joined[index_cor]);     // P*j = joined position vector
	if(P1j.is_nil()) {
		QplugAppendTextDisplay("Position is not vector sequence");
		return false;
	}
	
	Pos1 = QdRMatrixSeq_var::narrow(QarConvertUnitsTo(P1j, "1.0e3>m"));       // Pos* = position vector for this time interval in Cluster units
	if(Pos1.is_nil()){
		QplugAppendTextDisplay("Position 1 is not in units of distance");
		return false;
	}
	return result;
}

bool getPositionFromAllSatellites(QplugArgList *call_list, QdObjectSeq_var &joined, 
								 const unsigned short nr_spacecrafts, cluster::ClusterAnalysis &nabla)
{
	bool result=true;

	QdRMatrixSeq_var Pos1;
	int nestsize;
	for (unsigned short satellite=0; satellite<nr_spacecrafts; satellite++)
	{
		if (getPositionOneSatellite(call_list, joined,  satellite, Pos1)>0)
			return false;
		if (satellite==0)
		{
			nestsize = Pos1->sequence_size();
			nabla.position_TS_ = new QdRMatrixSeq(nestsize, RMatrix(Dim(nr_spacecrafts,3)));
		}

		for (int t=0; t<nestsize; t++) {
			for (int i=0; i<3; i++) {
				(*nabla.position_TS_)[t][satellite][i] = (double)(*Pos1)[t][i];
			}
		}
	}
	return result;
}

