/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/
/*************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT:    Plug In library
* 
* MODULE:	FFT.cc
*
* LANGUAGE:	C++
*
* FUNCTION:     QSAS FFT wrapper
*
* PURPOSE:      Wrapper for Fast Fourier Transform.
* 
* DESCRIPTION:  Plug In module checking input parameters and
*               constructing FFT arrays.
*
* ARGUMENTS:
*    IN:      QdCDOList  list of plug in objects containing
*              see FFT.qtpl file
*   OUT:
*
* GLOBALS:
*
* RETURN:
*    TYPE:    QplugReturnStatus
*    VALUES:
*
* FUNCTIONS_USED:	four1.c cdft.c rdft.c
*
* ALGORITHM:  see Press et al. 'Numerical Recipes in C' (1992)
*
* HISTORY: 1.0 created  20/03/2001 Markus Fraenz 
*          1.1 size of vector output corrected 31/07/2001 wjmf
*
***********************************************************************/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "qplug_if.h"
#include "qdutil.h"
#include "Qdos.h"
#include "qar.h"

extern "C" void four1(double data[], unsigned long nn, int isign);
extern "C" void cdft(int n, int isgn, double *a, int *ip, double *w);
extern "C" void rdft(int n, int isgn, double *a, int *ip, double *w);

extern "C" QplugReturnStatus FFTStart(QplugArgList *);

using namespace QSAS;

QplugReturnStatus FFTStart( QplugArgList *call_list )
{

// define some C variables

int	        attribute_ok,list_length;
unsigned long	Ndata,Ndata2=2,Ndata4,i,k,np,nm,Ndatah;
int             method,*ip_tak,inverseset=1,imfac;
bool            imset=FALSE,vectorset=FALSE;
int qar_status;

string attr_ptr;
char str_out[150];
double  *data,deltat,T,*w_tak;

// define QSAS data objects

QdTimeInterval_var start_end;
QdObject_var ts_out,ts_outi;
QdTimeSeq_var tt_out,tt_outi;

QdRScalarSeq_var  ts_real,ts_im;
QdRScalarSeq_var  out_freq_series;
QdRScalarSeq_var  out_remag_series,out_immag_series;
QdObject_var  out_revec_series,out_imvec_series;
QdRMatrixSeq_var ts_realv,ts_imv;



//
// ...first, unpack the input data objects
//

QplugAppendTextDisplay("FFT Plugin 1.1 31 Jul 2001\n");

// ...first item is input Time Series

QdObject_var ts_in = (* call_list)[0];

//  check QSAS datatype

QdTimeSeq_var tt_in = get_timetags(ts_in);
if( tt_in.is_nil() ) {
  QplugAppendTextDisplay ("FFT>> Failed to get time tags\n");
  return QPLUG_FAILURE;
}
if ( !QuIsTSRegular(tt_in, &deltat) ) {
	QplugAppendTextDisplay ("FFT>> ERROR - time series not evenly spaced.\n");
	return QPLUG_FAILURE; 
}
vectorset = QuIsThreeVector(ts_in);

// ...second item is imaginary part

QdObject_var ts_ini = (* call_list)[1];
QdTimeSeq_var tt_ini = get_timetags(ts_ini);
imset = !tt_ini.is_nil();

//  check QSAS datatype
if (imset) {  
 if(QarTestConformal(ts_in, ts_ini) == QAR_ERR) {
	QplugAppendTextDisplay ("FFT>> ERROR - Imaginary Input TS not conformal with real TS.\n");
	return QPLUG_FAILURE; 
 }
 if(vectorset && !QuIsThreeVector(ts_ini)) {
	QplugAppendTextDisplay ("FFT>> ERROR - Imaginary Input TS not vector.\n");
	return QPLUG_FAILURE; 
 }
}  


//	..3rd item is option 'time_interval'

 QdObject_var dobj1 = (* call_list)[2];
 start_end = get_timeinterval(dobj1);
 
//	..4th item is option 'Method'
method = (int) *QdRScalar_var::narrow( (* call_list)[3]);


// now take the subset for the time interval chosen  ###SUBSETTING START

QplugAppendTextDisplay ("FFT>> Subsetting Time Series...\n");
SequenceMask msk(tt_in->get_mask(QdTimeSeq::InRange(start_end->start(), start_end->end())));
 
tt_out = new QdTimeSeq();
  
tt_in->get_subsequence(*tt_out, msk);

  

if(!vectorset ){
    QdRScalarSeq_var ss = QdRScalarSeq_var::narrow(ts_in);
    QdRScalarSeq_var  ss_out = new QdRScalarSeq();
    ss->get_subsequence(*ss_out, msk);
    ts_out = (QdObject_var) ss_out;
}
else{
    QdRMatrixSeq_var ms = QdRMatrixSeq_var::narrow(ts_in);
    if(ms.is_nil() ){
      QplugAppendTextDisplay ("FFT>> ERROR - sequence type not handled\n");
      return QPLUG_FAILURE; 
    }
    if(QuIsThreeVector(ms) ){
      QdRMatrixSeq_var  ms_out = new QdRMatrixSeq();
      ms->get_subsequence(*ms_out, msk);
      ts_out = (QdObject_var) ms_out;
    }
    else {
      QplugAppendTextDisplay ("FFT>> ERROR - sequence type not handled\n");
      return QPLUG_FAILURE; 
    }
}
 

if ( ts_out.is_nil() ){
    QplugAppendTextDisplay ("FFT>> ERROR - subset failed:\n");
    return QPLUG_FAILURE;
} 
Ndata = ts_out->sequence_size();

if ( Ndata < 2  ){
    QplugAppendTextDisplay ("FFT>> Less than 2 elements in subset.\n");
    return QPLUG_FAILURE;
}

ts_in->copy_xrefs_to(ts_out);
set_timetags(ts_out, tt_out);

// do imaginary part
if (imset) {
  if (vectorset) {
    QdRMatrixSeq_var msi = QdRMatrixSeq_var::narrow(ts_ini);
    QdRMatrixSeq_var  ms_outi = new QdRMatrixSeq();
    msi->get_subsequence(*ms_outi, msk);
    ts_outi = (QdObject_var) ms_outi;
  }
  else {
    QdRScalarSeq_var ssi = QdRScalarSeq_var::narrow(ts_ini);
    QdRScalarSeq_var  ss_outi = new QdRScalarSeq();
    ssi->get_subsequence(*ss_outi, msk);
    ts_outi = (QdObject_var) ss_outi;
  }
}


//	..extract the data from the data series object

if ( vectorset ) {
  QplugAppendTextDisplay ("FFT>> Decomposing Vector Timeseries...\n");
  ts_real = QdRScalarSeq_var::narrow(QuGetVectorComponent(ts_out, "r(mag)"));
  ts_realv = QdRMatrixSeq_var::narrow(ts_out);
  if (imset)  {
    ts_im = QdRScalarSeq_var::narrow(QuGetVectorComponent(ts_outi, "r(mag)"));
    ts_imv = QdRMatrixSeq_var::narrow(ts_outi);
  }
}
else {
 ts_real = QdRScalarSeq_var::narrow(ts_out);
 if (imset)  ts_im = QdRScalarSeq_var::narrow(ts_outi);
}


//	..extract the time tags from the data series object ###SUBSETTING END



if (Ndata <= 1) {
	QplugAppendTextDisplay ("FFT>> ERROR - Series must have at least 2 data points.\n");
	return QPLUG_FAILURE; }

//      (notice that NR routines start indices with 1!!) 
if (imset) strcpy(str_out,"FFT>> complex FFT ");
else strcpy(str_out,"FFT>> real FFT ");
if (method > 0) strcat(str_out,"using Ooura CDFT8G.\n");
else strcat(str_out,"using NR four1.\n");
QplugAppendTextDisplay (str_out);

if (imset || (method == 0)) imfac=2;
else imfac=1;

// zero - padding :
while (Ndata2 < Ndata) Ndata2 *= 2;
Ndata4=imfac*Ndata2;
Ndatah=Ndata2/2;

data    = new double[Ndata4]; 
for (i=0; i<Ndata4; i++) data[i] = 0.0;

for (i=0; i<Ndata; i++) data[imfac*i] =  (double) ts_real[i];
if (imset) for (i=0; i<Ndata; i++) data[2*i+1] =  (double) ts_im[i];

// four1 still uses Fortran array indices

if (method > 0) { // initialize Tak data arrays
 ip_tak = new int[(long) sqrt((float) Ndatah*imfac) +3];
 ip_tak[0]=0;
 w_tak = new double[Ndatah*imfac];
 if (imset) cdft((int) Ndata4, inverseset, data, ip_tak, w_tak);
 else rdft((int) Ndata4, inverseset, data, ip_tak, w_tak);
}
else four1(data-1, Ndata2, inverseset);

//	..create the output data series objects
// we want the output sorted by frequency, so we have to resort
// according to NRp508.

out_freq_series  = new QdRScalarSeq(Ndatah*imfac);
out_remag_series = new QdRScalarSeq(Ndatah*imfac);
out_immag_series = new QdRScalarSeq(Ndatah*imfac);


T  = (double)Ndata2 * deltat;

if (imset||(method==0)){
 for (i =0; i<Ndatah; i++) {
  np=Ndatah+i;
  out_freq_series[np] =  (double)i / T;
  out_remag_series[np] = (double) data[2*i];
  out_immag_series[np] = (double) data[2*i+1];
  out_freq_series[i] = -(double)(Ndatah-i) / T; 
  out_remag_series[i] = (double) data[2*i+Ndata2];
  out_immag_series[i] = (double) data[2*i+1+Ndata2];
 }
}
else {
 for (i =0; i<Ndatah; i++) {
  out_freq_series[i] =  (double)i / T;
  out_remag_series[i] = (double) data[2*i];
  out_immag_series[i] = (double) data[2*i+1];
 }
}



for (i =0; i<Ndata2; i++) {
}

if(vectorset) { 
 QdRMatrixSeq_var revec_series = new QdRMatrixSeq(Ndatah*imfac,RMatrix(Dim(3)));
 QdRMatrixSeq_var imvec_series = new QdRMatrixSeq(Ndatah*imfac,RMatrix(Dim(3)));
 for (k=0;k<3;k++) {
 for (i=0; i<Ndata4; i++) data[i] = 0.0;
 for(i=0; i<Ndata; i++)  data[imfac*i] = (double) (ts_realv[i][k]);
 if (imset) for(i=1; i<Ndata; i++)  data[2*i+1] = (double) (ts_imv[i][k]);
 if (method > 0) {
  if (imset) cdft((int) Ndata4, inverseset, data, ip_tak, w_tak);
  else rdft((int) Ndata4, inverseset, data, ip_tak, w_tak);
 }
 else four1(data-1, Ndata2, inverseset);
 if (imset|| (method==0)) {
  for (i =0; i<Ndatah; i++) {
   np=Ndatah+i;
   revec_series[np][k] = (double) data[2*i];
   revec_series[i][k] = (double) data[2*i+Ndata2];
   imvec_series[np][k] = (double) data[2*i+1];
   imvec_series[i][k] = (double) data[2*i+1+Ndata2];
  }
 }
 else {
  for (i =0; i<Ndatah; i++) {
   revec_series[i][k] = (double) data[2*i];
   imvec_series[i][k] = (double) data[2*i+1];
  }
 }
 }

 out_revec_series = (QdObject_var) revec_series;
 out_imvec_series = (QdObject_var) imvec_series;
}




//	...set some attributes

 // SI_conversion
   attr_ptr = QarSIConv_inverse(ts_in);
   if( !attr_ptr.empty() ){
     QuSetTxtAttr("SI_conversion", attr_ptr.c_str(), out_remag_series);
     QuSetTxtAttr("SI_conversion", attr_ptr.c_str(), out_immag_series);
     if(vectorset) {
       QuSetTxtAttr("SI_conversion", attr_ptr.c_str(), out_revec_series);
       QuSetTxtAttr("SI_conversion", attr_ptr.c_str(), out_imvec_series);
     }
   }

  sprintf(str_out,"ReFFT(%s)", QuGetAttrText("LABLAXIS", ts_in).c_str());
  QuSetTxtAttr("LABLAXIS", str_out, (QdObject_var) out_remag_series);
  sprintf(str_out,"ReFFT(%s)", QuGetAttrText("FIELDNAM", ts_in).c_str());
  QuSetTxtAttr("FIELDNAM", str_out, (QdObject_var) out_remag_series);
  QuSetFrame((QdObject_var) out_remag_series, "scalar");
  sprintf(str_out,"%s", QuGetAttrText("UNITS", ts_in).c_str());
  QuSetTxtAttr("UNITS", str_out, (QdObject_var) out_remag_series);
  out_remag_series->set_xref("DEPEND_0", out_freq_series);
  
  sprintf(str_out,"ImFFT(%s)", QuGetAttrText("LABLAXIS", ts_in).c_str());
  QuSetTxtAttr("LABLAXIS", str_out, (QdObject_var) out_immag_series);
  sprintf(str_out,"ImFFT(%s)", QuGetAttrText("FIELDNAM", ts_in).c_str());
  QuSetTxtAttr("FIELDNAM", str_out, (QdObject_var) out_immag_series);
  QuSetFrame((QdObject_var) out_remag_series, "scalar");
  sprintf(str_out,"%s", QuGetAttrText("UNITS", ts_in).c_str());
  QuSetTxtAttr("UNITS", str_out, (QdObject_var) out_immag_series);
  out_immag_series->set_xref("DEPEND_0", out_freq_series);

  if(vectorset) {
    sprintf(str_out,"ReFFT(%s)", QuGetAttrText("LABLAXIS", ts_in).c_str());
    QuSetTxtAttr("LABLAXIS", str_out, (QdObject_var) out_revec_series);
    sprintf(str_out,"ReFFT(%s)", QuGetAttrText("FIELDNAM", ts_in).c_str());
    QuSetTxtAttr("FIELDNAM", str_out, (QdObject_var) out_revec_series);
    QuSetFrame( (QdObject_var) out_revec_series, QuGetFrameAttr( ts_in).c_str() );
    sprintf(str_out,"%s", QuGetAttrText("UNITS", ts_in).c_str());
    QuSetTxtAttr("UNITS", str_out, (QdObject_var) out_revec_series);
    out_revec_series->set_xref("DEPEND_0", out_freq_series);

    sprintf(str_out,"ImFFT(%s)", QuGetAttrText("LABLAXIS", ts_in).c_str());
    QuSetTxtAttr("LABLAXIS", str_out, (QdObject_var) out_imvec_series);
    sprintf(str_out,"ImFFT(%s)", QuGetAttrText("FIELDNAM", ts_in).c_str());
    QuSetTxtAttr("FIELDNAM", str_out, (QdObject_var) out_imvec_series);
    QuSetFrame((QdObject_var) out_imvec_series,QuGetFrameAttr( ts_in).c_str() );
    sprintf(str_out,"%s", QuGetAttrText("UNITS", ts_in).c_str());
    QuSetTxtAttr("UNITS", str_out, (QdObject_var) out_imvec_series);
    out_imvec_series->set_xref("DEPEND_0", out_freq_series);
  }


  QuSetTxtAttr("SI_conversion", " ", (QdObject_var) out_freq_series);
  QuSetTxtAttr("LABLAXIS", "frequency", (QdObject_var) out_freq_series);
  QuSetTxtAttr("FIELDNAM", "frequency", (QdObject_var) out_freq_series);
  QuSetFrame( (QdObject_var) out_freq_series, "scalar");
  QuSetTxtAttr("UNITS", "Hz", (QdObject_var) out_freq_series);


//	..add them to the list

  call_list->push_back( (QdObject_var) out_remag_series);     
  call_list->push_back( (QdObject_var) out_immag_series);     
  call_list->push_back( (QdObject_var) out_revec_series);     
  call_list->push_back( (QdObject_var) out_imvec_series);     


//	..some messages to the analysis interface window..

sprintf(str_out, "FFT>> %ld data points\n", Ndata);
QplugAppendTextDisplay (str_out);
sprintf(str_out, "FFT>> %ld zero padded data points\n", Ndata2);
QplugAppendTextDisplay (str_out);
sprintf(str_out,"FFT>> %ld frequencies\n", Ndata2);
QplugAppendTextDisplay (str_out);
QplugAppendTextDisplay ("FFT>> finished...\n");


//	..clean up and exit

delete[] data;
if (method > 0) { 
 delete ip_tak;
 delete w_tak;
}

return QPLUG_SUCCESS;


}



