// -----------------------------------------------------------------------------------------
//
// Calculate Projection matrix for constrained MVA
//
// SEH, Sep 12, 2003
// 
// Inputs :
//     B = magnetic field   (n x 3)  for Bn=0 constrain
//     V = plasma velocity  (n x 3) 
//     U = MP velocity vector  (3 comp) for (<V> - U)n = 0 constrain
//     V_alfven = alfven velocity (3 comp)
//     alpha = Walen regression slope form HT analysis
//     Evec = eigenvectormatrix from MVAB for (n perp to Evec_MVAB_max) constrain
//     nestsize = number of points in data sets
//     constrain = 1 for Bn=0, 2 for (<V> - U)n etc - see below.
//
// SEH, 12 Sep 2003
//      23 OCt 2005, tested <Valf> = <B> / sqrt(u_0 <rho>)
//
// -----------------------------------------------------------------------------------------


#include "qdutil.h"
#include "qplug_if.h"
#include "GenericResidue.h"
#include "Qdos.h"

QdRMatrix_var Projection(QdRMatrixSeq_var B, QdRMatrixSeq_var V, 
			 QdRMatrix_var U, QdRMatrixSeq_var V_Alfven,
			 QdRMatrix_var Evec, double alpha, int nestsize, int constrain, const double density) 
{  
  int i,j,k;
  double m_p= 1.67e-27;      // rho = mass dens with proton_mass m_p
  double mu_0 = 1.25664e-06;     // modified mu_0 (due to B in nT and dens in 1/ccm)
  double normfac = (double) nestsize;
  QdRMatrix_var C =  new QdRMatrix(Dim(3,3), 0.0);                // C = projection matrix for constrain
  QdRMatrix_var A  = new QdRMatrix(Dim(3), 0.0);
  QdRMatrix_var ave_B  = new QdRMatrix(Dim(3), 0.0);
  QdRMatrix_var ave_V  = new QdRMatrix(Dim(3), 0.0);
  QdRMatrix_var ave_Valf  = new QdRMatrix(Dim(3), 0.0);

  bool doB = !B.is_nil() && ( B->sequence_size() > 0 );
  bool doV = !V.is_nil() && ( V->sequence_size() > 0 );
  bool doVA = !V_Alfven.is_nil() && ( V_Alfven->sequence_size() > 0 );
  
//   if (doB) printf("B size = %d\n",B->sequence_size());
//   if (doV) printf("V size = %d\n",V->sequence_size());
//   if (doVA) printf("VAB size = %d\n",V_Alfven->sequence_size());

//   printf("in Projection...%d %d %d \n",doB, doV, doVA);


  try {
    for (k=0; k<nestsize; k++) {
      for (i=0; i<3; i++) {
	if (doB) (*ave_B)[i] += (*B)[k][i];
	if (doV) (*ave_V)[i] += (*V)[k][i];
	//      (*ave_Valf)[i] += alpha*(*V_Alfven)[k][i];
	if (density > 0.00 && doV && doB && doVA) {
	  (*ave_Valf)[i] += alpha*1.0e-3 * (double) (*B)[k][i] * 1e-9 / sqrt(mu_0 * density *1e06 * m_p);
	} else  if (doVA) (*ave_Valf)[i] += alpha*(*V_Alfven)[k][i];
      }
    }
    
    for (i=0; i<3; i++) {
      (*ave_B)[i] /= normfac;
      if (doV) (*ave_V)[i] /= normfac;
      if (doVA) (*ave_Valf)[i] /= normfac;
    } 
  }
  catch(Exception& e) {
    printf("Exception in Nested Generic, initial MVAB\n");
	fflush(stdout);
    return NULL;
  }



  //     printf("in Constraints.. \n<B> = [%10.2f  %10.2f  %10.2f]\n",(double)(*ave_B)[0],(double)(*ave_B)[1],(double)(*ave_B)[2]);
//   printf("<V>  = [%10.2f  %10.2f  %10.2f]\n",(double)(*ave_V)[0],(double)(*ave_V)[1],(double)(*ave_V)[2]);
//   printf("<Va> = [%10.2f  %10.2f  %10.2f]\n",(double)(*ave_Valf)[0],(double)(*ave_Valf)[1],(double)(*ave_Valf)[2]);
//   printf("<U>  = [%10.2f  %10.2f  %10.2f]\n",(double)(*U)[0],(double)(*U)[1],(double)(*U)[2]);

  switch(constrain) {
  case 1:
    for (i=0; i<3; i++) (*A)[i] = (*ave_B)[i];
    break;
  case 2:
    for (i=0; i<3; i++) {
      (*A)[i] = 1000.0*(double)(*ave_V)[i] - (double)(*U)[i];
      printf("U,V = %f    %f\n",(double)(*U)[i], (double)(*ave_V)[i]);
	  fflush(stdout);
    }
    break;
  case 3:
    for (i=0; i<3; i++) {
//       printf("Va, U,V = %10.3f  %10.3f    %10.3f\n",(double)(*ave_Valf)[i],(double)(*U)[i], (double)(*ave_V)[i]);
      (*A)[i] = -1000.0*(double)(*ave_Valf)[i] + 1000.0*(double)(*ave_V)[i] - (double)(*U)[i];
    }
    break;
  case 4:
//     for (i=0; i<3; i++) (*VC1)[i] = (double)(*VC)[i] - (double)(*V_ave)[i];
//     sz2 = sqrt(vecsum2((double)(*VC1)[0],(double)(*VC1)[1],(double)(*VC1)[2]));
//     for (i=0; i<3; i++) (*VC1)[i] /= sz2;
//     for (i=0; i<3; i++) (*VC2)[i] = (double)(*B_ave)[i];
//     sz2 = sqrt(vecsum2((double)(*VC2)[0],(double)(*VC2)[1],(double)(*VC2)[2]));
//     for (i=0; i<3; i++) (*A)[i] = (double)(*VC1)[i] + (double)(*VC2)[i]/sz2;
    break;
  case 5 :
    for (i=0; i<3; i++) (*A)[i] = (*Evec)[2][i];
    break;
  default:
    for (i=0; i<3; i++) (*A)[i] = 0.0;
    break;
  }
    
  // ---- projection matrix for constrain  ------------
  
  double sz2 = vecsum2(A);

  
  (*C)[0][0] = 1.0 - ((*A)[0]*(*A)[0]) / sz2;
  (*C)[0][1] = - ((*A)[0]*(*A)[1]) / sz2;
  (*C)[0][2] = - ((*A)[0]*(*A)[2]) / sz2;
  
  (*C)[1][0] = - ((*A)[1]*(*A)[0]) / sz2;
  (*C)[1][1] = 1.0 - ((*A)[1]*(*A)[1]) / sz2;
  (*C)[1][2] = - ((*A)[1]*(*A)[2]) / sz2;
  
  (*C)[2][0] = - ((*A)[2]*(*A)[0]) / sz2;
  (*C)[2][1] = - ((*A)[2]*(*A)[1]) / sz2;
  (*C)[2][2] = 1.0 - ((*A)[2]*(*A)[2]) / sz2;

  return (C);
}

