
// -----------------------------------------------------------------------------------------
//
// Calculate Q-matrix for MLMR - minimum linear momentum flux residue
//
// SEH, Sep 5, 2003
//  
// Inputs :
//     B = magnetic field  (n x 3)     [nT]
//     V = plasma velocity (n x 3)     [km/s]
//     P = pressure tensor (n x 3 x 3) [eV/cm^3]
//     N = plasma density  (n)         [1/cm^3]  (proton mass used to get mass density)
//
// For simplification, all units are converted to SI units
//
// -----------------------------------------------------------------------------------------

#include "GenericResidue.h"
#include "Qdos.h"


// -----------------------------------------------------------------------------------------
//
// Returns the Q_matrix for MLMR (Minimum Linear Momentum Residue)
//
// SEH, 07 Sep, 2001
//      19 Sep, 2001, bug fixes - U returned
//      15 Oct, 2003, Doxygen DOC
//
// Doxygen info :
/*! 
  \brief Returns the covariance matrix for the Minimum Linear Momentum Residue 
  (MLMR) method.

  \param V - vector time series with plasma velocity in units of km/s
  \param B - vector time series with the magnetic field in units of nT
  \param P - matrix time series with the full 3x3 pressure tensor
  \param N - scalar time series with the plasma density in units of 1/cm3
  \param U - vector - holds the calculated velocity in units of km/s on successful return.
  \param nestsize - number of elements in V,B,P,E,N and H.

*/ 
// -----------------------------------------------------------------------------------------
QdRMatrix_var Q_MLMR_calc(QdRMatrixSeq_var V, QdRMatrixSeq_var B, QdRMatrixSeq_var P,
			 QdRScalarSeq_var N, QdRMatrix_var U, int nestsize)
{
  int i,j,k,m;
  double m_p = 1.67e-27;                     // mass of proton [kg] )we assume ions are == protons)
  double Mu0  = 12.566370614e-07;  
  double dirac=0.00;                         // holds the dirac delta function (1 for i==j, 0 else)
  double normfac = (double) nestsize;
  double btot,d_eta2=0.00,p_ave=0.00;
  bool PO = false;                          // PrintOut to screen ? 

  QdRMatrix_var    B_k = new QdRMatrix(Dim(3), 0.0);                    // holds the B(k) vector
  QdRMatrixSeq_var q_ij = new QdRMatrixSeq(nestsize,RMatrix(Dim(3,3)));
  QdRMatrix_var    ave_q_ij =  new QdRMatrix(Dim(3,3), 0.0);
  QdRMatrixSeq_var eta = new QdRMatrixSeq(nestsize,RMatrix(Dim(3)));
  QdRMatrix_var    ave_eta = new QdRMatrix(Dim(3), 0.00);
  QdRMatrix_var    Q =  new QdRMatrix(Dim(3,3), 0.0);

  p_ave = aveTotalPressure(B, P, nestsize);
  if (PO) printf("Q_MLMR > p_ave = %10.4g %f\n",p_ave,Mu0);
  if(PO) fflush(stdout);

  for (k=0; k<nestsize; k++) {
    //    if (k > 9) m_p = 2.0e-27;
    for (i=0; i<3; i++) (*B_k)[i] = (*B)[k][i];
    btot = vecsum2(B_k) * 1E-18 / (2.0 * Mu0);                         // 1E-18 -> to get SI units T^2
    if (PO)
	{
		printf("enter Q_MLMR.k = %d. \n",k);
		fflush(stdout);
	}
       
    for (i=0; i<3; i++) {                                              // loop over three components
      (*eta)[k][i] = m_p * (*N)[k] * (*V)[k][i] * 1E9;                 // 1E9 -> to get SI units kg/s m^2
      (*ave_eta)[i] += (*eta)[k][i];

      for (j=0; j<3; j++) {
        dirac = (double)(i==j);  
        (*q_ij)[k][i][j] = 
	  m_p * (*N)[k] * (*V)[k][i] * (*V)[k][j] * 1E12 -             // 1E12 -> V ion km/s, N in /cm^3
	  (*B)[k][i] *  (*B)[k][j] * 1E-18 / Mu0 +                     // 1E-18 -> to get B in T 
	  dirac * btot +                                               // only diagonal elements herer
       	  (*P)[k][i][j] * 1.60219E-13;                                 // 1.60219E-13 -> from eV/cm^3 to kg/ms^2
	  //	            p_ave - ( vecsum2(B_k) * 1E-18 / (2.0*Mu0));
	(*ave_q_ij)[i][j] += (*q_ij)[k][i][j];
      }
    }
  }
  
  // ----- get <q_ij> and <eta> and initialize U* --------

  if (PO)
  {
	printf("get <q_ij>.. \n");
	fflush(stdout);
  }
  for (i=0; i<3; i++) {
    for (j=0; j<3; j++) {
      (*ave_q_ij)[i][j] /= normfac;
    }
    (*ave_eta)[i] /= normfac;
    (*U)[i] = 0.0;
  }

  if (PO) printf("Q_MLMR:\nq_ij matrix [km/(m s^2)] : \n");
  if (PO) for (i=0; i<3; i++) {
    printf("<q_%dj>     = [%E  %E  %E];\n",i,(double)(*ave_q_ij)[i][0], (double)(*ave_q_ij)[i][1], (double)(*ave_q_ij)[i][2]);
	fflush(stdout);
  } 

  if (PO)
  {
	printf("\n<eta>      = [%E  %E  %E]  [kg/(m^2 s)];\n", (double)(*ave_eta)[0], (double)(*ave_eta)[1], (double)(*ave_eta)[2]);
	fflush(stdout);
  }


  // ---   get delta_q_ij = q_ij - <q_ij> --------

  for (k=0; k<nestsize; k++) {
    for (i=0; i<3; i++) {
      for (j=0; j<3; j++) {
	(*q_ij)[k][i][j] -= (*ave_q_ij)[i][j];
      }
      (*eta)[k][i] -= (*ave_eta)[i];
      d_eta2 += ((double) (*eta)[k][i] * (double) (*eta)[k][i]);
    }
  }

 
  //  printf("\n<|deta|^2> = [%E  %E  %E];\n", (double)(*d_eta2)[0], (double)(*d_eta2)[1], (double)(*d_eta2)[2]);
  if (PO)
  {
	printf("\n<|deta|^2> = %E  [kg^2/(m^2s^4)]\n", d_eta2);
	fflush(stdout);
  }



  // ---- construct U* vector ---------------

  for (k=0; k<nestsize; k++) {
    for (j=0; j<3; j++) {
      for (i=0; i<3; i++) {
	(*U)[j] += (*eta)[k][i] * (*q_ij)[k][i][j];
      }
    }
  }

  for (i=0; i<3; i++) (*U)[i] /= d_eta2;
  if (PO)
  {
	printf("\nU*         = [%E  %E  %E] [m/s];  (eq 12 in manuscript)\n\n", (double)(*U)[0], (double)(*U)[1], (double)(*U)[2]);
	fflush(stdout);
  }


  // --- construct Q_ij matrix --------------  

  for (k=0; k<nestsize; k++) {
    for (i=0; i<3; i++) {
      for (j=0; j<3; j++) {
	for (m=0; m<3; m++) {
	  (*Q)[i][j] += ((*q_ij)[k][m][i] - (*U)[i]*(*eta)[k][m]) * ((*q_ij)[k][m][j] - (*U)[j]*(*eta)[k][m]);
	}
      }
    }
  }

  for (i=0; i<3; i++) {
    for (j=0; j<3; j++) {
      (*Q)[i][j] /= normfac;
    }
  }

  if (PO) printf("Q_ij [km/(m s^2)] (eq 15 in manuscript)\n");
  if (PO) for (i=0; i<3; i++) {
    printf("Q_i%d       = [%E  %E  %E];\n",i, (double)(*Q)[i][0], (double)(*Q)[i][1], (double)(*Q)[i][2]);
	fflush(stdout);
  } 


  return (Q);

}

