
// -----------------------------------------------------------------------------------------
//
// Calculate Q-matrix for MTER - minimum total energy residue
//
// SEH, Sep 7, 2003
// 
// Inputs :
//     B = magnetic field   (n x 3)     [nT]
//     V = plasma velocity  (n x 3)     [km/s]
//     P = pressure tensor  (n x 3 x 3) [eV/cm^3]
//     N = plasma density   (n)         [1/cm^3]  (proton mass used to get mass density)
//     H = heat flux vector (n x 3)  
//
// For simplification, all units are converted to SI units
//
// -----------------------------------------------------------------------------------------

#include "qdutil.h"
#include "qplug_if.h"
#include "GenericResidue.h"
#include "Qdos.h"


// -----------------------------------------------------------------------------------------
//
// Returns the Q_matrix for MTER (Minimum Total Energy Residue)
//
// SEH, 07 Sep, 2001
//      19 Sep, 2001, bug fixes - U returned
//      15 Oct, 2003, Doxygen DOC
//      18 May, 2005, quality control through manual calc by BS
//
// Doxygen info :
/*! 
  \brief Returns the covariance matrix for the Minimum Total Energy Residue 
  (MTER) method.
.
  \param V - vector time series with plasma velocity in units of km/s
  \param B - vector time series with the magnetic field in units of nT
  \param P - matrix time series with the full 3x3 pressure tensor
  \param E - vector time series with the electric field in units of mV/m
  \param N - scalar time series with the plasma density in units of 1/cm3
  \param H - vector time series heat flux vector in units of erg/s cm3
  \param U - vector - holds the calculated velocity in units of km/s on successful return.
  \param nestsize - number of elements in V,B,P,E,N and H.

*/ 
// -----------------------------------------------------------------------------------------
QdRMatrix_var Q_MTER_calc(QdRMatrixSeq_var V, QdRMatrixSeq_var B, QdRMatrixSeq_var P,
			  QdRMatrixSeq_var E, QdRScalarSeq_var N, QdRMatrixSeq_var H, 
			  QdRMatrix_var U ,int nestsize)

{
  int i,j,k;
  double m_p = 1.67e-27;                     // mass of proton [kg] (we assume ions == protons)
  double Mu0  = 12.566370614e-07;  
  double normfac = (double) nestsize;
  double p,p_ave = 0.0, rhov2=0.0, d_eta2=0.00, ave_eta=0.00;
  bool PO = false;                          // PrintOut to screen ?

  QdRMatrix_var    E_k = new QdRMatrix(Dim(3), 0.0);                    // holds the E^2(k) vector
  QdRMatrix_var    V_k = new QdRMatrix(Dim(3), 0.0);                    // holds the V^2(k) vector
  QdRMatrix_var    B_k = new QdRMatrix(Dim(3), 0.0);                    // holds the B^2(k) vector
  QdRMatrix_var    v_P = new QdRMatrix(Dim(3), 0.0);                    // holds the v_P(k) vector
  QdRMatrix_var    H_k = new QdRMatrix(Dim(3), 0.0);                    // holds the heatflux vector
  QdRMatrix_var    S =   new QdRMatrix(Dim(3), 0.0);                    // holds the Poynting flux vector
  QdRScalarSeq_var eta = new QdRScalarSeq(nestsize);
  QdRMatrix_var    Q =   new QdRMatrix(Dim(3,3), 0.0);
  QdRMatrixSeq_var q_i = new QdRMatrixSeq(nestsize,RMatrix(Dim(3)));
  QdRMatrix_var    ave_q_i = new QdRMatrix(Dim(3), 0.0);

  p_ave = aveTotalPressure(B, P, nestsize);
  if (PO)
  {
	printf("Q_MTER > p_ave = %10.4g\n",p_ave);
	fflush(stdout);
  }

  if (PO) {
     printf(" k           V_ * P_ii/2                       V_ * rho V^2/2                     V_ dot P__ ");
     printf("                      Heat flux                     S (Poynting flux)\n");
     for (i=0; i<169; i++) printf("=");
     printf("\n");
	 fflush(stdout);
  }

  for (k=0; k<nestsize; k++) {
    //   if (k > 9) m_p = 2.0e-27;
    p = 0.00;
    for (i=0; i<3; i++) {
      (*V_k)[i] = (*V)[k][i] * 1.0E3;                                                         // V vector in [m/s]
      (*E_k)[i] = (*E)[k][i] * 1.0E-3;                                                        // E field in  [V/m]
      (*B_k)[i] = (*B)[k][i] * 1.0E-9;                                                        // B field vec [T]
      (*H_k)[i] = (*H)[k][i] * 1.0E-3;                                                        // heat flux in [J/m^2 s] = [kg /s^3]
      (*v_P)[i] = 0.00;                                                                       // initialize to 0.00
      for (j=0; j<3; j++) (*v_P)[i] += (*V_k)[j] * ((*P)[k][i][j] * 1.60219E-13);           // 1.60219E-13 for eV/cm3 to J
// Experiment  
//      for (j=0; j<3; j++) (*v_P)[i] += (*V_k)[j] * ( vecsum2(B_k)/(2.0*Mu0) - p_ave);         // 1.60219E-13 for eV/cm3 to J
      p += ((double)(*P)[k][i][i])/3.0 * 1.60219E-13;                                         // scalar pressure = trace(P)/3 [ kg/ms^2 ]
    }
    S = CrossVec(CrossVec(B_k,V_k),B_k);                                            // S = E x B == (-V x B) x B = (B x V) x B [V/m * nT] = [kg/s^3];
// experiment :
//    p = p_ave - (vecsum2(B_k)/(2.0*Mu0));
    rhov2 = m_p * (*N)[k] * 1E6 * vecsum2(V_k);                                     // 1E6 due to density [1/cm^3] -> [1/m^3]
    (*eta)[k] = (3.0*p + rhov2 + vecsum2(B_k)/Mu0) / 2.0;
    ave_eta += (double) (*eta)[k];
    
    for (i=0; i<3; i++) {
      (*q_i)[k][i] = (*V_k)[i] * (3.0*p/2.0 + rhov2/2.0) + (*v_P)[i] + (*H_k)[i] + (*S)[i]/Mu0;         // see eq. 37
      (*ave_q_i)[i] += (*q_i)[k][i];
    }
 
//    printf("%2d   %8.3e ", k+1,p);

    if (PO) {
     printf("%2d   %8.1e  %8.1e  %8.1e      ", k+1,3*p/2.0*(float)(*V_k)[0], 3*p/2.0*(float)(*V_k)[1], 3*p/2.0*(float)(*V_k)[2]);
     printf("%8.1e  %8.1e  %8.1e      ",rhov2/2.0*(float)(*V_k)[0],rhov2/2.0*(float)(*V_k)[1], rhov2/2.0*(float)(*V_k)[2]);
     printf("%8.1e  %8.1e  %8.1e      ",(float)(*v_P)[0], (float)(*v_P)[1], (float)(*v_P)[2]);
     printf("%8.1e  %8.1e  %8.1e      ",(float)(*H_k)[0], (float)(*H_k)[1], (float)(*H_k)[2]);
     printf("%8.1e  %8.1e  %8.1e\n",    (float)(*S)[0]/Mu0,(float)(*S)[1]/Mu0, (float)(*S)[2]/Mu0);
	 fflush(stdout);
    }
  }

  if (PO)
  {
	printf("\n\n");
	fflush(stdout);
  }

  if (PO) {
   printf(" k     eta                 q_i \n");
   for (i=0; i<44; i++) printf("=");
   printf("\n");
   for (k=0; k<nestsize; k++) {
     printf("%2d   %9.2e     %9.2e  %9.2e  %9.2e\n", k+1,(double)(*eta)[k], (double)(*q_i)[k][0], (double)(*q_i)[k][1],(double)(*q_i)[k][2]);
    }
	fflush(stdout);
  }

  if (PO)
  {
	printf("\n\n");
	fflush(stdout);
  }


  // ----- initialize U, get <q_i> and <eta>--------

  for (i=0; i<3; i++) {
    (*ave_q_i)[i] /= normfac;
    (*U)[i] = 0.0;
  }
  ave_eta /= normfac;

  if (PO) printf("Q_MTER:\nq_i vector [kg/s^3] : \n");
  if (PO) printf("<q_j>     = [%E  %E  %E]; \n",(double)(*ave_q_i)[0], (double)(*ave_q_i)[1], (double)(*ave_q_i)[2]);
  if (PO) printf("\n<eta>      = %E [kg/ms^2]\n", ave_eta);
  if (PO) fflush(stdout);

  // ---   get delta_q_i = q_i - <q_i> and delta_eta --------

  for (k=0; k<nestsize; k++) {
    for (i=0; i<3; i++) (*q_i)[k][i] -= (*ave_q_i)[i];
    (*eta)[k] -= ave_eta;
    d_eta2 += ((double) (*eta)[k] * (double) (*eta)[k]);
  }


  if (PO)
  {
	printf("\n<|deta|^2> = %E  [kg^2/(m^2s^4)]\n", d_eta2);
	fflush(stdout);
  }

  // ---- construct U* vector ---------------

  if (d_eta2 != 0.0) {
     for (k=0; k<nestsize; k++) {
       for (i=0; i<3; i++) (*U)[i] += (double)(*eta)[k] * (double)(*q_i)[k][i] / d_eta2;
      }  
  } else {
     printf("MTER: singularity - setting U* = <V_cis>\n");
	 fflush(stdout);
     for (k=0; k<nestsize; k++) {
       for (i=0; i<3; i++) (*U)[i] +=  (double)(*V)[k][i] * 1E3 / normfac;
      }
  }

  if (PO)
  {
	printf("\nU*         = [%E  %E  %E];  [m/s] (eq 12 in manuscript)\n\n", (double)(*U)[0], (double)(*U)[1], (double)(*U)[2]);
	fflush(stdout);
  }


  // --- construct Q_ij matrix --------------

  for (k=0; k<nestsize; k++) {
    for (i=0; i<3; i++) {
      for (j=0; j<3; j++) {
	(*Q)[i][j] += ((*q_i)[k][i] - (*U)[i]*(*eta)[k]) * ((*q_i)[k][j] - (*U)[j]*(*eta)[k]);
      }
    }
  }
  

  for (i=0; i<3; i++) {
    for (j=0; j<3; j++) {
      (*Q)[i][j] /= normfac;
    }
  }

  if (PO)  printf("Q_ij  [kg/(s^3)] (eq 15 in manuscript)\n");
  if (PO)  for (i=0; i<3; i++) {
    printf("Q_i%d       = [%19.15E  %19.15E  %16.15E];\n",i, (double)(*Q)[i][0], (double)(*Q)[i][1], (double)(*Q)[i][2]);
  }
 if(PO)  fflush(stdout);

 return(Q);
}
