// ------------------------------------------------------------------
//
//  ConstrainedMinVariance.h
//  Qdos2 - Constrained Minimum Variance Algorithm. Based on 
//  the original MinVaraince.h.
//
//  Constrained MVA is described in the Blue ISSI book
//  "Analysis Methods for Multi-Spacecraft Data",
//  G. Paschmann, P.Daly (eds), published by ESA, 1998.
//
//
//
//  Janet Barnes
//  May 2000
//  Cloned & modified by SEH, May 2002
//      08 Aug, 2005, allow selection of covariance matrix type
//
//
// ------------------------------------------------------------------

#ifndef QDOS_CONSTRAINEDMINVARIANCE_H
#define QDOS_CONSTRAINEDMINVARIANCE_H

#include "Matrix.cc"

namespace QDOS{

class ConstrainedMinVariance
{
    Matrix<double> _eigvals;       // Matrix Dim(3) of eigenvalues
    Matrix<double> _eigvecs;       // Matrix Dim(3,3) of eigenvectors (in rows)
    Matrix<double> _eigqual;       // Matrix Dim(3) eigenvalue quality
    double _orgqual;               // orthogonality quality
    Matrix<double> _covariance;    // covariance matrix
    Matrix<double> P;              // projection matrix P (see ISSI book, p 195)
    Matrix<double> temp;           // temporary storage matrix
    int i,j;
    
public:
    ConstrainedMinVariance(const MatrixSeq<double>& ms, const MatrixSeq<double>& ms2, int constrained, const int Qsel=0):_eigvals(Matrix<double>(Dim(3))), \
      _eigvecs(Matrix<double>(Dim(3,3))),_covariance(Matrix<double>(Dim(3,3))),P(Matrix<double>(Dim(3,3))), \
      _eigqual(Matrix<double>(Dim(3)))
    {
        if(ms.empty())
            return;
        if (!ms.is_homogeneous_seq())
            throw MatrixDomainError();
        if(ms.elt(0).rank()!=1)
            throw MatrixRankError();
        if(ms.elt(0).dimdata()!=Dim(3))
             throw MatrixLengthError();
             
        int sz=ms.size();
        
        // compute the variance matrix 
       
        ScalarSeq<double> x=ms.get_element_seq(vector<int>((size_t)1,0));
        ScalarSeq<double> y=ms.get_element_seq(vector<int>((size_t)1,1));
        ScalarSeq<double> z=ms.get_element_seq(vector<int>((size_t)1,2));
   
        double sum_x=x.sum();
        double sum_y=y.sum();
        double sum_z=z.sum();

	  ScalarSeq<double> x2=ms2.get_element_seq(vector<int>((size_t)1,0));
	  ScalarSeq<double> y2=ms2.get_element_seq(vector<int>((size_t)1,1));
	  ScalarSeq<double> z2=ms2.get_element_seq(vector<int>((size_t)1,2));
   
	  double sum_x2=x2.sum();
	  double sum_y2=y2.sum();
	  double sum_z2=z2.sum();
 

        _covariance.resize(Dim(3,3));
   
	switch(Qsel) 
	  {
	  case 2:  // Weimer type covariance matrix
	    printf("MVA matrix with Weimer type covariance matrix\n");
	    _covariance[0][0]=((x*x).sum() - sum_x*sum_x)/sz;
	    _covariance[0][1]=((x*y).sum() - sum_x*sum_y)/sz;
	    _covariance[0][2]=((x*z).sum() - sum_x*sum_z)/sz;
	    _covariance[1][1]=((y*y).sum() - sum_y*sum_y)/sz;
	    _covariance[1][2]=((y*z).sum() - sum_y*sum_z)/sz;
	    _covariance[2][2]=((z*z).sum() - sum_z*sum_z)/sz;
	    constrained = 0;
	    break;
	  case 3:  // Siscoe type covariance matrix
	    printf("MVA matrix with Siscoe type covariance matrix\n");
	    _covariance[0][0]=(x*x).sum()/sz;
	    _covariance[0][1]=(x*y).sum()/sz;
	    _covariance[0][2]=(x*z).sum()/sz;
	    _covariance[1][1]=(y*y).sum()/sz;
	    _covariance[1][2]=(y*z).sum()/sz;
	    _covariance[2][2]=(z*z).sum()/sz;
	    constrained = 0;
	    break;
	  case 4: // extreme Weimer (for N -> infinity) NB! for record only - gives nonsense
	    printf("MVA matrix with extreme Weimer type covariance matrix\n");
	    _covariance[0][0]=-sum_x*sum_x/sz;
	    _covariance[0][1]=-sum_x*sum_y/sz;
	    _covariance[0][2]=-sum_x*sum_z/sz;
	    _covariance[1][1]=-sum_y*sum_y/sz;
	    _covariance[1][2]=-sum_y*sum_z/sz;
	    _covariance[2][2]=-sum_z*sum_z/sz;;
	    constrained = 0;
	    break;
	  default:
	    printf("MVA matrix with classic covariance matrix\n");
	    _covariance[0][0]=((x*x).sum() - sum_x*sum_x/sz)/sz;
	    _covariance[0][1]=((x*y).sum() - sum_x*sum_y/sz)/sz;
	    _covariance[0][2]=((x*z).sum() - sum_x*sum_z/sz)/sz;
	    _covariance[1][1]=((y*y).sum() - sum_y*sum_y/sz)/sz;
	    _covariance[1][2]=((y*z).sum() - sum_y*sum_z/sz)/sz;
	    _covariance[2][2]=((z*z).sum() - sum_z*sum_z/sz)/sz;
	    break;
	  }

        _covariance[1][0] = _covariance[0][1];
        _covariance[2][0] = _covariance[0][2];
        _covariance[2][1] = _covariance[1][2];
        
	// 
	// Constrained MVA : find eigenvec of [P # C # P] * n = lamda * n
	// instead of [C] * n = lambda * n. 
	//
	//   0 - standard MVA
	//   1 - constrain so that <ts2> * dot n == 0 (see p 209 in blue ISSI book)
	//
	// SEH, 14 May 2002
        //      24 Aug 2002 (constrain against ts2)
	//

	if (constrained == 1) {
          temp.resize(Dim(3,3));
	  double sz2 = (double)(sum_x2*sum_x2 + sum_y2*sum_y2 + sum_z2*sum_z2);

	  P.resize(Dim(3,3));

	  P[0][0] = 1.0 - (sum_x2*sum_x2) / sz2;
	  P[0][1] = - (sum_x2*sum_y2) / sz2;
	  P[0][2] = - (sum_x2*sum_z2) / sz2;

	  P[1][0] = - (sum_y2*sum_x2) / sz2;
	  P[1][1] = 1.0 - (sum_y2*sum_y2) / sz2;
	  P[1][2] = - (sum_y2*sum_z2) / sz2;

	  P[2][0] = - (sum_z2*sum_x2) / sz2;
	  P[2][1] = - (sum_z2*sum_y2) / sz2;
	  P[2][2] = 1.0 - (sum_z2*sum_z2) / sz2;

	  temp = _covariance*P;
	  _covariance = P*temp;

  	// force symmetric covariance matrix again
	  _covariance[1][0] = _covariance[0][1];
	  _covariance[2][0] = _covariance[0][2];
	  _covariance[2][1] = _covariance[1][2];
	}

        // get eigenvectors and values  
        MatrixEig eig(_covariance);
        _eigvecs=eig.eigenvectors().seq_to_matrix();
        _eigvals=eig.eigenvalues().seq_to_matrix();


	// If constrained : eigval 0 == 0. Permute eigenvecs and eigenvals.
	// For e.g., finding boundary normals, one has to look at the 
	// eigenvector corresponding to the intermediate eigenvalue
	// to get a proxy for the normal.
	// See blue ISSI book, ch8 for more info.

        if (constrained == 1) {
	  _eigvals[0] = 0.0;
	}

        
        // check the positive sense of the eigenvector associated with
        // the minimum eigenvalue and if necessary reverse the signs
        // of all three components i.e. vint x vmax = vector perpendicular
        // to both vint and vmax. If vmin is perpendicular to vint 
        // and vmax (i.e. system is ortogonal), then (vinx x vmax).vmin=1
        // If < 0 then not a right-handed system, so reverse signs of vmin
        double rh_factor=(_eigvecs[1].cross(_eigvecs[2])).dot(_eigvecs[0]);
        if(rh_factor<0.)
            _eigvecs[0]*=-1;
        _orgqual=std::abs(std::abs(rh_factor)-1.0);
                
        // ensure that the sign of the average minimum variance component
        // is positive by computing average of sum of B-field data.min
        // eigenvector. If less than zero negate both vmin and vint (so
        // preserving right-handed system)
        // double avg_minvar_comp=(ms*_eigvecs[0]).sum();
	//        if(avg_minvar_comp<0) 

        // SEH 9 May 2005 : ennforce positive x-dir for minvar direction

        int cindex = 1 * (constrained == 1);

        if (_eigvecs[cindex][0] < 0.0 )
        {
            _eigvecs[0]*=-1;
            _eigvecs[1]*=-1;
        }
        _eigqual=_eigvals/_eigvals[2];
            
    }



    // methods for returning members and calculating variance
    Matrix<double>& covariance() 
    { return _covariance; }

    Matrix<double>& projection() 
    { return P; }
    
    Matrix<double>& eigenvalues() 
    { return _eigvals;}
     
    Matrix<double>& eigenvectors()
    { return _eigvecs;}   
    
    Matrix<double>& eigenquality() 
    { return _eigqual;}
    
    double& orthogonality_quality()
    { return _orgqual;}    // orthogonality quality     

    //    printf("CMVA > Exiting.. \n");    
        
};
    
}  // end namespace QDOS
#endif
