// ------------------------------------------------------------------
//
// ISSI QSAS PlugIn : Minimum Faraday Residue calculation
// 
// This plugin takes a magnetic field vector and a plasma velocity
// vector as inputs, calculates the E-field (from V x B) and 
// thereafter calculates the MFR velocity and MFR normal
//
// See Khrabov & Sonnnerup (KS), GRL, p2374, 1998 
//
// Inputs :
// 	bfield		- vector time series with magnetic field
//	velocity	- vector time series with plasma velocity
//
// Outputs
//
// SEH, 29 May, 2002
//
// ------------------------------------------------------------------

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "qplug_if.h"
#include "qdutil.h"
#include "Qdos.h"
#include "qar.h"
#include "ConstrainedMinVarMFR.h"


float vecsum2(float x, float y, float z);

extern "C" QplugReturnStatus  MinFar(QplugArgList *);

using namespace QSAS;

QplugReturnStatus  MinFar  ( QplugArgList *call_list )
{ 
 long    nestsize;
 char    txtbuf[200];
 float   dB2=0.0, dE2=0.0,avg_width,u_n;
 double  deltat;
 int     nseq=2,constrained=0;
 int     i,j,k;

 // ---- QSAS types (see also Khrabov & Sonnerup, GRL, p2374, 1998) ---------

 QdRMatrixSeq_var  B;                                       // B = Magnetic field series
 QdRMatrixSeq_var  V;                                       // V = Velocity series
 QdRMatrixSeq_var  E;                                       // E = Electric field series
 QdRMatrix_var     B_ave;                                   // <B> = averaged B
 QdRMatrix_var     E_ave;                                   // <E> = averaged E
 QdRMatrix_var     dB;                                      // <dB> 
 QdRMatrix_var     dE;                                      // <dE> 
 QdRMatrix_var     lB;                                      // temporary <dB> used in (3) below
 QdRMatrix_var     lE;                                      // temporary <dE> used in (3) below 
 QdRMatrix_var     P;                                       // P = Poynting flux
 QdRMatrix_var     Q;                                       // Q = (see KS pap)
 QdRMatrix_var     n_MFR;                                   // the MFR normal;
 QdRScalar_var     mvatype_do;                              // Constrained (1) or unconstrained (0) MVA 



 //  --- first we join B measurements to V measurements -----------------

 QdObjectSeq_var ttIn    = new QdObjectSeq(nseq);
 QdObjectSeq_var toJoin  = new QdObjectSeq(nseq);
 QdObjectSeq_var joined  = new QdObjectSeq(nseq);
 QdObjectSeq_var restSig = new QdObjectSeq(nseq);
 QdStringSeq_var joinMethod = new QdStringSeq(nseq);
 vector <double> gapValue(nseq);
 QdStringSeq_var gapHandler = new QdStringSeq(nseq);
  
  // added by AJA
 // test if frames are same for B and V
 
 int frame_status =QarTestConformalForVecProds((*call_list)[0], (*call_list)[1]);
 
 if ( frame_status == QAR_ERR ){
   QplugAppendTextDisplay("B field and velocity in different frames\n");
   return QPLUG_FAILURE;
 }
 else if(frame_status == QAR_WARN){
   QplugAppendTextDisplay("Caution: Frame unknown for either B or V\n"); 
 }

 
 QdTimeSeq_var ttJoined = get_timetags((*call_list)[1]);    // tt of joined V and B
 QdTimeSeq_var tt_out;                                      // tt of masked V and B

 if (ttJoined.is_nil()) {
   QplugAppendTextDisplay("Bad or missing time tags in input data\n");
   return QPLUG_FAILURE;
 }

 QuIsTSRegular(ttJoined,&deltat) ;
 
 avg_width = 1.5 * deltat;
 if (VERBOSE_MV)
 {
	printf("\nMinFar > delta t = %f  ->  gap_fill if t > %f\n",deltat,avg_width);
	fflush(stdout);
  }

 for (i=0; i<2; i++) {
   toJoin[i] = ((*call_list)[i]);

   joinMethod[i] = QdString(LINEAR);
   gapValue[i] = avg_width;
   gapHandler[i] = QdString(LINEAR);   
 }

 //  ---- get Time interval and mask ----------------

 QdTimeInterval_var start_end = get_timeinterval((* call_list)[2]);
 SequenceMask msk(ttJoined->get_mask(QdTimeSeq::InRange(start_end->start(), start_end->end())));
 tt_out = new QdTimeSeq();
 ttJoined->get_subsequence(*tt_out, msk);

 try {
   if (VERBOSE_MV)
   {
	printf("MinFar > AvgWidth = %f\n",avg_width);
	fflush(stdout);
   }
   QuMultiJoin(toJoin, tt_out, joined, gapValue, gapHandler, joinMethod);
   if (VERBOSE_MV)
   {
	printf("MinFar > AvgWidth = %f\n",avg_width);
	fflush(stdout);
   }
 }
 catch(Exception &e) {
   e.print_msg();
   if (VERBOSE_MV)
   {
	printf("MinFar > AvgWidth =%f\n",avg_width);
	fflush(stdout);
   }
   return QPLUG_FAILURE;
 }


 QdRMatrixSeq_var masked_B = QdRMatrixSeq_var::narrow(joined[0]);

 if(masked_B.is_nil())
 {
	printf("B is nil");
	fflush(stdout);
 }
  // added by AJA to ensure B field units are correct (nT)
   string SIcB = QuGetAttrText("SI_conversion", ((* call_list)[0]));
   if (SIcB.empty())  {
     QplugAppendTextDisplay("CAUTION: unknown units (SI_conversion) for B field");
     B = (QdRMatrixSeq_var)  masked_B;
   }
   else if (QarSameBaseSI(SIcB.c_str(), "1.0e-9>T")){
     // base units the same, ensure conversion is too
     // take care as masked_B does not have orig xref
     masked_B->change_xref("SI_conversion", new QdString(SIcB));
     B = QdRMatrixSeq_var::narrow(QarConvertUnitsTo(masked_B, "1.0e-9>T"));
   }
   else {
     QplugAppendTextDisplay("ERROR: cannot correct units of B field");
     return QPLUG_FAILURE;
   }

 if( is_empty(B) ){
     QplugAppendTextDisplay("ERROR: no B field data in selected interval");
     return QPLUG_FAILURE;
 }

 nestsize = B->sequence_size();
 if (VERBOSE_MV)
 {
	printf("MinFar > Nestsize B = %ld\n",nestsize);
	fflush(stdout);
 }

 QdRMatrixSeq_var masked_V = QdRMatrixSeq_var::narrow(joined[1]);
 
   // added by AJA to ensure velocity units are correct (km s^-1)
   string SIcVel = QuGetAttrText("SI_conversion", ((* call_list)[1]));
   if (SIcVel.empty())  {
     QplugAppendTextDisplay("CAUTION: unknown units (SI_conversion) for velocity");
     V = (QdRMatrixSeq_var)  masked_V;
   }
   else if (QarSameBaseSI(SIcVel.c_str(), "1.0e3>m s^-1")){
     // base units the same, ensure conversion is too
     masked_V->change_xref("SI_conversion", new QdString(SIcVel.c_str()));
     V = QdRMatrixSeq_var::narrow(QarConvertUnitsTo(masked_V, "1.0e3>m s^-1"));
   }
   else {
     QplugAppendTextDisplay("ERROR: cannot correct units of velocity");
     return QPLUG_FAILURE;
   }
 
 if( is_empty(V) ){
     QplugAppendTextDisplay("ERROR: no V data in selected interval");
     return QPLUG_FAILURE;
 }

  nestsize = V->sequence_size();
 if (VERBOSE_MV)
 {
	printf("MinFar > Nestsize V = %ld\n",nestsize);
	fflush(stdout);
 }



 // ---- Get type MVA (constrained or normal) ------------------

 QdObject_var in_3 = (* call_list)[3];
 mvatype_do = QdRScalar_var::narrow(in_3);
 if( mvatype_do.is_nil() ) {
   QplugAppendTextDisplay("ERROR: cannot understand constrained choice");  
   return QPLUG_FAILURE;
 }
 constrained = ((int) *mvatype_do);
 if( constrained > 1 || constrained < 0 ){
   QplugAppendTextDisplay("ERROR: constrained choice must be 0 or 1");  
   return QPLUG_FAILURE;
 }


 // ----- Start calculation as described in the K & S  paper ------



 // ---- (1)  Calculate E = -V x B = B x V  and <E> and <B> ---------------- 

 E = new QdRMatrixSeq(nestsize,RMatrix(Dim(3)));   
 E_ave = new QdRMatrix(Dim(3), 0.0);
 B_ave = new QdRMatrix(Dim(3), 0.0);

 float cc = 1.0e-03;   // conversion due to B in [nT] and V in [km/s]  

 for (k=0; k<nestsize; k++) {
   (*E)[k][0] = cc*(float)((*B)[k][1] * (*V)[k][2]  -  (*B)[k][2] * (*V)[k][1]);
   (*E)[k][1] = cc*(float)((*B)[k][2] * (*V)[k][0]  -  (*B)[k][0] * (*V)[k][2]);
   (*E)[k][2] = cc*(float)((*B)[k][0] * (*V)[k][1]  -  (*B)[k][1] * (*V)[k][0]);
   for (j=0; j<3; j++) {
      (*B_ave)[j] += (*B)[k][j];
      (*E_ave)[j] += (*E)[k][j];
   }
 }   

 for (j=0; j<3; j++) {
    (*B_ave)[j] = (float)(*B_ave)[j] / (float) nestsize;
    (*E_ave)[j] = (float)(*E_ave)[j] / (float) nestsize;
 }

 if (VERBOSE_MV) {
   printf("MinFar > B_ave = [%9.3f %9.4f %9.4f]\n",(float)(*B_ave)[0],(float)(*B_ave)[1],(float)(*B_ave)[2]);
   printf("MinFar > E_ave = [%9.3f %9.4f %9.4f]\n",(float)(*E_ave)[0],(float)(*E_ave)[1],(float)(*E_ave)[2]);
   fflush(stdout);
 }


 // ---- (2) Calculate electric field variance matrix M^E -------- 

 QdRMatrix_var ME = new QdRMatrix(Dim(3,3), 0.00);

 for (k=0; k<nestsize; k++) {
   for (i=0; i<3; i++) {
     for (j=0; j<3; j++) {  
       (*ME)[i][j] += (((float)(*E)[k][i]) - (float)(*E_ave)[i]) *  (((float)(*E)[k][j])-(float)(*E_ave)[j]);
     }
   }
 }
 
 for (i=0; i<3; i++) for (j=0; j<3; j++) (*ME)[i][j] /= (float) nestsize;

 if (VERBOSE_MV) {
   printf("MinFar > ME =\n");
   for (i=0; i<3; i++) {
     for (j=0; j<3; j++) printf("%10.5f ",(float)(*ME)[i][j]);
     printf("\n");
   }
   printf("\n");
   fflush(stdout);
 }
 

 // ---- (3) Calculate <|dB|^2> and <|dE|^2>  -------- 

 lE = new QdRMatrix(Dim(3), 0.0);
 lB = new QdRMatrix(Dim(3), 0.0);
 dB2 = 0.0;
 dE2 = 0.0;

 for (k=0; k<nestsize; k++) {
   for (i=0; i<3; i++) {
     (*lB)[i] = (float)(*B)[k][i] - (float)(*B_ave)[i];
     (*lE)[i] = (float)(*E)[k][i] - (float)(*E_ave)[i];
   }
   dB2 += vecsum2((float)(*lB)[0],(float)(*lB)[1], (float)(*lB)[2]);
   dE2 += vecsum2((float)(*lE)[0],(float)(*lE)[1], (float)(*lE)[2]);
 }

 dB2 /= (float) nestsize;
 dE2 /= (float) nestsize;

 if (VERBOSE_MV)
 {
	printf("MinFar > dB2 = %10.4f, dE2 = %10.4f\n",dB2,dE2);
	fflush(stdout);
 }



 // ---- (4) Calculate Poynting vector P = <dB x dE> --------------

 P  = new QdRMatrix(Dim(3), 0.0);
 dB = new QdRMatrix(Dim(3), 0.0);
 dE = new QdRMatrix(Dim(3), 0.0);

 if (VERBOSE_MV)
 {
	printf("MinFar > P, dB and dE initialized\n");
	fflush(stdout);
 }

 for (k=0; k<nestsize; k++) {
   for (i=0; i<3; i++) {
     (*dB)[i] = (float)(*B)[k][i] - (float)(*B_ave)[i];
     (*dE)[i] = (float)(*E)[k][i] - (float)(*E_ave)[i];
   }
   (*P)[0] += ((((float)(*dE)[1] * (float)(*dB)[2]) - ((float)(*dE)[2] * (float)(*dB)[1])));
   (*P)[1] += ((((float)(*dE)[2] * (float)(*dB)[0]) - ((float)(*dE)[0] * (float)(*dB)[2])));
   (*P)[2] += ((((float)(*dE)[0] * (float)(*dB)[1]) - ((float)(*dE)[1] * (float)(*dB)[0])));
 }

 for (j=0; j<3; j++) (*P)[j] /= (float) nestsize;

 if (VERBOSE_MV) 
 {
	printf("MinFar > P = [%10.4f %10.4f %10.4f]\n",(float)(*P)[0],(float)(*P)[1],(float)(*P)[2]);
	fflush(stdout);
 }



 // ---- (5) Calculate Q matrix ------------------------------------

 Q  = new QdRMatrix(Dim(3,3));

 for(i=0; i<3; i++)
   for(j=0; j<3; j++)
     (*Q)[i][j] = -1.0 * (float)(*ME)[i][j] - (float)(*P)[i]*(float)(*P)[j]/dB2;

 for(i=0; i<3; i++) (*Q)[i][i] += dE2;                  // add the <|dE|^2>d_ij term
 
 if (VERBOSE_MV) {
   printf("MinFar > Q =\n");
   for (i=0; i<3; i++) {
     for (j=0; j<3; j++) printf("%10.5f ",(float)(*Q)[i][j]);
     printf("\n");
   }
   printf("\n");
   fflush(stdout);
 }


 // ---- (6) Find eigenvals + eigenvecs of Q --------------------- 

 QdRMatrix_var Rmatx = new QdRMatrix(Dim(3,3));
 QdRMatrix_var Eigenvalues = new QdRMatrix(Dim(3));
 QdRMatrix_var Proj_matx = new QdRMatrix(Dim(3,3));;

 try{
  ConstrainedMinVarMFR minvar(*B,constrained,*Q);   // qdos minvariance call
  *Rmatx = minvar.eigenvectors().duplicate(SHALLOW);
  *Eigenvalues = minvar.eigenvalues().duplicate(SHALLOW); 
 }
 catch(Exception& e) 
   { return QPLUG_FAILURE; }

 if (VERBOSE_MV) {
   printf("MinFar > Eigenvalues, eigenvectors are :\n");
   for (i=0; i<3; i++){
     printf("%10.5g  (%10.5g, %10.5g, %10.5g )\n", 
	     (double)(*Eigenvalues)[i],
	     (double)(*Rmatx)[i][0],
	     (double)(*Rmatx)[i][1], 
	     (double)(*Rmatx)[i][2]);
   }   
   fflush(stdout);
 }



 // ----- (7) Calculate V_MFR along normal --------

 u_n = 0.0;
 for (i=0; i<3; i++){
   u_n += ((float)(*P)[i]*(float) (*Rmatx)[0][i] / dB2);
 }
  
 u_n *= 1000.0;                  // Since P [nT * mV/m], and we want u_n in [km/s]

 if (VERBOSE_MV)
 {
	printf("\nMinFar > u_n = %10.4f\n",u_n);
	fflush(stdout);
 }

 // ----- Print output values ---------------------


 n_MFR = new QdRMatrix(Dim(3), 0.0);

 sprintf(txtbuf,"MinFar > Nestsize = : %ld \n",nestsize);
 QplugAppendTextDisplay ( txtbuf);

 sprintf(txtbuf,"MinFar > Eigenvalues, eigenvectors are :\n");
 QplugAppendTextDisplay ( txtbuf);
 for (i=0; i<3; i++){
   sprintf(txtbuf,"          %10.5g  [%10.5g, %10.5g, %10.5g ]\n", 
	   (double)(*Eigenvalues)[i],
	   (double)(*Rmatx)[i][0],
	   (double)(*Rmatx)[i][1], 
	   (double)(*Rmatx)[i][2]);
     QplugAppendTextDisplay ( txtbuf);
     (*n_MFR)[i] = (double)(*Rmatx)[0][i];
 }   

 sprintf(txtbuf,"MinFar > U_n = %10.4f km/s\n",u_n);
 QplugAppendTextDisplay ( txtbuf);



 // ----- Output normal and Efield ------

// QuSetTxtAttr("Frame","vector>xyz_xyz",(QdObject_var) n_MFR);
// QuSetTxtAttr("Frame","vector>xyz_xyz",(QdObject_var) E);

 string frame_in = QuGetFrameAttr(((* call_list)[0]), ((* call_list)[1]));
 QuSetFrame((QdObject_var) n_MFR, frame_in.c_str());
 QuSetFrame((QdObject_var) E, frame_in.c_str());
 QuSetTxtAttr("UNITS","mV/m",(QdObject_var) E);
 QuSetTxtAttr("SI_conversion","0.001>V m^-1",(QdObject_var) E);
 QuSetTxtAttr("LABLAXIS","E",(QdObject_var) E);

 call_list->push_back((QdObject_var) E);
 call_list->push_back((QdObject_var) n_MFR);

 return QPLUG_SUCCESS;

}



// ---------------------------------------------------------
//
// Calculates | vec |^2
//
// ---------------------------------------------------------

float vecsum2(float x, float y, float z)
{
    return (x*x +y*y +z*z);
}



// ---------------------------------------------------------
//
// Does nothing, but there needs to be a main() fuction
//
// ---------------------------------------------------------

int main()
{
  // --- do nothing ---
}

