/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/

#ifndef __cplusplus
#define __cplusplus
#endif


#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "MinVar.h"

//#define VERBOSE_MV 1

extern "C" QplugReturnStatus  MinVar(QplugArgList *);

using namespace QSAS;

QplugReturnStatus  MinVar  ( QplugArgList *arglist )
{
// QDOS Objects:

QdRMatrixSeq_var	 vts_in;
QdRMatrixSeq_var	 vts_out;
QdRMatrix_var	     Rmatx;
QdRMatrix_var	     Eigenvalues;
QdTimeInterval_var   start_end;

char attrtxt[150];

QplugReturnStatus     MV_return;  // used to return from sub-modules

int    i;
char   evalsevecs[80];



//	 get the input data objects

QdObject_var in_0 = (*arglist)[0];
vts_in = QdRMatrixSeq_var::narrow(in_0);
if( vts_in.is_nil() ) return QPLUG_FAILURE;

// check input is cartesian

string rep = QuGetRep(vts_in);
if( QuToUpper(rep).find("XYZ") == string::npos ){
    QplugAppendTextDisplay ("MinVar Failed >> vector is not cartesian\n");
    return QPLUG_FAILURE;
}

QdObject_var in_1 = (*arglist)[1];
start_end = get_timeinterval(in_1);
if( start_end.is_nil() ) return QPLUG_FAILURE;

Rmatx=new QdRMatrix(Dim(3,3));
Eigenvalues=new QdRMatrix(Dim(3));

//	...next, subset the time range and data array

 QdTimeSeq_var tt_in = get_timetags(vts_in);

 SequenceMask msk(tt_in->get_mask(QdTimeSeq::InRange(start_end->start(), start_end->end())));
 
 QdTimeSeq_var tt_out = new QdTimeSeq();
  
 tt_in->get_subsequence(*tt_out, msk);
 
 vts_out = new QdRMatrixSeq();
 vts_in->get_subsequence(*vts_out, msk);
  
 vts_out->copy_xrefs_from(vts_in);
 set_timetags( (QdObject_var) vts_out, tt_out);


#ifdef VERBOSE_MV
  printf(">>MinVar.cc Subsetting done successful\n");
  fflush(stdout);
#endif

//      ...next do the minimum variance analysis

  MV_return = MV_analyse(vts_out, Eigenvalues, Rmatx);
	
#ifdef VERBOSE_MV
  printf(">>MinVar.cc Minimum Variance Analysis done \n");
  printf("Eigenvalues, eigenvectors are:\n");
  for (i=0; i<3; i++){
    sprintf(evalsevecs,"%10.4g  (%10.4g, %10.4g, %10.4g )\n", 
      (double)(*Eigenvalues)[i],
      (double)(*Rmatx)[i][0],
      (double)(*Rmatx)[i][1], 
      (double)(*Rmatx)[i][2]);
    printf ("%s",evalsevecs);
  }   // end for
  fflush(stdout);
#endif

if (MV_return != QPLUG_SUCCESS) return MV_return; 

// rotate vector TS into MV frame

 for (int k=0; k<vts_out->size() ; k++){
   vts_out[k] =  (*Rmatx) * vts_out[k];
 }

//      ...next do the plotting

MV_return = MV_plot_results(vts_out, Eigenvalues, Rmatx);
if (MV_return != QPLUG_SUCCESS) return MV_return;

#ifdef VERBOSE_MV
  printf(">>MinVar.cc Plotting done successfully\n");
  fflush(stdout);
#endif

 // set attributes

  vts_out->change_xref("Eigenvalues", Eigenvalues);
  vts_out->change_xref("Eigenvectors", Rmatx);
  QuSetFrame((QdObject_var) vts_out, "vector>mv_xyz");
  string inFrame = QuGetFrame(vts_in);
  QuSetFrame((QdObject_var) Rmatx, "rot", inFrame.c_str(), "xyz");

 try{
   QdString_var ToFrameXref = new QdString("vector>mv_xyz");
   QdObject_var name_xref = vts_in->get_xref(OBJECT_NAME);
   Rmatx->change_xref("InputSeriesName", name_xref);
   Rmatx->change_xref("ToFrame", ToFrameXref);
   vts_out->change_xref("InputSeriesName", name_xref);
 }
 catch(Exception &e){
   e.print_msg();
   return QPLUG_FAILURE;   
 }
 
 sprintf(attrtxt, "%d", (int)(vts_out->size()) );
 QuSetTxtAttr("NumberOfDataValues", attrtxt, Rmatx);
 QuSetTxtAttr("NumberOfDataValues", attrtxt, Eigenvalues);

 QuSetTxtAttr("FIELDNAM", "Rotation matrix", Rmatx);
 QuSetTxtAttr("CATDESC", "Matrix of Eigenvectors as rows", Rmatx);
 QuSetTxtAttr("FIELDNAM", "Eigenvalues", Eigenvalues);
 QuSetTxtAttr("CATDESC", "Eigenvectors are corresponding rows in Rmatx, rotates vector to MV frame", Eigenvalues);

 Rmatx->change_xref("TimeInterval", start_end);
 Eigenvalues->change_xref("TimeInterval", start_end);

 Rmatx->change_xref("Eigenvalues", Eigenvalues);

// put new objects on working list

  arglist->push_back(vts_out);     
  arglist->push_back(Rmatx);   

 
//	..clean up and exit

QplugAppendTextDisplay ("MinVar >> - exiting successfully\n");
return  QPLUG_SUCCESS;
 
}  // end MinVar




/***********************************************************************
************************************************************* MV_analyse
***********************************************************************/

QplugReturnStatus  MV_analyse(QdRMatrixSeq_var ts_out,
                              QdRMatrix_var Eigenvalues,
                              QdRMatrix_var Rmatx){

char evalsevecs[80];
try{
  MinVariance minvar(*ts_out);   // qdos minvariance call
  double dot=minvar.orthogonality_quality();

  if(dot>0.001){
     QplugAppendTextDisplay(
    "MinVar >> Eigenvectors failed simple orthonormal test at level 0.001\n");
  }
  *Rmatx = minvar.eigenvectors().duplicate(SHALLOW); 
  *Eigenvalues = minvar.eigenvalues().duplicate(SHALLOW); 

}
catch(Exception& e)
{ 
	e.print_msg();
	return QPLUG_FAILURE; 
}

QplugAppendTextDisplay("Eigenvalues, eigenvectors are:\n");
for (int i=0; i<3; i++){
  sprintf(evalsevecs,"%10.4g  (%10.4g, %10.4g, %10.4g )\n",  (double)(*Eigenvalues)[i], (double)(*Rmatx)[i][0],
	  (double) (*Rmatx)[i][1], (double)(*Rmatx)[i][2]);
  QplugAppendTextDisplay (evalsevecs);
}   // end for

return QPLUG_SUCCESS;
}

/***********************************************************************
********************************************************* MV_plot_results
***********************************************************************/

QplugReturnStatus  MV_plot_results(QdRMatrixSeq_var vts_out, 
                                   QdRMatrix_var Eigenvalues, 
				   QdRMatrix_var Rmatx)   {

mvplot (vts_out, Eigenvalues, Rmatx);

return QPLUG_SUCCESS;
}






