/****************************************************************************
							LICENSE

  sf_Pomme -  plug-in for the QSAS software.
  Copyright (C) 2010 Eugen Sorbalo

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  A copy of the GNU General Public License is available on the internet at:

  http://www.gnu.org/copyleft/gpl.html

  or you can write to:

  The Free Software Foundation, Inc.
  675 Mass Ave
  Cambridge, MA 02139, USA
****************************************************************************/

/**
 * \file sf_pomme.cc
 * 
 * \brief Computes the magnetic field model using POMME-7. Based on the code of Stefan Maus.
 * 
 * 
 */

#include <stdio.h>
//the next line is needed for visual studio to find M_PI and others.
#define _USE_MATH_DEFINES
#include <math.h>
#include <string.h>

#include "qdutil.h"
#include "qar.h"
#include "Qdos.h"

#include "./pomme.h"
#include "./POMME7_ext.h"
#include "qrtn.h"  //coordinates rotation

using namespace QSAS;
using namespace std;

/**
 * The class Status has the purpose to print messages either to user or to the Console.
 */
class Status
{
public:
	static const bool kPrintToUserAndConsole = true;
	static const bool kPrintToConsoleOnly = false;
	static void PrintMessage(bool to_user, const char fmt[], ...);
public:
	enum ReturnStatus
	{
		kOk = 0, kUnknownError, kWrongInput, kNoInput, kVectorFieldRequired
	} code;
	std::string Message() const;
	inline void set_message(const std::string message)
	{
		message_ = message;
	}
	;
	Status();
	virtual ~Status();
private:
	std::string message_;
};

std::string QplugNewSTR_2string(const char *in_str)
{
	char * str = QplugNewSTR(in_str);
	std::string out_string = str;
	free(str);
	return out_string;
}

// ==========================================================================

const int kCallListIndexOfPosition  = 0;
const int kCallListIndexOfTimeRange = 1;
const int kCallListIndexOfIMFBy		= 2;
const int kCallListIndexOfEm		= 3;
const int kCallListIndexOfEst 		= 4;
const int kCallListIndexOfIst 		= 5;
const int kCallListIndexOff107 		= 6;
const int kCallListIndexOfMaxOrder	= 7;

struct TExternalParameters
{
	//printf("IMF-By 35 minutes earlier (enter '0' if unknown)? ");
	double IMF_By;
	//printf("Merging electric field, 60 minutes earlier (enter '0.5' if unknown)? ");
	double Em;
	//printf("E_st (enter '0' if unknown)? ");
	double E_st;
	//printf("I_st (enter '0' if unknown)? ");
	double I_st;
	//printf("F10.7, 81-day average, 20 months earlier (enter '120' if unknown)? ");
	double f107;
	unsigned int model_number;
	int oder_internal_field;
};

QplugReturnStatus GetPositions(QplugArgList *call_list, int call_list_index
		, SequenceMask &mask, QdRMatrixSeq_var &V);

QplugReturnStatus ComputePommeModel(QdRMatrixSeq_var &pos, TExternalParameters &parameters, QdRMatrixSeq_var &B_field);

QplugReturnStatus GetExternalParameters(QplugArgList *call_list, TExternalParameters &parameters);

QplugReturnStatus  ConvertToDesiredFrame (QdRMatrixSeq_var &ts_in, int frame, QdRMatrixSeq_var &ts_out);

// ==========================================================================
extern "C" QplugReturnStatus sf_pomme(QplugArgList *call_list)
{
//	   INPUT  VECTOR_TIME_SERIES Pos_sat,
//	   INPUT  TIME_INTERVAL      start_end,
//	   INPUT  FLOAT              IMF_By,
//	   INPUT  FLOAT              Em,
//	   INPUT  FLOAT              E_st,
//	   INPUT  FLOAT              I_st,
//	   INPUT  FLOAT              f107,
//	   INPUT  INTEGER            model_no,
//	   OUTPUT VECTOR_TIME_SERIES B_field)
	QplugReturnStatus status = QPLUG_SUCCESS;

	// 1. Importing data into memory --------------------------------------------------------------
	Status::PrintMessage(Status::kPrintToConsoleOnly, "1. Extracting data from the working list\n");

	QdTimeInterval_var start_end = get_timeinterval((*call_list)[kCallListIndexOfTimeRange]);
	if (start_end.is_nil()) {
		status = QPLUG_FAILURE;
		Status::PrintMessage(Status::kPrintToUserAndConsole, "error: Could not get the time interval.\n");
	}
	QdRMatrixSeq_var P_in = NULL;
	QdTimeSeq_var tt_out = NULL;
	int init_frame = UNKNOWN;
	if (status == QPLUG_SUCCESS) {
		QdTimeSeq_var time_tags = get_timetags((*call_list)[kCallListIndexOfPosition]);
		SequenceMask mask(time_tags->get_mask(QdTimeSeq::InRange(start_end->start(), start_end->end())));
		tt_out = new QdTimeSeq(); // tt of masked V and Pos
		time_tags->get_subsequence(*tt_out, mask);

		QdRMatrixSeq_var P_in_orig_frame = NULL;
		status = GetPositions(call_list, kCallListIndexOfPosition, mask, P_in_orig_frame);
		if ((status == QPLUG_SUCCESS) && (P_in_orig_frame.is_nil())) {
			status = QPLUG_FAILURE;
			Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: could not read data. Contact the developer.\n");
		}
		//convert to GEOC before continuing
		if (status == QPLUG_SUCCESS) {
			init_frame = QrGetFrameFromAttr(P_in_orig_frame);
			if (init_frame == UNKNOWN) {
				status = QPLUG_FAILURE;
				Status::PrintMessage(Status::kPrintToUserAndConsole,
						"error: Could not understand the frame. Do both FRAME and COORDINATE_SYSTEM exist?\n");
			}
			else {
				status = ConvertToDesiredFrame(P_in_orig_frame, GEOC, P_in);
			}
		}
	}
	// Extract external parameters
	TExternalParameters external_params;
	if (status == QPLUG_SUCCESS)
		status = GetExternalParameters(call_list, external_params);

	// 2. Computing model -------------------------------------------------------------------------
	Status::PrintMessage(Status::kPrintToConsoleOnly, "2. Computing the POMME model.\n");
	QdRMatrixSeq_var B_field_geoc = NULL;
	if (status == QPLUG_SUCCESS) {
		try {
			ComputePommeModel(P_in, external_params, B_field_geoc);
		}
		catch (...) {
			status = QPLUG_FAILURE;
		}
	}

	// 3. Preparing output ----------------------------------------------------
	if ((status == QPLUG_SUCCESS) && (!B_field_geoc.is_nil()))
	{
		Status::PrintMessage(Status::kPrintToConsoleOnly, "3. Placing output into the working list\n");
		try{
			//Units and SI conversion:
			QuSetFrame((QdObject_var) B_field_geoc, "vector>geoc_xyz");
			QuSetTxtAttr(UNITS, "nT", (QdObject_var) B_field_geoc);
			QuSetTxtAttr(SI_CONVERSION, "1.0E-9>T",	(QdObject_var) B_field_geoc);
			QuSetTxtAttr(LABLAXIS, "POMME6 B field", (QdObject_var) B_field_geoc);
			QuSetTxtAttr(FIELDNAM, "POMME6 B field", (QdObject_var) B_field_geoc);
			set_timetags((QdObject_var) B_field_geoc, tt_out);
		}
		catch (...) {
			status = QPLUG_FAILURE;
		}
	}

	// 4. Coordinate transformation to the frame of the position vector ---------------------------
	QdRMatrixSeq_var B_field = NULL;
	if ((status == QPLUG_SUCCESS) && (!B_field_geoc.is_nil()))
	{
		Status::PrintMessage(Status::kPrintToConsoleOnly, "4. Rotate to the desire frame.\n");
		status = ConvertToDesiredFrame(B_field_geoc, init_frame, B_field);
	}

	// 5. Placing output into the working list ----------------------------------------------------
	if (status == QPLUG_SUCCESS) {
		try {
			Status::PrintMessage(Status::kPrintToConsoleOnly, "5. Push back\n");
			call_list->push_back((QdObject_var) B_field);
		}
		catch (...) {
			status = QPLUG_FAILURE;
		}
	}
	return status;
}

QplugReturnStatus GetPositions(QplugArgList *call_list, int call_list_index
		, SequenceMask &mask, QdRMatrixSeq_var &position)
{
	QplugReturnStatus status = QPLUG_SUCCESS; //success

	QdRMatrixSeq_var P_full = QdRMatrixSeq_var::narrow((*call_list)[call_list_index]);
	if (P_full.is_nil()) {
		status = QPLUG_FAILURE;
		Status::PrintMessage(Status::kPrintToUserAndConsole, "error: Input is not a sequence of vectors.\n");
	}
	else {
		QdRMatrixSeq_var V = new QdRMatrixSeq();
		P_full->get_subsequence(*V, mask);
		V->copy_xrefs_from((*call_list)[call_list_index]);
		position = QdRMatrixSeq_var::narrow(QarConvertUnitsTo(V, "1.0e3>m"));
		if(position.is_nil()){
			Status::PrintMessage(true, "Position is not in units of distance.\n");
			status = QPLUG_FAILURE;
		}
	}
	// 3. Compare reference frame (must be same)  -----------------------------
	if (status == QPLUG_SUCCESS) {
		string frame_P = QuGetFrameAttr( ((* call_list)[call_list_index])) ;
		if (frame_P.empty()) {
			Status::PrintMessage(true, "The position vector must have a frame.");
			status = QPLUG_FAILURE;
		}
	}
	return status;
}

QplugReturnStatus GetExternalParameters(QplugArgList *call_list, TExternalParameters &parameters)
{
	QplugReturnStatus status = QPLUG_SUCCESS;
	//IMF By
	QdRScalar_var temp_scalar  = QdRScalar_var::narrow((*call_list)[kCallListIndexOfIMFBy]);
	if (temp_scalar.is_nil()) {
		status = QPLUG_FAILURE;
		Status::PrintMessage(Status::kPrintToUserAndConsole,
				"error: Could not get IMF By. Check that it is a scalar.\n");
	}
	else {
		parameters.IMF_By = *temp_scalar;
	};
	//Em
	if (status != QPLUG_FAILURE) {
		temp_scalar
				= QdRScalar_var::narrow((*call_list)[kCallListIndexOfEm]);
		if (temp_scalar.is_nil()) {
			status = QPLUG_FAILURE;
			Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: Could not get Em. Check that it is a scalar.\n");
		} else {
			parameters.Em = *temp_scalar;
		}
	};
	//Est
	if (status != QPLUG_FAILURE) {
		temp_scalar
				= QdRScalar_var::narrow((*call_list)[kCallListIndexOfEst]);
		if (temp_scalar.is_nil()) {
			status = QPLUG_FAILURE;
			Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: Could not get Est. Check that it is a scalar.\n");
		} else {
			parameters.E_st = *temp_scalar;
		}
	};
	//Ist
	if (status != QPLUG_FAILURE) {
		temp_scalar
				= QdRScalar_var::narrow((*call_list)[kCallListIndexOfIst]);
		if (temp_scalar.is_nil()) {
			status = QPLUG_FAILURE;
			Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: Could not get Ist. Check that it is a scalar.\n");
		} else {
			parameters.I_st = *temp_scalar;
		}
	};
	//F107
	if (status != QPLUG_FAILURE) {
		temp_scalar
				= QdRScalar_var::narrow((*call_list)[kCallListIndexOff107]);
		if (temp_scalar.is_nil()) {
			status = QPLUG_FAILURE;
			Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: Could not get F107. Check that it is a scalar.\n");
		} else {
			parameters.f107 = *temp_scalar;
		}
	};
	//Main field order.
	if (status != QPLUG_FAILURE) {
		temp_scalar = QdRScalar_var::narrow((*call_list)[kCallListIndexOfMaxOrder]);
		if (temp_scalar.is_nil()) {
			status = QPLUG_FAILURE;
			Status::PrintMessage(Status::kPrintToUserAndConsole,
					"error: Could not get the max order of the main field. Check that it is an integer.\n");
		} else {
			parameters.oder_internal_field = static_cast<unsigned int>(*temp_scalar);
		}
	};
	return status;
}

/**
 * Converts a vector object to the desired new frame.
 * \param[in] ts_in is a vector TS object, which has a well-defined frame
 * \param[in] frame is an integer. Acronyms are set to integer values by #define in qrtn.h
 *        0  GSE,  Geocentric Solar Ecliptic
 *        1  GSM,  Geocentric Solar Magnetic
 *        2  GEOC, Geographic (geocentric)
 *        3  SM,   Solar Magnetic
 *        4  GEI,  Geocentric Inertial
 *        5  MAGD,  Geomagnetic (dipole)
 *        6  HAE,  Heliospheric Aries Ecliptic
 *        7  HEE,  Heliospheric Earth Ecliptic
 *        8  HEEQ, Heliospheric Earth Equatorial
 *        9  GSEQ, Geocentric Solar Equatorial
 * \param[out] ts_out is the converted (rotated) TS object.
 */
QplugReturnStatus ConvertToDesiredFrame (QdRMatrixSeq_var &ts_in, int frame, QdRMatrixSeq_var &ts_out)
{
	QplugReturnStatus	status = QPLUG_SUCCESS;
	QrVector vec;
	vec.n_recs = ts_in->sequence_size();
	vec.in_frame = QrGetFrameFromAttr(ts_in);
	vec.out_frame = frame;

	vec.IsPositionVector = QrIsPositionVector(ts_in);
	if(vec.IsPositionVector == 1) {
		vec.IsPositionVector = 0; /* issue warning & continue, delta r rotates */
		if(QrOriginsDiffer(vec.in_frame, vec.out_frame)) {
			QplugAppendTextDisplay
			("WARNING: CoordRotation plug-in provides frame rotation only.\n");
			QplugAppendTextDisplay
			("Transformation is between geocentric and heliocentric frames.\n");
			QplugAppendTextDisplay
			("Position vector coordinate transformations will be WRONG.\n");
			QplugAppendTextDisplay
			("Note: Spacecraft separations will rotate OK.\n");
		}
	}

	ts_out = (QdRMatrixSeq_var) Factory::clone(ts_in);
	ts_out->copy_xrefs_from(ts_in);
	vec.ptr = ts_out;

	vec.t = get_timetags(ts_in);
	if( (vec.t).is_nil() ) {
		QplugAppendTextDisplay ("Cannot find time tags, some frames are time dependent\n");
		return QPLUG_FAILURE;
	}

	// call C function to do rotation
	long check = QrDoRotation(&vec);
	//Check that the COORDINATE_SYSTEM metadata is correctly written after the coordinate transformation.
	if (check == QRTN_OK || check == SAME_FRAME) {
		QdTimeSeq_var new_tt = Factory::clone(vec.t);
		if(new_tt.is_nil()){
			QplugAppendTextDisplay ("Cannot create new time tags\n");
			return QPLUG_FAILURE;
		}
		set_timetags((QdObject_var)vec.ptr, new_tt);

		ts_out = vec.ptr;
	}
	else{
		ts_out = NULL;
		QplugAppendTextDisplay (QrErrorMessage(check));
		status = QPLUG_FAILURE;
	} // end if on check
	return status;
}

// =========  Model functions  =================================================

extern void pomme_geocentric (int control[MAXCONTROL],
		double fday, double r, double lat, double lon,
		double est, double ist, double imf_by, double f107, double Em,  /* time lagged but not peaked or subtracted mean */
		double *bx, double *by, double *bz);

/**
 *\param[in] x, y, z : cartesian coordinate of the position with respect to geo-center.
 *\param[out] lon, lat, r : position in spherical coordinates. lon and lat are given in RADIANS!
 */
void ConvertGeoXYZ2LongLatR(double x, double y, double z, double &lon, double &lat, double &r)
{
	double temp = x*x + y*y;
	r = sqrt(temp + z*z);
	lon = atan2(y, x);
	lat = atan2(z, sqrt(temp));
}

/**
 * \param[in] B_nec is an array with 3 elements
 * \param[in] lon, lat (radian) define geographic position of the origin of the vector. These must be in radians..
 * \param[out] B_geo is an array with 3 elements
 */
void ConvertNEC2Geo(double *B_nec, double lon, double lat, double *B_geo)
{
	double Bsph[3];  //frame defined as: [in colatitude direction; radially; eastward]
	Bsph[0] = -B_nec[2];
	Bsph[1] = -B_nec[0];
	Bsph[2] =  B_nec[1];
	double colat = M_PI/2 - lat;
	double sin_col = sin(colat), cos_col = cos(colat), sin_lon = sin(lon), cos_lon = cos(lon);
	B_geo[0] = Bsph[0] * sin_col * cos_lon  +  Bsph[1] * cos_col * cos_lon  +  Bsph[2] * (-sin_lon);
	B_geo[1] = Bsph[0] * sin_col * sin_lon  +  Bsph[1] * cos_col * sin_lon  +  Bsph[2] * ( cos_lon);
	B_geo[2] = Bsph[0] * cos_col            +  Bsph[1] * (-sin_col);
}

QplugReturnStatus ComputePommeModel(QdRMatrixSeq_var &pos, TExternalParameters &parameters, QdRMatrixSeq_var &B_field)
{
	QplugReturnStatus status = QPLUG_SUCCESS;
	B_field = NULL;
	if (pos.is_nil())
	{
		status = QPLUG_FAILURE;
	}
	else
	{
		size_t nr_measurements = pos->sequence_size();
		B_field = new QdRMatrixSeq(nr_measurements, RMatrix(Dim(3), 0.0));

		//Initialize model's variables:
		tcontrol control;        /* array of user-control parameters */

		/* mainfield_open_file_mp initializes the control parameters. You can now modify these: */
		control[POS_0] =  parameters.oder_internal_field;  //133;   /* default: model.deg_int, 0=off */
		control[POS_1] =         15;    /* default: model.deg_1st, 0=off */

		control[POS_SM] =        DEG_SM;       /* degree of SM field, default = (*model).deg_sm, 0=off */
		control[POS_SM_IND] =    DEG_SM;       /* degree of field induced by SM field, default = (*model).deg_sm, 0=off */
		control[POS_F107] =      DEG_F107;     /* degree of SM field, default = (*model).deg_sm, 0=off */
		control[POS_F107_IND] =  DEG_F107;     /* degree of field induced by SM field, default = (*model).deg_sm, 0=off */
		control[POS_EST] =       1;                  /* Est dependent field: 1=on, 0=off */
		control[POS_IST] =       1;                  /* Ist dependent field: 1=on, 0=off */
		control[POS_EST_FAC] =   1;                  /* 1 = use model Est factor, 0 = use factor 1.0 */

		control[POS_GSM] =       DEG_GSM;     /* degree of GSM field, default = (*model).deg_gsm, 0=off */
		control[POS_GSM_IND] =   DEG_GSM;     /* degree of field induced by GSM field, default = (*model).deg_gsm, 0=off */
		control[POS_IMFBY] =     1;                 /* IMF-By dependent field: 1=on, 0=off */
		control[POS_IMFBY_IND] = 1;                 /* Induction by IMF-By dependent field: 1=on, 0=off */
		control[POS_EM] =        DEG_EM;      /* degree of GSM field, default = (*model).deg_gsm, 0=off */
		control[POS_EM_IND] =    DEG_EM;      /* degree of field induced by GSM field, default = (*model).deg_gsm, 0=off */


		// Computation of the model for each position
		QmjdTime day_Jan2000(2000, 1, 1, 0, 0, 0);
		QdTimeSeq_var times = get_timetags(pos);
		double lon = 0, lat = 0, r = 0;
		double B_nec[3], B_geo[3];  /* Resulting field */
		for (size_t t = 0; (t < nr_measurements) && (status != QPLUG_FAILURE); ++t) {
			ConvertGeoXYZ2LongLatR((double)(*pos)[t][0], (double)(*pos)[t][1], (double)(*pos)[t][2], lon, lat, r);
			// Universal time in floating point days since 00:00:00 on 1/1/2000
			double fday = times[t].getDiffDays(day_Jan2000);

			pomme_geocentric (control, fday, r, lat * 180.0 / M_PI, lon * 180.0 / M_PI, parameters.E_st
					, parameters.I_st, parameters.IMF_By, parameters.f107
					, parameters.Em  /* time lagged but not peaked or subtracted mean */
					, &(B_nec[0]), &(B_nec[1]), &(B_nec[2]) );
			ConvertNEC2Geo(B_nec, lon, lat, B_geo);
			(*B_field)[t][0] = B_geo[0];
			(*B_field)[t][1] = B_geo[1];
			(*B_field)[t][2] = B_geo[2];
		}
	}
	return status;
}

// =========  class Status  =================================================

Status::Status()
{
	message_ = "";
}

Status::~Status()
{
}

string Status::Message() const
{
	string result = "sf_Pomme: ";
	result += message_;
	return result;
}

void Status::PrintMessage(bool to_user, const char fmt[], ...)
{
	try {
		va_list args;
		char buffer[2048];
		va_start(args, fmt);
		vsprintf(buffer, fmt, args);
		va_end(args);
		if (to_user)
			QplugAppendTextDisplay(buffer);
		printf("%s", buffer);
		fflush(stdout);
	}
	catch (...) {
	}
}
