/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/
/*************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT:    Plug In library
* 
* MODULE:	Q4SC_Normal.cc
*
* LANGUAGE:	C++
*
* FUNCTION:     QSAS 4 SC Determination of Boundary Normal and Velocity wrapper
*
* PURPOSE:      Calculate the normal and boundary velocity based on the
*               crossing of the boundary (assumed planar) by 4 spacecraft.
*               Based on the methd in Ch 10.4.3 of the ISSI book.
*
* DESCRIPTION:  .
*
* ARGUMENTS:
*    IN:      QdCDOList  list of plug in objects containing
*
*   INPUT   VECTOR    dr1  pos of s/c1 as const vec or time ser
    INPUT   VECTOR    dr2 ,
    INPUT   VECTOR    dr3 ,
    INPUT   VECTOR    dr4 ,
    INPUT   TIME      tic,  time used if dr's are time series
    INPUT   TIME    ti1 , event time as s/c 1 in secs 
    INPUT   TIME    ti2 ,
    INPUT   TIME    ti3 ,
    INPUT   TIME    ti4 ,

    Out:

    OUTPUT  MATRIX             n   , normal vector with same frame as dr's
                                     n_x forced positive
    OUTPUT  FLOAT              Vn  , velocity along normal

* GLOBALS:    none
*
* RETURN:
*    TYPE:    QplugReturnStatus
*    VALUES:
*
*
* FUNCTIONS USED: Hacked from Boundary Normal routine
*
* HISTORY: November 2003 Steve Schwartz 
*          January 2004 AJA added error traps
*
***********************************************************************/

//#define Q4SC_DIAG

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <float.h>
#include "qplug_if.h"
#include "qrtn.h"

#include "qdutil.h"
#include "Qdos.h"
#include "qar.h"
#include "Xrefs.h"

//const double Re = 6.3712e6; //Earth radius [m]  - a global	

using namespace QSAS;  

// local prototypes

extern "C" QplugReturnStatus Q4SC_Normal(QplugArgList *);
QplugReturnStatus Q4SCPosition(QrVector *vecp,
                              QdObject_var pos_in,
			                  QdTime_var time_c);
QplugReturnStatus Q4SCConstPosition(QrVector *vecp, 
                                    QdRMatrix_var pos_in, 
				                    QdTime_var time_c);
QplugReturnStatus Q4SCCentrePosition(QrVector *vecp, 
                                    QdRMatrixSeq_var pos_in_seq, 
				                    QdTime_var time_c);

double SIConv_factor(QdObject_var in1_ptr);
void QrVector_diag_print(QrVector *vecp);
 
#define MAT_INV_EPS 10*DBL_EPSILON

typedef enum MAT_INV_RET_E{
MAT_INV_OK,
MAT_INV_SINGULAR,
MAT_INV_UNKNOWN
} MAT_INV_RET;

MAT_INV_RET mat_inv(double a[3][3], double inverse[3][3]);
double determinant(double mat[3][3]);
void multiply_mat(double a[3][3],double b[3][3],double prod[3][3]);
void multiply_mat_vec(double a[3][3],double b[3],double prod[3]);
void print_mat(double a[3][3]);
 
  
/**********************************************************************/
QplugReturnStatus Q4SC_Normal(  QplugArgList *call_list){
/**********************************************************************/

// define some C variables

int i,j,sign_nx;
double dt[4];
double  A[3][3], A_INV[3][3], B[3];
double  XX[3],n[3];
double drvec[4][3], signedmodX;
MAT_INV_RET ret_mat_inv;

// call-list arguments
 
QdObject_var         drin[4];        //  0-3
QdTime_var           tic;           //  4
QdTime_var           ti_tvar[4];    //  5-8 alt
QdRMatrix_var        n_var;          //  9
QdRScalar_var        Vn_var;         //  10

// define Qdos data objects

QrVector  dr[4];

QplugReturnStatus ret_status;

#ifdef Q4SC_DIAG
   printf("Q4SC Plugin Diagnostics On; undefine in .cc file and recompile\n");
   fflush(stdout);
#endif

// Ensure no input slots are empty

 if(call_list->size() != 9 ){
    QplugAppendTextDisplay("At least one input slot is empty");
    return QPLUG_FAILURE;
 }
 
// ...4 items are the s/c positions

for (i=0; i<4; i++){
  drin[i] = (*call_list)[i];
  if (drin[i].is_nil()){
    char pos_err_string[200]; 
    sprintf(pos_err_string,"Q4SC>> Error getting pos dr[%d] object\n",i+1);
    QplugAppendTextDisplay(pos_err_string);
  }
  #ifdef Q4SC_DIAG
     printf("Q4SC got input s/c %d position\n",i);
	 fflush(stdout);
  #endif
}

tic = QdTime_var::narrow((*call_list)[4]);
if( tic.is_nil() ) {
  QplugAppendTextDisplay ("Q4SC>> Time object for tic invalid.\n");
  return QPLUG_FAILURE;
}

#ifdef Q4SC_DIAG
   printf("Q4SC got input time interval\n",i);
   fflush(stdout);
#endif


for (i=0; i<4; i++){
  ret_status = Q4SCPosition(&(dr[i]),drin[i],tic);
  if (ret_status != QPLUG_SUCCESS){
    char pos_err_string[200]; 
    sprintf(pos_err_string,"Q4SC>> Q4SCPosition failed for sc[%d]\n",i);
    QplugAppendTextDisplay(pos_err_string);
    return QPLUG_FAILURE;
  }
  #ifdef Q4SC_DIAG
     printf("Q4SC calculated s/c %d position...\n",i);
	 fflush(stdout);
  #endif

  drvec[i][0] = 
   (double)*(QdRScalar_var::narrow(QuGetVectorComponent(dr[i].ptr,"x")));
  drvec[i][1] =
   (double)*(QdRScalar_var::narrow(QuGetVectorComponent(dr[i].ptr,"y")));
  drvec[i][2] =
   (double)*(QdRScalar_var::narrow(QuGetVectorComponent(dr[i].ptr,"z")));
  #ifdef Q4SC_DIAG
     printf("and filled drvec = (%g, %g,%g)\n",
               drvec[i][0],drvec[i][1],drvec[i][2]);
	fflush(stdout);
  #endif
}

// Whew! dr[i]'s now have a sequence with 1 element, the position
// of the crossing at s/c i with one timetag, and drvecs are C-arrays 

for (i=0; i<4; i++){
	ti_tvar[i]  = QdTime_var::narrow((*call_list)[i+5]);
	if (!ti_tvar[i].is_nil()) {
    	dt[i] = *(ti_tvar[i]) - *tic;
  	} 
  	else {
    	char time_err_string[200];
    	sprintf(time_err_string, "Q4SC ERROR>> t[%d] not time object\n",i+1);
    	QplugAppendTextDisplay(time_err_string);
    	return QPLUG_FAILURE;
  	}
  #ifdef Q4SC_DIAG
     printf("Q4SC calculated s/c %d dt as %g\n",i,dt[i]);
	 fflush(stdout);
  #endif
}
// dt[i] now contains times in secs for all 4 s/c

// Now fill dr matrix and right hand sides; s/c 4 is the reference
// [it doesn't matter which s/c is the ref; there are only 3 indep pairs]

for(i=0; i<3; i++){
  for(j=0; j<3; j++){
    A[i][j] = drvec[i][j] - drvec[3][j];
  }
  B[i]      = dt[i] - dt[3];
}

#ifdef Q4SC_DIAG
   printf("Q4SC A matrix is\n");
   print_mat(A);
   fflush(stdout);
#endif
ret_mat_inv = mat_inv(A,A_INV);
switch (ret_mat_inv) {
   case MAT_INV_OK:
     break;
   case MAT_INV_SINGULAR:
     QplugAppendTextDisplay("Q4SC ERROR>> dr matrix singular\n");
     return QPLUG_FAILURE;
     break;
   case MAT_INV_UNKNOWN: default:
     QplugAppendTextDisplay("Q4SC ERROR>> Matrix Inv unknown error\n");
     return QPLUG_FAILURE;
     break;
}
#ifdef Q4SC_DIAG
   printf("Q4SC A inv matrix found successfully as\n");
   print_mat(A_INV);
   fflush(stdout);
#endif


multiply_mat_vec(A_INV, B, XX);
sign_nx = (XX[0] < 0 ) ? -1 : 1;  
Vn_var = new QdRScalar();
signedmodX = (float)sign_nx*sqrt(XX[0]*XX[0] + XX[1]*XX[1] + XX[2]*XX[2]);
*Vn_var = 1.0/signedmodX;
n_var  = new QdRMatrix(Dim(3));
n_var->copy_xrefs_from( dr[3].ptr); // copy  frame, etc.

// added by AJA Feb 2004
set_timetags((QdObject_var)n_var, dr[3].t); // copy  time tag

#ifdef Q4SC_DIAG
   printf("Q4SC normal velocity is %g\n", (float)*Vn_var);
   fflush(stdout);
#endif

for (i=0; i<3; i++) {
  n[i]  = XX[i]/signedmodX;
 (* n_var)[i] = n[i];
}
#ifdef Q4SC_DIAG
   printf("Q4SC normal  is (%g, %g, %g)\n", n[0],n[1],n[2]);
   fflush(stdout);
#endif

// set the attributes for the output objects

QuSetTxtAttr("FIELDNAM", "Normal", n_var);
QuSetTxtAttr("FIELDNAM", "Vn", Vn_var);

#ifdef Q4SC_DIAG
   printf("Q4SC FIELDNAMs set\n");
   fflush(stdout);
#endif

QuSetTxtAttr("SI_conversion", "1>()", n_var);
char si_string[200];
sprintf(si_string, "%s s^-1",QuGetAttrText("SI_conversion",drin[3]).c_str());
//sprintf(si_string,"dummy SI insideQ4SC");
QuSetTxtAttr("SI_conversion",si_string,Vn_var);

#ifdef Q4SC_DIAG
   printf("Q4SC SI_conversions set\n");
   fflush(stdout);
#endif

QuSetTxtAttr("UNITS", " ", n_var);
char units_str[200];
sprintf(units_str,"%s/s",QuGetAttrText("UNITS",drin[3]).c_str());
//sprintf(si_string,"dummy UNITS insideQ4SC");
QuSetTxtAttr("UNITS",units_str,Vn_var);

#ifdef Q4SC_DIAG
   printf("Q4SC UNITS set\n");
   fflush(stdout);
#endif

char crossingt_str[200];
sprintf(crossingt_str,"%s", (dr[3].t)[0].iso_srep().c_str());
QuSetTxtAttr("Crossing Time",crossingt_str,Vn_var);
QuSetTxtAttr("Crossing Time",crossingt_str,n_var);

#ifdef Q4SC_DIAG
   printf("Q4SC Crossing Time set\n");
   fflush(stdout);
#endif

QuSetFrame(n_var, "vector", QuGetFrame(drin[3]).c_str(), QuGetRep(drin[3]).c_str());

#ifdef Q4SC_DIAG
   printf("Q4SC pushing back n_var result object\n");
   fflush(stdout);
#endif

call_list->push_back( (QdObject_var) n_var);     
#ifdef Q4SC_DIAG
   printf("Q4SC pushing back Vn result object\n");
   fflush(stdout);
#endif
call_list->push_back( (QdObject_var) Vn_var);     

// put some results to plug-in text window
char outstr[200];
sprintf(outstr,"Q4SC>> Crossing at %s\n",  crossingt_str);
QplugAppendTextDisplay(outstr);
sprintf(outstr,"Q4SC>> Normal is (%9.3g, %9.3g, %9.3g) ",
   n[0], n[1], n[2]);
QplugAppendTextDisplay(outstr);
sprintf(outstr,"Q4SC>> Velocity along normal is %g %s\n", 
    (float)*Vn_var, QuGetAttrText("UNITS",Vn_var).c_str());
QplugAppendTextDisplay(outstr);

QplugAppendTextDisplay("Q4SC>> finished successfully\n");

return QPLUG_SUCCESS;

}

/**********************************************************************/
QplugReturnStatus Q4SCPosition(QrVector *vecp,
                              QdObject_var pos_in,
							  QdTime_var time_c){
/**********************************************************************/

long  check_rtn;
QdRMatrixSeq_var pos_in_seq;
QdRMatrix_var    pos_in_const;
QplugReturnStatus ret_status;

// sort out position of crossing. Input can either be a time series or
// a constant vector, so this a bit complicated. Separate routines
// return in vecp a sequence of 1 cartesian vector whose timetag
// is the time  time_c, and whose frame is
// that of pos_in. 

pos_in_seq       = QdRMatrixSeq_var::narrow(pos_in);
pos_in_const     = QdRMatrix_var::narrow(pos_in);

if( !pos_in_seq.is_nil() ) {
  QplugAppendTextDisplay ("Q4SCCoords>> Vector Sequence. Interpolating ...\n");
  ret_status = Q4SCCentrePosition(vecp,pos_in_seq,time_c);
  if (ret_status != QPLUG_SUCCESS){
   QplugAppendTextDisplay("Q4SCCoords>> Failed to calculate crossing position\n");
   return QPLUG_FAILURE;
  }
} 
else if (!pos_in.is_nil()){
  ret_status = Q4SCConstPosition(vecp, pos_in_const, time_c);
  if (ret_status != QPLUG_SUCCESS){
    QplugAppendTextDisplay("Q4SC>> Failed to calc const. position\n");
    return QPLUG_FAILURE;
  }
} else {
  QplugAppendTextDisplay("Q4SC>> pos_in neither const nor sequence!?\n");
  return QPLUG_FAILURE;
}


return QPLUG_SUCCESS;
}

/**********************************************************************/
QplugReturnStatus  Q4SCConstPosition(QrVector *vecp, 
                                    QdRMatrix_var pos_in, 
				                    QdTime_var time_c){
/**********************************************************************/

QplugReturnStatus ret_status;
const char * xyz_comp[] = {"x", "y", "z"};
int i;

// grab constant vector and fill vecp.

if (!QuIsThreeVector(pos_in)){
   QplugAppendTextDisplay("Q4SC>> Const pos_in not a 3Vector\n");
   return QPLUG_FAILURE;
}

QdTimeSeq_var tt = new QdTimeSeq(1, *time_c);
vecp->ptr = new QdRMatrixSeq(1,RMatrix(Dim(3)));
(vecp->ptr)->copy_xrefs_from(pos_in);
QdRMatrixSeq_var pos_seq = new QdRMatrixSeq(1, (*pos_in) );
pos_seq->copy_xrefs_from(pos_in);
for (i=0; i<3; i++){
  QdRScalarSeq_var comp = 
     QdRScalarSeq_var::narrow(QuGetVectorComponent(pos_seq,xyz_comp[i]));
  if (comp.is_nil()){
     QplugAppendTextDisplay("Q4SC>> Const pos comp narrow failed\n");
     return QPLUG_FAILURE;
  }
  (vecp->ptr)[0][i] = comp[0];
}
vecp->n_recs = 1;
vecp->t = tt;

vecp->in_frame = vecp->QrGetFrameFromAttr(pos_in);
if (vecp->in_frame == UNKNOWN){
  QplugAppendTextDisplay("Q4SC>> Const pos_in unknown frame\n");
  return QPLUG_FAILURE;
}

return QPLUG_SUCCESS;
}


/**********************************************************************/
QplugReturnStatus  Q4SCCentrePosition(QrVector *vecp, 
                                    QdRMatrixSeq_var pos_in_seq, 
				                    QdTime_var time_c){
/**********************************************************************/

int i,nttin,ntrue,il,ih;
QplugReturnStatus ret_status;
double cl,ch, frac;
const char * xyz_comp[] = {"x", "y", "z"};

if (!QuIsThreeVector(pos_in_seq)){
   QplugAppendTextDisplay("Q4SC>> Sequence pos_in not a 3Vector\n");
   return QPLUG_FAILURE;
}
// prepare vecp 
QmjdTime tc = *time_c;
QdTimeSeq_var tt = new QdTimeSeq(1, tc);
vecp->ptr = new QdRMatrixSeq(1,RMatrix(Dim(3)));
(vecp->ptr)->copy_xrefs_from(pos_in_seq);
vecp->t = tt;
vecp->n_recs = 1;
vecp->in_frame = vecp->QrGetFrameFromAttr(pos_in_seq);
if (vecp->in_frame == UNKNOWN){
  QplugAppendTextDisplay("Q4SC>> Sequence pos_in unknown frame\n");
  return QPLUG_FAILURE;
}

// linearly interpolate crossing position from bracketting time tags
// If any problems, default to 1st vector in pos_in_seq
// There are probably much simpler ways to do this. Note, though, that
// Qrtn routines only give correct answers if given x,y,z representation

QdTimeSeq_var tt_in = get_timetags(pos_in_seq);
if (tt_in.is_nil()){
  QplugAppendTextDisplay(
     "Q4SC>> No timetags for input pos seq. Taking 1st pt\n");
     il = ih = 0;
} else {
  nttin = tt_in->size();
  if (nttin <= 0){
     QplugAppendTextDisplay("Q4SC>> Input seq 0 length\n");
     return QPLUG_FAILURE;
  } else if (nttin == 1){
     il = ih = 0;
  } else {
     SequenceMask msk(tt_in->get_mask( QdTimeSeq::LE( tc)));
     ntrue = msk.num_true();
     if (ntrue <= 0) il = ih = 0;
     else {
       il = ntrue -1;
       ih =  (ntrue < nttin) ?   il + 1 : il; 
     }  
  }
}

if(ih == il) frac = 0.0;   // safe
else frac = (tc-tt_in[il])/(tt_in[ih] - tt_in[il]); // calc lin interpol factor

/* printf("tl: %s\ntc: %s\nth: %s\nfrac = %g\n",
                tt_in[il].iso_srep().c_str(),
		tc.iso_srep().c_str(),
		tt_in[ih].iso_srep().c_str(),
		frac);
*/

for (i=0; i<3; i++){                          // fill crossing position
  QdRScalarSeq_var comp;
  if(!pos_in_seq.is_nil()) comp = QdRScalarSeq_var::narrow(QuGetVectorComponent(pos_in_seq, xyz_comp[i]));
  if (comp.is_nil()){
     QplugAppendTextDisplay("Q4SC>> Centre Pos comp narrow failed\n");
     return QPLUG_FAILURE;
  }
  if (ih==il) {
    vecp->ptr[0][i] = (double)comp[il];
  } else {
    cl = (double)comp[il];
    ch = (double)comp[ih];
    vecp->ptr[0][i] = cl + frac*(ch-cl);
  }
}

return QPLUG_SUCCESS;
}


/***********************************************************************/
double SIConv_factor(QdObject_var in1_ptr){ 
/***********************************************************************/

 double value1;

  string siconv1 = QuGetAttrText("SI_conversion", in1_ptr);
  
  if( siconv1.empty() ) {
    siconv1 = "1 > ()";
  } 
  
  size_t ptr1 = siconv1.find(">");
  if( ptr1 == string::npos) {
     return 0.0;
  }
  value1 = (double) atof(siconv1.c_str());
   
  return value1;
  
} // end SIConv_factor

/**************************************************************************/
void QrVector_diag_print(QrVector *vecp){
  printf("IsPositionVector = %d, in_frame = %d, out_frame = %d, n_recs = %d\n",
        vecp->IsPositionVector,vecp->in_frame,vecp->out_frame, vecp->n_recs);
  printf("First Vector: (%9.3g, %9.3g, %9.3g) at %s\n",
     (double)vecp->ptr[0][0], (double)vecp->ptr[0][1], (double)vecp->ptr[0][2],
     vecp->t[0].iso_srep().c_str());
	 fflush(stdout);
  return;
}



MAT_INV_RET mat_inv(double a[3][3], double inverse[3][3]){

double det;


det = determinant(a);
if (fabs(det) < MAT_INV_EPS) return MAT_INV_SINGULAR;

inverse[0][0] =  ( a[1][1]*a[2][2] - a[1][2]*a[2][1] )/det;
inverse[0][1] = -( a[0][1]*a[2][2] - a[0][2]*a[2][1] )/det;
inverse[0][2] =  ( a[0][1]*a[1][2] - a[0][2]*a[1][1] )/det;
inverse[1][0] = -( a[1][0]*a[2][2] - a[1][2]*a[2][0] )/det;
inverse[1][1] =  ( a[0][0]*a[2][2] - a[0][2]*a[2][0] )/det;
inverse[1][2] = -( a[0][0]*a[1][2] - a[0][2]*a[1][0] )/det;
inverse[2][0] =  ( a[1][0]*a[2][1] - a[1][1]*a[2][0] )/det;
inverse[2][1] = -( a[0][0]*a[2][1] - a[0][1]*a[2][0] )/det;
inverse[2][2] =  ( a[0][0]*a[1][1] - a[0][1]*a[1][0] )/det;

return MAT_INV_OK;


}



double determinant(double m[3][3]){

double d;

d  = m[0][0]*m[1][1]*m[2][2] + m[0][1]*m[1][2]*m[2][0] +m [0][2]*m[1][0]*m[2][1];
d -= m[2][0]*m[1][1]*m[0][2] + m[2][1]*m[1][2]*m[0][0] + m[2][2]*m[1][0]*m[0][1];

return d;
}

void multiply_mat(double a[3][3],double b[3][3], double prod[3][3]){
int i,j,k;

for (i=0; i<3; i++){
  for (j=0; j<3; j++){
    for (prod[i][j]=0.0,k=0; k<3; k++) prod[i][j]+=a[i][k]*b[k][j];
  }
}

return;

}

void multiply_mat_vec(double a[3][3],double b[3], double prod[3]){
int i,j;

for (i=0; i<3; i++){
    for (prod[i]=0.0,j=0; j<3; j++) prod[i]+=a[i][j]*b[j];
}


return;

}

void print_mat(double a[3][3]){

int i,j;

for (i=0; i<3; i++){
   for (j=0; j<3; j++) printf("%10.4g ", a[i][j]);
   printf("\n");
   fflush(stdout);
}
return;
}


