/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/

/**********************************************************************
*
*  QSAS Plugin to stretch timetags of a time series; also displaces
*
*
*  Inputs:  Time Series
*           Original Time Interval
*           Time Interval to Stretch To
*
*  Output:  Cloned Time series with displaced & stretched tags and all xrefs
*
*  History:  sjs September 2001
***********************************************************************/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "qplug_if.h"
#include "qdutil.h"
#include "Qdos.h"

using namespace QSAS;

// local prototype
extern "C" QplugReturnStatus TimeStretch (QplugArgList *);

typedef enum SeqType_e{SCALAR_SEQ, MATRIX_SEQ, UNKNOWN_SEQ} SeqType;

QplugReturnStatus TimeStretch  ( QplugArgList *call_list )
{

// call-list arguments

QdObject_var tsin_obj;
QdTimeInterval_var tint_in, tint_out;
QdObject_var tsout_obj;

// internal variables
QdRScalarSeq_var tsin_sseq, tsout_sseq;
QdRMatrixSeq_var tsin_mseq, tsout_mseq;
SeqType seq_type = UNKNOWN_SEQ;
int i, nsub;

QplugAppendTextDisplay("\n\n");

//	Unpack the input data objects
//  set pointer to objects and check objects

tsin_obj = QdObject_var::narrow((*call_list)[0]);
tsin_sseq = QdRScalarSeq_var::narrow(tsin_obj);
tsin_mseq = QdRMatrixSeq_var::narrow(tsin_obj);
tint_in   = get_timeinterval((*call_list)[1]);
tint_out  = get_timeinterval((*call_list)[2]);

if (tsin_obj.is_nil() ){
   QplugAppendTextDisplay("TimeStretch> Input TS is nil\n");
   return QPLUG_FAILURE;
}
if (tsin_sseq.is_nil() && tsin_mseq.is_nil()){
   QplugAppendTextDisplay("TimeStretch> Input TS not sequence\n");
   return QPLUG_FAILURE;
}
if (!tsin_sseq.is_nil() && !tsin_mseq.is_nil()){
   QplugAppendTextDisplay("TimeStretch> Input TS sequence confused\n");
   return QPLUG_FAILURE;
}
if (!tsin_sseq.is_nil()) {seq_type = SCALAR_SEQ;
} else if (!tsin_mseq.is_nil()) {seq_type = MATRIX_SEQ;
} else {
   QplugAppendTextDisplay("TimeStretch> Input TS seq; shouldn't get here!\n");
   return QPLUG_FAILURE;
}
if (tint_in.is_nil()){
   QplugAppendTextDisplay("TimeStretch> Input Interval is nil\n");
   return QPLUG_FAILURE;
}
if (tint_out.is_nil()){
   QplugAppendTextDisplay("TimeStretch> Output Interval is nil\n");
   return QPLUG_FAILURE;
}
if (!tsin_obj->xref_exists(DEPEND_0)){
   QplugAppendTextDisplay("TimeShift> Input TS not a time series!\n");
   return QPLUG_FAILURE;
}

// get timetags and sequence mask to subset  input series to tint_in

QdTimeSeq_var tt_in = get_timetags(tsin_obj);
SequenceMask msk(tt_in->get_mask(
      QdTimeSeq::InRange(tint_in->start(), tint_in->end())));
QdTimeSeq_var ttin_sub = new QdTimeSeq();
tt_in->get_subsequence(*ttin_sub, msk);
nsub = ttin_sub->size();
if (nsub <= 0){
   QplugAppendTextDisplay("TimeStretch>No data points in Orig. Interval\n");
   return QPLUG_FAILURE;
}

// create output object and copy values using mask
switch (seq_type) {
   case SCALAR_SEQ:
     tsout_sseq = new QdRScalarSeq();
     tsin_sseq->get_subsequence(*tsout_sseq,msk);
     tsout_obj = tsout_sseq;
     break;
   case MATRIX_SEQ:
     tsout_mseq = new QdRMatrixSeq();
     tsin_mseq->get_subsequence(*tsout_mseq,msk);
     tsout_obj = tsout_mseq;
     break;
   case UNKNOWN_SEQ: default:
     QplugAppendTextDisplay("TimeStretch>Tried to create UNKNOWN_SEQ\n");
     return QPLUG_FAILURE;
}    

tsout_obj->copy_xrefs_from(tsin_obj);

// create new tags and stretch from old ones
QdTimeSeq_var tt_out = new QdTimeSeq(nsub);
double stretch_factor = 
   (tint_out->end() - tint_out->start())/(tint_in->end() - tint_in->start());
QdTime tout_start(tint_out->start());
QdTime tin_start(tint_in->start());

for(i=0; i<nsub; i++){
  tt_out[i] = tout_start + (ttin_sub[i] - tin_start) * stretch_factor;
}

set_timetags( (QdObject_var) tsout_obj, tt_out);
char old_int[200], new_int[200];
sprintf(old_int,"%s - %s",tint_in->start().iso_srep().c_str(),
           tint_in->end().iso_srep().c_str());
sprintf(new_int,"%s - %s",tint_out->start().iso_srep().c_str(),
           tint_out->end().iso_srep().c_str());
QuSetTxtAttr("Original Time Interval",old_int,tsout_obj);
QuSetTxtAttr("New Stretched Time Interval",new_int,tsout_obj);

QplugAppendTextDisplay("Old Interval = ");
QplugAppendTextDisplay(old_int);QplugAppendTextDisplay("\n");
QplugAppendTextDisplay("Stretched to New Interval = ");
QplugAppendTextDisplay(new_int);QplugAppendTextDisplay("\n");

call_list->push_back(tsout_obj);

QplugAppendTextDisplay ("TimeStretch: Done\n");
  
return QPLUG_SUCCESS;

}

