/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/
/* PROJECT: Cluster UK CDHF
*
* COMPONENT:    VectorTSDecomp
* 
* MODULE:	QSAS Decomposition of Vector Time Series
*
* LANGUAGE:	C++
*
* FUNCTION:
*
* PURPOSE: Decompose a vector time series into a Scalar Time Series of
*                    one of its components (x,y,z,r,theta,phi)
* 
* DESCRIPTION:
*
* ARGUMENTS:
*    IN:
*     VECTOR_TIME_SERIES    ts_in                 input vector time series 
*                                                  data object
*     INTEGER               componentdo           flag for which component   
*    OUT:
*        SCALAR_TIME_SERIES ts_out
* GLOBALS:
*
* RETURN:
*    TYPE:     QplugReturnStatus
*    VALUES:   see qplug_if.h
*
* FUNCTIONS_USED:	 QpAxisLabelMakeStem - from QPUI to make label stem
*
* ALGORITHM:   pretty obvious, most of the work is in the attributes
*
* HISTORY:		Steve Schwartz, July 1996
*   Amended to use Qfui interface instead of Qai - A R Hare, 12 Sep 96.
*   SJS  19 May 97 - amended to use Qplug instead of Qplug 
*   SJS  30 June 97 - fixed some memory leaks
*
***********************************************************************/


#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "qplug_if.h"
#include "Qdos.h"
#include "qdutil.h"
#include "VectorTSDecomp.h"

using namespace QSAS;

QplugReturnStatus  VectorTSDecomp  ( QplugArgList *call_list )

{

QdObject_var  ts_in;
QdObject_var  ts_out;
QdRScalar_var  componentdo;
int             icomponent;

string fieldnam;                  // attributes
string siconversion;
string units;
string lablaxis;
string stem;
double hs,hsp;

int    ndata_in;                                  /* number of data points  */
                                                  /* array new'ed below  */

//
//	...first, unpack the input data objects
//

//	..first item is input Time Series

ts_in = (*call_list)[0];

if ( !QuIsThreeVector(ts_in) ) {	
   QplugAppendTextDisplay (
"VectorTSDecomp>> ERROR - input not a Vector Time Series!\n");
	return QPLUG_FAILURE; 
}        

ndata_in = ts_in->sequence_size();


//	2nd item integer flag for component (order same as PUI):
//           0 = x
//           1 = y
//           2 = z
//           3 = r(mag)
//           4 = Lat(deg)
//           5 = theta(deg)
//           6 = phi(deg)
//           7 = Lat(rad)
//           8 = theta(rad)
//           9 = phi(rad)


componentdo = QdRScalar_var::narrow((*call_list)[1]);

 if ( componentdo.is_nil() ) { 
    QplugAppendTextDisplay (
        "VectorTSDecomp>> ERROR - argument not an integer flag!\n");
    return QPLUG_FAILURE; 
}  // end if

 icomponent=(int)(*componentdo);

 stem = QpAxisLabelMakeStem(ts_in);            // Make stem for label

 fieldnam = QuGetAttrText("FIELDNAM", ts_in);
       
       
switch(icomponent){
     case(0):            // x component
        units = QuGetAttrText("UNITS", ts_in);
             
        siconversion = QuGetAttrText("SI_conversion", ts_in);
        if( siconversion.empty() )
		{                                    // set SI_conversion
          siconversion = "1>(unknown)";     //  empty if fail
        }      
        
        lablaxis = stem;
        lablaxis += "_x";
        fieldnam += "_x";
        ts_out = QuGetVectorComponent(ts_in, "x");
        break;
     case(1):            // y component
        units = QuGetAttrText("UNITS", ts_in);

        siconversion = QuGetAttrText("SI_conversion", ts_in);
        if( siconversion.empty())
		{
                                             // set SI_conversion
          siconversion = "1>(unknown)";     //  empty if fail
        }      
        
        lablaxis = stem;
        lablaxis += "_y";
        fieldnam += "_y";
        ts_out = QuGetVectorComponent(ts_in, "y");
        break;
     case(2):            // z component
        units = QuGetAttrText("UNITS", ts_in);

        siconversion = QuGetAttrText("SI_conversion", ts_in);
        if( siconversion.empty())
		{
                                             // set SI_conversion
          siconversion = "1>(unknown)";     //  empty if fail
        }      

        lablaxis =stem;
        lablaxis += "_z";
        fieldnam += "_z";
        ts_out = QuGetVectorComponent(ts_in, "z");
        break;
     case(3):            // r component
        units = QuGetAttrText("UNITS", ts_in);

        siconversion = QuGetAttrText("SI_conversion", ts_in);
        if( siconversion.empty())
		{
                                             // set SI_conversion
          siconversion = "1>(unknown)";     //  empty if fail
        }      

        lablaxis = "|";
        lablaxis += stem;
        lablaxis += "|";
        fieldnam += "_magnitude";
        ts_out = QuGetVectorComponent(ts_in, "r(mag)");
        break;
     case(4):            // Lat(deg) component
        units = "deg"; 
        siconversion = "1>degree";
        lablaxis = "Lat_";
        lablaxis += stem;
        fieldnam += "_Lat(deg)";
        ts_out = QuGetVectorComponent(ts_in, "Lat(deg)");
        break;
     case(5):            // theta(deg) component
        units = "deg"; 
        siconversion = "1>degree";
        lablaxis = "theta_";
        lablaxis += stem;
        fieldnam += "_theta(deg)";
        ts_out = QuGetVectorComponent(ts_in, "theta(deg)");
        break;
     case(6):            // phi(deg) component
        units = "deg"; 
        siconversion = "1>degree";
        lablaxis = "phi_";
        lablaxis += stem;
        fieldnam += "_phi(deg)";
        ts_out = QuGetVectorComponent(ts_in, "phi(deg)");
        break;
     case(7):            // Lat(rad) component
        units = "rad"; 
        siconversion = "1>rad";
        lablaxis = "Lat_";
        lablaxis += stem;
        fieldnam += "_Lat(rad)";
        ts_out = QuGetVectorComponent(ts_in, "Lat(rad)");
        break;
     case(8):            // theta(rad) component
        units = "rad"; 
        siconversion = "1>rad";
        lablaxis = "theta_";
        lablaxis += stem;
        fieldnam += "_theta(rad)";
        ts_out = QuGetVectorComponent(ts_in, "theta(rad)");
        break;
     case(9):            // phi(rad) component
        units = "rad"; 
        siconversion = "1>rad";
        lablaxis = "phi_";
        lablaxis += stem;
        fieldnam += "_phi(deg)";
        ts_out = QuGetVectorComponent(ts_in, "phi(rad)");
        break;
     default:                     // Shouldn't get here!   
        QplugAppendTextDisplay 
           ("VectorTSDecompAVG>> Unknown component requested; Failure\n");
        return QPLUG_FAILURE;
        break;
}                                 // end switch(icomp)


// Create the output Time Series and fill it

 QdObject_var qd0 = get_Depend_0( ts_in );
 if ( !qd0.is_nil() ) ts_out->change_xref(DEPEND_0 , qd0 );

// load all the attributes, and the half spacing/sample period:

 QuSetTxtAttr("FIELDNAM", fieldnam.c_str(), ts_out);
 QuSetFrame(ts_out, "scalar");
 QuSetTxtAttr("SI_conversion", siconversion.c_str(), ts_out);
 QuSetTxtAttr("UNITS", units.c_str(), ts_out);
 QuSetTxtAttr("LABLAXIS", lablaxis.c_str(), ts_out);

//	..add  the output series to the list

  call_list->push_back(ts_out);     

// Clean up:


QplugAppendTextDisplay ("VectorTSDecomp>> data object created - exiting successfully\n");

return QPLUG_SUCCESS;

}
