/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/
/*************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT:    Plug In library
* 
* MODULE:	WindowSpectrum.cc
*
* LANGUAGE:	C++
*
* FUNCTION:     QSAS WindowSpectrum wrapper
*
* PURPOSE:      FFT windowed powerspectrum.
* 
* DESCRIPTION:  Plug In module checking input parameters.
*
* ARGUMENTS:
*    IN:      QdCDOList  list of plug in objects containing
*              QdCTimeSeries*   vector or scalar regular time series object
*              QdCTimeInterval inter vall object for subsetting
*              nfreq  INT number of frequencies
*              overlap 0/1 set to 1 the windows will overlap by half their length
*              WindowType  0-3 square/Hanning/Welch/Bartlett
*
*   OUT:
*         double *frequency       frequency tags		
*         double *power_mag       power spectral density of magnitude
*         vector series *power_vec       power spectral density of components
*
* GLOBALS:
*
* RETURN:
*    TYPE:    QplugReturnStatus
*    VALUES:
*
* FUNCTIONS_USED:	QSpctrm.c QWindow.c 
*
* ALGORITHM:  see Press et al. 'Numerical Recipes in C' (1992)
*             We adopted the NR spctrm routine for direct datainput
*             with C-indexing and replaced the NR four1 by Ooura's cdft.
*
* HISTORY: created  05/06/97 Markus Fraenz 
*          migrated to qdos 2 - AJ Allen Oct 2000
*          vector spectrum and subsetting added Markus Fraenz 18/01/2001
*          changed from float to double precision
*          changed to 0-offset indexing
*          replaced NR four1 by Takuya OOURA cdft Markus Fraenz 10/07/2001
*
***********************************************************************/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "qplug_if.h"
#include "qdutil.h"
#include "Qdos.h"

extern "C" QplugReturnStatus  WindowSpectrumStart( QplugArgList *);

double	 QWindow( int WType, int j, double a, double b );
void cdft(int n, int isgn, double *a, int *ip, double *w);
void makewt(int nw, int *ip, double *w);
void bitrv2(int n, int *ip, double *a);
void cftbsub(int n, double *a, double *w);
void bitrv2conj(int n, int *ip, double *a);
void cftfsub(int n, double *a, double *w);
void cftmdl(int n, int l, double *a, double *w);
void cft1st(int n, double *a, double *w);

void QSpctrm(double data[], double p[], int m, int k, int ovrlap, int   WindowType, double w1[], double w2[]);

using namespace QSAS;

static double sqrarg;
#define SQR(a) ((sqrarg=(a)) == 0.0 ? 0.0 : sqrarg*sqrarg)
#define SWAP(a,b) tempr=(a);(a)=(b);(b)=tempr



QplugReturnStatus  WindowSpectrumStart( QplugArgList *call_list )
{


// define some C variables


int	        attribute_ok,list_length,overlap=0,WindowType=0,vectorset=0;
unsigned long	Ndata,Ndata0=0;
int             nfreq, nensemble=0, power2=2;


char str_out[150];


double  *data,*d,*pf,*freq,*w1,*w2;
double deltat=0.0, deltat2=0.0;

// define QSAS data objects

QdTimeInterval_var start_end;
QdObject_var ts_out;
QdTimeSeq_var tt_out;

QdRScalarSeq_var  ts_inr;
QdRScalarSeq_var  out_freq_series;
QdRScalarSeq_var  out_spmag_series;
QdObject_var  out_spvec_series;
QdRMatrixSeq_var ts_inm;



//
// ...first, unpack the input data objects
//


// ...first item is input Time Series

QdObject_var ts_in = (* call_list)[0];

//  check QSAS datatype

QdTimeSeq_var tt_in = get_timetags(ts_in);
if( tt_in.is_nil() ) {
  QplugAppendTextDisplay ("WindowSpectrumFFT>> Failed to get time tags\n");
  return QPLUG_FAILURE;
}
if ( !QuIsTSRegular(tt_in, &deltat) ) {
	QplugAppendTextDisplay ("WindowSpectrum>> ERROR - time series not evenly spaced\n");
	return QPLUG_FAILURE; 
}


//	..2nd item is option 'time_interval'

 QdObject_var dobj1 = (* call_list)[1];
 start_end = get_timeinterval(dobj1);
 
//	..3rd item is option 'nfreq'


QdObject_var dobj2 = (* call_list)[2];
QdRScalar_var option_2 = QdRScalar_var::narrow(dobj2);

if ( option_2.is_nil() ) { 
	QplugAppendTextDisplay ("WindowSpectrum>> ERROR - nfreq must be integer.\n");
	return QPLUG_FAILURE; 
}

nfreq = (int)(*option_2);
if (nfreq <= 1) {
	QplugAppendTextDisplay ("WindowSpectrum>> ERROR - nfreq must be greater than 1.\n");
	return QPLUG_FAILURE; }

//	..4th item is option 'overlap'


QdObject_var dobj3 = (* call_list)[3];
QdRScalar_var option_3 = QdRScalar_var::narrow(dobj3);

if ( option_3.is_nil() ) { 
	QplugAppendTextDisplay ("WindowSpectrum>> ERROR - overlap must be 0 or 1.\n");
	return QPLUG_FAILURE; 
}

overlap = (int)(*option_3);
if (overlap < 0) {
	QplugAppendTextDisplay ("WindowSpectrum>> ERROR - overlap must be 0 or 1.\n");
	return QPLUG_FAILURE; 
}


//	..5th item is option 'window type'

QdObject_var dobj4 = (* call_list)[4];
QdRScalar_var option_4 = QdRScalar_var::narrow(dobj4);

if ( option_4.is_nil() ) { 
	QplugAppendTextDisplay ("WindowSpectrumFFT>> ERROR - Window Type must be integer.\n");
	return QPLUG_FAILURE; 
}

WindowType = (int)(*option_4);

if ((WindowType > 3) || (WindowType < 0)) {
	QplugAppendTextDisplay ("WindowSpectrumFFT>> ERROR - Window Type must be 0,1,2 or 3.\n");
	return QPLUG_FAILURE; 
}



// now take the subset for the time interval chosen  ###SUBSETTING START

QplugAppendTextDisplay ("WindowSpectrum>> Subsetting Time Series...\n");
SequenceMask msk(tt_in->get_mask(QdTimeSeq::InRange(start_end->start(), start_end->end())));
 
tt_out = new QdTimeSeq();
  
tt_in->get_subsequence(*tt_out, msk);

  
QdRScalarSeq_var ss = QdRScalarSeq_var::narrow(ts_in);
if(!ss.is_nil() ){
    QdRScalarSeq_var  ss_out = new QdRScalarSeq();
    ss->get_subsequence(*ss_out, msk);
    ts_out = (QdObject_var) ss_out;
}
else{
    QdRMatrixSeq_var ms = QdRMatrixSeq_var::narrow(ts_in);
    if(ms.is_nil() ){
      QplugAppendTextDisplay ("WindowSpectrumFFT>> ERROR - sequence type not handled\n");
      return QPLUG_FAILURE; 
    }
    if(QuIsThreeVector(ms) ){
      QdRMatrixSeq_var  ms_out = new QdRMatrixSeq();
      ms->get_subsequence(*ms_out, msk);
      ts_out = (QdObject_var) ms_out;
    }
    else {
      QplugAppendTextDisplay ("WindowSpectrumFFT>> ERROR - sequence type not handled\n");
      return QPLUG_FAILURE; 
    }
}
 

if ( ts_out.is_nil() ){
    QplugAppendTextDisplay ("WindowSpectrumFFT>> ERROR - subset failed:\n");
    return QPLUG_FAILURE;
} 
if ( ts_out->sequence_size() <=1 ){
    QplugAppendTextDisplay ("WindowSpectrumFFT>> Empty subset:\n");
    return QPLUG_FAILURE;
}

ts_in->copy_xrefs_to(ts_out);
set_timetags(ts_out, tt_out);


//	..extract the time tags from the data series object ###SUBSETTING END

Ndata = ts_out->sequence_size();

if (Ndata <= 1) {
	QplugAppendTextDisplay ("WindowSpectrum>> ERROR - Series must have at least 2 data points.\n");
	return QPLUG_FAILURE; }

if (nfreq > Ndata) {
	QplugAppendTextDisplay ("WindowSpectrum>> ERROR - nfreq must be smaller than number of data points.\n");
	return QPLUG_FAILURE; }


// calculate true nfreq as power of 2

while (power2 < nfreq) power2 *= 2;
if (power2 > nfreq) nfreq = power2/2;


// calculate full number of data points


while (Ndata0 < Ndata) {
  nensemble++;
  if (overlap > 0) {
    Ndata0=(2*nensemble+1)*nfreq ;
  }
  else {
    Ndata0=4*nensemble*nfreq ;
  }
}


QplugAppendTextDisplay ("WindowSpectrum>> Calculating FFT Spectrum.\n");
//	call the NR memcof power spectrum calculation
//      (notice that NR routines start indices with 1!!) 


pf = new double[nfreq];
freq = new double[nfreq];

w1 = new double[4*nfreq];
w2 = new double[nfreq];

deltat2=2.0*(double)deltat*nfreq;
for(int i=0;i<nfreq;i++)  freq[i]=(double)(((double)i)/deltat2);



data  = new double[Ndata0];

// pad with zeroes

for (int i=Ndata;i<Ndata0;i++) data[i]=0.0;

//	..extract the data from the data series object

if ( QuIsThreeVector(ts_out) ) {
  vectorset=1;	
  QplugAppendTextDisplay ("WindowSpectrum>> Decomposing Vector Timeseries...\n");
  ts_inm = QdRMatrixSeq_var::narrow(ts_out);
  if ( ts_inm.is_nil() ) { return QPLUG_FAILURE; }
  ts_inr = QdRScalarSeq_var::narrow(QuGetVectorComponent(ts_out, "r(mag)")); 
  if ( ts_inr.is_nil() ) { return QPLUG_FAILURE; } 
}
else
 ts_inr = QdRScalarSeq_var::narrow(ts_out);


for(long i=0; i<Ndata; i++)  data[i] = (double) (ts_inr[i]);


// Qspctrm still uses Fortran array indices

 QSpctrm(data, pf , nfreq, nensemble, overlap, WindowType, w1, w2);

//	..create the output data series objects

out_spmag_series = new QdRScalarSeq(nfreq);
for (long i =0; i<nfreq; i++) out_spmag_series[i] = (double) pf[i];


if(vectorset) { 
 QdRMatrixSeq_var spvec_series = new QdRMatrixSeq(nfreq,RMatrix(Dim(3)));
 for(long i=0; i<Ndata; i++)  data[i] = (double) (ts_inm[i][0]);
 QSpctrm(data, pf , nfreq, nensemble, overlap, WindowType, w1, w2);
 for (long i =0; i<nfreq; i++) spvec_series[i][0] = (double) pf[i];

 for(long i=0; i<Ndata; i++)  data[i] = (double) (ts_inm[i][1]);
 QSpctrm(data, pf , nfreq, nensemble, overlap, WindowType, w1, w2);
 for (long i =0; i<nfreq; i++) spvec_series[i][1] = (double) pf[i];

 for(long i=0; i<Ndata; i++)  data[i] = (double) (ts_inm[i][2]);
 QSpctrm(data, pf , nfreq, nensemble, overlap, WindowType, w1, w2);
 for (long i =0; i<nfreq; i++) spvec_series[i][2] = (double) pf[i];
 out_spvec_series = (QdObject_var) spvec_series;
}

out_freq_series  = new QdRScalarSeq(nfreq);
for (long i =0; i<nfreq; i++) out_freq_series[i] = (double) freq[i];

//	...set some attributes

  QuSetTxtAttr("SI_conversion", " ", (QdObject_var) out_spmag_series);
  QuSetTxtAttr("LABLAXIS", "Window Spectrogram", (QdObject_var) out_spmag_series);
  QuSetTxtAttr("FIELDNAM", "Window Spectrogram", (QdObject_var) out_spmag_series);
  QuSetFrame((QdObject_var) out_spmag_series, "scalar");
  sprintf(str_out,"(%s)^2/Hz", QuGetAttrText("UNITS", ts_in).c_str());
  QuSetTxtAttr("UNITS", str_out, (QdObject_var) out_spmag_series);
  out_spmag_series->set_xref("DEPEND_0", out_freq_series);


  if(vectorset) {
    QuSetTxtAttr("SI_conversion", " ", (QdObject_var) out_spvec_series);
    QuSetTxtAttr("LABLAXIS", "Window Spectrogram", (QdObject_var) out_spvec_series);
    QuSetTxtAttr("FIELDNAM", "Window Spectrogram", (QdObject_var) out_spvec_series);
    QuSetFrame( (QdObject_var) out_spvec_series, QuGetOrder(ts_in).c_str(), QuGetFrame(ts_in).c_str(), QuGetRep(ts_in).c_str()); 
    QuSetTxtAttr("UNITS", str_out, (QdObject_var) out_spvec_series);
    out_spvec_series->set_xref("DEPEND_0", out_freq_series);
  }


  QuSetTxtAttr("SI_conversion", " ", (QdObject_var) out_freq_series);
  QuSetTxtAttr("LABLAXIS", "frequency", (QdObject_var) out_freq_series);
  QuSetTxtAttr("FIELDNAM", "frequency", (QdObject_var) out_freq_series);
  QuSetFrame((QdObject_var) out_freq_series,  "scalar>na");
  QuSetTxtAttr("UNITS", "Hz", (QdObject_var) out_freq_series);


//	..add them to the list

//   call_list->push_back( (QdObject_var) out_freq_series);    
  call_list->push_back( (QdObject_var) out_spmag_series);     
  call_list->push_back( (QdObject_var) out_spvec_series);     


//	..some messages to the analysis interface window..

sprintf(str_out, "WindowSpectrum>> %ld data points\n", Ndata);
QplugAppendTextDisplay (str_out);
sprintf(str_out, "WindowSpectrum>> %ld zero padded data points\n", Ndata0-Ndata);
QplugAppendTextDisplay (str_out);
sprintf(str_out,"WindowSpectrum>> %i frequencies\n", nfreq);
QplugAppendTextDisplay (str_out);
sprintf(str_out,"WindowSpectrum>> %i Segments\n", 2*nensemble);
QplugAppendTextDisplay (str_out);
QplugAppendTextDisplay ("WindowSpectrum>> finished...\n");


//	..clean up and exit

delete[] data;

delete[] w1;
delete[] w2;
delete[] pf;
delete[] freq;

return QPLUG_SUCCESS;


}




/***************************************************************************/

double	 QWindow( int WType, int j, double a, double b )
{

double fji;
fji=(double)j;

	switch( WType )

	{


	default:
	return -1.0;

	case 0:		/*	No Window	*/
	return 1.0;		

	case 1:		/*	Hanning window	*/
         return  0.5*(1.0-cos(6.2831853*fji*b));	

	case 2:			/*	Welch  window	*/
         return	(1.0 - SQR((fji-a)*b));

	case 3:		/*	Bartlett or Parzen window */
         return	(1.0 - fabs((fji-a)*b ));

        }
}


/***************************************************************************/


void QSpctrm(double data[], double p[], int m, int k, int ovrlap, int WindowType, double w1[], double w2[])
{
	int mm,m44,m43,m4,kk,joffn,joff,j2,j,dpointer,*ip_tak;
	double w,facp,facm,sumw=0.0,den=0.0,*w_tak;

	mm=m+m;
	m43=(m4=mm+mm)+3;
	m44=m43+1;

	facm=m;
	facp=1.0/m;
        dpointer=1;
	ip_tak = new int[(long) sqrt((double) mm) +3];
        ip_tak[0]=0;
        w_tak = new double[mm];

	for (j=1;j<=mm;j++) sumw += SQR(QWindow(WindowType,j,facm,facp));
	for (j=0;j<m;j++) p[j]=0.0;
	if (ovrlap) {
		for (j=0;j<m;j++) w2[j]=data[j]; 
                dpointer=m;
                }
	for (kk=1;kk<=k;kk++) {
		for (joff = -1;joff<=0;joff++) {
			if (ovrlap) {
				for (j=0;j<m;j++) w1[joff+j+j+1]=w2[j]; 
				for (j=0;j<m;j++) w2[j]=data[dpointer+j]; 
                                dpointer+=m;
				joffn=joff+mm;
				for (j=0;j<m;j++) w1[joffn+j+j+1]=w2[j];
			} else {
			      // AJA changed j<=m4 to j<m4 to avoid array bound overrun
				for (j=joff+1;j<m4;j+=2) w1[j]=data[dpointer+j/2-1] ; 
                                dpointer+=m4/2+2*joff+1;
			}
		}
		for (j=0;j<mm;j++) {
			j2=j+j;
			w=QWindow(WindowType,j+1,facm,facp);
			w1[j2+1] *= w;
			w1[j2] *= w;
		}
//	        printf("w1[2]:% 22.15e,m4:%6i,k%6i\n",w1[2],m4,k);
		cdft(m4, 1, w1, ip_tak, w_tak);
//	        printf("w1[2]:% 22.15e,w1[3]:% 22.15e\n",w1[2],w1[3]);
        p[0] += SQR(w1[0]);
        p[0] += SQR(w1[1]);
        for (j=1;j<m;j++) {
            j2=j+j+1;
            p[j] += SQR(w1[j2]);
            p[j] += SQR(w1[j2-1]);
            p[j] += SQR(w1[m44-j2]);
            p[j] += SQR(w1[m43-j2]);
		}
		den += sumw;
	}
	den *= mm; //factor 2 added for Eriksson normalization
	for (j=0;j<m;j++) p[j] /= den;
//      printf("data[1]:% 22.15e,data[4*k*m]:%22.15e\n",data[1],data[4*k*m]);
//	printf("p[1]:% 22.15e,den:% 22.15e,data[0]:% 22.15e\n",p[1],den,data[0]);
        delete ip_tak;
        delete w_tak;


}


void cdft(int n, int isgn, double *a, int *ip, double *w)
{   
    if (n > (ip[0] << 2)) {
        makewt(n >> 2, ip, w);
    }
    if (n > 4) {
        if (isgn >= 0) {
            bitrv2(n, ip + 2, a);
            cftfsub(n, a, w);
        } else {
            bitrv2conj(n, ip + 2, a);
            cftbsub(n, a, w);
        }
    } else if (n == 4) {
        cftfsub(n, a, w);
    }
}


void makewt(int nw, int *ip, double *w)
{
    void bitrv2(int n, int *ip, double *a);
    int j, nwh;
    double delta, x, y;
    
    ip[0] = nw;
    ip[1] = 1;
    if (nw > 2) {
        nwh = nw >> 1;
        delta = atan(1.0) / nwh;
        w[0] = 1;
        w[1] = 0;
        w[nwh] = cos(delta * nwh);
        w[nwh + 1] = w[nwh];
        if (nwh > 2) {
            for (j = 2; j < nwh; j += 2) {
                x = cos(delta * j);
                y = sin(delta * j);
                w[j] = x;
                w[j + 1] = y;
                w[nw - j] = y;
                w[nw - j + 1] = x;
            }
            for (j = nwh - 2; j >= 2; j -= 2) {
                x = w[2 * j];
                y = w[2 * j + 1];
                w[nwh + j] = x;
                w[nwh + j + 1] = y;
            }
            bitrv2(nw, ip + 2, w);
        }
    }
}

// -------- child routines -------- 


void bitrv2(int n, int *ip, double *a)
{
    int j, j1, k, k1, l, m, m2;
    double xr, xi, yr, yi;
    
    ip[0] = 0;
    l = n;
    m = 1;
    while ((m << 3) < l) {
        l >>= 1;
        for (j = 0; j < m; j++) {
            ip[m + j] = ip[j] + l;
        }
        m <<= 1;
    }
    m2 = 2 * m;
    if ((m << 3) == l) {
        for (k = 0; k < m; k++) {
            for (j = 0; j < k; j++) {
                j1 = 2 * j + ip[k];
                k1 = 2 * k + ip[j];
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += m2;
                k1 += 2 * m2;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += m2;
                k1 -= m2;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += m2;
                k1 += 2 * m2;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
            }
            j1 = 2 * k + m2 + ip[k];
            k1 = j1 + m2;
            xr = a[j1];
            xi = a[j1 + 1];
            yr = a[k1];
            yi = a[k1 + 1];
            a[j1] = yr;
            a[j1 + 1] = yi;
            a[k1] = xr;
            a[k1 + 1] = xi;
        }
    } else {
        for (k = 1; k < m; k++) {
            for (j = 0; j < k; j++) {
                j1 = 2 * j + ip[k];
                k1 = 2 * k + ip[j];
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += m2;
                k1 += m2;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
            }
        }
    }
}



void bitrv2conj(int n, int *ip, double *a)
{
    int j, j1, k, k1, l, m, m2;
    double xr, xi, yr, yi;
    
    ip[0] = 0;
    l = n;
    m = 1;
    while ((m << 3) < l) {
        l >>= 1;
        for (j = 0; j < m; j++) {
            ip[m + j] = ip[j] + l;
        }
        m <<= 1;
    }
    m2 = 2 * m;
    if ((m << 3) == l) {
        for (k = 0; k < m; k++) {
            for (j = 0; j < k; j++) {
                j1 = 2 * j + ip[k];
                k1 = 2 * k + ip[j];
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += m2;
                k1 += 2 * m2;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += m2;
                k1 -= m2;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += m2;
                k1 += 2 * m2;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
            }
            k1 = 2 * k + ip[k];
            a[k1 + 1] = -a[k1 + 1];
            j1 = k1 + m2;
            k1 = j1 + m2;
            xr = a[j1];
            xi = -a[j1 + 1];
            yr = a[k1];
            yi = -a[k1 + 1];
            a[j1] = yr;
            a[j1 + 1] = yi;
            a[k1] = xr;
            a[k1 + 1] = xi;
            k1 += m2;
            a[k1 + 1] = -a[k1 + 1];
        }
    } else {
        a[1] = -a[1];
        a[m2 + 1] = -a[m2 + 1];
        for (k = 1; k < m; k++) {
            for (j = 0; j < k; j++) {
                j1 = 2 * j + ip[k];
                k1 = 2 * k + ip[j];
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += m2;
                k1 += m2;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
            }
            k1 = 2 * k + ip[k];
            a[k1 + 1] = -a[k1 + 1];
            a[k1 + m2 + 1] = -a[k1 + m2 + 1];
        }
    }
}


void cftfsub(int n, double *a, double *w)
{
    int j, j1, j2, j3, l;
    double x0r, x0i, x1r, x1i, x2r, x2i, x3r, x3i;
    
    l = 2;
    if (n >= 16) {
        cft1st(n, a, w);
        l = 16;
        while ((l << 3) <= n) {
            cftmdl(n, l, a, w);
            l <<= 3;
        }
    }
    if ((l << 1) < n) {
        for (j = 0; j < l; j += 2) {
            j1 = j + l;
            j2 = j1 + l;
            j3 = j2 + l;
            x0r = a[j] + a[j1];
            x0i = a[j + 1] + a[j1 + 1];
            x1r = a[j] - a[j1];
            x1i = a[j + 1] - a[j1 + 1];
            x2r = a[j2] + a[j3];
            x2i = a[j2 + 1] + a[j3 + 1];
            x3r = a[j2] - a[j3];
            x3i = a[j2 + 1] - a[j3 + 1];
            a[j] = x0r + x2r;
            a[j + 1] = x0i + x2i;
            a[j2] = x0r - x2r;
            a[j2 + 1] = x0i - x2i;
            a[j1] = x1r - x3i;
            a[j1 + 1] = x1i + x3r;
            a[j3] = x1r + x3i;
            a[j3 + 1] = x1i - x3r;
        }
    } else if ((l << 1) == n) {
        for (j = 0; j < l; j += 2) {
            j1 = j + l;
            x0r = a[j] - a[j1];
            x0i = a[j + 1] - a[j1 + 1];
            a[j] += a[j1];
            a[j + 1] += a[j1 + 1];
            a[j1] = x0r;
            a[j1 + 1] = x0i;
        }
    }
}


void cftbsub(int n, double *a, double *w)
{
    int j, j1, j2, j3, j4, j5, j6, j7, l;
    double wn4r, x0r, x0i, x1r, x1i, x2r, x2i, x3r, x3i, 
        y0r, y0i, y1r, y1i, y2r, y2i, y3r, y3i, 
        y4r, y4i, y5r, y5i, y6r, y6i, y7r, y7i;
    
    l = 2;
    if (n > 16) {
        cft1st(n, a, w);
        l = 16;
        while ((l << 3) < n) {
            cftmdl(n, l, a, w);
            l <<= 3;
        }
    }
    if ((l << 2) < n) {
        wn4r = w[2];
        for (j = 0; j < l; j += 2) {
            j1 = j + l;
            j2 = j1 + l;
            j3 = j2 + l;
            j4 = j3 + l;
            j5 = j4 + l;
            j6 = j5 + l;
            j7 = j6 + l;
            x0r = a[j] + a[j1];
            x0i = -a[j + 1] - a[j1 + 1];
            x1r = a[j] - a[j1];
            x1i = -a[j + 1] + a[j1 + 1];
            x2r = a[j2] + a[j3];
            x2i = a[j2 + 1] + a[j3 + 1];
            x3r = a[j2] - a[j3];
            x3i = a[j2 + 1] - a[j3 + 1];
            y0r = x0r + x2r;
            y0i = x0i - x2i;
            y2r = x0r - x2r;
            y2i = x0i + x2i;
            y1r = x1r - x3i;
            y1i = x1i - x3r;
            y3r = x1r + x3i;
            y3i = x1i + x3r;
            x0r = a[j4] + a[j5];
            x0i = a[j4 + 1] + a[j5 + 1];
            x1r = a[j4] - a[j5];
            x1i = a[j4 + 1] - a[j5 + 1];
            x2r = a[j6] + a[j7];
            x2i = a[j6 + 1] + a[j7 + 1];
            x3r = a[j6] - a[j7];
            x3i = a[j6 + 1] - a[j7 + 1];
            y4r = x0r + x2r;
            y4i = x0i + x2i;
            y6r = x0r - x2r;
            y6i = x0i - x2i;
            x0r = x1r - x3i;
            x0i = x1i + x3r;
            x2r = x1r + x3i;
            x2i = x1i - x3r;
            y5r = wn4r * (x0r - x0i);
            y5i = wn4r * (x0r + x0i);
            y7r = wn4r * (x2r - x2i);
            y7i = wn4r * (x2r + x2i);
            a[j1] = y1r + y5r;
            a[j1 + 1] = y1i - y5i;
            a[j5] = y1r - y5r;
            a[j5 + 1] = y1i + y5i;
            a[j3] = y3r - y7i;
            a[j3 + 1] = y3i - y7r;
            a[j7] = y3r + y7i;
            a[j7 + 1] = y3i + y7r;
            a[j] = y0r + y4r;
            a[j + 1] = y0i - y4i;
            a[j4] = y0r - y4r;
            a[j4 + 1] = y0i + y4i;
            a[j2] = y2r - y6i;
            a[j2 + 1] = y2i - y6r;
            a[j6] = y2r + y6i;
            a[j6 + 1] = y2i + y6r;
        }
    } else if ((l << 2) == n) {
        for (j = 0; j < l; j += 2) {
            j1 = j + l;
            j2 = j1 + l;
            j3 = j2 + l;
            x0r = a[j] + a[j1];
            x0i = -a[j + 1] - a[j1 + 1];
            x1r = a[j] - a[j1];
            x1i = -a[j + 1] + a[j1 + 1];
            x2r = a[j2] + a[j3];
            x2i = a[j2 + 1] + a[j3 + 1];
            x3r = a[j2] - a[j3];
            x3i = a[j2 + 1] - a[j3 + 1];
            a[j] = x0r + x2r;
            a[j + 1] = x0i - x2i;
            a[j2] = x0r - x2r;
            a[j2 + 1] = x0i + x2i;
            a[j1] = x1r - x3i;
            a[j1 + 1] = x1i - x3r;
            a[j3] = x1r + x3i;
            a[j3 + 1] = x1i + x3r;
        }
    } else {
        for (j = 0; j < l; j += 2) {
            j1 = j + l;
            x0r = a[j] - a[j1];
            x0i = -a[j + 1] + a[j1 + 1];
            a[j] += a[j1];
            a[j + 1] = -a[j + 1] - a[j1 + 1];
            a[j1] = x0r;
            a[j1 + 1] = x0i;
        }
    }
}

void cft1st(int n, double *a, double *w)
{
    int j, k1;
    double wn4r, wtmp, wk1r, wk1i, wk2r, wk2i, wk3r, wk3i, 
        wk4r, wk4i, wk5r, wk5i, wk6r, wk6i, wk7r, wk7i;
    double x0r, x0i, x1r, x1i, x2r, x2i, x3r, x3i, 
        y0r, y0i, y1r, y1i, y2r, y2i, y3r, y3i, 
        y4r, y4i, y5r, y5i, y6r, y6i, y7r, y7i;
    
    wn4r = w[2];
    x0r = a[0] + a[2];
    x0i = a[1] + a[3];
    x1r = a[0] - a[2];
    x1i = a[1] - a[3];
    x2r = a[4] + a[6];
    x2i = a[5] + a[7];
    x3r = a[4] - a[6];
    x3i = a[5] - a[7];
    y0r = x0r + x2r;
    y0i = x0i + x2i;
    y2r = x0r - x2r;
    y2i = x0i - x2i;
    y1r = x1r - x3i;
    y1i = x1i + x3r;
    y3r = x1r + x3i;
    y3i = x1i - x3r;
    x0r = a[8] + a[10];
    x0i = a[9] + a[11];
    x1r = a[8] - a[10];
    x1i = a[9] - a[11];
    x2r = a[12] + a[14];
    x2i = a[13] + a[15];
    x3r = a[12] - a[14];
    x3i = a[13] - a[15];
    y4r = x0r + x2r;
    y4i = x0i + x2i;
    y6r = x0r - x2r;
    y6i = x0i - x2i;
    x0r = x1r - x3i;
    x0i = x1i + x3r;
    x2r = x1r + x3i;
    x2i = x1i - x3r;
    y5r = wn4r * (x0r - x0i);
    y5i = wn4r * (x0r + x0i);
    y7r = wn4r * (x2r - x2i);
    y7i = wn4r * (x2r + x2i);
    a[2] = y1r + y5r;
    a[3] = y1i + y5i;
    a[10] = y1r - y5r;
    a[11] = y1i - y5i;
    a[6] = y3r - y7i;
    a[7] = y3i + y7r;
    a[14] = y3r + y7i;
    a[15] = y3i - y7r;
    a[0] = y0r + y4r;
    a[1] = y0i + y4i;
    a[8] = y0r - y4r;
    a[9] = y0i - y4i;
    a[4] = y2r - y6i;
    a[5] = y2i + y6r;
    a[12] = y2r + y6i;
    a[13] = y2i - y6r;
    if (n > 16) {
        wk1r = w[4];
        wk1i = w[5];
        x0r = a[16] + a[18];
        x0i = a[17] + a[19];
        x1r = a[16] - a[18];
        x1i = a[17] - a[19];
        x2r = a[20] + a[22];
        x2i = a[21] + a[23];
        x3r = a[20] - a[22];
        x3i = a[21] - a[23];
        y0r = x0r + x2r;
        y0i = x0i + x2i;
        y2r = x0r - x2r;
        y2i = x0i - x2i;
        y1r = x1r - x3i;
        y1i = x1i + x3r;
        y3r = x1r + x3i;
        y3i = x1i - x3r;
        x0r = a[24] + a[26];
        x0i = a[25] + a[27];
        x1r = a[24] - a[26];
        x1i = a[25] - a[27];
        x2r = a[28] + a[30];
        x2i = a[29] + a[31];
        x3r = a[28] - a[30];
        x3i = a[29] - a[31];
        y4r = x0r + x2r;
        y4i = x0i + x2i;
        y6r = x0r - x2r;
        y6i = x0i - x2i;
        x0r = x1r - x3i;
        x0i = x1i + x3r;
        x2r = x1r + x3i;
        x2i = x3r - x1i;
        y5r = wk1i * x0r - wk1r * x0i;
        y5i = wk1i * x0i + wk1r * x0r;
        y7r = wk1r * x2r + wk1i * x2i;
        y7i = wk1r * x2i - wk1i * x2r;
        x0r = wk1r * y1r - wk1i * y1i;
        x0i = wk1r * y1i + wk1i * y1r;
        a[18] = x0r + y5r;
        a[19] = x0i + y5i;
        a[26] = y5i - x0i;
        a[27] = x0r - y5r;
        x0r = wk1i * y3r - wk1r * y3i;
        x0i = wk1i * y3i + wk1r * y3r;
        a[22] = x0r - y7r;
        a[23] = x0i + y7i;
        a[30] = y7i - x0i;
        a[31] = x0r + y7r;
        a[16] = y0r + y4r;
        a[17] = y0i + y4i;
        a[24] = y4i - y0i;
        a[25] = y0r - y4r;
        x0r = y2r - y6i;
        x0i = y2i + y6r;
        a[20] = wn4r * (x0r - x0i);
        a[21] = wn4r * (x0i + x0r);
        x0r = y6r - y2i;
        x0i = y2r + y6i;
        a[28] = wn4r * (x0r - x0i);
        a[29] = wn4r * (x0i + x0r);
        k1 = 4;
        for (j = 32; j < n; j += 16) {
            k1 += 4;
            wk1r = w[k1];
            wk1i = w[k1 + 1];
            wk2r = w[k1 + 2];
            wk2i = w[k1 + 3];
            wtmp = 2 * wk2i;
            wk3r = wk1r - wtmp * wk1i;
            wk3i = wtmp * wk1r - wk1i;
            wk4r = 1 - wtmp * wk2i;
            wk4i = wtmp * wk2r;
            wtmp = 2 * wk4i;
            wk5r = wk3r - wtmp * wk1i;
            wk5i = wtmp * wk1r - wk3i;
            wk6r = wk2r - wtmp * wk2i;
            wk6i = wtmp * wk2r - wk2i;
            wk7r = wk1r - wtmp * wk3i;
            wk7i = wtmp * wk3r - wk1i;
            x0r = a[j] + a[j + 2];
            x0i = a[j + 1] + a[j + 3];
            x1r = a[j] - a[j + 2];
            x1i = a[j + 1] - a[j + 3];
            x2r = a[j + 4] + a[j + 6];
            x2i = a[j + 5] + a[j + 7];
            x3r = a[j + 4] - a[j + 6];
            x3i = a[j + 5] - a[j + 7];
            y0r = x0r + x2r;
            y0i = x0i + x2i;
            y2r = x0r - x2r;
            y2i = x0i - x2i;
            y1r = x1r - x3i;
            y1i = x1i + x3r;
            y3r = x1r + x3i;
            y3i = x1i - x3r;
            x0r = a[j + 8] + a[j + 10];
            x0i = a[j + 9] + a[j + 11];
            x1r = a[j + 8] - a[j + 10];
            x1i = a[j + 9] - a[j + 11];
            x2r = a[j + 12] + a[j + 14];
            x2i = a[j + 13] + a[j + 15];
            x3r = a[j + 12] - a[j + 14];
            x3i = a[j + 13] - a[j + 15];
            y4r = x0r + x2r;
            y4i = x0i + x2i;
            y6r = x0r - x2r;
            y6i = x0i - x2i;
            x0r = x1r - x3i;
            x0i = x1i + x3r;
            x2r = x1r + x3i;
            x2i = x1i - x3r;
            y5r = wn4r * (x0r - x0i);
            y5i = wn4r * (x0r + x0i);
            y7r = wn4r * (x2r - x2i);
            y7i = wn4r * (x2r + x2i);
            x0r = y1r + y5r;
            x0i = y1i + y5i;
            a[j + 2] = wk1r * x0r - wk1i * x0i;
            a[j + 3] = wk1r * x0i + wk1i * x0r;
            x0r = y1r - y5r;
            x0i = y1i - y5i;
            a[j + 10] = wk5r * x0r - wk5i * x0i;
            a[j + 11] = wk5r * x0i + wk5i * x0r;
            x0r = y3r - y7i;
            x0i = y3i + y7r;
            a[j + 6] = wk3r * x0r - wk3i * x0i;
            a[j + 7] = wk3r * x0i + wk3i * x0r;
            x0r = y3r + y7i;
            x0i = y3i - y7r;
            a[j + 14] = wk7r * x0r - wk7i * x0i;
            a[j + 15] = wk7r * x0i + wk7i * x0r;
            a[j] = y0r + y4r;
            a[j + 1] = y0i + y4i;
            x0r = y0r - y4r;
            x0i = y0i - y4i;
            a[j + 8] = wk4r * x0r - wk4i * x0i;
            a[j + 9] = wk4r * x0i + wk4i * x0r;
            x0r = y2r - y6i;
            x0i = y2i + y6r;
            a[j + 4] = wk2r * x0r - wk2i * x0i;
            a[j + 5] = wk2r * x0i + wk2i * x0r;
            x0r = y2r + y6i;
            x0i = y2i - y6r;
            a[j + 12] = wk6r * x0r - wk6i * x0i;
            a[j + 13] = wk6r * x0i + wk6i * x0r;
        }
    }
}


void cftmdl(int n, int l, double *a, double *w)
{
    int j, j1, j2, j3, j4, j5, j6, j7, k, k1, m;
    double wn4r, wtmp, wk1r, wk1i, wk2r, wk2i, wk3r, wk3i, 
        wk4r, wk4i, wk5r, wk5i, wk6r, wk6i, wk7r, wk7i;
    double x0r, x0i, x1r, x1i, x2r, x2i, x3r, x3i, 
        y0r, y0i, y1r, y1i, y2r, y2i, y3r, y3i, 
        y4r, y4i, y5r, y5i, y6r, y6i, y7r, y7i;
    
    m = l << 3;
    wn4r = w[2];
    for (j = 0; j < l; j += 2) {
        j1 = j + l;
        j2 = j1 + l;
        j3 = j2 + l;
        j4 = j3 + l;
        j5 = j4 + l;
        j6 = j5 + l;
        j7 = j6 + l;
        x0r = a[j] + a[j1];
        x0i = a[j + 1] + a[j1 + 1];
        x1r = a[j] - a[j1];
        x1i = a[j + 1] - a[j1 + 1];
        x2r = a[j2] + a[j3];
        x2i = a[j2 + 1] + a[j3 + 1];
        x3r = a[j2] - a[j3];
        x3i = a[j2 + 1] - a[j3 + 1];
        y0r = x0r + x2r;
        y0i = x0i + x2i;
        y2r = x0r - x2r;
        y2i = x0i - x2i;
        y1r = x1r - x3i;
        y1i = x1i + x3r;
        y3r = x1r + x3i;
        y3i = x1i - x3r;
        x0r = a[j4] + a[j5];
        x0i = a[j4 + 1] + a[j5 + 1];
        x1r = a[j4] - a[j5];
        x1i = a[j4 + 1] - a[j5 + 1];
        x2r = a[j6] + a[j7];
        x2i = a[j6 + 1] + a[j7 + 1];
        x3r = a[j6] - a[j7];
        x3i = a[j6 + 1] - a[j7 + 1];
        y4r = x0r + x2r;
        y4i = x0i + x2i;
        y6r = x0r - x2r;
        y6i = x0i - x2i;
        x0r = x1r - x3i;
        x0i = x1i + x3r;
        x2r = x1r + x3i;
        x2i = x1i - x3r;
        y5r = wn4r * (x0r - x0i);
        y5i = wn4r * (x0r + x0i);
        y7r = wn4r * (x2r - x2i);
        y7i = wn4r * (x2r + x2i);
        a[j1] = y1r + y5r;
        a[j1 + 1] = y1i + y5i;
        a[j5] = y1r - y5r;
        a[j5 + 1] = y1i - y5i;
        a[j3] = y3r - y7i;
        a[j3 + 1] = y3i + y7r;
        a[j7] = y3r + y7i;
        a[j7 + 1] = y3i - y7r;
        a[j] = y0r + y4r;
        a[j + 1] = y0i + y4i;
        a[j4] = y0r - y4r;
        a[j4 + 1] = y0i - y4i;
        a[j2] = y2r - y6i;
        a[j2 + 1] = y2i + y6r;
        a[j6] = y2r + y6i;
        a[j6 + 1] = y2i - y6r;
    }
    if (m < n) {
        wk1r = w[4];
        wk1i = w[5];
        for (j = m; j < l + m; j += 2) {
            j1 = j + l;
            j2 = j1 + l;
            j3 = j2 + l;
            j4 = j3 + l;
            j5 = j4 + l;
            j6 = j5 + l;
            j7 = j6 + l;
            x0r = a[j] + a[j1];
            x0i = a[j + 1] + a[j1 + 1];
            x1r = a[j] - a[j1];
            x1i = a[j + 1] - a[j1 + 1];
            x2r = a[j2] + a[j3];
            x2i = a[j2 + 1] + a[j3 + 1];
            x3r = a[j2] - a[j3];
            x3i = a[j2 + 1] - a[j3 + 1];
            y0r = x0r + x2r;
            y0i = x0i + x2i;
            y2r = x0r - x2r;
            y2i = x0i - x2i;
            y1r = x1r - x3i;
            y1i = x1i + x3r;
            y3r = x1r + x3i;
            y3i = x1i - x3r;
            x0r = a[j4] + a[j5];
            x0i = a[j4 + 1] + a[j5 + 1];
            x1r = a[j4] - a[j5];
            x1i = a[j4 + 1] - a[j5 + 1];
            x2r = a[j6] + a[j7];
            x2i = a[j6 + 1] + a[j7 + 1];
            x3r = a[j6] - a[j7];
            x3i = a[j6 + 1] - a[j7 + 1];
            y4r = x0r + x2r;
            y4i = x0i + x2i;
            y6r = x0r - x2r;
            y6i = x0i - x2i;
            x0r = x1r - x3i;
            x0i = x1i + x3r;
            x2r = x1r + x3i;
            x2i = x3r - x1i;
            y5r = wk1i * x0r - wk1r * x0i;
            y5i = wk1i * x0i + wk1r * x0r;
            y7r = wk1r * x2r + wk1i * x2i;
            y7i = wk1r * x2i - wk1i * x2r;
            x0r = wk1r * y1r - wk1i * y1i;
            x0i = wk1r * y1i + wk1i * y1r;
            a[j1] = x0r + y5r;
            a[j1 + 1] = x0i + y5i;
            a[j5] = y5i - x0i;
            a[j5 + 1] = x0r - y5r;
            x0r = wk1i * y3r - wk1r * y3i;
            x0i = wk1i * y3i + wk1r * y3r;
            a[j3] = x0r - y7r;
            a[j3 + 1] = x0i + y7i;
            a[j7] = y7i - x0i;
            a[j7 + 1] = x0r + y7r;
            a[j] = y0r + y4r;
            a[j + 1] = y0i + y4i;
            a[j4] = y4i - y0i;
            a[j4 + 1] = y0r - y4r;
            x0r = y2r - y6i;
            x0i = y2i + y6r;
            a[j2] = wn4r * (x0r - x0i);
            a[j2 + 1] = wn4r * (x0i + x0r);
            x0r = y6r - y2i;
            x0i = y2r + y6i;
            a[j6] = wn4r * (x0r - x0i);
            a[j6 + 1] = wn4r * (x0i + x0r);
        }
        k1 = 4;
        for (k = 2 * m; k < n; k += m) {
            k1 += 4;
            wk1r = w[k1];
            wk1i = w[k1 + 1];
            wk2r = w[k1 + 2];
            wk2i = w[k1 + 3];
            wtmp = 2 * wk2i;
            wk3r = wk1r - wtmp * wk1i;
            wk3i = wtmp * wk1r - wk1i;
            wk4r = 1 - wtmp * wk2i;
            wk4i = wtmp * wk2r;
            wtmp = 2 * wk4i;
            wk5r = wk3r - wtmp * wk1i;
            wk5i = wtmp * wk1r - wk3i;
            wk6r = wk2r - wtmp * wk2i;
            wk6i = wtmp * wk2r - wk2i;
            wk7r = wk1r - wtmp * wk3i;
            wk7i = wtmp * wk3r - wk1i;
            for (j = k; j < l + k; j += 2) {
                j1 = j + l;
                j2 = j1 + l;
                j3 = j2 + l;
                j4 = j3 + l;
                j5 = j4 + l;
                j6 = j5 + l;
                j7 = j6 + l;
                x0r = a[j] + a[j1];
                x0i = a[j + 1] + a[j1 + 1];
                x1r = a[j] - a[j1];
                x1i = a[j + 1] - a[j1 + 1];
                x2r = a[j2] + a[j3];
                x2i = a[j2 + 1] + a[j3 + 1];
                x3r = a[j2] - a[j3];
                x3i = a[j2 + 1] - a[j3 + 1];
                y0r = x0r + x2r;
                y0i = x0i + x2i;
                y2r = x0r - x2r;
                y2i = x0i - x2i;
                y1r = x1r - x3i;
                y1i = x1i + x3r;
                y3r = x1r + x3i;
                y3i = x1i - x3r;
                x0r = a[j4] + a[j5];
                x0i = a[j4 + 1] + a[j5 + 1];
                x1r = a[j4] - a[j5];
                x1i = a[j4 + 1] - a[j5 + 1];
                x2r = a[j6] + a[j7];
                x2i = a[j6 + 1] + a[j7 + 1];
                x3r = a[j6] - a[j7];
                x3i = a[j6 + 1] - a[j7 + 1];
                y4r = x0r + x2r;
                y4i = x0i + x2i;
                y6r = x0r - x2r;
                y6i = x0i - x2i;
                x0r = x1r - x3i;
                x0i = x1i + x3r;
                x2r = x1r + x3i;
                x2i = x1i - x3r;
                y5r = wn4r * (x0r - x0i);
                y5i = wn4r * (x0r + x0i);
                y7r = wn4r * (x2r - x2i);
                y7i = wn4r * (x2r + x2i);
                x0r = y1r + y5r;
                x0i = y1i + y5i;
                a[j1] = wk1r * x0r - wk1i * x0i;
                a[j1 + 1] = wk1r * x0i + wk1i * x0r;
                x0r = y1r - y5r;
                x0i = y1i - y5i;
                a[j5] = wk5r * x0r - wk5i * x0i;
                a[j5 + 1] = wk5r * x0i + wk5i * x0r;
                x0r = y3r - y7i;
                x0i = y3i + y7r;
                a[j3] = wk3r * x0r - wk3i * x0i;
                a[j3 + 1] = wk3r * x0i + wk3i * x0r;
                x0r = y3r + y7i;
                x0i = y3i - y7r;
                a[j7] = wk7r * x0r - wk7i * x0i;
                a[j7 + 1] = wk7r * x0i + wk7i * x0r;
                a[j] = y0r + y4r;
                a[j + 1] = y0i + y4i;
                x0r = y0r - y4r;
                x0i = y0i - y4i;
                a[j4] = wk4r * x0r - wk4i * x0i;
                a[j4 + 1] = wk4r * x0i + wk4i * x0r;
                x0r = y2r - y6i;
                x0i = y2i + y6r;
                a[j2] = wk2r * x0r - wk2i * x0i;
                a[j2 + 1] = wk2r * x0i + wk2i * x0r;
                x0r = y2r + y6i;
                x0i = y2i - y6r;
                a[j6] = wk6r * x0r - wk6i * x0i;
                a[j6 + 1] = wk6r * x0i + wk6i * x0r;
            }
        }
    }
}



