//  qslot.cc
//
// slot Utility Widgets
//
// Aug 2007: Tony Allen
//

/// \file qgu.cc
///  Implementation file for the main GUI classes that are used within QSAS which are: \n
///  \arg \c QguListItem \n
///  \arg \c QguListView \n
///  \arg \c QguWLinput  \n
///  \arg \c QguWindow   \n \n
///  Those classes specialise respectively the QT3 built-in classes as listed below\n
///  \arg \c Q_LIST_VIEW_ITEM \n
///  \arg \c QListView     \n
///  \arg \c QLineEdit     \n
///  \arg \c QMainWindow   \n 
///  \author Tony Allen    \n
///  \version 2.1         \n
///  \date    2007

#include <qcolor.h>
#include <qpalette.h>
#include <qglobal.h>
#include <qpaintdevice.h>
#include <qfont.h>

#include "qslot.h"
#include "qwui.h"
#include "qcalc.h"
#include "qprofile.h"

#include <stdio.h>

extern qwuiWindow* qsasWindow;


QColor col_White("#ffffff");
QColor col_Light_Grey("#eeeeee");
QColor col_Grey("#8c8c8c");
QColor col_Black("#000000");
QColor col_Green("#228b22");
QColor col_Red("#cc0000");
QColor col_DarkRed("#990000");
QColor col_Yellow("#eeee00");
QColor col_Amber("#bbaa33");
QColor col_LightYellow("#ffffaa");
QColor col_Blue("#000099");



// QguWLinput class
//----------------------------

/// \fn QguWLinput::QguWLinput(QWidget *parent, const char *name)
/// \brief Constructor of QguWLinput class
/// \param     parent  of type QWidget      
/// \param     name    of type const char *  
/// \exception None
/// \return    None
///
QguWLinput::QguWLinput(QWidget *parent, const char *name)
	:  QWidget(parent)
{
//	if  ( QuSlotDebug() ){
//		printf(" QguWLinput::QguWLinput %ld of %s \n", (long)this, name);
//	}

	QHBoxLayout * layout=new QHBoxLayout(this);

	this->setLayout(layout);

	WLI_palette = new QPalette(qApp->palette());
	QFont slot_font = QFont(qApp->font());
	QFontMetrics slot_font_mt(slot_font);
	int font_width = slot_font_mt.width("n");
	int font_height = slot_font_mt.height();
	this->setFont(slot_font);

	active = true;       // turn this slot on
	aliasList = NULL;
	ObjectName_nil = QString("");
	ObjectName = ObjectName_nil;
	
    if(QSAS_profiles->current()->slotLR == 1)
	{
		dropDownArray = new QToolButton(this);
		dropDownArray->setArrowType(Qt::DownArrow);
		dropDownArray->setFixedSize(24,20);
		dropDownArray->setPopupMode(QToolButton::InstantPopup);
		layout->addWidget(dropDownArray);
		
		lineBox = new QguLineEdit(this, "InputLineBox");
		lineBox->WLinput = this;
		layout->addWidget(lineBox);
	}
	else
	{
		lineBox = new QguLineEdit(this, "InputLineBox");
		lineBox->WLinput = this;
		layout->addWidget(lineBox);
 
		dropDownArray = new QToolButton(this);
		dropDownArray->setArrowType(Qt::DownArrow);
		dropDownArray->setFixedSize(24,20);
		dropDownArray->setPopupMode(QToolButton::InstantPopup);
		layout->addWidget(dropDownArray);
	  
	}
  // initialise the dropDown menu possibilities
    
	defaultPopup = 0;
	TrangePopup = 0;
	TimePopup = 0;
	TimeFromELPopup = 0;
	TivlFromEvtTablePopup = 0;
	TivlSlicePopup = 0;
	vectorPopup = 0;
	arrayPopup = 0;
  
	layout->setMargin(0);
	layout->setSpacing(0);
  
	dropDownArray->hide();
	setColour(COL_OK);
	connect(lineBox, SIGNAL(hadDropEvent()), this, SIGNAL(hadDropEvent()));
	connect(lineBox, SIGNAL(textEdited(const QString &)), this, SLOT(validateInput(const QString &)));
	connect(qsasWindow->wlview->wlmodel(), SIGNAL(objChanged(const QString&)), this, SLOT(validateObjChanged(const QString&)));
	connect(qsasWindow->wlview->wlmodel(), SIGNAL(deleted(const QString&)), this, SLOT(validateDeletion(const QString&)));
	connect(qsasWindow->wlview->wlmodel(), SIGNAL(pathChanged(const QString&, const QString&)), this, SLOT(validatePathChanged(const QString&, const QString&)));
	connect(qsasWindow->wlview->wlmodel(), SIGNAL(created(const QString&)), this, SLOT(validateCreation(const QString&)));
  
	this->resize(font_width*23,(int)(font_height*1.5)); // large enough for a time string
	this->setMinimumWidth(font_width*23);
  
}


/// \fn QguWLinput::~QguWLinput()
/// \brief Destructor of QguWLinput class
/// \exception None
/// \return    None
///
QguWLinput::~QguWLinput()
{
	delete lineBox;
	if(WLI_palette) delete WLI_palette;
	if(defaultPopup) delete defaultPopup;
	if(TrangePopup) delete TrangePopup;
	if(TimePopup) delete TimePopup;
	if(TivlFromEvtTablePopup) delete TivlFromEvtTablePopup;
	if(TivlSlicePopup) delete TivlSlicePopup;
	if(vectorPopup) delete vectorPopup;
	if(arrayPopup) delete arrayPopup;

	delete dropDownArray;
  
}

// --------------------------
void QguWLinput::set3DSubsetDef(const char *setting, bool sliceTime){
	rtpDefault = string(setting);
	if(sliceTime) rtpDefault.append("0");
	
	// used to override defaults for subsample of particle data with preferred options
}

// --------------------------
QString QguWLinput::get3DSubsetDef(QdObject_var dobj){
	// used to override defaults for subsample of particle data with preferred options
	QdObject_var res;
	QString compTxt = "";
	if(dobj.is_nil()) return compTxt;
	
	QdRMatrix_var mv = QdRMatrix_var::narrow(dobj);
	QdRMatrixSeq_var ms = QdRMatrixSeq_var::narrow(dobj);
	int nDims=0;
	
	if(!mv.is_nil()) nDims = (mv->dims()).size();
	if(!ms.is_nil()) nDims = (ms[0].dims()).size();
	
	if(nDims != 3) return compTxt;
	
	if(rtpDefault.empty()) return compTxt;

	compTxt += " [ ";
	bool getR = (rtpDefault.find("r") != string::npos);
	bool getPhi = (rtpDefault.find("p") != string::npos);
	bool getTheta = (rtpDefault.find("t") != string::npos);
	
	char strDepend[30];

	for(int i=1; i<=3; i++) {
        sprintf(strDepend, "DEPEND_%d", i);
        if(dobj->xref_exists(strDepend)){
			// find what this dimension is
        	QdObject_var dep_i;
			dep_i = dobj->get_xref(strDepend);
			
			bool isDeg = false;
			bool isAngle = QuIsAngle(dep_i, isDeg);
			bool isR = !isAngle;
			
			double minVal;
			double fillVal = -1.;
			int kStart = 0;
			int kEnd = 0;
			bool isPhi = QuIsPhiAngle(dep_i, isDeg, i, minVal, kStart, kEnd);
			bool isTheta = !isPhi;
			
			// handle each component, skips bins with fills
			if(isR){
				if(getR){
					compTxt +=  QString(" (%1,%2) ").arg(kStart).arg(kEnd);
				}
				else{
					compTxt +=  QString(" Sum(%1,%2) ").arg(kStart).arg(kEnd);
				}
			}
			else if(isPhi){
				if(getPhi){
					compTxt +=  QString(" (%1,%2) ").arg(kStart).arg(kEnd);
				}
				else{
					compTxt +=  QString(" Sum(%1,%2) ").arg(kStart).arg(kEnd);
				}
			}
			else if(isTheta){
				if(getTheta){
					compTxt +=  QString(" (%1,%2) ").arg(kStart).arg(kEnd);
				}
				else{
					compTxt +=  QString(" Sum(%1,%2) ").arg(kStart).arg(kEnd);
				}
			}
		}
	}
	compTxt += " ]";

	if(rtpDefault.find("0") != string::npos) compTxt += " <0>";
	
	
	return compTxt;
	
}

   
// --------------------------
QMenu * QguWLinput::getDefaultPopup(int seqChoice)
{
	if(this->defaultPopup == 0){

		defaultPopup = new QMenu(this);
		defaultPopup->addAction("This Object", this, SLOT(setDropdownText_dobj()));
		defaultPopup->addSeparator();
		if(seqChoice == QS_SEQ ) defaultPopup->addAction("Time Subsample... ", this, SLOT(openSliceS_seq()));
		else if(seqChoice == QS_NOT_SEQ ) defaultPopup->addAction("Time Subsample... ", this, SLOT(openSliceS_not_seq()));
		else if(seqChoice == QS_ANY) defaultPopup->addAction("Time Subsample... ", this, SLOT(openSliceS()));

		defaultPopup->addSeparator();
		defaultPopup->addAction("Metadata", this, SLOT(openSelectMetadata()));
	}
	this->defaultPopup->hide();
	return this->defaultPopup;
}

// --------------------------
QMenu * QguWLinput::getTrangePopup()
{
	if(this->TrangePopup == 0){

		TrangePopup = new QMenu(this);

		TrangePopup->addAction("From Object", this, SLOT(setDropdownText_dobj()));
		TrangePopup->addSeparator();
		TrangePopup->addAction("Subset...", this, SLOT(openSliceT_seq()));
		TrangePopup->addSeparator();
		TrangePopup->addAction("Metadata", this, SLOT(openSelectMetadata()));
	}
	this->TrangePopup->hide();
	return this->TrangePopup;
}

// --------------------------
QMenu * QguWLinput::getTimePopup()
{
	if(this->TimePopup == 0){
		TimePopup = new QMenu(this);
		TimePopup->addAction("Default", this, SLOT(setTimeDropdownDefault()));
		TimePopup->addSeparator();
		TimePopup->addAction("Start Time", this, SLOT(setDropdownTstart()));
		TimePopup->addAction("Centre Time", this, SLOT(setDropdownTcentre()));
		TimePopup->addAction("End Time", this, SLOT(setDropdownTend()));
		TimePopup->addSeparator();
		TimePopup->addAction("Time at Record", this, SLOT(openSliceT_not_seq()));
		TimePopup->addSeparator();
		TimePopup->addAction("Metadata", this, SLOT(openSelectMetadata()));	
	}

	this->TimePopup->hide();
	return this->TimePopup;
}

// --------------------------
QMenu * QguWLinput::getTimeFromELPopup()
{
	if(this->TimeFromELPopup == 0){
		TimeFromELPopup = new QMenu(this);
		TimeFromELPopup->addAction("Default", this, SLOT(setETDropdownDefault()));
		TimeFromELPopup->addSeparator();
		TimeFromELPopup->addAction("Select Event and time", this, SLOT(selectEventAndTime()));
		TimeFromELPopup->addSeparator();
		TimeFromELPopup->addAction("Metadata", this, SLOT(openSelectMetadata()));	
	}

	this->TimeFromELPopup->hide();
	return this->TimeFromELPopup;
}

// --------------------------

QMenu * QguWLinput::getTivlFromEvtTablePopup()
{
	if(this->TivlFromEvtTablePopup == 0){
		TivlFromEvtTablePopup = new QMenu(this);
		TivlFromEvtTablePopup->addAction("Default Event", this, SLOT(setDropdownText_event()));
		TivlFromEvtTablePopup->addAction("Select Event", this, SLOT(selectTivlEvent()));
	}
	this->TivlFromEvtTablePopup->hide();
	return this->TivlFromEvtTablePopup;
}

// --------------------------

QMenu * QguWLinput::getTivlSeqSlicePopup()
{
	if(this->TivlSlicePopup == 0){
		TivlSlicePopup = new QMenu(this);
		TivlSlicePopup->addAction("This Object", this, SLOT(setDropdownText_dobj()));
		TivlSlicePopup->addSeparator();
		TivlSlicePopup->addAction("Select Subset", this, SLOT(selectTivlSeqSubset()));
		TivlSlicePopup->addAction("Select Event", this, SLOT(selectTivlEvent()));

	}
	this->TivlSlicePopup->hide();
	return this->TivlSlicePopup;
}

 // --------------------------
 
QMenu * QguWLinput::getVectorPopup()
{
	if(this->vectorPopup == 0){
		vectorPopup = new QMenu(this);

		vectorPopup->addAction("Default", this, SLOT(setVectorDropdownDefault()));
		vectorPopup->addAction(" r (mag) ", this, SLOT(setDropdownText_r()));
		vectorPopup->addAction(" x ", this, SLOT(setDropdownText_x()));
		vectorPopup->addAction(" y ", this, SLOT(setDropdownText_y()));
		vectorPopup->addAction(" z ", this, SLOT(setDropdownText_z()));
		vectorPopup->addAction(" Lat (deg) ", this, SLOT(setDropdownText_ld()));
		vectorPopup->addAction(" theta (deg) ", this, SLOT(setDropdownText_td()));
		vectorPopup->addAction(" phi (deg) ", this, SLOT(setDropdownText_pd()));
		vectorPopup->addAction(" Lat (rad) ", this, SLOT(setDropdownText_lr()));
		vectorPopup->addAction(" theta (rad) ", this, SLOT(setDropdownText_tr()));
		vectorPopup->addAction(" phi (rad) ", this, SLOT(setDropdownText_pr()));
		vectorPopup->addSeparator();
		vectorPopup->addAction(" Time and Component Slice ", this, SLOT(openSliceV()));
		vectorPopup->addAction(" Time and Array Slice ", this, SLOT(openArrayReduce()));
		vectorPopup->addSeparator();
		vectorPopup->addAction("Metadata", this, SLOT(openSelectMetadata()));
	}
	this->vectorPopup->hide();
	return this->vectorPopup;
}

 // --------------------------
 
QMenu * QguWLinput::getArrayPopup()
{
	if(this->arrayPopup == 0){
		arrayPopup = new QMenu(this);

		arrayPopup->addAction("Default", this, SLOT(setArrayDropdownDefault()));
		arrayPopup->addSeparator();
		arrayPopup->addAction("Reduce Array Dimensions", this, SLOT(openArrayReduce()));
		arrayPopup->addSeparator();
		arrayPopup->addAction("Metadata", this, SLOT(openSelectMetadata()));
	}
	this->arrayPopup->hide();
	return this->arrayPopup;
}

 // --------------------------
 
bool QguWLinput::hasFocus(){
	return lineBox->hasFocus();
}

 // --------------------------
 
void QguWLinput::clearFocus(){
	return lineBox->clearFocus();
}

// --------------------------
QguComp_e QguWLinput::getComponent(){

	QguComp_e component;
	QString StrCmp = QString( this->text().toLatin1().data() );

	int compEndAt = StrCmp.indexOf(" [");
	int sliceEndAt = StrCmp.indexOf(" <");
	
	if (StrCmp.contains(T_STR_0_START)) component = T_0_START_COMP;
	else if(compEndAt != -1)
	{
		StrCmp.remove(0,compEndAt); // get " [ x ]" etc
		if (StrCmp.contains( "mag"))       component = R_COMP;
		else if (StrCmp.contains( "x"))  component = X_COMP;
		else if (StrCmp.contains( "y"))  component = Y_COMP;
		else if (StrCmp.contains( "z"))  component = Z_COMP;
		else if (StrCmp.contains( "Lat") && StrCmp.contains( "deg")) component = LAT_DEG_COMP;
		else if (StrCmp.contains( "Lat") && StrCmp.contains( "rad")) component = LAT_RAD_COMP;
		else if (StrCmp.contains( "theta") && StrCmp.contains( "deg")) component = THETA_DEG_COMP;
		else if (StrCmp.contains( "theta") && StrCmp.contains( "rad")) component = THETA_RAD_COMP;
		else if (StrCmp.contains( "phi") && StrCmp.contains( "deg")) component = PHI_DEG_COMP;
		else if (StrCmp.contains( "phi") && StrCmp.contains( "rad")) component = PHI_RAD_COMP;
		else if (StrCmp.contains("Start")) component = T_START_COMP;
		else if (StrCmp.contains("Centre")) component = T_CENTRE_COMP;
		else if (StrCmp.contains( "End"))   component = T_END_COMP;
		else if (StrCmp.contains("[=>") ) component = XREF_COMP;
		else component = ARRAY_COMP;
	}
	else if(sliceEndAt !=-1)
		component=SLICE_COMP;
	else
	{
		component = WL_OBJ;
	}

	return component;
}


void QguWLinput::openSliceT_seq(){

	if(QuSlotDebug())
		std::cout << "QguWLinput::openSliceT "  << endl;

	QString objName = this->textObject();

	TimeSliceWindow * tSliceWindow;
	tSliceWindow = new TimeSliceWindow(objName, "T", this, "tSliceWindow", true, false);
	
	tSliceWindow->setPalette(qApp->palette());

	tSliceWindow->parentSlot = this;
	tSliceWindow->move(QCursor::pos());
	
	tSliceWindow->show();
	tSliceWindow->setToLastChoice();
  
}

void QguWLinput::openSliceT_not_seq(){

	if(QuSlotDebug())
		std::cout << "QguWLinput::openSliceT "  << endl;

	QString objName = this->textObject();

	TimeSliceWindow * tSliceWindow;
	tSliceWindow = new TimeSliceWindow(objName, "T", this, "tSliceWindow", false, true);
	
	tSliceWindow->setPalette(qApp->palette());

	tSliceWindow->parentSlot = this;
	tSliceWindow->move(QCursor::pos());

	tSliceWindow->show();
	tSliceWindow->setToLastChoice();
  
}

void QguWLinput::openSliceS_seq(){

	if(QuSlotDebug())
		std::cout << "QguWLinput::openSliceS_seq "  << endl;

	QString objName = this->textObject();

	TimeSliceWindow * tSliceWindow;
	tSliceWindow = new TimeSliceWindow(objName, "S", this, "tSliceWindow", true, false);
	
	tSliceWindow->setPalette(qApp->palette());

	tSliceWindow->parentSlot = this;
	tSliceWindow->move(QCursor::pos());

	tSliceWindow->show();
	tSliceWindow->setToLastChoice();
  
}

void QguWLinput::openSliceS_not_seq(){

	if(QuSlotDebug())
		std::cout << "QguWLinput::openSliceS "  << endl;

	QString objName = this->textObject();

	TimeSliceWindow * tSliceWindow;
	tSliceWindow = new TimeSliceWindow(objName, "S", this, "tSliceWindow", false, true);
	
	tSliceWindow->setPalette(qApp->palette());

	tSliceWindow->parentSlot = this;
	tSliceWindow->move(QCursor::pos());

	tSliceWindow->show();
	tSliceWindow->setToLastChoice();
  
}

void QguWLinput::openSliceS(){

	if(QuSlotDebug())
		std::cout << "QguWLinput::openSliceS "  << endl;

	QString objName = this->textObject();

	TimeSliceWindow * tSliceWindow;
	tSliceWindow = new TimeSliceWindow(objName, "S", this, "tSliceWindow", true, true);
	
	tSliceWindow->setPalette(qApp->palette());

	tSliceWindow->parentSlot = this;
	tSliceWindow->move(QCursor::pos());

	tSliceWindow->show();
	tSliceWindow->setToLastChoice();
  
}

void QguWLinput::openSliceV(){

	if(QuSlotDebug())
		std::cout << "QguWLinput::openSliceV "  << endl;

	QString objName = this->textObject();

	TimeSliceWindow * tSliceWindow;
	tSliceWindow = new TimeSliceWindow(objName, "V", this, "tSliceWindow", true, true);

	
	tSliceWindow->setPalette(qApp->palette());

	tSliceWindow->parentSlot = this;
	tSliceWindow->move(QCursor::pos());

	tSliceWindow->show();
	tSliceWindow->setToLastChoice();
  
}



void QguWLinput::setDropdownText_event(){

	if(QuSlotDebug())
		std::cout << "QguWLinput::setDropdownText_event " << EVENT_0_STR << endl;

	this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(EVENT_0_STR));
    this->lineBox->end(false);
	this->lineBox->setFocus();

	this->reValidate(true);
}

void QguWLinput::setDropdownText_r(){

	if(QuSlotDebug())
		std::cout << "QguWLinput::setDropdownText_r " << R_STR_CMP << endl;

	this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(R_STR_CMP));
    this->lineBox->end(false);
	this->lineBox->setFocus();

	emit this->componentSelected(R_STR_CMP, "VECTOR");
	this->reValidate(true);
}

void QguWLinput::setDropdownText_x(){

	if(QuSlotDebug())
		std::cout << "QguWLinput::setDropdownText_x " << X_STR_CMP << endl;

	this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(X_STR_CMP));
    this->lineBox->end(false);
	this->lineBox->setFocus();

	emit this->componentSelected(X_STR_CMP, "VECTOR");
	this->reValidate(true);
}

void QguWLinput::setDropdownText_y(){
  
	if(QuSlotDebug())
		std::cout << "QguWLinput::setDropdownText_y " << Y_STR_CMP << endl;

	this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(Y_STR_CMP));
    this->lineBox->end(false);
	this->lineBox->setFocus();

	emit this->componentSelected(Y_STR_CMP, "VECTOR");
	this->reValidate(true);
}

void QguWLinput::setDropdownText_z(){
  
	if(QuSlotDebug())
		std::cout << "QguWLinput::setDropdownText_z " << Z_STR_CMP << endl;

	this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(Z_STR_CMP));
    this->lineBox->end(false);
	this->lineBox->setFocus();

	emit this->componentSelected(Z_STR_CMP, "VECTOR");
	this->reValidate(true);
}

void QguWLinput::setDropdownText_ld(){
  
	if(QuSlotDebug())
		std::cout << "QguWLinput::setDropdownText_ld " << Ld_STR_CMP << endl;

	this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(Ld_STR_CMP));
    this->lineBox->end(false);
	this->lineBox->setFocus();

	emit this->componentSelected(Ld_STR_CMP, "VECTOR");
	this->reValidate(true);
}

void QguWLinput::setDropdownText_td(){
  
	if(QuSlotDebug())
		std::cout << "QguWLinput::setDropdownText_td " << Td_STR_CMP << endl;

	this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(Td_STR_CMP));
    this->lineBox->end(false);
	this->lineBox->setFocus();

	emit this->componentSelected(Td_STR_CMP, "VECTOR");
	this->reValidate(true);
}

void QguWLinput::setDropdownText_pd(){
  
	if(QuSlotDebug())
		std::cout << "QguWLinput::setDropdownText_pd " << Pd_STR_CMP << endl;

	this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(Pd_STR_CMP));
    this->lineBox->end(false);
	this->lineBox->setFocus();

	emit this->componentSelected(Pd_STR_CMP, "VECTOR");
	this->reValidate(true);
}

void QguWLinput::setDropdownText_lr(){
  
	if(QuSlotDebug())
		std::cout << "QguWLinput::setDropdownText_lr " << Lr_STR_CMP << endl;

	this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(Lr_STR_CMP));
    this->lineBox->end(false);
	this->lineBox->setFocus();

	emit this->componentSelected(Lr_STR_CMP, "VECTOR");
	this->reValidate(true);
}

void QguWLinput::setDropdownText_tr(){
  
	if(QuSlotDebug())
		std::cout << "QguWLinput::setDropdownText_tr " << Tr_STR_CMP << endl;

	this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(Tr_STR_CMP));
    this->lineBox->end(false);
	this->lineBox->setFocus();

	emit this->componentSelected(Tr_STR_CMP, "VECTOR");
	this->reValidate(true);
}

void QguWLinput::setDropdownText_pr(){
  
	if(QuSlotDebug())
		std::cout << "QguWLinput::setDropdownText_pr " << Pr_STR_CMP << endl;

	this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(Pr_STR_CMP));
    this->lineBox->end(false);
	this->lineBox->setFocus();

	emit this->componentSelected(Pr_STR_CMP, "VECTOR");
	this->reValidate(true);
}

void QguWLinput::setDropdownText_dobj(){
  
	if(QuSlotDebug())
		std::cout << "QguWLinput::setDropdownText_dobj " << "this obj" << endl;

	this->lineBox->setText(QString(ObjectName.toLatin1().data()));
    this->lineBox->end(false);
	this->lineBox->setFocus();
	
	emit this->componentSelected("", "VECTOR");
	this->reValidate(true);
}

void QguWLinput::setDropdownTstart(){
  
	if(QuSlotDebug())
		std::cout << "QguWLinput::setDropdownTstart " << T_STR_START << endl;

	this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(T_STR_START));
    this->lineBox->end(false);
	this->lineBox->setFocus();

	emit this->TimeChosen();
	emit this->componentSelected(T_STR_START, "TIME");
	this->reValidate(true);
}

void QguWLinput::setDropdownTcentre(){
  
	if(QuSlotDebug())
		std::cout << "QguWLinput::setDropdownTcentre " << T_STR_CENTRE << endl;

	this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(T_STR_CENTRE));
    this->lineBox->end(false);
	this->lineBox->setFocus();

	emit this->TimeChosen();
	emit this->componentSelected(T_STR_CENTRE, "TIME");
	this->reValidate(true);
}
void QguWLinput::setDropdownTend(){
  
	if(QuSlotDebug())
		std::cout << "QguWLinput::setDropdownTend " << T_STR_END << endl;

	this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(T_STR_END));
    this->lineBox->end(false);
	this->lineBox->setFocus();

	emit this->TimeChosen();
	emit this->componentSelected(T_STR_END, "TIME");
	this->reValidate(true);
}
void QguWLinput::setTimeDropdownDefault(){
	
	if(QuSlotDebug()) std::cout << "QguWLinput::setTimeDropdownDefault "  << endl;

	QdObject_var dobj = QuGetObjByName(ObjectName);
	QdTime_var tobj = QdTime_var::narrow(dobj);
	
	if(!tobj.is_nil())
	{
		this->lineBox->setText(ObjectName);
		this->setColour(COL_OK);
		emit this->TimeChosen();
		this->reValidate(true);
	}
	else
	{
		// set start time
		this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(T_STR_START));
        this->lineBox->end(false);
		this->lineBox->setFocus();

		emit this->TimeChosen();
		emit this->componentSelected(T_STR_START, "TIME");
		this->reValidate(true);
	}
}

void QguWLinput::setETDropdownDefault(){
	
	if(QuSlotDebug()) std::cout << "QguWLinput::setETDropdownDefault "  << endl;

	QdObject_var dobj = QuGetObjByName(ObjectName);
	QdTime_var tobj = QdTime_var::narrow(dobj);
	
	if(!tobj.is_nil())
	{
		this->lineBox->setText(ObjectName);
		this->setColour(COL_OK);
		emit this->TimeChosen();
		this->reValidate(true);
	}
	else
	{
		// set start time
		this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(T_STR_0_START));
        this->lineBox->end(false);
		this->lineBox->setFocus();

		emit this->TimeChosen();
		emit this->componentSelected(T_STR_0_START, "TIME");
		this->reValidate(true);
	}
}

void QguWLinput::setVectorDropdownDefault(){
	
	if(QuSlotDebug()) std::cout << "QguWLinput::setVectorDropdownDefault "  << endl;
	
	QdObject_var dobj = QuGetObjByName(ObjectName);
	
	bool takesVector = false;
	for(int i=0; i<acceptConditions.size(); i++)
	{
		if(acceptConditions[i].condDim == QS_VECTOR
		|| acceptConditions[i].condData == QS_TI
		|| acceptConditions[i].condObj == QS_WL_OBJ) takesVector = true;
	}
	
	if(takesVector)
	{
		setDropdownText_dobj();		
	}
	else
	{
		setDropdownText_r();
	}
	// reValidate done in setDropdown calls
}

void QguWLinput::setArrayDropdownDefault(){
	
	if(QuSlotDebug()) std::cout << "QguWLinput::setArrayDropdownDefault "  << endl;
		
	int targetDim = 0;
	int targetSeq = QS_SEQ;
	int cDim = 3;
	for(int i=0; i<acceptConditions.size(); i++)
	{
		// take largest acceptable dim, otherwise it may be impossible to get to base object
		if(acceptConditions[i].condDim == QS_0D) cDim = 0;
		else if(acceptConditions[i].condDim == QS_1D) cDim = 1;
		else if(acceptConditions[i].condDim == QS_2D) cDim = 2;
		else if(acceptConditions[i].condDim == QS_ARRAY) cDim = 3;
		else if(acceptConditions[i].condDim == QS_ANY) cDim = 3;
		if(targetDim < cDim) 
		{
			targetDim = cDim;
			targetSeq = acceptConditions[i].condSeq;
		}
		
	}
	
	QString defText = getDefaultComp(targetDim, targetSeq);
	this->setText(defText);
	reValidate(true);
}


void QguWLinput::openSelectMetadata(){

	QString objName = this->textObject();

	QdObject_var dobj = QuGetObjByName(objName);
	if( !dobj.is_nil() )
	{
		selectMetadataWindow * metaWindow =	new selectMetadataWindow(dobj, this);
		metaWindow->parentSlot = this;
		metaWindow->move(QCursor::pos());
		metaWindow->show();
	}
	else setColour(COL_NOT_OK);
}


void QguWLinput::openArrayReduce(){

	if(QuSlotDebug())
		std::cout << "QguWLinput::openArrayReduce () " << this << endl;

	QString objName = this->textObject();
	Dim_var mdim;
	QcalcListView *qview = NULL;
	
	if(this->isAlias)
	{
     // defensive test that list exists
		qview = dynamic_cast <QcalcListView*>(this->aliasList);
		if(qview == NULL) return;
		qview->getAliasArrDim(objName,mdim);
	}
	else{
		QdObject_var dobj = QuGetObjByName(objName);
		if(dobj.is_nil() ){
			cout << "unresolved object: notify csc-support" << endl;
			setColour(COL_NOT_OK);
			return;
		}
		QdRMatrixSeq_var ms_dobj = QdRMatrixSeq_var::narrow(dobj);
		if (ms_dobj.is_nil() ){
			QdRMatrix_var m_dobj = QdRMatrix_var::narrow(dobj);
			if (m_dobj.is_nil() ) {
				cout << "Cannot resolve matrix object - notify csc_upport" << endl;
				setColour(COL_NOT_OK);
				return;
			}
			mdim = new Dim(m_dobj->dimdata()); // only handles sequences of arrays of same dimensions
		}
		else{
			mdim = new Dim((ms_dobj[0]).dimdata()); // only handles sequences of arrays of same dimensions
		}
	}
	if(mdim.is_nil()) {
		QuMsgBar("Cannot get dimension of object", true);
		return;
	}
	
	ArrayReduceWindow * arrWindow = new ArrayReduceWindow(objName, *mdim, this, "ArrayWindow", false);
	arrWindow->setPalette(qApp->palette());

	arrWindow->parentSlot = this;
	arrWindow->move(QCursor::pos());
	arrWindow->show();
	arrWindow->setToLastChoice();
	if(qview) qview->selectedObjDim = mdim; // keep here for multiple selection propagation

}

void QguWLinput::setEnabled(bool enabled){
	lineBox->setEnabled(enabled);
	if( enabled) setColour(COL_OK);
	else setColour(COL_INACTIVE);
}
 
 
/// \fn QguWLinput::setReadOnly()
/// \brief passes flag to setReadOnly of contained QLineEdit
/// \exception None
/// \return    None
///  
void QguWLinput::setReadOnly(bool readOnly)
{
	lineBox->setReadOnly(readOnly);
}

/// \fn QguWLinput::clear()
/// \brief passes clear to contained QLineEdit
/// \exception None
/// \return    None
///
void QguWLinput::clear()
{
	lineBox->clear();
}


void QguWLinput::validateObjChanged(const QString& context)
{
	if(this->text().startsWith(context))
	{
		reValidate(true);
	}
}

void QguWLinput::validateDeletion(const QString& path)
{
	const QString txt = this->text();
	if(txt.startsWith(path))
	{
		reValidate();
	}
}

void QguWLinput::validatePathChanged(const QString& from, const QString& to)
{
	const QString txt = this->text();
	if(txt.startsWith(from) || txt.startsWith(to))
	{
		reValidate();
	}
}

void QguWLinput::validateCreation(const QString& path)
{
	const QString txt = this->text();
	if(txt.startsWith(path))
	{
		reValidate(true);
	}
}

void QguWLinput::resetColour(){

	this->setPalette(*WLI_palette);
	this->lineBox->setPalette(*WLI_palette);
	this->dropDownArray->setPalette(*WLI_palette);
}

void QguWLinput::selectTivlEvent(){

	if(QuSlotDebug())
		std::cout << "QguWLinput::selectTivlEvent "  << endl;

	QString objName = this->textObject();

	SelectTivlWindow * selectWindow = new SelectTivlWindow(objName, this, "tSliceWindow", false);
	selectWindow->setPalette(qApp->palette());

	selectWindow->parentSlot = this;
	selectWindow->move(QCursor::pos());

	selectWindow->exec();
}

void QguWLinput::selectEventAndTime(){

	if(QuSlotDebug())
		std::cout << "QguWLinput::selectEventAndTime "  << endl;

	QString objName = this->textObject();

	SelectETWindow * selectWindow = new SelectETWindow(objName, this, "etWindow", false);
	selectWindow->setPalette(qApp->palette());

	selectWindow->parentSlot = this;
	selectWindow->move(QCursor::pos());

	selectWindow->exec();
}



void QguWLinput::selectTivlSeqSubset(){
	if(QuSlotDebug())
		std::cout << "QguWLinput::selectTivlSeqSubset "  << endl;

	QString objName = this->textObject();

	SelectTivlSubsetWindow * selectWindow = new SelectTivlSubsetWindow(objName, this, "tSliceWindow", false);
	selectWindow->setPalette(qApp->palette());

	selectWindow->parentSlot = this;
	selectWindow->move(QCursor::pos());

	selectWindow->show();
}

/// \fn void QguWLinput::setColour(int validity)
/// \brief Method that sets the colour of the line edit object on the basis of the validity state of the content of this object
/// \param     validity of type const QString
/// \exception None
/// \return    None
///
void QguWLinput::setColour(int validity){
//	if(QuSlotDebug()) std::cout << "QguWLinput::setColour (validity) " << this << "( "<< validity << " )" <<endl;

	WLI_palette->setBrush(QPalette::Active, QPalette::Button, WLI_palette->light());
	WLI_palette->setBrush(QPalette::Inactive, QPalette::Button, WLI_palette->light());
	WLI_palette->setBrush(QPalette::Inactive, QPalette::Text, WLI_palette->mid());
	WLI_palette->setBrush(QPalette::Inactive, QPalette::Base, WLI_palette->midlight());

	_validity=validity;

// palettes need to be changed BEFORE signals are emitted, to guarantee that
// any fields subsequently changed by slots connected to these signals,
// are not undone on return to this function

	if( validity == COL_OK){
		WLI_palette->setColor(QPalette::Active, QPalette::Base, col_White);
		WLI_palette->setColor(QPalette::Active, QPalette::Text, col_Black);
		WLI_palette->setColor(QPalette::Inactive, QPalette::Base, col_White);
		WLI_palette->setColor(QPalette::Inactive, QPalette::Text, col_Black);

		this->setPalette(*WLI_palette);
		this->lineBox->setPalette(*WLI_palette);
		this->dropDownArray->setPalette(*WLI_palette);
	}
	else if( validity == COL_INACTIVE){
		WLI_palette->setColor(QPalette::Active, QPalette::Base, col_White);
		WLI_palette->setColor(QPalette::Active, QPalette::Text, col_Grey);
		WLI_palette->setColor(QPalette::Inactive, QPalette::Base, col_White);
		WLI_palette->setColor(QPalette::Inactive, QPalette::Text, col_Grey);

		this->setPalette(*WLI_palette);
		this->lineBox->setPalette(*WLI_palette);
		this->dropDownArray->setPalette(*WLI_palette);
	}
	else if( validity == COL_NOT_OK){
		WLI_palette->setColor(QPalette::Active, QPalette::Base, col_Red);
		WLI_palette->setColor(QPalette::Active, QPalette::Text, col_Grey);
		WLI_palette->setColor(QPalette::Inactive, QPalette::Base, col_DarkRed);
		WLI_palette->setColor(QPalette::Inactive, QPalette::Text, col_Grey);

		this->setPalette(*WLI_palette);
		this->lineBox->setPalette(*WLI_palette);
		this->dropDownArray->setPalette(*WLI_palette);
	}
	else if( validity == COL_INCOMPLETE){
		WLI_palette->setColor(QPalette::Active, QPalette::Base, col_Amber);
		WLI_palette->setColor(QPalette::Active, QPalette::Text, col_Black);
		WLI_palette->setColor(QPalette::Inactive, QPalette::Base, col_Yellow);
		WLI_palette->setColor(QPalette::Inactive, QPalette::Text, col_Black);

		this->setPalette(*WLI_palette);
		this->lineBox->setPalette(*WLI_palette);
		this->dropDownArray->setPalette(*WLI_palette);
	}
	else if( validity == COL_COMP){
		WLI_palette->setColor(QPalette::Active, QPalette::Base, col_Yellow);
		WLI_palette->setColor(QPalette::Active, QPalette::Text, col_Blue);
		WLI_palette->setColor(QPalette::Inactive, QPalette::Base, col_LightYellow);
		WLI_palette->setColor(QPalette::Inactive, QPalette::Text, col_Blue);

		this->setPalette(*WLI_palette);
		this->lineBox->setPalette(*WLI_palette);
		this->dropDownArray->setPalette(*WLI_palette);

	}
	return;
}

bool QguWLinput::isValid() const
{
	return (_validity==COL_OK || _validity==COL_COMP);
}

// -----------------------------------------------------------------

bool QguWLinput::reValidate(bool propagate)
{
	if(QuSlotDebug())
	{
		printf("QguWLinput::reValidate %ld %d\n", (long)this, propagate);
		fflush(stdout);
	}

	// This method allows the calling method to force slot validation

	if(!(this->active))
	{
		// grey out if inactive
		ObjectName = ObjectName_nil;
		setColour(COL_INACTIVE);
		return false;
	}

	QString txt = this->text();
	
	// if slot is empty, only validate once to set colour and don't propagate
	if(txt.isEmpty() && acceptConditions.size() > 0){
		bool OK = validateCondition(txt, acceptConditions[0]);
		return OK;	
	}
	

	if(propagate) 
	{
		return validateInput(txt); // this can emit edited_OK() signal
	}
	
	// non-propagating validation when program sets text
	for(int i=0; i<acceptConditions.size(); i++)
	{
		bool OK = validateCondition(txt, acceptConditions[i]);
		if( OK ) 
		{
			// no signal is emitted, this version just provides visual feedback
			return OK;
		}
	}
	
	return false;
}


bool QguWLinput::validateInput(const QString &txt)
{
	// slot that always causes propagation and doesn't need slot text to be set (e.g. maybe from textEdited() signal)
	for(int i=0; i<acceptConditions.size(); i++)
	{
		bool OK = validateCondition(txt, acceptConditions[i]);
		if( OK ) 
		{
			emit edited_OK();
			return OK;
		}
	}
	return false;
}

bool QguWLinput::validateCondition(const QString &txt, Qcondition &cond)
{

	if(QuSlotDebug())
	{
		printf("QguWLinput::validateCondition %ld '%s'\n", (long)this, txt.toLatin1().data());
		fflush(stdout);
	}
	isAlias = false;
	
	QString noWhite = txt.trimmed();
	if(noWhite.isEmpty())
	{
		emit slotEmpty();
		ObjectName = ObjectName_nil;
		setColour(COL_OK);
		if ( dropDownArray->isVisibleTo(this)) dropDownArray->hide();
		return false; // will not emit signal edited_OK
	}
		
	// Does this resolve to a folder
	WLItem *WLitem = qsasWindow->wlview->wlmodel()->itemFromPath(ObjectName);
	if(WLitem != NULL && WLitem->isFolder() && cond.condData != QS_TEXT)
	{
		ObjectName = ObjectName_nil;
		setColour(COL_INCOMPLETE);
		if ( dropDownArray->isVisibleTo(this)) dropDownArray->hide();
		return false;
	}

	ObjectName = QString(baseObjectText(txt).toLatin1().data()); // deep copy
	QdObject_var dobj = QuGetObjByName(ObjectName);


	// QS_WL_OBJ
	// ----------
	if(cond.condObj == QS_WL_OBJ)
	{
		if(dobj.is_nil())
		{
			if ( dropDownArray->isVisibleTo(this)) dropDownArray->hide();
			setColour(COL_NOT_OK);
			return false;
		}

		QdRScalar_var sobj = QdRScalar_var::narrow(dobj);
		QdRMatrix_var mobj = QdRMatrix_var::narrow(dobj);
		QdRMatrixSeq_var ms = QdRMatrixSeq_var::narrow(dobj); 
		QdTimeIntervalSeq_var elobj = QdTimeIntervalSeq_var::narrow(dobj);
		QdTimeInterval_var tiobj = QdTimeInterval_var::narrow(dobj);

		if(sobj.is_nil() && mobj.is_nil() && ms.is_nil() && elobj.is_nil() && tiobj.is_nil())
		{
			setDropdown(getDefaultPopup(QS_ANY));
		}
		else if( !sobj.is_nil() )
		{
			setDropdown(getDefaultPopup(QS_NOT_SEQ));
		}
		else if( !mobj.is_nil() || !ms.is_nil() )
		{
			if( QuIsThreeVector(dobj) ) 
			{
				setDropdown(getVectorPopup());
			}
			else
			{
				setDropdown(getArrayPopup());
			}
		}
		else if( !elobj.is_nil()){
			setDropdown(getTivlFromEvtTablePopup());
		}
		else if( !tiobj.is_nil()){
			setDropdown(getTimePopup());
		}
		
		QguComp_e component = this->getComponent();
		if(component == WL_OBJ) {
			emit validated_WL_OBJ(ObjectName);
			setColour(COL_OK);
		}
		else setColour(COL_COMP);

		return true;
	}
	
	// QS_NARROW
	// ----------
	else if(cond.condObj == QS_NARROW)
	{
		if(dobj.is_nil())
		{
			if ( dropDownArray->isVisibleTo(this)) dropDownArray->hide();
			
			// only value typed in slot possible
			if(cond.condTT == QS_TT ||
			   cond.condSeq == QS_SEQ ||
			   (cond.condDim != QS_0D && cond.condDim != QS_ANY) )
			{
				setColour(COL_NOT_OK);
				return false;
			}
			
			if(cond.condData == QS_DBL)
			{
				bool OK;
				txt.toDouble(&OK); 
				if(OK) 
				{
					setColour(COL_OK);
					return true;
				}
				else
				{
					setColour(COL_NOT_OK);
					return false;
				}
			}
			else if(cond.condData == QS_INT)
			{
				bool OK;
				txt.toInt(&OK); 
				if(OK) 
				{
					setColour(COL_OK);
					return true;
				}
				else
				{
					setColour(COL_NOT_OK);
					return false;
				}
			}
			else if(cond.condData == QS_TIME)
			{
				QDOS::QmjdTime timeValue;
				if(timeValue.setFromISOstring(txt.toLatin1().data()))
				{
					setColour(COL_OK);
					return true;
				}
				else
				{
					setColour(COL_INCOMPLETE);
					return false;
				}
			}
			else if(cond.condData == QS_TI)
			{
				QDOS::QmjdTimeInterval TIvalue;
				if(TIvalue.setFromISOstring(txt.toLatin1().data()))
				{
					setColour(COL_OK);
					return true;
				}
				else
				{
					setColour(COL_INCOMPLETE);
					return false;
				}
			}
			else
			{
				// always accept  QS_ANY or QS_TEXT
				setColour(COL_OK);
				return true;
			}
		}
		else
		{
			// object exists, can we narrow or subsample to required condition
				
			QguComp_e component = this->getComponent();
			if(component == WL_OBJ) {
				 setColour(COL_OK);
			}
			else 
			{
				setColour(COL_COMP);
			}

			// sequence conditions
			int seqLen = dobj->sequence_size();
			if( seqLen < 1 && cond.condSeq == QS_SEQ    ) 
			{
				setColour(COL_NOT_OK);
				return true;
			}
			
			// TT conditions
			QdTimeSeq_var ts = get_timetags(dobj); // object or its DEPEND_0
			if( ts.is_nil()  && cond.condTT == QS_TT)  
			{
				setColour(COL_NOT_OK);
				return false;
			}
			
			QdTimeIntervalSeq_var elobj = QdTimeIntervalSeq_var::narrow(dobj);
			if(cond.condData != QS_TIME && cond.condData != QS_TI && !elobj.is_nil()){
				setColour(COL_NOT_OK);
				return false;
			}
			
			// data type conditions
			if(cond.condData == QS_TIME)
			{
				QdTime_var tobj = QdTime_var::narrow(dobj);
				QdTimeInterval_var tiobj = QdTimeInterval_var::narrow(dobj);
				
				if(tobj.is_nil() && ts.is_nil() && tiobj.is_nil() && elobj.is_nil())
				{
					setColour(COL_NOT_OK);
					return false;
				}
				
				if(cond.condSeq == QS_NOT_SEQ)
				{
					if( !tobj.is_nil())
					{
						setColour(COL_OK);
					}
					else if(!elobj.is_nil()) setDropdown(getTimeFromELPopup());
					else 					 setDropdown(getTimePopup());
				}
				else
				{
					if( ts.is_nil())
					{
						setColour(COL_NOT_OK);
						return false;
					}
					setDropdown(getDefaultPopup(cond.condSeq));
				}
				return true;
			}
	
			if(cond.condData == QS_TI)
			{
				QdTimeInterval_var tiobj = get_timeinterval(dobj);
				QdTimeIntervalSeq_var tis = QdTimeIntervalSeq_var::narrow(dobj);
				if(tiobj.is_nil() && tis.is_nil() )
				{
					setColour(COL_NOT_OK);
					return false;
				}
				
				if(cond.condSeq == QS_NOT_SEQ)
				{
					if(!tis.is_nil())
					{
						setDropdown(getTivlFromEvtTablePopup());
						if(component == WL_OBJ )
						{
							this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(EVENT_0_STR));
							setColour(COL_COMP);
						}
					}
					else
					{
						setDropdown(getTrangePopup());
					}
				}
				else
				{
					setDropdown(getTivlSeqSlicePopup());
				}
				return true;
				
			}
			
			if(cond.condData == QS_TEXT)
			{
				setDropdown(getDefaultPopup(cond.condSeq));
				return true;
			}
			
			if(cond.condData == QS_DBL || cond.condData == QS_INT)
			{
				QdRScalar_var sobj = QdRScalar_var::narrow(dobj);
				QdRScalarSeq_var ss = QdRScalarSeq_var::narrow(dobj);
				QdRMatrix_var mobj = QdRMatrix_var::narrow(dobj);
				QdRMatrixSeq_var ms = QdRMatrixSeq_var::narrow(dobj); 
				if(sobj.is_nil() && ss.is_nil() && mobj.is_nil() && ms.is_nil() )
				{
				
					setColour(COL_NOT_OK);
					return false;
				}
				
				if( !sobj.is_nil() || !ss.is_nil() )
				{
					setDropdown(getDefaultPopup(cond.condSeq));
					return true;
				}
				else if( !mobj.is_nil() || !ms.is_nil() )
				{
					if( QuIsThreeVector(dobj) ) 
					{
						setDropdown(getVectorPopup());
						if(component == WL_OBJ && cond.condDim == QS_0D) 
						{
							this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(R_STR_CMP));
							setColour(COL_COMP);
						}
						return true;
					}
					else
					{

						setDropdown(getArrayPopup());
						if(component == WL_OBJ )
						{
							QString defComp = getDefaultComp(cond.condDim, cond.condSeq);

							this->lineBox->setText(defComp);
							setColour(COL_COMP);
						}
						return true;
					}
				}
			}
			
			if(cond.condData == QS_ANY )
			{
				// just do dropdown for now 
				// could handle each data type differently
				setDropdown(getDefaultPopup(cond.condSeq));
				return true;
			}
			
		}
		
	}
	
	// QS_ALIAS (for calculator)
	// -------------------------
	
	else if(cond.condObj == QS_ALIAS)
	{

		// defensive test that list exists
		QcalcListView *qview = dynamic_cast<QcalcListView*>(this->aliasList);
		if(qview == NULL) return false;

		QcalcAliasObjType aliasObjType=qview->getAliasObjectType(ObjectName);
		if(aliasObjType!=ALIAS_IS_UNKNOWN)
		{
			// handle individual alias types
			// Note alias may be virtual, so rely on getIsVector() and
			// getIsArray()  constructed from inputs
			// dobj itself may not exist yet

			isAlias = true;
			QguComp_e component = this->getComponent();
			
			if (cond.condDim == QS_0D)
			{
				if(aliasObjType==ALIAS_IS_VECTOR)
				{
					setDropdown(getVectorPopup());
					if(component == WL_OBJ)
					{					
						this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(R_STR_CMP));
						setColour(COL_COMP);
					}
				}
				else if(aliasObjType==ALIAS_IS_ARRAY)
				{
					// array sub-sampling
					// if reduction already made this may now be a scalar really
					Dim_var vDim = this->getSlotArrDim();
					if (vDim.is_nil())
						setColour(COL_INCOMPLETE); // original array in slot
					else if ( vDim->rank() > 0)
						setColour(COL_INCOMPLETE);  // still has dimensions
					else
						setColour(COL_COMP);  // reduced to scalar

					setDropdown(getArrayPopup());
				}
				else
				{
					// arithmetic not handling text and other data types 
	       				// >>> update for plugin interface
					setDropdown(getDefaultPopup(cond.condSeq));
					
					setText(ObjectName);
	       				// reset name in case input virtual obj has gone from vector to scalar
					setColour(COL_OK);
				}

				return true;
			}
			else if (cond.condDim == QS_VECTOR)
			{
				if(aliasObjType==ALIAS_IS_VECTOR)
				{
					setColour(COL_OK);
				}
				if ( dropDownArray->isVisibleTo(this)) dropDownArray->hide();
				
				return true;
			}
			else
			{  // generic acceptAlias
				if(aliasObjType==ALIAS_IS_VECTOR)
				{
					setDropdown(getVectorPopup());
					
					if (component != WL_OBJ) setColour(COL_COMP);
					else setColour(COL_OK);
				}
				else if(aliasObjType==ALIAS_IS_ARRAY)
				{
					// general case of any object or array object, so
					// dobj may be acceptable
					setColour(COL_OK); // original array in slot
					setDropdown(getArrayPopup());
					
					if (component != WL_OBJ) setColour(COL_COMP);
				}
				else
				{
					// reset if component no longer allowed
					dropDownArray->hide();
					if(ObjectName != txt) setText(ObjectName);
					setColour(COL_OK);
				}

				return true;

			}
		}
	} // end if on ALIAS
	
	return false;
}

void  QguWLinput::setDropdown(QMenu *dropdownChoice)
{
	dropDownArray->setMenu(dropdownChoice);
	dropDownArray->hide();
	if ( !(dropDownArray->isVisibleTo(this)))
	{
		dropDownArray->show();
	}
	return;
}

// --------------------------------

void QguWLinput::setDefaultComp(int targetDim, int targetIsSeq, QString &obj_name)
{
	this->setText(obj_name);
	QString compText = this->getDefaultComp(targetDim, targetIsSeq);
	this->setText(compText);
}

// --------------------------------

QString QguWLinput::getDefaultComp(int targetDim, int targetIsSeq)
{
	
	if(QuSlotDebug()) cout << "QguWLinput::getDefaultComp:  "  << targetDim <<  " " << targetIsSeq << endl;
	
	QString outText = baseObjectText(this->text());
	
	if(targetDim >= 3) return outText;
	
	if(outText.isEmpty())  return this->text();

	 QString compDef;
	 if(targetDim == 0 ) compDef = QSAS_profiles->current()->Array3Dto0D;
	 if(targetDim == 1 ) compDef = QSAS_profiles->current()->Array3Dto1D;
	 if(targetDim == 2 ) compDef = QSAS_profiles->current()->Array3Dto2D;
	
	if(compDef.isEmpty()) return outText;
	
	QdObject_var dobj = QuGetObjByName(outText);
	
	QString guiChoice = get3DSubsetDef(dobj);
	if( !guiChoice.isEmpty() ) {
		outText += guiChoice;
		return outText;
	}

	QString compTxt = " [";
	
	
	Dim_var d;
	QdRMatrixSeq_var MatSeq_in;
	QdRMatrix_var Mat_in;
	Mat_in = QdRMatrix_var::narrow(dobj);
	if( !Mat_in.is_nil() )
	{
		d = new Dim(Mat_in->dims());
	}
	else 
	{
		MatSeq_in = QdRMatrixSeq_var::narrow(dobj);
		if( !MatSeq_in.is_nil()) 
		{
			d = new Dim(MatSeq_in[0].dims());
		}
		
	}

	if(d.is_nil() || d->rank() <= targetDim) return outText;
	
	int di;
	if(compDef.contains("[") || compDef.contains("<"))
	{
		compTxt = QString(" ")+compDef;
	}
	else 
	{
		for (int i=0; i<compDef.length(); i++)
		{
			if( d->rank() <= i) return outText;
		
			if(compDef[i] == QChar('i'))
			{
				di = (d->dims())[i]; 
				compTxt +=  QString("(1,%1) ").arg(di);
			}
			else if(compDef[i] == QChar('S'))
			{
				di = (d->dims())[i];
				compTxt +=  QString("Sum(1,%1) ").arg(di);
			}
			else if(compDef[i] == QChar('A'))
			{
				di = (d->dims())[i];
				compTxt +=  QString("Ave(1,%1) ").arg(di);
			}

			else return outText;
		}
		if(targetIsSeq == QS_NOT_SEQ  && !MatSeq_in.is_nil()) compTxt +=  "] < 0 >";
		else compTxt +=  "]";
	}
	
	outText += QString(compTxt);
	
	return outText;

}


/// \fn QdObject_var QguWLinput::getSlotObject()
/// \brief Method that gets the QdObject from the WL on the basis of the name that is held within the line edit text
///
/// Will create an object from input text (not on WL, so volatile) if slot accepts non-object input.
/// Will return a QdObject_var containing a dimension reduced object if pulldown is used to select dimension reduction
/// \param     None
/// \exception None
/// \return    Returns QdObject_var, nil if conversion fails
///
QdObject_var QguWLinput::getSlotObject()
{

	// always returns a QdObject_var pointer.
	// Returns nil pointer on failure (test on obj.is_nil() ) and must be narrowed to object type required

	if ( QuSlotDebug() )
	{
		printf( " QguWLinput::getSlotObject for '%s' ObjectName= '%s' \n", text().toLatin1().data(), ObjectName.toLatin1().data());
		fflush(stdout);
	}
	QdObject_var slotObj;
	// get slot text
	QString txt =this->text().trimmed();
	
	
	// give up now if slot empty
	if(txt.isEmpty()) return slotObj;

	// try to resolve to one of the conditions in the order they are set
	for(int i=0; i<acceptConditions.size(); i++)
	{
		Qcondition cond = acceptConditions[i];


		// QS_WL_OBJ
		// ----------
	
		if(cond.condObj == QS_WL_OBJ)
		{ 
		
		//slotObj = QuGetObjByName(baseObjectText(txt));
			slotObj = QuGetObjByNameAndComp(txt);
			if( !slotObj.is_nil() ) return  slotObj;
		}
	
		// QS_NARROW
		// ----------
	
		else if(cond.condObj == QS_NARROW)
		{
			QguComp_e component = this->getComponent();
			QdObject_var baseObj = QuGetObjByName(baseObjectText(txt));
		
			if(baseObj.is_nil())
			{
			
				// only value typed in slot possible
				if(cond.condTT == QS_TT || cond.condSeq == QS_SEQ || (cond.condDim != QS_0D && cond.condDim != QS_ANY) ) return slotObj;	
			
				if(cond.condData == QS_DBL || cond.condData == QS_INT)
				{
					// qdos does not handle int
					bool OK;
					double value = txt.toDouble(&OK); 
					if(OK) 
					{
						QdRScalar_var s_obj = new QdRScalar(value);
						slotObj = s_obj;
						return slotObj;
					}
					
				}
			
				else if(cond.condData == QS_TIME)
				{
					QDOS::QmjdTime timeValue;
					if(timeValue.setFromISOstring(txt.toLatin1().data()))
					{
						QdTime_var t_obj = new QdTime(timeValue);
						slotObj = t_obj;
						return slotObj;
					}
				}

				else if(cond.condData == QS_TI)
				{
					QDOS::QmjdTimeInterval TIvalue;
					if(TIvalue.setFromISOstring(txt.toLatin1().data()))
					{
						QdTimeInterval_var ti_obj = new QdTimeInterval(TIvalue);
						slotObj = ti_obj;
						return slotObj;
					}
				}

				else if(cond.condData == QS_TEXT)
				{
					QdString_var str_obj = new QdString(txt.toLatin1().data());
					slotObj = str_obj;
					return slotObj;
				}

			}
			else
			{
				// base object exists, can we narrow or subsample to required condition

				// sequence conditions
				int seqLen = baseObj->sequence_size();
				if( seqLen <= 1 && cond.condSeq == QS_SEQ ) continue; // try next condition
			
				// TT conditions
				if( !baseObj->xref_exists(DEPEND_0)  && cond.condTT == QS_TT)  continue; // try next condition
			
				// data type conditions
				if(cond.condData == QS_TIME)
				{
					slotObj = QuGetTimeByNameAndComp(txt);
				}
				
				else if(cond.condData == QS_TI)
				{
					slotObj = QuGetIntervalsByNameAndComp(txt);
				}
			
				else if(cond.condData == QS_TEXT)
				{
					slotObj = QuGetObjByNameAndComp(txt);
				}
			
				else if(cond.condData == QS_DBL || cond.condData == QS_INT)
				{

					slotObj = QuGetObjByNameAndComp(txt);
				}
			
				else if(cond.condData == QS_ANY )
				{
					slotObj = QuGetObjByNameAndComp(txt);
				}
				
				if( !slotObj.is_nil() ) return slotObj;
			
			}
		
		}
	
	
		// QS_ALIAS (for calculator)
		// -------------------------
	
		else if(cond.condObj == QS_ALIAS)
		{	
			QguComp_e component = this->getComponent();
			
			QdObject_var baseObj;
			QcalcListView *qview = dynamic_cast <QcalcListView*>(this->aliasList);
			if(qview == NULL) return slotObj;
	
			baseObj=qview->getAliasResultObject(ObjectName);
	
			if(component == WL_OBJ){
				slotObj = baseObj;
			}
			else if(component == ARRAY_COMP){
				dimTripleStack dimStack = this->parseArrayReduce();
				slotObj = QuGetArrayComponent(baseObj, dimStack );
			}
			else if(component == R_COMP){
				slotObj = QuGetVectorComponent(baseObj, "r(mag)" );
			}
			else if(component == X_COMP){
				slotObj = QuGetVectorComponent(baseObj, "x" );
			}
			else if(component == Y_COMP){
				slotObj = QuGetVectorComponent(baseObj, "y" );
			}
			else if(component == Z_COMP){
				slotObj = QuGetVectorComponent(baseObj, "z" );
			}
			else if(component == LAT_DEG_COMP){
				slotObj = QuGetVectorComponent(baseObj, "Lat(deg)" );
			}
			else if(component == THETA_DEG_COMP){
				slotObj = QuGetVectorComponent(baseObj, "theta(deg)" );
			}
			else if(component == PHI_DEG_COMP){
				slotObj = QuGetVectorComponent(baseObj, "phi(deg)" );
			}
			else if(component == LAT_RAD_COMP){
				slotObj = QuGetVectorComponent(baseObj, "Lat(rad)" );
			}
			else if(component == THETA_RAD_COMP){
				slotObj = QuGetVectorComponent(baseObj, "theta(rad)" );
			}
			else if(component == PHI_RAD_COMP){
				slotObj = QuGetVectorComponent(baseObj, "phi(rad)" );
			}
			else if(component == T_START_COMP){
				slotObj = QuGetTimeComponent(baseObj, "Tstart" );
			}
			else if(component == T_CENTRE_COMP){
				slotObj = QuGetTimeComponent(baseObj, "Tcentre" );
			}
			else if(component == T_END_COMP){
				slotObj = QuGetTimeComponent(baseObj, "Tend" );
			}
			else if(component == XREF_COMP){
				QString xref_name = getXrefName(text().toLatin1().data());
				try{
					QdObject_var xrefObj = baseObj->get_xref(xref_name.toLatin1().data());
					slotObj = xrefObj;
				}
				catch(Exception &e){
					QuMsgBar("Object Xref not found: ", xref_name.toLatin1().data(), true);
					slotObj = nilObject;
				}
			}
			if( !slotObj.is_nil() ) return slotObj;
		
		} // end if on ALIAS
  
	} // end for on condition list
	
	return slotObj; // probably is_nil if we get here
}

/// \fn double QguWLinput::value()
/// \brief Method that returns the double value in the input slot, if acceptNumber or acceptTime is true 
/// \param     None
/// \exception None
/// \return    double
///
double QguWLinput::value()
{
	bool acceptNumber=false;
	bool acceptTime=false;
	bool acceptTI=false;
	for (int i=0; i<acceptConditions.size(); i++)
	{
		Qcondition cond = acceptConditions[i];
		if( cond.condDim == QS_0D && cond.condSeq == QS_NOT_SEQ)
		{
			if( cond.condData == QS_DBL || cond.condData == QS_INT ) acceptNumber = true;
			if( cond.condData == QS_TIME ) acceptTime = true;
			if( cond.condData == QS_TI ) acceptTI = true;
		}
	}
	
	if (acceptNumber)
	{
		bool ok;
		return this->lineBox->text().toDouble(&ok);
	}
	else if (acceptTime)
	{
		QmjdTime slotTime;
		try{
			slotTime.setFromISOstring(this->lineBox->text().toLatin1().data());
		}
		catch(...){return 0;}

		return  slotTime.getEpoch2000(); // used for slots needing to get/set from double
	}
	else if (acceptTI)
	{
		QmjdTimeInterval slotTI;
		try{
			slotTI.setFromISOstring(this->lineBox->text().toLatin1().data());
		}
		catch(...){return 0;}

		return  slotTI.duration(); // used for slots needing to get/set from double
	}
	else return 0;
}

/// \fn void QguWLinput::setDouble()
/// \brief Method that sets the slot text as time string from double if acceptTime is true 
/// \param     None
/// \exception None
/// \return    none
///
void QguWLinput::setDouble(double value)
{

	bool acceptNumber=false;
	bool acceptTime=false;
	bool acceptTI=false;
	for (int i=0; i<acceptConditions.size(); i++)
	{
		Qcondition cond = acceptConditions[i];
		if( cond.condDim == QS_0D && cond.condSeq == QS_NOT_SEQ)
		{
			if( cond.condData == QS_DBL || cond.condData == QS_INT ) acceptNumber = true;
			if( cond.condData == QS_TIME ) acceptTime = true;
			if( cond.condData == QS_TI ) acceptTI = true;
		}
	}

	if (acceptTime)
	{
		QmjdTime slotTime(value);
		this->lineBox->setText(slotTime.getISOstring(0).c_str());
	}
	if(acceptNumber)
	{
		this->lineBox->setText(double_to_string(value).c_str());
	}
	return;
}
 
// -------------------------------------------

Dim_var QguWLinput::getSlotArrDim( )
{
	Dim_var vDim;

	// construct dim for selection in ArrayReduce via text() in slot

	dimTripleStack arrTriple = this->parseArrayReduce();
	if(arrTriple.size() == 0)  return vDim;  // no reduction done, return is_nil() object
		
	vector <int> dimList;
	for(int i=0; i<arrTriple.size(); i++){
		if(arrTriple[i][0] == 0 ) {
			// this dim not collapsed
			int elems = arrTriple[i][2] - arrTriple[i][1] + 1; // start > end not yet supported
			dimList.push_back(elems);
		}
	}
	vDim = new Dim(dimList);
	return vDim;
}

/// \fn void QguWLinput::dropEvent(QDropEvent *e)
/// \brief This event handler is called when the drag is dropped on this widget. It has been re-implemented from QWidget
///
/// Sets and validates dropped text
/// \param     e of type QDropEvent *
/// \exception None
/// \return    None
///
void QguWLinput::dropEvent(QDropEvent *e)
{
	if(QuSlotDebug())
	{
		printf("QguWLinput::dropEvent %ld \n", (long)this);
		fflush(stdout);
	}

	QApplication::setOverrideCursor( QCursor(Qt::WaitCursor) );

	qApp->processEvents();


	// use valid ObjectName in feedback

	if (e->mimeData()->hasText())
	{
		string newText = e->mimeData()->text().toLatin1().data();
		cleanWLtext(newText);
		setTextKeepComp(QString(newText.c_str()));
			
	}
	
	QApplication::restoreOverrideCursor();

	qApp->processEvents();
	lineBox->setFocus();
	emit hadDropEvent();
	reValidate(true); // causes edited_OK to be emitted on success
}




/// \fn void QguWLinput::setText(QString txt)
/// \brief Method that sets the text in the input slot
/// \param     txt of type QString
/// \exception None
/// \return    None
///
void QguWLinput::setText(QString txt)
{
	this->lineBox->setText(txt);
	return;

}

/// \fn void QguWLinput::setTextKeepComp(QString txt)
/// \brief Method that sets the text in the input slot but retains component choice if valid
/// \param     txt of type QString
/// \exception None
/// \return    None
///
void QguWLinput::setTextKeepComp(const QString &txt){
	
	if ( QuSlotDebug() )
	{
		printf( " QguWLinput::setTextKeepComp for  '%s'\n",  txt.toLatin1().data());
		fflush(stdout);
	}
	bool compOK = true;
	QString oldText = QString(this->text().toLatin1().data()); // deep copy
	int cptr = oldText.indexOf("[");
	int tptr = oldText.indexOf("<");

	// find any tokens to replace
	QString newText = QString(txt.toLatin1().data()); // deep copy of dropped name
	QuReplaceTokens(newText, oldText);
	
	if (cptr != -1 || tptr != -1 ){
	
	  	// array reduce component found
		int ptr = oldText.indexOf("[");
		if (ptr == -1) 	oldText = "";
		else {
			oldText.remove(0, ptr-1);
			ptr = oldText.indexOf("]");
			oldText.truncate(ptr+1);
		}

		QdObject_var dobj = QuGetObjByName(newText);

		bool vectorDrop = false;
		compOK = compatibleComp(dobj, oldText, vectorDrop);

		int maxSeq = 0;
		if( !dobj.is_nil() ) maxSeq = dobj->sequence_size();

		// now replace with components if still valid
		if( compOK ) 
		{
			newText +=  oldText;
			this->lineBox->setText(newText);
			return;
		}
	}

	// Empty slot or no component and subset.
	// Set new object and return
	bool acceptNumber=false;
	bool acceptTime=false;
	for (int i=0; i<acceptConditions.size(); i++)
	{
		Qcondition cond = acceptConditions[i];
		if( cond.condDim == QS_0D && cond.condSeq == QS_NOT_SEQ)
		{
			if( cond.condData == QS_DBL || cond.condData == QS_INT ) acceptNumber = true;
			if( cond.condData == QS_TIME ) acceptTime = true;
		}
	}
	this->lineBox->setText(newText);
	this->reValidate(); // needed to set objectName
	
	// now correct for special requests
	
	QdObject_var obj = QuGetObjByName(newText);
	QdTimeIntervalSeq_var ET = QdTimeIntervalSeq_var::narrow(obj);
	if(acceptTime && !ET.is_nil()) setETDropdownDefault();
	else if(acceptTime) setTimeDropdownDefault();
	else if(acceptNumber) 
	{
		if(QuIsThreeVector(obj)) setVectorDropdownDefault();
		else setArrayDropdownDefault();
	}
	
	emit dropDetected(txt.toLatin1().data());	
	return;

}


// ------------------------------------------------------
 
bool QguWLinput::compatibleComp(QdObject_var dobj, const QString &comp, bool &nowVector){
	
	const char *cmp = comp.toLatin1().data();
	bool vectorComp=false;
	nowVector = false;

	if(strcmp(cmp, R_STR_CMP) == 0)  vectorComp = true;
	if(strcmp(cmp, X_STR_CMP) == 0)  vectorComp = true;
	if(strcmp(cmp, Y_STR_CMP) == 0)  vectorComp = true;
	if(strcmp(cmp, Z_STR_CMP) == 0)  vectorComp = true;
	if(strcmp(cmp, Ld_STR_CMP) == 0) vectorComp = true;
	if(strcmp(cmp, Lr_STR_CMP) == 0) vectorComp = true;
	if(strcmp(cmp, Td_STR_CMP) == 0) vectorComp = true;
	if(strcmp(cmp, Tr_STR_CMP) == 0) vectorComp = true;
	if(strcmp(cmp, Pd_STR_CMP) == 0) vectorComp = true;
	if(strcmp(cmp, Pr_STR_CMP) == 0) vectorComp = true;

	// most things take time comp
	if(strcmp(cmp, T_STR_START) == 0) return true;
	if(strcmp(cmp, T_STR_CENTRE) == 0) return true;
	if(strcmp(cmp, T_STR_END) == 0) return true;
	
	if( QuIsThreeVector(dobj) )
	{
		nowVector = true;
		// look only for Vector comp types
		if(vectorComp) return true;
		return false;
	}

	// check array dimensions
	Dim_var dv;
	QdRMatrix_var rmatrix=QdRMatrix_var::narrow(dobj);
  	if(rmatrix.is_nil())
  	{
    	QdRMatrixSeq_var rmatrixseq=QdRMatrixSeq_var::narrow(dobj);
    	if(rmatrixseq.is_nil()) return false;
		
    	if(rmatrixseq->sequence_size() == 0 || !rmatrixseq->is_homogeneous_seq()) return false;
   		
		try
		{
      		dv = new Dim(rmatrixseq[0].dims());
    	}
    	catch(Exception &e)
    	{
      		 e.print_msg();
       		return false;
    	}
  	} 
	else{
	
  		try
  		{
    		dv = new Dim(rmatrix->dims());
  		}
  		catch(Exception &e)
  		{
   			e.print_msg();
    		return false;
		}
	}

	string unpack = comp.toLatin1().data();
	for(int i=0; i<dv->rank(); i++)
	{
		size_t ptr1 = unpack.find("(");
		if (ptr1 == string::npos) 	return false;
		
		size_t ptr2 = unpack.find(")");
		if (ptr2 == string::npos) 	return false;
		
		size_t ptrComma = unpack.find(",");
		int idx;
		if(ptrComma == string::npos){
			// single number, slice
			idx = string_to_int(unpack.substr(ptr1+1, ptr2-ptr1-1));
		}
		else{
			// two values, take second
			idx = string_to_int(unpack.substr(ptrComma+1, ptr2-ptrComma-1));
		}

		if(idx > (*dv)[i]) return false;
		
		// move to next comp
		unpack = unpack.substr(ptr2+1);
		
	}
	

	// must be matrix data 
	if(vectorComp) return false;

	return true;
	
}


// ------------------------------------------------------

void QguWLinput::setAliasKeepComp(const QString &txt){

	QString newTxt = QString(this->text().toLatin1().data()); // deep copy
	if (newTxt.isEmpty() ){
	   // set new object and return - should not happen
		this->lineBox->setText(txt);
		return;
	}

	int ptr = newTxt.indexOf(" [");
	if (ptr != -1)
	{
		newTxt.remove(0, ptr);
		newTxt = txt + newTxt;
	}
	else  newTxt = txt;
   
	
	this->lineBox->setText(newTxt);
	return;
}


/// \fn QString QguWLinput::text()
/// \brief Method that returns the text in the input slot
/// \param     None
/// \exception None
/// \return    text of type QString
///
QString QguWLinput::text()
{

	return this->lineBox->text();

}

/// \fn QString QguWLinput::textObject()
/// \brief Method that returns the text in the input slot with component text removed
/// \param     None
/// \exception None
/// \return    text of type QString
///
QString QguWLinput::textObject()
{
	QString baseTxt = QString( this->lineBox->text().toLatin1().data() );
	return baseObjectText(baseTxt);
}

/// \fn void QguWLinput::setCursorPosition()
/// \brief Method that sets the cursor position in the underlying lineEdit
/// \param     posn of type int
/// \exception None
/// \return    None
void QguWLinput::setCursorPosition(int posn)
{
	this->lineBox->setCursorPosition(posn);
	return;

}


/// \fn void QguWLinput::setAlignment(int flag)
/// \brief Method that sets the text alignment in the underlying lineEdit
/// \param     flag of type int
/// \exception None
/// \return    None

void QguWLinput::setAlignment(Qt::AlignmentFlag flag)
{
   // Takes int flag of form QT:AlignLeft of QT:AlignAuto
	this->lineBox->setAlignment(flag);

	return;

}


/// \fn void QguWLinput::setMaxLength(int len)
/// \brief Method that sets the max text length in the underlying lineEdit
/// \param     len of type int
/// \exception None
/// \return    None
void QguWLinput::setMaxLength(int len)
{

	this->lineBox->setMaxLength(len);

	return;

}



/// \fn vector <dimTriple *> QguWLinput::parseArrayReduce()
/// \brief Method that parses text in QguLineEdit and finds the array reduction requested
/// \param     None
/// \exception None
/// \return    vector <dimTriple *>

dimTripleStack QguWLinput::parseArrayReduce()
{
	if ( QuSlotDebug() )
		printf( "QguWLinput::parseArrayReduce this = %s \n", this->text().toLatin1().data() );

	dimTripleStack arrayTriples;

	QString compTxt = QString( this->lineBox->text().toLatin1().data() );
	int ptr = compTxt.indexOf(" [");
	if(ptr == -1) return arrayTriples;  // no component section found, arrayTriples is empty vector

	compTxt.remove(0, ptr+2); // remove " ["

	return getDimTripleStack(compTxt);

}

/// \fn vector <dimTriple *> QguWLinput::parseTimeSlice()
/// \brief Method that parses text in QguLineEdit and finds the array reduction requested
/// \param     None
/// \exception None
/// \return    vector <dimTriple *>

dimTriple QguWLinput::parseTimeSlice()
{
	if ( QuSlotDebug() )
		printf( "QguWLinput::parseTimeSlice this = %s \n", this->text().toLatin1().data() );

	QString cmpTxt = QString(this->lineBox->text().toLatin1().data());
	return getSliceTriple(cmpTxt);
}


/// \fn Vostream& operator<<(Vostream& v_os, QguWLinput& slot)
/// \brief Method that writes the content of slot object (object type/QdObject) to the text stream v_os
/// \param     v_os of type Vostream&
/// \param     slot of type QguWLinput&
/// \exception None
/// \return    None
///
Vostream& operator<<(Vostream& v_os, QguWLinput& slot)
{
	v_os << string(slot.text().toLatin1().data());

	if (slot.with == 1){
		QdObject_var obj = QuGetObjByName( slot.text().toLatin1().data() );
		if( !(obj.is_nil()) ){
			v_os << slot.with;
			string type=obj->object_type();
			v_os << type;
			v_os << obj;
		}
		else {
			slot.with = 0;
			v_os << slot.with;
		}
	}
	else v_os << slot.with;
   
	return v_os;
}


/// \fn Vistream& operator>>(Vistream& v_is, QguWLinput& slot)
/// \brief Method that reads the content of slot object (object type/QdObject) from the text stream v_is
/// \param     v_is of type Vistream&    
/// \param     slot of type QguWLinput& 
/// \exception None
/// \return    Returns Vistream& object
///
Vistream& operator>>(Vistream& v_is, QguWLinput& slot)
{
	string field_str;
	v_is >>  field_str;
   
	v_is >> slot.with;
	if (slot.with == 1){

		string obj_type;
		v_is >> obj_type;
		QdObject_var obj=QdObject::create(obj_type);
		v_is >> obj;
		Name name=field_str;
		name.delete_component(0);
		field_str=name.srep();
     
       		QwuiPutOnWL(obj,field_str.c_str());
		slot.setText(field_str.c_str());
	}
	else
		slot.setText(QString(field_str.c_str()));

	return v_is;
}

// QguLIneEdit Class

/// \fn void QguLineEdit::QguLineEdit(QWidget *parent, const char *name)
/// \brief QguLineEdit constructor
///
/// \param    parent of type QWidget*
/// \param    name of type const char*
/// \exception None
/// \return    None
///
QguLineEdit::QguLineEdit(QWidget *parent, const char *name)
	:QLineEdit(parent)
{
      // simple class to allow us to overload drop
	active=true;
	activeColor=QApplication::palette().color(QPalette::Active, QPalette::Light);
	inactiveColor=QApplication::palette().color(QPalette::Active, QPalette::Midlight);
	activeText=QApplication::palette().color(QPalette::Active, QPalette::Text);
	inactiveText=QApplication::palette().color(QPalette::Inactive, QPalette::Text);

	setAcceptDrops(true);
	setDragEnabled(true);

}

QguLineEdit::~QguLineEdit(){

}

void QguLineEdit::setActive(bool active)
{

	if(this->active!=active)
	{
		if(active)
		{
			QPalette palette(this->palette());
 			palette.setColor(this->backgroundRole(), activeColor);
 			palette.setColor(this->foregroundRole(), activeText);
 			this->setPalette(palette);
		}
		else
		{
			QPalette palette(this->palette());
 			palette.setColor(this->backgroundRole(), inactiveColor);
 			palette.setColor(this->foregroundRole(), inactiveText);
 			this->setPalette(palette);
		}
		this->active=active;
	}

}
/*
void QguLineEdit::focusInEvent(QFocusEvent* e)
{
	QLineEdit::focusInEvent(e);
	this->setFocus();
}
 */
/// \fn void QguLineEdit::dropEvent(QDropEvent *e)
/// \brief This event handler is called when the drag is dropped on this widget. It has been re-implemented from QWidget
///
/// Sets and validates dropped text
/// \param     e of type QDropEvent *
/// \exception None
/// \return    None
///
void QguLineEdit::dropEvent(QDropEvent *e)
{
	if(QuSlotDebug())
	{
		printf("QguLineEdit::dropEvent %ld \n", (long)this);
		fflush(stdout);
	}

	if( !(WLinput->active) ) return;
	
	WLinput->dropEvent(e);

}



// ------------------------------------------------------
void QguLineEdit::mouseDoubleClickEvent ( QMouseEvent * event )  
{
	this->selectAll();
}



/*
QMimeData* QguLineEdit::mimeData ( const QString slotText ) const 
{
	QMimeData *mimeData = new QMimeData();

	mimeData->setData("application/slotdata", slotText.toAscii());
	mimeData->setText(slotText);
	return mimeData;

}
*/
QStringList QguLineEdit::mimeTypes() const
{
	QStringList types;
	types << "application/slotdata";
	return types;
}

void QguLineEdit::focusInEvent(QFocusEvent *e)
{
	QLineEdit::focusInEvent(e);
	emit gotFocus();
}


