// qpui_page.cc
//
// Plot Interface Implementation - Page Object
//
// Janet Barnes
//
// July 2002

#include <stdlib.h>

#include "qpui.h"


// global to allow counting spectra colour tables
// vector <int> colourTablesUsed;

//     PageItem CLASS
//     --------------

// static members
double PageItem::_width_in_inches[6]={8.3,8.5,7.5,8.5,8.5,0.};
double PageItem::_height_in_inches[6]={11.694,9,10.,14,11,0.};

PageItem::PageItem(const QString& _name_, QpuiItem* parent, int position)
{
	if(parent==NULL)
	{
		std::cerr << "PageItem::PageItem: creating PageItem with NULL parent" << std::endl;
		return;
	}
	RootItem* rootParent=dynamic_cast<RootItem*>(parent);
	if(rootParent==NULL)
	{
		std::cerr << "PageItem::PageItem: wrong parent type" << std::endl;
		return;
	}
	if(_name_.isEmpty())
	{
		_name=QString("Page %1").arg(parent->numChildren()+1);
	}
	else
	{
		_name=_name_;
	}
	parent->addChild(this, position);
	
	
	tabWidget=parent->tabWidget;
	// Constructor
	I_Type = QP_PAGE;
	_print_id = 0;
	
	// initialise page tab widgets
	PageTab = new QpPageTab(tabWidget, "PageTab", this);
	PageTab->hide();
	
	LayoutTab = new QpLayoutTab(tabWidget, "LayoutTab", this);
	LayoutTab->hide();
		
	ValueReturnTab = new QpValueReturnTab(tabWidget, "ValueReturnTab", this);
	ValueReturnTab->hide();

    _conversion_factors[1]=1;
    _conversion_factors[2]=2.54;
    _conversion_factors[3]=25.4;
    
// 	tabWidget->adjustSize();
}

PageItem::PageItem(PageItem* page, QpuiItem* parent, int position)
{
	if(parent==NULL)
	{
		std::cerr << "PageItem::PageItem: creating PageItem with NULL parent" << std::endl;
		return;
	}
	RootItem* rootParent=dynamic_cast<RootItem*>(parent);
	if(rootParent==NULL)
	{
		std::cerr << "PageItem::PageItem: wrong parent type" << std::endl;
		return;
	}
	_children.clear();
	setName(page->_name);
	
	parent->addChild(this, position);
	
	tabWidget=parent->tabWidget;
	// Constructor
	I_Type = QP_PAGE;
	_print_id = page->_print_id;
	
	//_filename=page->_filename;
	
	// initialise page tab widgets
	PageTab = new QpPageTab(page->PageTab, this);
	PageTab->hide();
	

	LayoutTab = new QpLayoutTab(page->LayoutTab, this);
	LayoutTab->hide();
		
	ValueReturnTab = new QpValueReturnTab(page->ValueReturnTab, this);
	ValueReturnTab->hide();
	
	foreach(QpuiItem* item, page->_children)
	{
		FrameItem* frameItem=dynamic_cast<FrameItem*>(item);
		if(frameItem!=NULL)
		{
			new FrameItem(frameItem, this);
		}
	}
    _conversion_factors[1]=1;
    _conversion_factors[2]=2.54;
    _conversion_factors[3]=25.4;
}

PageItem::~PageItem()
{
	delete ValueReturnTab;
	delete PageTab;
	delete LayoutTab;
}

bool PageItem::setName(const QString& newName)
{
	QpuiItem::setName(newName);
	for(QList<PlotWindow*>::iterator i=PlotWindows.begin(); i!=PlotWindows.end(); ++i)
	{
#ifdef Q_OS_MACX
		(*i)->setTitle(QString("(double click on plot to launch interval editor)                 QSAS Plot:             %1   Window #%2       ").arg(_name).arg((*i)->getId()));
#else
		(*i)->setTitle(QString("(click on plot to launch interval editor)                 QSAS Plot:             %1   Window #%2       ").arg(_name).arg((*i)->getId()));
#endif

	}
    return true;
}

bool PageItem::addChild(QpuiItem* newChild, int position)
{
	if(newChild==NULL || _children.indexOf(newChild)>=0) return false;
	
	FrameItem* frameChild=dynamic_cast<FrameItem*>(newChild);
	PanelItem* panelChild=dynamic_cast<PanelItem*>(newChild);
	PlotItem* plotChild=dynamic_cast<PlotItem*>(newChild);
	
	if(frameChild!=NULL)
	{
		QString childName=newChild->_name;
		while(getChildByName(childName))
		{
			childName.append("A");
		}
		
		newChild->_parent=this;
		if(position<0 || position>=_children.size())
		{
			_children.push_back(newChild);
		}
		else
		{
			_children.insert(position, newChild);
		}
		newChild->_name=childName;
		newChild->refreshPath();
	}
	else if(panelChild!=NULL || plotChild!=NULL)
	{
		FrameItem* newFrame=new FrameItem("", this, QP_TS, position);
		newFrame->addChild(newChild);
	}
	else return false;
	
	return true;
}


bool PageItem::reparent(QpuiItem* newParent, int position)
{
	if(isParentOf(newParent)) return false;

	RootItem* rootParent=dynamic_cast<RootItem*>(newParent);
	if(rootParent==NULL) return false;
	
	if(_parent!=NULL)
	{
		_parent->removeChild(this, false);
	}
	if(rootParent!=NULL)
	{
		rootParent->addChild(this, position);
	}
	
	return true;
}


// ----------------------------------

void PageItem::showTabs()
{
	// ensure all tabs are hidden
	int Ntabs = tabWidget->count();
	for(int i = Ntabs-1; i >= 0; i--) 
	{
		if(tabWidget->widget(i))
		{
			tabWidget->widget(i)->hide();
			tabWidget->removeTab(i);
		}
	}

	tabWidget->addTab(PageTab, "Page");
	tabWidget->addTab(LayoutTab, "Layout");
	tabWidget->addTab(ValueReturnTab, "Value Return");

 	if(this->root()->I_PlotView->I_LastTab == QP_LAYOUT_TAB)  tabWidget->setCurrentIndex(tabWidget->indexOf(LayoutTab)); 
 	else	tabWidget->setCurrentIndex(tabWidget->indexOf(PageTab));

}

// ----------------------------------

void PageItem::setDefaultOptions()
{
	
	PageTab->_page_size_fld->setCurrentIndex(QPUI_A4);
	PageTab->_units_fld->setCurrentIndex(QPUI_PIXELS);
	PageTab->_orientation_fld->setCurrentIndex(QPUI_PORTRAIT);
	PageTab->_show_ident_fld->setChecked(true);
	PageTab->_xwin_num_fld->setText("-1");
		
	ValueReturnTab->_xaxis_rescale_fld->setChecked(true);
	ValueReturnTab->_xaxis_swap_fld->setChecked(true);
	ValueReturnTab->_yaxis_rescale_fld->setChecked(true);
	ValueReturnTab->_yaxis_swap_fld->setChecked(true);
	ValueReturnTab->_replot_fld->setChecked(true);

	LayoutTab->_autolayout_fld->setChecked(true);
	LayoutTab->_hgap_fld->setValue(QPUI_PAGE_HGAP);
	LayoutTab->_vgap_fld->setValue(QPUI_PAGE_VGAP);
	LayoutTab->_lmargin_fld->setValue(QPUI_PAGE_LMARGIN);
	LayoutTab->_rmargin_fld->setValue(QPUI_PAGE_RMARGIN);
	LayoutTab->_tmargin_fld->setValue(QPUI_PAGE_TMARGIN);
	LayoutTab->_bmargin_fld->setValue(QPUI_PAGE_BMARGIN);
}

double PageItem::aspectRatio() const
{
    double page_width=PageTab->_width_fld->text().toDouble();
    double page_height=PageTab->_height_fld->text().toDouble();
    double page_ratio=sqrt(2.);

    if(page_width>0. && page_height>0.)
    {
        page_ratio=page_height/page_width;
    }
    if(PageTab->_orientation_fld->currentIndex()==QPUI_PORTRAIT)
    {
        page_ratio=1./page_ratio;
    }

    return page_ratio;
}

// ----------------------------------

PlotWindow* PageItem::newWindow(double width, double height, int* id, int resolution, QPrinter::PageSize pageSize, QPrinter::Orientation orientation)
{
		// find free id
	
	bool found_id=false;
	bool this_one_ok;
	*id=1;
	while(!found_id)
	{
		this_one_ok=true;
		for(QList<PlotWindow*>::iterator i=PlotWindows.begin(); i!=PlotWindows.end() && this_one_ok; ++i)
		{
			this_one_ok=this_one_ok && ((*i)->getId()!=*id);
		}
		if(this_one_ok)
		{
			found_id=true;
		}
		else
		{
			(*id)++;
		}
	}

	PlotWindow* res=new PlotWindow(*id, (width<1000?(int)width: 1000), (height<700?(int)height:700), width, height, resolution, pageSize, orientation, NULL);

	PlotWindows.append(res);

#ifdef Q_OS_MACX
	res->setTitle(QString("(double click on plot to launch interval editor)                 QSAS Plot:             %1   Window #%2       ").arg(getName()).arg(*id));
#else
	res->setTitle(QString("(click on plot to launch interval editor)                 QSAS Plot:              %1   Window #%2       ").arg(getName()).arg(*id));
#endif
	res->show();
	return res;
}


// ----------------------------------

void PageItem::setSmooth(int id, bool smooth)
{
	PlotWindow* _plotWindow = getWindow(id);
	if(_plotWindow == 0)  return;

	_plotWindow->setSmooth(smooth);
	
}

// ----------------------------------

PlotWindow* PageItem::getWindow(int id)
{
	for(QList<PlotWindow*>::iterator i=PlotWindows.begin(); i!=PlotWindows.end(); ++i)
	{
		if((*i)->getId()==id)
		{
			(*i)->show();
			(*i)->setPLFocus();
			(*i)->activateWindow();
			(*i)->raise();
			return (*i);
		}
	}
	
	// If reached this point, the window does not exist.
	return 0;
}

// ----------------------------------

PlotWindow* PageItem::focusWindow(int id, double width, double height, int resolution, QPrinter::PageSize pageSize, QPrinter::Orientation orientation)
{
	for(QList<PlotWindow*>::iterator i=PlotWindows.begin(); i!=PlotWindows.end(); ++i)
	{
		if((*i)->getId()==id)
		{
			(*i)->show();
			if(width>0. && height > 0.)
			{
				(*i)->setSize(width, height);
				(*i)->resolution=resolution;
				(*i)->pageSize=pageSize;
				(*i)->orientation=orientation;
			}
			(*i)->setPLFocus();
			(*i)->activateWindow();
			(*i)->raise();
			return (*i);
		}
	}
	
	// If reached this point, the window does not exist.
	// Force creation of Window with this id
	PlotWindow* newPlotWin = new PlotWindow(id, (width<1000?(int)width: 1000), (height<700?(int)height:700), width, height, resolution, pageSize, orientation, NULL);
	PlotWindows.append(newPlotWin);

#ifdef Q_OS_MACX
	newPlotWin->setTitle(QString("(double click on plot to launch interval editor)                 QSAS Plot:             %1   Window #%2       ").arg(getName()).arg(id));
#else
	newPlotWin->setTitle(QString("(click on plot to launch interval editor)                 QSAS Plot:             %1   Window #%2       ").arg(getName()).arg(id));
#endif
	newPlotWin->show();

	return newPlotWin;
}

void PageItem::clearWindows()
{
	for(QList<PlotWindow*>::iterator i=PlotWindows.begin(); i!=PlotWindows.end(); ++i)
	{
		PlotWindow * temp=*i;
		temp->close();
		delete temp;
	}
	PlotWindows.clear();
}

// ----------------------------------

QpuiReturnStatus PageItem::plot(PlotWindow* win)
{
	if(QuSlotDebug())
	{
		printf("QpPageItem::plot\n");
		fflush(stdout);
	}

	if(!win) return QPUI_FAILURE;
	
	win->setPLFocus(); // just to be safe
	
	static PLFLT default_xwidth=0.;
	static PLFLT default_xheight=0.;
	
	plspause(0);
	
	// To make sure the new width will be validated:
	// plwid only stores changes in width
	// If no change from the previous plot, the default
	// width is not stored in the device buffer.
	// Thus we force a change, to store it.
	plwid(QtWidth2PlWidth(10));
	
	pladv(0);
	plvpor(0., 1., 0., 1.);

	// set background and default colour
	plscol0(0,255,255,255); // background temporarily set to white
	plscol0(1,0,0,0); // default set to black
	
	plwid(QtWidth2PlWidth(1));
	
	// clear colour table 
	
	getPGPColourIndex(QColor(0,0,0), true);
	
	win->setSmooth(true);
	
	// now handle each frame within this page
	
	if(this->hasChildren())
	{
		FrameItem* frameItem;
		for(int i=0; i<this->numChildren(); ++i)
		{
			frameItem = dynamic_cast<FrameItem*>(this->getChildByNumber(i));
			if(QPUI_FAILURE == frameItem->plot(win))
			{
				QuMsgBar(("Problem plotting frame "+frameItem->getName()).toLatin1().data(),true);
			}
		}
	}

	if(PageTab->_show_ident_fld->isChecked()) print_iden();
	
	qApp->processEvents();
	
	return QPUI_SUCCESS; 
}

// ---------------

void PageItem::set_colour_table(const QpuiColourTable& colour_table)
{
//     PLFLT *h, *s, *l, *i;
//     QString str;
//     QStringList sl, sl2;
//     int npoints;
//     
//     switch(colour_table)
//     {
// 
//         case QPUI_RAINBOW_EXT:
//             h=new PLFLT[4]; h[0]=300.; h[1]=300.; h[2]=-10.; h[3]=-10.;
//             s=new PLFLT[4]; s[0]=1.; s[1]=1.; s[2]=1.; s[3]=1.;
//             l=new PLFLT[4]; l[0]=1.; l[1]=.5; l[2]=.5; l[3]=0.;
//             
//             i=new PLFLT[4]; i[0]=0.; i[1]=0.1; i[2]=0.9; i[3]=1.;
//             
//             plscmap1l(0, 4, i, h, l, s, NULL);
//             
//             delete[] h;
//             delete[] s;
//             delete[] l;
//             break;
//             
//         case QPUI_GRAY_WHITE_HIGH:
//             h=new PLFLT[2]; h[0]=0.; h[1]=0.;
//             s=new PLFLT[2]; s[0]=0.; s[1]=0.;
//             l=new PLFLT[2]; l[0]=0.; l[1]=1.;
//             
//             i=new PLFLT[2]; i[0]=0.; i[1]=1.;
//             
//             plscmap1l(0, 2, i, h, l, s, NULL);
//             
//             delete[] h;
//             delete[] s;
//             delete[] l;
//             delete[] i;
//             break;
//             
//         case QPUI_GRAY_WHITE_LOW:
//             h=new PLFLT[2]; h[0]=0.; h[1]=0.;
//             s=new PLFLT[2]; s[0]=0.; s[1]=0.;
//             l=new PLFLT[2]; l[0]=1.; l[1]=0.;
//             
//             i=new PLFLT[2]; i[0]=0.; i[1]=1.;
//             
//             plscmap1l(0, 2, i, h, l, s, NULL);
//             
//             delete[] h;
//             delete[] s;
//             delete[] l;
//             delete[] i;
//             break;
//             
//         case QPUI_INV_RED_BLUE:
//             h=new PLFLT[4]; h[0]=230.; h[1]=230.; h[2]=-10.; h[3]=-10.;
//             s=new PLFLT[4]; s[0]=1.; s[1]=1.; s[2]=1.; s[3]=1.;
//             l=new PLFLT[4]; l[0]=.9; l[1]=.2; l[2]=.2; l[3]=.9;
//             
//             i=new PLFLT[4]; i[0]=0.; i[1]=0.499; i[2]=0.5; i[3]=1.;
//             
//             plscmap1l(0, 4, i, h, l, s, NULL);
//             
//             delete[] h;
//             delete[] s;
//             delete[] l;
//             delete[] i;
// 
//             break;
//             
//         case QPUI_RED_BLUE:
//             h=new PLFLT[4]; h[0]=230.; h[1]=230.; h[2]=-10.; h[3]=-10.;
//             s=new PLFLT[4]; s[0]=1.; s[1]=1.; s[2]=1.; s[3]=1.;
//             l=new PLFLT[4]; l[0]=.25; l[1]=.9; l[2]=.9; l[3]=.25;
//             
//             i=new PLFLT[4]; i[0]=0.; i[1]=0.499; i[2]=0.5; i[3]=1.;
//             
//             plscmap1l(0, 4, i, h, l, s, NULL);
//             
//             delete[] h;
//             delete[] s;
//             delete[] l;
//             delete[] i;
//             
//             break;
//             
//         case QPUI_EARTH:
//             h=new PLFLT[2]; h[0]=50.; h[1]=50.;
//             s=new PLFLT[2]; s[0]=0.2; s[1]=1.;
//             l=new PLFLT[2]; l[0]=0.9; l[1]=0.;
//             
//             i=new PLFLT[2]; i[0]=0.; i[1]=1.;
//             
//             plscmap1l(0, 2, i, h, l, s, NULL);
//             
//             delete[] h;
//             delete[] s;
//             delete[] l;
//             delete[] i;
//             break;
// 	    
// 	    case QPUI_USER_TABLE:
// 		    str=QSAS_profiles->current()->ColourTable;
// 		    str=str.trimmed();
// 		    bool isRGB;
// 		    if(str.isEmpty())
// 		    {
// 			    set_colour_table(QPUI_RAINBOW);
// 			    return;
// 		    }
// 
// 		    sl=str.split(QRegExp("\\(|\\)"), QString::SkipEmptyParts);
// 		    if(sl.size()<3)
// 		    {
// 			    set_colour_table(QPUI_RAINBOW);
// 			    return;
// 		    }
// 		    if(sl[0].contains("RGB", Qt::CaseInsensitive))
// 		    {
// 			    isRGB=true;
// 		    }
// 		    else if(sl[0].contains("HSV", Qt::CaseInsensitive))
// 		    {
// 			    isRGB=false;
// 		    }
// 
// 		    else
// 		    {
// 			    set_colour_table(QPUI_RAINBOW);
// 			    return;
// 		    }
// 		    npoints=sl.size()-1;
// 		    h=new PLFLT[npoints];
// 		    s=new PLFLT[npoints];
// 		    l=new PLFLT[npoints];
// 		    i=new PLFLT[npoints];
// 		    for(int index=0; index<npoints; ++index)
// 		    {
// 			    sl2=(sl[index+1]).split(",",  QString::SkipEmptyParts, Qt::CaseInsensitive);
// 
// 			    if(sl2.size()!=4)
// 			    {
// 				h[index]=0.; s[index]=0.; l[index]=0.; i[index]=0.;
// 			    }
// 			    else
// 			    {
// 				i[index]=sl2[0].toDouble();
// 				h[index]=sl2[1].toDouble();
// 				if(isRGB)
// 				{
// 					l[index]=sl2[2].toDouble();
// 					s[index]=sl2[3].toDouble();
// 				}
// 				else
// 				{
// 					s[index]=sl2[2].toDouble();
// 					l[index]=sl2[3].toDouble();
// 				}
// 			    }
// 		    }
// 		    plscmap1l(isRGB, npoints, i, h, l, s, NULL);
//             
// 		    delete[] h;
// 		    delete[] s;
// 		    delete[] l;
// 		    delete[] i;
// 		    break;
// 	    
// 	case QPUI_RAINBOW:
// 	    default:
// 		    h=new PLFLT[2]; h[0]=300.; h[1]=-10.;
// 		    s=new PLFLT[2]; s[0]=1.; s[1]=1.;
// 		    l=new PLFLT[2]; l[0]=.5; l[1]=.5;
//             
// 		    i=new PLFLT[2]; i[0]=0.; i[1]=1.;
//             
// 		    plscmap1l(0, 2, i, h, l, s, NULL);
//             
// 		    delete[] h;
// 		    delete[] s;
// 		    delete[] l;
// 		    delete[] i;
//             
// 		    break;
//     }
}

// -----------------


void PageItem::updateIsotropicFrames()
{
    // update the sizes of the frames with isotropic or square ratios
    if(this->hasChildren())
    {
        FrameItem* frameItem;
        for(int i=0; i<this->numChildren(); ++i)
        {
           frameItem=dynamic_cast<FrameItem*>(this->getChildByNumber(i));

            double page_ratio=aspectRatio();
            
            if(frameItem->PropertiesTab->_aspect_ratio_fld->currentIndex()==QPUI_SQUARE_ASPECT)
            {
                frameItem->PropertiesTab->_height_fld->setValue(frameItem->PropertiesTab->_width_fld->value()*page_ratio);
            } 
            else if(frameItem->PropertiesTab->_aspect_ratio_fld->currentIndex()==QPUI_ISOTROPIC)
            {
				/* double aspectRatio = frameItem->PropertiesTab->setIsotropicRatio();
				double newWidth = frameItem->PropertiesTab->_height_fld->value()/aspectRatio;
				double l_offset = frameItem->PropertiesTab->_loffset_fld->value();
				if (newWidth + l_offset > 0.99)
				{
					newWidth = 0.99 - l_offset;
					double newHeight = newWidth * aspectRatio;
					frameItem->PropertiesTab->_height_fld->setValue(newHeight);
				}
				frameItem->PropertiesTab->_width_fld->setValue(newWidth);
				*/
				frameItem->updateIsotropic();
            }
        }
    }
}

void PageItem::setLayout()
{
	// does the auto-layout for a page. Ignored for User layout.
	// update layout for page
	if(QuSlotDebug())
	{
		printf("PageItem::setLayout %ld\n", (long)this);
		fflush(stdout);
	}

	if( !(LayoutTab->_autolayout_fld->isChecked()) ) return;
	
	double _tmargin = LayoutTab->_tmargin_fld->value();
	double _lmargin = LayoutTab->_lmargin_fld->value();
	double _rmargin = LayoutTab->_rmargin_fld->value();
	double _bmargin = LayoutTab->_bmargin_fld->value();
	double _hgap = LayoutTab->_hgap_fld->value();
	double _vgap = LayoutTab->_vgap_fld->value();
	
	// calculate number of rows and columns
	int num_rows = LayoutTab->_row_num_fld->value();
	int num_cols = LayoutTab->_col_num_fld->value();
	// get widths of panels 
	float frameWidth = 1. - (_lmargin + _rmargin + _hgap*(num_cols - 1));
	frameWidth /= num_cols;
	
	int cnt = 0;
	int col = 0;
	int row = 0;
	float height = 0.;
	float width = 0.;
	
	// calculate left offest and top offset for frames in page
	
	double frameLoffset = _lmargin;
	double frameToffset = _tmargin;

	if(this->hasChildren())
	{
		FrameItem* frameItem;
		for(int i=0; i<this->numChildren(); ++i)
		{
			frameItem=dynamic_cast<FrameItem*>(this->getChildByNumber(i));
			// fetch number of panels in this frame
			int numPanels = frameItem->numberInStack();
			numPanels = (numPanels>1)?numPanels:1;
			
			// calculate height of this frame so that all panels are equal height
			float frameHeight = 1. - (_tmargin + _bmargin + _vgap*(num_rows - 1));
			frameHeight /= panelsInColumn(col);
			frameHeight *= numPanels;
			
			// set panel loffset, width, toffset, height
			frameItem->setPanelProperties(frameLoffset, frameWidth, frameToffset, frameHeight);
			
			int rowLast = row;
			int colLast = col;
			cnt++;
			col = (int)cnt  / num_rows;
			row = cnt - col * num_rows;
			
			// add widths of stack panels to loffset
			if(col > colLast) frameLoffset += frameWidth + _hgap;
			else if(col == 0)  frameLoffset = _lmargin; // reset for next row
			
			// add height of stack panel to toffset
			if(row > rowLast) frameToffset += frameHeight + _vgap;
			else if(row ==0) frameToffset = _tmargin;
		}
	}
}

// -----------------------------------
	
int PageItem::panelsInColumn(int colno)
{
	// number of panels for all frames in column
	if(QuSlotDebug())
	{
		printf("QpPageItem::panelsInColumn %ld\n", (long)this);
		fflush(stdout);
	}

  	int cnt = 0;
	int row = 0;
	int col = 0;
	int numPanels = 0;
	int num_rows =  LayoutTab->_row_num_fld->value();
	
	if(this->hasChildren())
	{
		FrameItem* frameItem;
		for(int i=0; i<this->numChildren(); ++i)
		{
			frameItem = dynamic_cast<FrameItem*>(this->getChildByNumber(i));
			col = (int)cnt  / num_rows;
			row = cnt - col * num_rows;
			cnt++;
			if(colno == col) 
			{	
				// increment by number of panels in this frame
				numPanels += frameItem->numChildren();
			}
		}
	}
	
	return (numPanels>0)?numPanels:1;
}


// -----------------------------------
	
void PageItem::setFramesAuto(bool active)
{
	if(this->hasChildren())
	{
		FrameItem* frameItem;
		for(int i=0; i<this->numChildren(); ++i)
		{
			frameItem=dynamic_cast<FrameItem*>(this->getChildByNumber(i));
			
			if(active)
			{
				LayoutTab->updateRowCol(this->numChildren());
				frameItem->PropertiesTab->_autolayout_lbl->setText("Autoset by Page Layout ");
			}
			else
			{
				frameItem->PropertiesTab->_autolayout_lbl->setText(" Page Layout not Autoset");
			}
			frameItem->PropertiesTab->_loffset_fld->setDisabled(active);
			frameItem->PropertiesTab->_toffset_fld->setDisabled(active);
			frameItem->PropertiesTab->_width_fld->setDisabled(active);
			frameItem->PropertiesTab->_height_fld->setDisabled(active);
			frameItem->PropertiesTab->_aspect_ratio_fld->setDisabled(active);

		}
	}
	
	LayoutTab->_hgap_fld->setEnabled(active);
	LayoutTab->_vgap_fld->setEnabled(active);
	LayoutTab->_lmargin_fld->setEnabled(active);
	LayoutTab->_rmargin_fld->setEnabled(active);
	LayoutTab->_tmargin_fld->setEnabled(active);
	LayoutTab->_bmargin_fld->setEnabled(active);
}
// -------------------------------

bool PageItem::containsSpectrum()
{
	if(this->hasChildren())
	{
		FrameItem* frameItem;
		for(int i=0; i<this->numChildren(); ++i)
		{
			frameItem = dynamic_cast<FrameItem*>(this->getChildByNumber(i));
			if(frameItem->containsSpectrum()) return true;
		}
	}
	return false;
}

// -------------------------------

void PageItem::displayIntervalWidget(double x, double y, int pos_x, int pos_y)
{
	PlotWindow* win=(PlotWindow *) sender();
	if(!win) return;
	
	win->setPLFocus(); // ensure we are plotting to and reading from same window
	
	// Search if clicked inside a frame
	FrameItem * selectedFrame=0;
	int nPanel=0;
	if(this->hasChildren())
	{
		FrameItem* frameItem;
		for(int i=0; i<this->numChildren() && !selectedFrame; ++i)
		{
			frameItem=dynamic_cast<FrameItem*>(this->getChildByNumber(i));
			
			int numPanels = frameItem->numChildren();
			bool hasSpectrum = containsSpectrum();
			double factor=(hasSpectrum?0.9:1.);
			if(frameItem->PropertiesTab->_loffset_fld->value() <= x &&  
						x <= frameItem->PropertiesTab->_loffset_fld->value() + frameItem->PropertiesTab->_width_fld->value()*factor && 
						frameItem->PropertiesTab->_toffset_fld->value()<= y &&
						y <= frameItem->PropertiesTab->_toffset_fld->value()+ frameItem->PropertiesTab->_height_fld->value())
			{
				selectedFrame=frameItem;
			}
		}
	}

	if(!selectedFrame)
	{
		// No selected frame, return
		return;
	}
	
	if(selectedFrame->frameType()==QP_TS)
	{	
		double tstart = selectedFrame->tAxisTab->_axis_time_fld->getStart().getEpoch2000();
		double tend = selectedFrame->tAxisTab->_axis_time_fld->getEnd().getEpoch2000();
		
		// look for min and max for all the objects in the frame
		
		double objtmin=tstart;
		double objtmax=tend;

		// now set up widget

		win->disconnect(SIGNAL(rescaleScalar(double, double)));
		win->disconnect(SIGNAL(rescaleTime(const QmjdTimeInterval&)));
		
		connect(win, SIGNAL(rescaleTime(const QmjdTimeInterval&)), selectedFrame->tAxisTab, SLOT(interactiveRescale(const QmjdTimeInterval&)));
		
		win->intervalDialog->setWindowTitle(selectedFrame->TitleTab->_title_fld->text());

		win->intervalDialog->move(win->x()+win->plot->x()+pos_x-win->intervalDialog->width()/2,win->y()+win->plot->y()+pos_y);

		win->intervalDialog->setVisible(true);
		win->intervalDialog->resize(win->intervalDialog->minimumSizeHint());
		win->intervalDialog->raise();
		
		double fromMeta, toMeta;
		bool foundMeta = selectedFrame->getXmetaRange(fromMeta, toMeta);
		QmjdTime from, to;
		from=selectedFrame->tAxisTab->_axis_time_fld->getStart();
		to=selectedFrame->tAxisTab->_axis_time_fld->getEnd();
		if(foundMeta/* && found*/)
		{
			win->intervalDialog->setViewRange(new QdTimeInterval(QdTime(fromMeta), QdTime(toMeta)));
			win->intervalDialog->setAsTimeInterval(from, to, selectedFrame);
			win->intervalDialog->setAvailabilityRange(fromMeta, toMeta);
		}
	}
	else if(selectedFrame->frameType()==QP_XY)
	{
		double xstart = selectedFrame->xAxisTab->_axis_xy_fld->getStart();
		double xend = selectedFrame->xAxisTab->_axis_xy_fld->getEnd();
		// look for min and max for all the objects in the frame
		

		// now set up widget

		win->disconnect(SIGNAL(rescaleScalar(double, double)));
		win->disconnect(SIGNAL(rescaleTime(const QmjdTimeInterval&)));

		connect(win, SIGNAL(rescaleScalar(double, double)), selectedFrame->xAxisTab, SLOT(interactiveRescale(double, double)));
		
		win->intervalDialog->setWindowTitle(selectedFrame->TitleTab->_title_fld->text());
		win->intervalDialog->move(win->x()+win->plot->x()+pos_x-win->intervalDialog->width()/2,win->y()+win->plot->y()+pos_y);
		
		win->intervalDialog->setVisible(true);
		win->intervalDialog->resize(win->intervalDialog->minimumSizeHint());
		win->intervalDialog->raise();
		
		double fromMeta, toMeta;
		bool foundMeta = selectedFrame->getXmetaRange(fromMeta, toMeta);
		double from, to;
		from=selectedFrame->xAxisTab->_axis_xy_fld->getStart();
		to=selectedFrame->xAxisTab->_axis_xy_fld->getEnd();
		if(foundMeta)
		{
			win->intervalDialog->setViewRange(fromMeta, toMeta);
			win->intervalDialog->setAsXInterval(from, to);
			win->intervalDialog->setAvailabilityRange(fromMeta, toMeta);
		}
	}
}

// -------------------------------

QDomElement PageItem::toXMLElement(QDomDocument& doc) const
{
    QDomElement xml=doc.createElement("Page_Item");
    xml.appendChild(createTextXMLNode(doc, "Name", getName()));
//     xml.appendChild(createTextXMLNode(doc, "File_Name", _filename));
    xml.appendChild(PageTab->toXMLElement(doc));
    xml.appendChild(LayoutTab->toXMLElement(doc));

    int cnt = numChildren();
    xml.setAttribute("size", QString::number(cnt));
   
    FrameItem* frameItem;
    for(int i=0; i<cnt; ++i)
    {
        frameItem=dynamic_cast<FrameItem*>(getChildByNumber(i));
        xml.appendChild(frameItem->toXMLElement(doc));
    }
    return xml;
}

void PageItem::fromXMLElement(QDomElement& el)
{
    QString s;
    
    if(getXMLNodeText(el, "Name", s))
    {
        setName(s);
    }

    QDomElement page_tab=el.firstChildElement("Page_Tab");
    if(!page_tab.isNull()) PageTab->fromXMLElement(page_tab);

    QDomElement layout_tab=el.firstChildElement("Layout_Tab");
    if(!layout_tab.isNull()) LayoutTab->fromXMLElement(layout_tab);

    int cnt=el.attribute("size", "0").toInt();

    QDomElement frameElement=el.firstChildElement("Frame_Item");

    while(!frameElement.isNull())
    {
        FrameItem *frameItem = new FrameItem("", this);
        frameItem->fromXMLElement(frameElement);

        frameElement=frameElement.nextSiblingElement("Frame_Item");
    }

    if(getXMLNodeText(layout_tab, "Num_Rows", s))
    {
        LayoutTab->_row_num_fld->setValue(s.toInt());
    }
    if(getXMLNodeText(layout_tab, "Num_Cols", s))
    {
        LayoutTab->_col_num_fld->setValue(s.toInt());
    }
	
	setFramesAuto(LayoutTab->_autolayout_fld->isChecked());
}

Vistream& operator>>(Vistream& v_is, PageItem& page)
{
	if(QuSlotDebug()) cout <<  "PageItem restore (>>)" << endl;

	// streams from virtual input stream v_is into item
	QpuiItem& plot_item = page;
	v_is >> plot_item;

	// read in page specifics
	
	int _page_size;
	v_is >> _page_size;
	page.PageTab->_page_size_fld->setCurrentIndex(_page_size);
	
	double _width;
	v_is >> _width;
	double _height;
	v_is >> _height;
	int _units;
	v_is >> _units;
	page.PageTab->_units_fld->setCurrentIndex(_units);
	page.PageTab->oldUnits = _units;
	
	bool readonly = (_page_size != QPUI_CUSTOM);
	page.PageTab->_width_fld->setReadOnly(readonly);
	page.PageTab->_height_fld->setReadOnly(readonly);
	page.PageTab->_width_fld->setEnabled(!readonly);
	page.PageTab->_height_fld->setEnabled(!readonly);
	
	int _dpi;
	v_is >> _dpi;

	page.PageTab->_dpi_fld->setCurrentIndex(_dpi);

	switch(_dpi)
	{
		case QPUI_150_DPI:
			page._conversion_factors[QPUI_PIXELS]=150.;
			page._conversion_factors[QPUI_PSPOINTS]=150.;
			break;
			
		case QPUI_300_DPI:
			page._conversion_factors[QPUI_PIXELS]=300.;
			page._conversion_factors[QPUI_PSPOINTS]=300.;
			break;
			
		case QPUI_600_DPI:
			page._conversion_factors[QPUI_PIXELS]=600.;
			page._conversion_factors[QPUI_PSPOINTS]=600.;
			break;
			
		case QPUI_1200_DPI:
			page._conversion_factors[QPUI_PIXELS]=1200.;
			page._conversion_factors[QPUI_PSPOINTS]=1200.;
			break;
			
		default:
			page._conversion_factors[QPUI_PIXELS]=72.;
			page._conversion_factors[QPUI_PSPOINTS]=72.;
			break;
	}

	double conv = page._conversion_factors[_units];
	_width *= conv;
	page.PageTab->_width_fld->setText(double_to_string(_width).c_str());
	_height *= conv;
	page.PageTab->_height_fld->setText(double_to_string(_height).c_str());
	
	int _orientation;
	v_is >> _orientation;
	page.PageTab->_orientation_fld->setCurrentIndex(_orientation);

	int _xwin_num;
	v_is >> _xwin_num;
	page.PageTab->_xwin_num_fld->setText(int_to_string(_xwin_num).c_str());
	
	string temp_name;
	v_is >> temp_name;
	page.setName(QString(temp_name.c_str()));

    std::string _filename;//for backward compatibility
	v_is >> _filename;
	bool _show_ident;
	v_is >> _show_ident;
	page.PageTab->_show_ident_fld->setChecked(_show_ident);

	int _row_num;
	v_is >> _row_num; // set after frames read in
	
	int _col_num;
	v_is >> _col_num; // set after frames read in
	
	bool _autolayout;
	v_is >> _autolayout;
	page.LayoutTab->_autolayout_fld->setChecked(_autolayout);

	float _hgap;
	v_is >> _hgap;
	page.LayoutTab->_hgap_fld->setValue(_hgap);

	float _vgap;
	v_is >> _vgap;
	page.LayoutTab->_vgap_fld->setValue(_vgap);

	float _lmargin;
	v_is >> _lmargin;
	page.LayoutTab->_lmargin_fld->setValue(_lmargin);
	
	float _rmargin;
	v_is >> _rmargin;
	page.LayoutTab->_rmargin_fld->setValue(_rmargin);
	
	float _tmargin;
	v_is >> _tmargin;
	page.LayoutTab->_tmargin_fld->setValue(_tmargin);
	
	float _bmargin;
	v_is >> _bmargin;
	page.LayoutTab->_bmargin_fld->setValue(_bmargin);


	// read in frames
	int cnt;
	v_is >> cnt;

	for(int i=0; i<cnt; i++)
	{
		FrameItem *frameItem = new FrameItem("", &page);
		v_is >> *frameItem;
	}
	
	page.LayoutTab->_row_num_fld->setValue(_row_num);
	page.LayoutTab->_col_num_fld->setValue(_col_num);
	page.setFramesAuto(_autolayout);
	
	return v_is;
}

// -----------------------------------------------

Vostream& operator<<(Vostream& v_os, PageItem& page)
{
	if(QuSlotDebug()) cout << "PageItem save (<<)" << endl;

	// streams item to virtual output stream v_os

	QpuiItem& plot_item = page;
// 	v_os << plot_item;
	v_os=save(v_os, plot_item);
	
	v_os << ListTypeGroup;
	v_os << page.I_Type;

	// page Details

	int _units = page.PageTab->_units_fld->currentIndex();
	double conv = page._conversion_factors[_units];
	
	v_os << (int) page.PageTab->_page_size_fld->currentIndex();
	
	double _width = string_to_double( page.PageTab->_width_fld->text().toLatin1().data() );
	_width /= conv;
	v_os << _width;
	
	double _height = string_to_double( page.PageTab->_height_fld->text().toLatin1().data() );
	_height /= conv;
	v_os << _height;
	v_os << _units;
	
	v_os << (int) page.PageTab->_dpi_fld->currentIndex();
	
	v_os << (int) page.PageTab->_orientation_fld->currentIndex();
	
	int _xwin_num = string_to_int( page.PageTab->_xwin_num_fld->text().toLatin1().data() );
	v_os << _xwin_num;
	v_os << string(page.getName().toLatin1().data());
	
	v_os << string("");//page._filename; // for backward compatibility
	
	bool _show_ident = page.PageTab->_show_ident_fld->isChecked();
	v_os << _show_ident;
	
	v_os << (int) page.LayoutTab->_row_num_fld->value();
	v_os << (int) page.LayoutTab->_col_num_fld->value();
	bool _autolayout = page.LayoutTab->_autolayout_fld->isChecked();
	v_os << _autolayout;
	v_os << (float) page.LayoutTab->_hgap_fld->value();
	v_os << (float) page.LayoutTab->_vgap_fld->value();

	v_os << (float) page.LayoutTab->_lmargin_fld->value();
	v_os << (float) page.LayoutTab->_rmargin_fld->value();
	v_os << (float) page.LayoutTab->_tmargin_fld->value();
	v_os << (float) page.LayoutTab->_bmargin_fld->value();

	// Frames in page

	int cnt = page.numChildren();
	v_os << cnt;
   
	FrameItem* frameItem;
	for(int i=0; i<cnt; ++i)
	{
		frameItem=dynamic_cast<FrameItem*>(page.getChildByNumber(i));
		v_os << *frameItem;
	}
	
	return v_os;
}


