// qpui_panel.cc
//
// Plot Interface Implementation - Panel Object
//
// Janet Barnes
// July 2002
//
//  Tony Allen
//  Feb 2008

#include "qpui.h"
#include "qsas_plplot_time_helpers.h"

extern QPalette QPUI_normal_palette;
extern int qsas_restore_release;

//     PanelItem CLASS
//     ----------------------

PanelItem::PanelItem(const QString& name, QpuiItem* parent, int position)
{
	if(parent==NULL)
	{
		std::cerr << "PanelItem::PanelItem: creating PanelItem with NULL parent" << std::endl;
		return;
	}
	RootItem* rootParent=dynamic_cast<RootItem*>(parent);
	PageItem* pageParent=dynamic_cast<PageItem*>(parent);
	FrameItem* frameParent=dynamic_cast<FrameItem*>(parent);
	if(rootParent==NULL && pageParent==NULL && frameParent==NULL)
	{
		std::cerr << "FrameItem::FrameItem: wrong parent type" << std::endl;
		return;
	}
	if(name.isEmpty())
	{
		_name=QString("Panel %1").arg(parent->numChildren()+1);
	}
	else
	{
		_name=name;
	}
	parent->addChild(this, position);
	
	tabWidget=parent->tabWidget;
	//I_Frame = dynamic_cast<FrameItem*>(_parent);
	I_Type = QP_PANEL;
	
	// initialise options tab widgets
	LegendTab = new QpLegendTab(tabWidget, "LegendTab", this);
	LegendTab->hide();
	
	yAxisTab = new QpYAxisTab(tabWidget, "yAxisTab", this);
	yAxisTab->hide();
}

PanelItem::PanelItem(PanelItem* panel, QpuiItem* parent, int position)
{
	if(parent==NULL)
	{
		std::cerr << "Error: PanelItem::PanelItem: creating PanelItem with NULL parent" << std::endl;
		return;
	}
	RootItem* rootParent=dynamic_cast<RootItem*>(parent);
	PageItem* pageParent=dynamic_cast<PageItem*>(parent);
	FrameItem* frameParent=dynamic_cast<FrameItem*>(parent);
	if(rootParent==NULL && pageParent==NULL && frameParent==NULL)
	{
		std::cerr << "FrameItem::FrameItem: wrong parent type" << std::endl;
		return;
	}
	_children.clear();
	setName(panel->_name);
	parent->addChild(this, position);
	
	tabWidget=parent->tabWidget;
	//I_Frame = dynamic_cast<FrameItem*>(_parent);
	I_Type = QP_PANEL;
	
	// initialise options tab widgets
	LegendTab = new QpLegendTab(panel->LegendTab, this);
	LegendTab->hide();
	
	yAxisTab = new QpYAxisTab(panel->yAxisTab, this);
	yAxisTab->hide();
	
	foreach(QpuiItem* item, panel->_children)
	{
		TraceItem* traceItem=dynamic_cast<TraceItem*>(item);
		EventItem* eventItem=dynamic_cast<EventItem*>(item);
		SpectrumItem* spectrumItem=dynamic_cast<SpectrumItem*>(item);
		if(traceItem!=NULL)
		{
			new TraceItem(traceItem, this);
		}
		if(eventItem!=NULL)
		{
			new EventItem(eventItem, this);
		}
		if(spectrumItem!=NULL)
		{
			new SpectrumItem(spectrumItem, this);
		}
	}

}

PanelItem::~PanelItem()
{

	delete LegendTab;
	delete yAxisTab;
}

bool PanelItem::addChild(QpuiItem* newChild, int position)
{
	if(newChild==NULL || _children.indexOf(newChild)>=0) return false;
	PlotItem* plotChild=dynamic_cast<PlotItem*>(newChild);
	
	if(plotChild!=NULL)
	{
		QString childName=newChild->_name;
		while(getChildByName(childName))
		{
			childName.append("A");
		}
		
		newChild->_parent=this;
		if(position<0 || position>=_children.size())
		{
			_children.push_back(newChild);
		}
		else
		{
			_children.insert(position, newChild);
		}
		newChild->_name=childName;
		newChild->refreshPath();
	}
	else return false;
	
	return true;
}

bool PanelItem::reparent(QpuiItem* newParent, int position)
{
	if(isParentOf(newParent)) return false;

	RootItem* rootParent=dynamic_cast<RootItem*>(newParent);
	PageItem* pageParent=dynamic_cast<PageItem*>(newParent);
	FrameItem* frameParent=dynamic_cast<FrameItem*>(newParent);
	if(rootParent==NULL && pageParent==NULL && frameParent==NULL)
	{
		bool found=false;
		// find the first FrameItem which is a parent of newParent
		while(!found && newParent->parent()!=NULL)
		{
			newParent=newParent->parent();
			frameParent=dynamic_cast<FrameItem*>(newParent);
			if(frameParent!=NULL) found=true;
		}
		if(!found) return false;
	}
	
	if(_parent!=NULL)
	{
		_parent->removeChild(this, false);
	}
	if(newParent!=NULL)
	{
		newParent->addChild(this, position);
	}
	
	return true;
}

// ----------------------------------

void PanelItem::setDefaultOptions()
{

	yAxisTab->_axis_linlog_fld->setCurrentIndex(QP_LINEAR);
	
	yAxisTab->_label_fld->setText("");
		
	yAxisTab->handleAutoScale(QP_FROM_META);
	yAxisTab->_axis_major_ticks_num_fld->setValue(0);
	yAxisTab->_axis_minor_ticks_num_fld->setValue(0);
	yAxisTab->_axis_major_tickmarks_fld->setCurrentIndex(QPUI_IN);
	yAxisTab->_axis_minor_tickmarks_fld->setCurrentIndex(QPUI_IN);
	yAxisTab->_axis_autolabel_fld->setChecked(true);
	yAxisTab->_label_char_size_fld->setValue(QSAS_profiles->current()->BaseFontSize+QPUI_LABELS_CHAR_SIZE_ADD);
	yAxisTab->_label_disp_fld->setValue(QPUI_LABELS_DISP);
	yAxisTab->_label_rotation_fld->setValue(0);
	yAxisTab->_label_font_fld->setSansFont();
	yAxisTab->_label_font_fld->setItalic(false);
	yAxisTab->_label_font_fld->setBold(false);

	QPalette line_palette = QPalette(QPUI_normal_palette);
	line_palette.setColor(QPalette::Active ,QPalette::ButtonText, Qt::black);
	line_palette.setColor(QPalette::Inactive, QPalette::ButtonText, Qt::black);
	yAxisTab->_label_colour_fld->setPalette(line_palette);
    yAxisTab->_label_colour = Qt::black;
	
	LegendTab->_show_legend_fld->setChecked(false);
	LegendTab->_legend_border_fld->setCurrentIndex(QPUI_LINE);
	LegendTab->_legend_background_fld->setCurrentIndex(QPUI_TRANSPARENT);
	LegendTab->_legend_num_rows_fld->setValue(1);
	LegendTab->_legend_num_cols_fld->setValue(1);
	LegendTab->_legend_width_fld->setValue(QPUI_LEGEND_WIDTH);
	LegendTab->_legend_height_fld->setValue(QPUI_LEGEND_HEIGHT);
	LegendTab->_legend_max_char_length_fld->setValue(15);
	LegendTab->_legend_line_length_fld->setValue(0.2);
	LegendTab->_legend_char_size_fld->setValue(0.5);
	LegendTab->_legend_just_fld->setCurrentIndex(QPUI_VCENTRE);
	LegendTab->_legend_font_fld->setSansFont();
	LegendTab->_legend_font_fld->setItalic(false);
	LegendTab->_legend_font_fld->setBold(false);
	LegendTab->_legend_align_fld->setCurrentIndex(QPUI_INSIDE_TOP_RIGHT);
	LegendTab->_legend_loffset_fld->setValue(0.75);
	LegendTab->_legend_toffset_fld->setValue(0.);
}

// -------------------------------------------

void PanelItem::showTabs()
{	

	// ensure all tabs are hidden
	int Ntabs = tabWidget->count();
	for(int i = Ntabs-1; i >= 0; i--) 
	{
		if(tabWidget->widget(i))
		{
			tabWidget->widget(i)->hide();
			tabWidget->removeTab(i);
		}
	}

	tabWidget->addTab(yAxisTab, "Y Axis");
	tabWidget->addTab(LegendTab, "Legend");
	// value return is always shown
	tabWidget->addTab(page()->ValueReturnTab, "Value Return");
	
 	if(this->root()->I_PlotView->I_LastTab == QP_LEGEND_TAB)  tabWidget->setCurrentIndex(tabWidget->indexOf(LegendTab));
 	else tabWidget->setCurrentIndex(tabWidget->indexOf(yAxisTab));
}

// -------------------------------------------

int PanelItem::getStackNumber()
{
	// get position in frame stack, 1 = first
	
	return childNumber()+1;

}

// -------------------------------------------

int PanelItem::numberTraces()
{
	// get number of plot items that are traces in this panel
	
	int cnt = 0;
	TraceItem* item;
	for(int i=0; i<this->numChildren(); ++i)
	{
		item=dynamic_cast<TraceItem*>(this->getChildByNumber(i));
		if(item && item->DataTab->_show_legend_fld->isChecked()) cnt++;
	}
	return cnt;

}

// -------------------------------------------

int PanelItem::numberSpectra()
{
	// get number of plot items that are traces in this panel
	
	int cnt = 0;
	SpectrumItem* item;
	for(int i=0; i<this->numChildren(); ++i)
	{
		item=dynamic_cast<SpectrumItem*>(this->getChildByNumber(i));
		if(item) cnt++;
	}
	return cnt;

}

int PanelItem::numberThisSpectrum(SpectrumItem *thisItem)
{
	// get number of plot items that are traces in this panel
	
	SpectrumItem* item;
	for(int i=0; i<this->numChildren(); ++i)
	{
		item=dynamic_cast<SpectrumItem*>(this->getChildByNumber(i));
		if(item == thisItem) return i;
	}
	return 0;

}

bool PanelItem::isInterlaced()
{
	SpectrumItem* item;
	for(int i=0; i<this->numChildren(); ++i)
	{
		item=dynamic_cast<SpectrumItem*>(this->getChildByNumber(i));
		if(item && item->DataTab->interlace_toggle->isChecked()) return true;
	}
	return false;

}

string PanelItem::getTitle()
{
	if(QuSlotDebug())
	{
		printf("QpPanelItem::getTitle\n");
		fflush(stdout);
	}

	string title="";
	// get panel title from plot elements
	
	PlotItem* item;
	for(int i=0; i<this->numChildren(); ++i)
	{
		item=dynamic_cast<PlotItem*>(this->getChildByNumber(i));
		string item_title;
		TraceItem* traceItem = dynamic_cast<TraceItem*>(item);
		SpectrumItem* spectrumItem = dynamic_cast<SpectrumItem*>(item);      

		if(traceItem) item_title = traceItem->getTitle();
		else if(spectrumItem) item_title = spectrumItem->getTitle();
		
		if( title != item_title) 
		{
			if( title != "" && title != " ") title += "/";
			title += item_title;
		}
	}
	
	title = SafePGPString(title.c_str());
	return title;
}

// ---------------------------------------------

void PanelItem::prePlot()
{
	// pre plots panel (i.e. called BEFORE plot component is plotted)
	
	plwid(0); // force reset

	PLFLT xmin, xmax;
	if(frameType() == QP_TS)
	{
		xmin = 0.0; // scale window to range
		xmax = frame()->tAxisTab->_axis_time_fld->getDuration();
	}
	else
	{
		xmin = frame()->xAxisTab->_axis_xy_fld->getStart();
		xmax = frame()->xAxisTab->_axis_xy_fld->getEnd();
	}
	
	if(frame()->xtAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
	{
		xmin = SafeLog10(xmin);
		xmax = SafeLog10(xmax);
	}
	
	PLFLT ymin = yAxisTab->_axis_xy_fld->getStart();
	PLFLT ymax = yAxisTab->_axis_xy_fld->getEnd();
	if(yAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
	{
		ymin = (PLFLT)SafeLog10(ymin);
		ymax = (PLFLT)SafeLog10(ymax);
	}
	
	plwind(xmin, xmax, ymin, ymax);
	
	if(LegendTab->_show_legend_fld->isChecked())
	{
		// calculate width and height of legend entry
// 		string legend_text = SafePGPString(yAxisTab->_label_fld->text().toLatin1().data());
		plschr(QPUI_CHAR_HEIGHT, LegendTab->_legend_char_size_fld->value());
// 		PLFLT xch, ych;
// 		plgchr(&xch, &ych);
		_legend_entry = 0;
// 		PLFLT loffset, roffset, toffset, boffset;
// 		PLFLT x0vp_save, x1vp_save, y0vp_save, y1vp_save;
// 		plgvpd(&x0vp_save, &x1vp_save, &y0vp_save, &y1vp_save);

	}

	plschr(QPUI_CHAR_HEIGHT, frame()->DisplayTab->_ticks_length_fld->value());
}

// ------------------------------------------------

void PanelItem::postPlot(bool show_top_axis, bool show_bottom_axis)
{
	PLINT family, style, weight;
	
	double ymin = yAxisTab->_axis_xy_fld->getStart();
	double ymax = yAxisTab->_axis_xy_fld->getEnd();
// 	if(ymin >= ymax) return;
	
	// draw frame

		// set colours, font, character size, line style and line width
	PLINT  line_colour_index = getPGPColourIndex(frame()->_line_colour);
	plcol0(line_colour_index);
	PLFLT _ticks_length = frame()->DisplayTab->_ticks_length_fld->value();
	PLFLT _ticks_label_size = frame()->DisplayTab->_ticks_char_size_fld->value();
// 		plschr(40., _ticks_length);
	pllsty(QP_SOLID);
	plwid(QtWidth2PlWidth(frame()->DisplayTab->_line_width_fld->value()));
		
	string xopt, yopt;
	string x_subopt, y_subopt;
	if(frame()->DisplayTab->_show_bottom_axis_fld->isChecked())
	{
		xopt+="b";
		x_subopt+="b";
	}
	if(frame()->DisplayTab->_show_left_axis_fld->isChecked())
	{
		yopt+="b";
		y_subopt+="b";
	}
	if(frame()->DisplayTab->_show_top_axis_fld->isChecked())
	{
		xopt+="c";
		x_subopt+="c";
	}
	if(frame()->DisplayTab->_show_right_axis_fld->isChecked())
	{
		yopt+="c";
		y_subopt+="c";
	}
	if(frame()->DisplayTab->_show_lr_alternating_axis_fld->isChecked())
	{
		if(getStackNumber()%2==1)
		{
			yopt+="b";
			y_subopt+="b";
		}
		else
		{
			yopt+="c";
			y_subopt+="c";
		}
	}
	if(frame()->DisplayTab->_show_rl_alternating_axis_fld->isChecked())
	{
		if(getStackNumber()%2==0)
		{
			yopt+="b";
			y_subopt+="b";
		}
		else
		{
			yopt+="c";
			y_subopt+="c";
		}
	}
				
	if(frame()->DisplayTab->_show_bottom_axis_fld->isChecked() || frame()->DisplayTab->_show_top_axis_fld->isChecked() || frame()->DisplayTab->_show_yiszero_fld->isChecked() )
	{
#if QT_VERSION >= 0x040300
		frame()->xtAxisTab->_label_font_fld->getFontParams(family, style, weight);
		plsfont(family, style, weight);
#else
        family=frame()->xtAxisTab->_label_font_fld->getFontIndex();
        plfont(family);
#endif
		switch(frame()->xtAxisTab->_axis_minor_tickmarks_fld->currentIndex())
		{
			case QPUI_IN:
				x_subopt+="st";
				break;
			case QPUI_OUT: default:
				x_subopt+="ist";
				break;
			case QPUI_NOTICKS:
				break;
		}
		switch(frame()->xtAxisTab->_axis_major_tickmarks_fld->currentIndex())
		{
			case QPUI_IN: default:
				xopt += "t";
				break;
			case QPUI_OUT:
				xopt += "it";
				break;
			case QPUI_GRID:
				xopt+="gt";
				break;
			case QPUI_NOTICKS:
				break;
		}

		if(show_bottom_axis && frame()->DisplayTab->_show_bottom_ticks_labels_fld->isChecked()) xopt+="n";
			
		if(show_top_axis && frame()->DisplayTab->_show_top_ticks_labels_fld->isChecked()) xopt+="m";
			
		PLFLT xmin, xmax;
		PLINT nsubx = frame()->xtAxisTab->_axis_minor_ticks_num_fld->value();
		PLFLT xtick = 0.;
		int nx=frame()->xtAxisTab->_axis_major_ticks_num_fld->value();
		char fmt_ticks[50], fmt_label[50];

        PLLabelDefaults x_label_opt; // do nothing if QP_TS. Else used to move exponential label
        
		if(frameType() == QP_TS)
		{
			PLINT auto_nsubx;
			PLFLT auto_xtick;
			char label[50];	
			QmjdTime start=frame()->tAxisTab->_axis_time_fld->getStart();
			QmjdTime end=frame()->tAxisTab->_axis_time_fld->getEnd();
// 			struct tm tm_start, tm_end;
				
			xopt += "d";
			x_subopt += "d";
				
// 			start=*(frame()->tAxisTab->_axis_time_fld->getStartTime());
// 			end=*(frame()->tAxisTab->_axis_time_fld->getEndTime());

// 			getXminXmaxTS(start, end, &xmin, &xmax, &tm_start, &tm_end);
			xmin = start.getEpoch1970();
			xmax = end.getEpoch1970();
				
			int separator=(QSAS_profiles->current()->TimeRep=="Colon"?1:0);
			computeFormat(fabs(xmax-xmin), fmt_ticks, fmt_label, &auto_xtick, &auto_nsubx, separator);
				
			ymin = yAxisTab->_axis_xy_fld->getStart();
			ymax = yAxisTab->_axis_xy_fld->getEnd();
			if(yAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
			{
				ymin = SafeLog10(ymin);
				ymax = (PLFLT)SafeLog10(ymax);
			}

			plwind(xmin, xmax, (PLFLT)ymin, (PLFLT)ymax);
				
			if(nx!=0)
			{
				xtick=(xmax-xmin)/(PLFLT)nx;
			}
			else
			{
				xtick=auto_xtick;
			}
				
			if(nsubx==0) nsubx=auto_nsubx;
				
			if(frame()->tAxisTab->tick_time_fmt_auto_fld->isChecked())
			{
				pltimefmt(fmt_ticks);
				frame()->tAxisTab->tick_time_fmt_fld->setText(QString(fmt_ticks));
			}
			else
			{
				pltimefmt(frame()->tAxisTab->tick_time_fmt_fld->text().toLatin1().data());
			}
				
			// print labels

#if QT_VERSION > 0x040300
			frame()->tAxisTab->_label_font_fld->getFontParams(family, style, weight);
			plsfont(family, style, weight);
#else
            family=frame()->tAxisTab->_label_font_fld->getFontIndex();
            plfont(family);
#endif
				

			plschr(0., _ticks_label_size);
				
			if(frame()->tAxisTab->time_labels_fmt_auto_fld->isChecked())
			{
				start.strfMJD(label,49,fmt_label);
				frame()->tAxisTab->time_labels_fmt_fld->setText(QString(fmt_label));
			}
			else
			{
				QString format=frame()->tAxisTab->time_labels_fmt_fld->text();
				start.strfMJD(label,49,format.toLatin1().data());
			}
			if(show_bottom_axis && frame()->DisplayTab->_show_bottom_ticks_labels_fld->isChecked())
				plmtex("b",3.,0.0,0.3,label);
			if(show_top_axis && frame()->DisplayTab->_show_top_ticks_labels_fld->isChecked())
				plmtex("t",3.,0.0,0.3,label);
		}
		else
		{
            x_label_opt.exp_label_disp=3.2+frame()->xAxisTab->_exp_label_offset_y_fld->value();
            x_label_opt.exp_label_pos=1.02+frame()->xAxisTab->_exp_label_offset_x_fld->value()/5.;
            x_label_opt.exp_label_just=.5;
            
			xmin = frame()->xAxisTab->_axis_xy_fld->getStart();
			xmax = frame()->xAxisTab->_axis_xy_fld->getEnd();
			if(nx>0)
			{
				xtick=(xmax-xmin)/(PLFLT)nx;
			}
		}
			
		if(frame()->xtAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG){
			xopt+="l";
			x_subopt+="l";
			xmin = SafeLog10(xmin);
			xmax = SafeLog10(xmax);
		}

        if(frameType() == QP_XY)
        {
            plslabelfunc(NULL, &x_label_opt);
        }
		// fix for plplot putting grid and major ticks in different place (suppress major tick)		
		plsmaj(0., _ticks_length);
		plschr(0., _ticks_label_size);
		plbox(xopt.c_str(), xtick, nsubx, "", 0, 0);
		plsmaj(0., 0.);
		plsmin(0., _ticks_length*0.6);
		plbox(x_subopt.c_str(), xtick, nsubx, "", 0, 0);
        plslabelfunc(NULL, NULL);
			
		// draw y=0
		if(frame()->DisplayTab->_show_yiszero_fld->isChecked() )
		{
			PLFLT yzero_x[2] = {xmin, xmax};
			PLFLT yzero_y[2]={0., 0.};
			plline(2,yzero_x,yzero_y);
		}
	}

#if QT_VERSION >= 0x040300
	yAxisTab->_label_font_fld->getFontParams(family, style, weight);
	plsfont(family, style, weight);
#else
    family=yAxisTab->_label_font_fld->getFontIndex();
    plfont(family);
#endif

	switch(yAxisTab->_axis_minor_tickmarks_fld->currentIndex())
	{
		case QPUI_IN:
		default:
			y_subopt += "st";
			break;
		case QPUI_OUT:
			y_subopt += "ist";
			break;
		case QPUI_NOTICKS:
			break;
	}
		
	if(yAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
	{
		yopt += "l";
		y_subopt+="l";
		ymin = SafeLog10(ymin);
		ymax = SafeLog10(ymax);
	}
		
	switch(yAxisTab->_axis_major_tickmarks_fld->currentIndex())
	{
		case QPUI_IN: default:
			yopt += "t";
			break;
		case QPUI_OUT:
			yopt += "it";
			break;
		case QPUI_GRID:
			yopt += "gt";
			break;
		case QPUI_NOTICKS:
			break;
	}

    bool drawLabelsLeftAxis=false;
    bool drawLabelsRightAxis=false;
    
	if(frame()->DisplayTab->_show_left_ticks_labels_fld->isChecked())
	{
		yopt += "nv";
        drawLabelsLeftAxis=true;
	}
	if(frame()->DisplayTab->_show_right_ticks_labels_fld->isChecked())
	{
		yopt += "mv";
        drawLabelsRightAxis=true;
	}
	if(frame()->DisplayTab->_show_lr_alternating_ticks_labels_fld->isChecked())
	{
		if(getStackNumber()%2==1)
		{
			yopt += "nv";
            drawLabelsLeftAxis=true;
		}
		else
		{
			yopt += "mv";
            drawLabelsRightAxis=true;
		}
	}
	if(frame()->DisplayTab->_show_rl_alternating_ticks_labels_fld->isChecked())
	{
		if(getStackNumber()%2==0)
		{
			yopt += "nv";
            drawLabelsLeftAxis=true;
		}
		else
		{
			yopt += "mv";
            drawLabelsRightAxis=true;
		}
	}
		
	PLINT nsuby = yAxisTab->_axis_minor_ticks_num_fld->value();
	int ny=yAxisTab->_axis_major_ticks_num_fld->value();
	PLFLT ytick;
	if(ny!=0)
	{
		ytick=(ymax-ymin)/(PLFLT)ny;
	}
	else ytick=0.;

    if(drawLabelsRightAxis && drawLabelsLeftAxis)
    {
        // Draw axes in 2 times because the exponential label offsets differ for each axis
        std::string bkp=yopt;
        size_t position;
        PLLabelDefaults y_label_opt;
        y_label_opt.exp_label_disp=2.+yAxisTab->_exp_label_offset_y_fld->value();
        // draw left axis
        // remove "mv" commands
        do
        {
            position=yopt.find("mv");
            if(position!=std::string::npos) yopt.replace(position, 2, "");
        }
        while(position!=std::string::npos);
        y_label_opt.exp_label_pos=-0.02-yAxisTab->_exp_label_offset_x_fld->value()/5.;
        y_label_opt.exp_label_just=1.;
        plslabelfunc(NULL, &y_label_opt);
		// fix for plplot putting grid and major ticks in different place (suppress major tick)
        plsmaj(0., _ticks_length);
        plschr(0., _ticks_label_size);
        plbox("", 0., 0, yopt.c_str(), ytick, nsuby);
        plsmaj(0., 0.);
        plsmin(0., _ticks_length*0.6);
        plbox("", 0., 0, y_subopt.c_str(), ytick, nsuby);
        plsmaj(0., _ticks_length);
        plslabelfunc(NULL, NULL);

        yopt=bkp;
        // draw right axis
        // remove "nv" commands
        do
        {
            position=yopt.find("nv");
            if(position!=std::string::npos) yopt.replace(position, 2, "");
        }
        while(position!=std::string::npos);
        y_label_opt.exp_label_pos=1.02+yAxisTab->_exp_label_offset_x_fld->value()/5.;
        y_label_opt.exp_label_just=0.;
        plslabelfunc(NULL, &y_label_opt);
        plsmaj(0., _ticks_length);
        plschr(0., _ticks_label_size);
        plbox("", 0., 0, yopt.c_str(), ytick, nsuby);
        plsmaj(0., 0.);
        plsmin(0., _ticks_length*0.6);
        plbox("", 0., 0, y_subopt.c_str(), ytick, nsuby);
        plsmaj(0., _ticks_length);
        plslabelfunc(NULL, NULL);
    }
    else
    {
        PLLabelDefaults y_label_opt;
        y_label_opt.exp_label_disp=2.+yAxisTab->_exp_label_offset_y_fld->value();
        if(drawLabelsLeftAxis)
        {
            y_label_opt.exp_label_pos=-0.02+yAxisTab->_exp_label_offset_x_fld->value()/5.;
            y_label_opt.exp_label_just=1.;
        }
        else
        {
            y_label_opt.exp_label_pos=1.02+yAxisTab->_exp_label_offset_x_fld->value()/5.;
            y_label_opt.exp_label_just=0.;
        }
        plslabelfunc(NULL, &y_label_opt);
        plsmaj(0., _ticks_length);
        plschr(0., _ticks_label_size);
        plbox("", 0., 0, yopt.c_str(), ytick, nsuby);
        plsmaj(0., 0.);
        plsmin(0., _ticks_length*0.6);
        plbox("", 0., 0, y_subopt.c_str(), ytick, nsuby);
        plsmaj(0., _ticks_length);
        plslabelfunc(NULL, NULL);
    }
		
	//draw x=0
	if(frame()->DisplayTab->_show_xiszero_fld->isChecked())
	{
		PLFLT xzero_x[2] = {0.,0.};
		PLFLT xzero_y[2] = {ymin, ymax};
		plline(2, xzero_x, xzero_y);
	}
		
	// draw labels - text line width needs to be adjustable
	int _text_line_width = 1;
	plwid(QtWidth2PlWidth(_text_line_width));
				
	PLFLT x0win_save, x1win_save, y0win_save, y1win_save;
	PLFLT x0vp_save, x1vp_save, y0vp_save, y1vp_save;
			
	plgvpw(&x0win_save, &x1win_save, &y0win_save, &y1win_save);
	plgvpd(&x0vp_save, &x1vp_save, &y0vp_save, &y1vp_save);
				
	float xch_ndc = yAxisTab->_label_char_size_fld->value()/40.;
	float ych_ndc = frame()->xtAxisTab->_label_char_size_fld->value()/40.;
			
	plvpor(0., 1., 0., 1.);
	plwind(0., 1., 0., 1.);
				
	if(show_top_axis || show_bottom_axis)
	{
#if QT_VERSION >= 0x040300
		frame()->xtAxisTab->_label_font_fld->getFontParams(family, style, weight);
		plsfont(family, style, weight);
#else
        family=frame()->xtAxisTab->_label_font_fld->getFontIndex();
        plfont(family);
#endif
		plschr(QPUI_CHAR_HEIGHT, frame()->xtAxisTab->_label_char_size_fld->value());
		double p_def, p_ht;
		plgchr (&p_def, &p_ht);	

		PLINT labels_colour_index = getPGPColourIndex(frame()->xtAxisTab->_label_colour);
		plcol0(labels_colour_index);
			
		if(frame()->DisplayTab->_show_top_labels_fld->isChecked() && show_top_axis)
		{
			PLFLT dx, dy;
			dx=cos(frame()->xtAxisTab->_label_rotation_fld->value()*M_PI/180.);
			dy=sin(frame()->xtAxisTab->_label_rotation_fld->value()*M_PI/180.);
			plptex((x0vp_save+x1vp_save)/2.,
					y1vp_save+getCharHeight()*(1.+frame()->xtAxisTab->_label_disp_fld->value()),	
					dx, dy, 0.5,SafePGPString(frame()->xtAxisTab->_label_fld->text().toLatin1().data()).c_str());
		}
		if(frame()->DisplayTab->_show_bottom_labels_fld->isChecked() && show_bottom_axis)
		{
			PLFLT dx, dy;
			dx=cos(frame()->xtAxisTab->_label_rotation_fld->value()*M_PI/180.);
			dy=sin(frame()->xtAxisTab->_label_rotation_fld->value()*M_PI/180.);
			plptex((x0vp_save+x1vp_save)/2.,
					y0vp_save-getCharHeight()*(1.+frame()->xtAxisTab->_label_disp_fld->value()),
					 dx, dy
					,0.5,
					SafePGPString(frame()->xtAxisTab->_label_fld->text().toLatin1().data()).c_str());
		}
	}
				
	if(frame()->DisplayTab->_show_right_labels_fld->isChecked() || frame()->DisplayTab->_show_left_labels_fld->isChecked() || frame()->DisplayTab->_show_lr_alternating_labels_fld->isChecked() || frame()->DisplayTab->_show_rl_alternating_labels_fld->isChecked())
	{
		float y_label_just = 13./(1. + yAxisTab->_label_char_size_fld->value() * yAxisTab->_label_fld->text().length());
		if(y_label_just > 1.) y_label_just = 1.;
		
		PLINT family, style, weight;
#if QT_VERSION >= 0x040300
		yAxisTab->_label_font_fld->getFontParams(family, style, weight);
		plsfont(family, style, weight);
#else
        family=yAxisTab->_label_font_fld->getFontIndex();
        plfont(family);
#endif
		plschr(QPUI_CHAR_HEIGHT, yAxisTab->_label_char_size_fld->value());
				
		PLINT labels_colour_index = getPGPColourIndex(yAxisTab->_label_colour);
		plcol0(labels_colour_index);
				
		if(frame()->DisplayTab->_show_right_labels_fld->isChecked() ||
		(frame()->DisplayTab->_show_lr_alternating_labels_fld->isChecked() && getStackNumber()%2==0) ||
		(frame()->DisplayTab->_show_rl_alternating_labels_fld->isChecked() && getStackNumber()%2==1))
		{
			PLFLT dx, dy;
			dx=cos(yAxisTab->_label_rotation_fld->value()*M_PI/180.);
			dy=sin(yAxisTab->_label_rotation_fld->value()*M_PI/180.);
			plptex(x1vp_save+getCharHeight()*(1.+yAxisTab->_label_disp_fld->value()),
					(y0vp_save+y1vp_save)/2.,
					dx, dy,
					0.5,
					SafePGPString(yAxisTab->_label_fld->text().toLatin1().data()).c_str());
		}
		if(frame()->DisplayTab->_show_left_labels_fld->isChecked() ||
		(frame()->DisplayTab->_show_lr_alternating_labels_fld->isChecked() && getStackNumber()%2==1) ||
		(frame()->DisplayTab->_show_rl_alternating_labels_fld->isChecked() && getStackNumber()%2==0))
		{
			PLFLT dx, dy;
			dx=cos(yAxisTab->_label_rotation_fld->value()*M_PI/180.);
			dy=sin(yAxisTab->_label_rotation_fld->value()*M_PI/180.);
			plptex(x0vp_save-getCharHeight()*(1.+yAxisTab->_label_disp_fld->value()),
					(y0vp_save+y1vp_save)/2.,
					dx, dy,
					0.5,
				SafePGPString(yAxisTab->_label_fld->text().toLatin1().data()).c_str());
		}
	}
				
	plvpor(x0vp_save, x1vp_save, y0vp_save, y1vp_save);
	plwind(x0win_save, x1win_save, y0win_save, y1win_save);
			
	if(show_top_axis)
	{
#if QT_VERSION >= 0x040300
		frame()->TitleTab->_title_font_fld->getFontParams(family, style, weight);
		plsfont(family, style, weight);
#else
        family=frame()->TitleTab->_title_font_fld->getFontIndex();
        plfont(family);
#endif
		plschr(QPUI_CHAR_HEIGHT, frame()->TitleTab->_title_char_size_fld->value());
		PLINT title_colour_index = getPGPColourIndex(frame()->_title_colour);
		plcol0(title_colour_index);
		PLFLT fjust;
		PLFLT coord;
		switch(frame()->TitleTab->_title_just_fld->currentIndex())
		{
			case QPUI_HCENTRE:
			default:
				coord = 0.5;
				fjust = 0.5;
				break;
			case QPUI_LEFT:
				coord = 0.0;
				fjust = 0.0;
				break;
			case QPUI_RIGHT:
				coord = 1.0;
				fjust = 1.0;
				break;
		}
				
		plmtex("t", frame()->TitleTab->_title_disp_fld->value(), coord, fjust, SafePGPString(frame()->TitleTab->_title_fld->text().toLatin1().data()).c_str());
				
		plschr(QPUI_CHAR_HEIGHT, frame()->TitleTab->_subtitle_char_size_fld->value());
				
		plmtex("t", frame()->TitleTab->_subtitle_disp_fld->value(), coord, fjust, SafePGPString(frame()->TitleTab->_subtitle_fld->text().toLatin1().data()).c_str());
	}
	
	// draw legend if show_legend set
	if(LegendTab->_show_legend_fld->isChecked() && LegendTab->_legend_border_fld->currentIndex()  != QPUI_NO_BORDER && numberTraces() > 0)
	{
		PLFLT x0win_save, x1win_save, y0win_save, y1win_save;
		PLFLT x0vp_save, x1vp_save, y0vp_save, y1vp_save;
	
		plgvpw(&x0win_save, &x1win_save, &y0win_save, &y1win_save);
		plgvpd(&x0vp_save, &x1vp_save, &y0vp_save, &y1vp_save);
	
	//         QPlot_window->save();
	
		if(LegendTab->_legend_border_fld->currentIndex() == QPUI_LINE) pllsty(QP_SOLID);
		else pllsty(QP_NOLINE);
			
		plschr(QPUI_CHAR_HEIGHT, LegendTab->_legend_char_size_fld->value());
		PLFLT xch, ych;
		plgchr(&xch, &ych);
		PLFLT x0_vp = x0vp_save + LegendTab->_legend_loffset_fld->value() * (x1vp_save - x0vp_save);
		PLFLT x1_vp = x0_vp + LegendTab->_legend_width_fld->value() * (x1vp_save - x0vp_save) ;
		PLFLT y1_vp = y1vp_save - LegendTab->_legend_toffset_fld->value() * (y1vp_save - y0vp_save);
		PLFLT y0_vp = y1_vp - LegendTab->_legend_height_fld->value() * (y1vp_save - y0vp_save);
		if(x1_vp > 0.999) x1_vp = 0.999;

		plvpor(x0_vp, x1_vp, y0_vp, y1_vp);
		plwind(0., 1., 0., 1.);

		plwid(QtWidth2PlWidth(0.6*frame()->DisplayTab->_line_width_fld->value()));
		plbox("bc", 0., 0, "bc", 0., 0);
	
		plscol0(0, 255, 255, 255);
		plscol0(1, 0, 0, 0);
		plcol0(1);
		//         QPlot_window->unsave();
		plvpor(x0vp_save, x1vp_save, y0vp_save, y1vp_save);
		plwind(x0win_save, x1win_save, y0win_save, y1win_save);
	}
}

// ------------------------------------

void PanelItem::updateAxes()
{
	if(QuSlotDebug()) cout << "QpPanelItem::updateAxes " << endl;
	// this is called from QpFrame::updateAxes to ensure x or t axis set first
	// Note range done first to ensure _ydo exists for other calls
	
	if( yAxisTab->_axis_autoscale_fld->currentIndex() != QP_FROM_USER)  yAxisTab->setAxisMinMax();

	if( yAxisTab->_axis_autolabel_fld->isChecked())  yAxisTab->setAutoLabelText();
	
	LegendTab->updateNumTraces();

}

// ---------------------------------------

void PanelItem::getLabel(const QpAxes& axis, string& label)
{
	// get label for axis

	
	label = "-";
	string itemLabel;
	
	PlotItem* plotItem;
	for(int i=0; i<this->numChildren(); ++i)
	{
		plotItem=dynamic_cast<PlotItem*>(this->getChildByNumber(i));
		TraceItem* traceItem=dynamic_cast<TraceItem*>(plotItem);
		SpectrumItem* spectrumItem=dynamic_cast<SpectrumItem*>(plotItem);
		if(traceItem)
		{
			traceItem->getLabel(axis, itemLabel);
			if(label == "-") label = itemLabel.c_str();
			else{
				label = QuCommonLabl(label, itemLabel);
			}
		}
		else if(spectrumItem)
		{
			// only allowed one spectrum in a panel
			spectrumItem->getLabel(axis, label);
			break;
		}
	}
	
	QString greek = QString(label.c_str());
	greek.replace("theta","#gh");
	greek.replace("phi","#gf");
	label = greek.toLatin1().data();
}

// ---------------------------------------

int PanelItem::getXmetaIsLog()
{	
	int isLog=0;
	
	PlotItem* plotItem;
	for(int i=0; i<this->numChildren(); ++i)
	{
		plotItem=dynamic_cast<PlotItem*>(this->getChildByNumber(i));
		if(isLog == 0) isLog = plotItem->getXmetaIsLog(); // update until log is set 
	}
	return isLog;
}

// ---------------------------------------

bool PanelItem::getXmetaRange(double& panel_xmin, double& panel_xmax)
{
	// get range of values for x axis

	double xmin = DBL_MAX;
	double xmax = -DBL_MAX;
	
	
	double item_xmin, item_xmax;
	bool rangeFound=false;
	
	// descend thru' panel hierarchy
	
	PlotItem* plotItem;
	for(int i=0; i<this->numChildren(); ++i)
	{
		plotItem=dynamic_cast<PlotItem*>(this->getChildByNumber(i));
		plotItem->get_dobjs();
		
		if(plotItem->getXmetaRange(item_xmin, item_xmax))
		{
			xmin = min(xmin, item_xmin);
			xmax = max(xmax, item_xmax);
			rangeFound = true;
		}
	}
	
	if(rangeFound)
	{
		panel_xmin = xmin;
		panel_xmax = xmax;
	}

	return rangeFound;
}


// ---------------------------------------

void PanelItem::getYObjects()
{
	// Get data objects so get Y range has them.
	// Called when XT axis is user set.
	
	
		
	// descend thru' panel hierarchy
	
	PlotItem* plotItem;
	for(int i=0; i<this->numChildren(); ++i)
	{
		plotItem=dynamic_cast<PlotItem*>(this->getChildByNumber(i));
		plotItem->get_dobjs();
	}
	
	return;
}

// ---------------------------------------

bool PanelItem::getXRange(double& panel_xmin, double& panel_xmax)
{
	// get range of values for x axis

	double xmin = DBL_MAX;
	double xmax = -DBL_MAX;
	
	double item_xmin, item_xmax;
	bool rangeFound=false;
	
	// descend thru' panel hierarchy
	
	PlotItem *plotItem;
	for(int i=0; i<this->numChildren(); ++i)
	{
		plotItem=dynamic_cast<PlotItem*>(this->getChildByNumber(i));
		plotItem->get_dobjs();
		if(plotItem->getXRange(item_xmin, item_xmax))
		{
			xmin = min(xmin, item_xmin);
			xmax = max(xmax, item_xmax);
			rangeFound = true;
		}
	}
	if(rangeFound)
	{
		panel_xmin = xmin;
		panel_xmax = xmax;
	}

	return rangeFound;
}

// ---------------------------------------

bool PanelItem::getYRange(double& panel_ymin, double& panel_ymax)
{
	// get range of values for y axis
	if ( QuSlotDebug() )
	{
		printf( "QpPanelItem::getYRange this = %ld \n", (long)this );
		fflush(stdout);
	}
	
	double ymin = DBL_MAX;
	double ymax =- DBL_MAX;
	
	double item_ymin, item_ymax;
	bool rangeFound = false;
	
	PlotItem* plotItem;
	for(int i=0; i<this->numChildren(); ++i)
	{
		plotItem=dynamic_cast<PlotItem*>(this->getChildByNumber(i));
		TraceItem* traceItem=dynamic_cast<TraceItem*>(plotItem);
		SpectrumItem* spectrumItem=dynamic_cast<SpectrumItem*>(plotItem);
		if(traceItem)
		{
			if(traceItem->getYRange(item_ymin,item_ymax))
			{
				ymin = min(ymin, item_ymin);
				ymax = max(ymax, item_ymax);
				rangeFound = true;
			}
		}
		else if(spectrumItem)
		{
			if(spectrumItem->getYRange( item_ymin, item_ymax))
			{
				ymin = min(ymin, item_ymin);
				ymax = max(ymax, item_ymax);
				rangeFound = true;
			}
		}
	}

	if(!rangeFound) return rangeFound;
	
	if(yAxisTab->_axis_y_nudge_fld->isChecked())
	{
		if(yAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
		{
			double nudgeUp = pow(fabs(ymax / (ymin +1e-29)), 0.01);
			double nudgeDown = pow(fabs(ymax / (ymin +1e-29)), 0.02);
			ymax *= nudgeUp;
			ymin /= nudgeDown;
		}
		else
		{
			double nudge = fabs(ymax - ymin)*0.02;
			ymax += nudge;
			ymin -= nudge;
		}
	}
	panel_ymin = ymin;
	panel_ymax = ymax;
	return rangeFound;
}

// ---------------------------------------

bool PanelItem::getYmetaRange(double& panel_ymin, double& panel_ymax)
{
	// get range of values for y axis
	double ymin = DBL_MAX;
	double ymax =- DBL_MAX;
	
	double item_ymin, item_ymax;
	bool rangeFound = false;
	int isLog = 0;
	
	PlotItem* plotItem;
	for(int i=0; i<this->numChildren(); ++i)
	{
		plotItem=dynamic_cast<PlotItem*>(this->getChildByNumber(i));
		if(plotItem->getYmetaRange(item_ymin, item_ymax))
		{
			ymin = min(ymin, item_ymin);
			ymax = max(ymax, item_ymax);
			rangeFound = true;
		}

		if(isLog == 0) isLog = plotItem->getYmetaIsLog();
	}
	if(isLog == 1) yAxisTab->_axis_linlog_fld->setCurrentIndex(QP_LOG);
	else yAxisTab->_axis_linlog_fld->setCurrentIndex(QP_LINEAR);

	if(yAxisTab->_axis_y_nudge_fld->isChecked())
	{
		if(yAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
		{
			double nudgeUp = pow(fabs(ymax / (ymin +1e-29)), 0.01);
			double nudgeDown = pow(fabs(ymax / (ymin +1e-29)), 0.02);
			ymax *= nudgeUp;
			ymin /= nudgeDown;
		}
		else
		{
			double nudge = fabs(ymax - ymin)*0.02;
			ymax += nudge;
			ymin -= nudge;
		}
	}
	
	if(rangeFound)
	{
		panel_ymin = ymin;
		panel_ymax = ymax;
	}
	
	
	return rangeFound;
}

// ---------------------------------------

QpuiReturnStatus PanelItem::plot(double _height, double _width, double _toffset, double _loffset, PlotWindow* win)
{
	// defend against signals out pof order, set set focus window, ensure plot in right window
	win->setPLFocus();
	
	// set viewport
    PLFLT x0, x1, y0, y1;
    x0 = _loffset;
    x1 = x0 + _width;
    y1 = 1 - _toffset;
    y0 = y1 - _height;
    plvpor(x0, x1, y0, y1);
    
    // preplot
    this->prePlot();
    
    // now handle each plot object within this panel

	PlotItem* plotItem;
	for(int i=0; i<this->numChildren(); ++i)
	{
		plotItem=dynamic_cast<PlotItem*>(this->getChildByNumber(i));
		QpuiReturnStatus plotRes;      
        win->setSmooth(false);
	    if(QPUI_FAILURE == plotItem->plot())
	    {
		    cout << "Problem plotting plot item " << (plotItem->getName().toLatin1().data()) << endl;
	    }
	    win->setSmooth(true);
	}

	for(int i=0; i<this->numChildren(); ++i)
	{
		plotItem=dynamic_cast<PlotItem*>(this->getChildByNumber(i));
		TraceItem* traceItem=dynamic_cast<TraceItem*>(plotItem);
		if(traceItem)
        {
			try{
				traceItem->plotLegend();
			}
			catch(...){;}
        }
	}

    bool show_top_axis = false;
    bool show_bottom_axis = false;
    int stackNumber = getStackNumber();
    if(stackNumber == 1) show_top_axis = true;
    if(stackNumber == frame()->numberInStack() ) show_bottom_axis = true;

    this->postPlot(show_top_axis, show_bottom_axis);

    return QPUI_SUCCESS;
}

void PanelItem::updateYZRange()
{
  	if ( QuSlotDebug() ) cout << "QpPanelItem::updateYZRange" << endl;

	// called when X or T axis range is changed
	PlotItem* plotItem;
	for(int i=0; i<this->numChildren(); ++i)
	{
		plotItem=dynamic_cast<PlotItem*>(this->getChildByNumber(i));
		plotItem->clearObj_id();
	}
	this->yAxisTab->setAxisMinMax();
	
	// update colourscale in spectra
	for(int i=0; i<this->numChildren(); ++i)
	{
		plotItem=dynamic_cast<PlotItem*>(this->getChildByNumber(i));
		
		SpectrumItem* spectrumItem = dynamic_cast<SpectrumItem*>(plotItem);
		if(spectrumItem) spectrumItem->ColourscaleTab->setColourscaleMinMax();
	}

}

bool PanelItem::removeChild(QpuiItem* child, bool alsoDelete)
{
    bool res;
    res=QpuiItem::removeChild(child, alsoDelete);
    updateAxes();
    return res;
}

bool PanelItem::removeChild(int i, bool alsoDelete)
{
    bool res;
    res=QpuiItem::removeChild(i, alsoDelete);
    updateAxes();
    return res;
}



// ---------------------------------------

QDomElement PanelItem::toXMLElement(QDomDocument& doc) const
{
    QDomElement xml=doc.createElement("Panel_Item");
    xml.appendChild(createTextXMLNode(doc, "Name", getName()));

    xml.appendChild(yAxisTab->toXMLElement(doc));// do bits common to all axes
    xml.appendChild(createTextXMLNode(doc, "Lin_Log", yAxisTab->_axis_linlog_fld->currentIndex()));

    xml.appendChild(LegendTab->toXMLElement(doc));
    
    int cnt = numChildren();
    xml.setAttribute("size", QString::number(cnt));
   
    PlotItem* plotItem;
    for(int i=0; i<cnt; ++i)
    {
        plotItem=dynamic_cast<PlotItem*>(getChildByNumber(i));
        if(plotItem->I_Type == QP_TRACE)
        {
            TraceItem *traceItem = dynamic_cast<TraceItem*>(plotItem);
            xml.appendChild(traceItem->toXMLElement(doc));
        }
        else if(plotItem->I_Type == QP_SPECTROGRAM)
        {
            SpectrumItem *spectrumItem = dynamic_cast<SpectrumItem*>(plotItem);
            xml.appendChild(spectrumItem->toXMLElement(doc));
        }
        else if(plotItem->I_Type == QP_CONST_Y)
        {
            TraceItem *constYItem = dynamic_cast<TraceItem*>(plotItem);
            xml.appendChild(constYItem->toXMLElement(doc));
        }
        else if(plotItem->I_Type == QP_CONST_XT)
        {
            TraceItem *constXTItem = dynamic_cast<TraceItem*>(plotItem);
            xml.appendChild(constXTItem->toXMLElement(doc));
        }
        else if(plotItem->I_Type == QP_REG_LINE)
        {
            TraceItem *regItem = dynamic_cast<TraceItem*>(plotItem);
            xml.appendChild(regItem->toXMLElement(doc));
        }
        else if(plotItem->I_Type == QP_EVENT)
        {
            EventItem *eventItem = dynamic_cast<EventItem*>(plotItem);
            xml.appendChild(eventItem->toXMLElement(doc));
        }
    }
    
    return xml;
}

void PanelItem::fromXMLElement(QDomElement& el)
{
    QString s;
    
    if(getXMLNodeText(el, "Name", s))
    {
        setName(s);
    }

    QDomElement axisElement=el.firstChildElement("Axis_Tab");
    if(!axisElement.isNull())
    {
        yAxisTab->fromXMLElement(axisElement);
    }
    if(getXMLNodeText(el, "Lin_Log", s))
    {
        yAxisTab->_axis_linlog_fld->setCurrentIndex((bool)s.toInt());
    }

    QDomElement legendElement=el.firstChildElement("Legend_Tab");
    if(!legendElement.isNull())
    {
        LegendTab->fromXMLElement(legendElement);
    }

    QDomElement plotElement=el.firstChildElement("Plot_Item");

    while(!plotElement.isNull())
    {
        QString type=plotElement.attribute("Type", "");
        if(type=="Trace")
        {
            TraceItem *traceItem = new TraceItem("", this, "", QP_TRACE);
            traceItem->fromXMLElement(plotElement);
        }
        else if(type=="Spectrogram")
        {
            SpectrumItem *spectrumItem = new SpectrumItem("", this);
            spectrumItem->fromXMLElement(plotElement);
        }
        else if(type=="Const_Y")
        {
            TraceItem *constYItem = new TraceItem("", this, "", QP_CONST_Y);
            constYItem->fromXMLElement(plotElement);
        }
        else if(type=="Const_XT")
        {
            TraceItem *constXTItem = new TraceItem("", this, "", QP_CONST_XT);
            constXTItem->fromXMLElement(plotElement);
        }
        else if(type=="Reg_Line")
        {
            TraceItem *regLine = new TraceItem("", this, "", QP_REG_LINE);
            regLine->fromXMLElement(plotElement);
        }
        else if(type=="Event")
        {
            EventItem *eventItem = new EventItem("", this);
            eventItem->fromXMLElement(plotElement);
        }
        
        plotElement=plotElement.nextSiblingElement("Plot_Item");
    }
    
}

Vistream& operator>>(Vistream& v_is, PanelItem& panel)
{
	if(QuSlotDebug()) cout <<  "QpPanelItem restore (>>)" << endl;

	// streams from virtual input stream v_is into panel object
	
	QpuiItem& list_item = panel;
	v_is >> list_item;
	
	QpAxisTab *axisTab = panel.yAxisTab;
	
	v_is >> *axisTab; // do bits common to all axes

	int linlog;
	v_is >> linlog;
	panel.yAxisTab->_axis_linlog_fld->setCurrentIndex(linlog);
	bool show_legend;
	v_is >> show_legend;
	panel.LegendTab->_show_legend_fld->setChecked(show_legend);
	
	int legend_border;
	v_is >> legend_border;
	panel.LegendTab->_legend_border_fld->setCurrentIndex(legend_border);
	
	int legend_background;
	v_is >> legend_background;
	panel.LegendTab->_legend_background_fld->setCurrentIndex(legend_background);
	
	int legend_num_rows;
	v_is >> legend_num_rows;
	panel.LegendTab->_legend_num_rows_fld->setValue(legend_num_rows);
	
	int legend_num_cols;
	v_is >> legend_num_cols;
	panel.LegendTab->_legend_num_cols_fld->setValue(legend_num_cols);
	
	float legend_char_size;
	v_is >> legend_char_size;
	panel.LegendTab->_legend_char_size_fld->setValue(legend_char_size);
	
	int legend_max_char_length;
	v_is >> legend_max_char_length;
	panel.LegendTab->_legend_max_char_length_fld->setValue(legend_max_char_length);
	
	float legend_line_length;
	v_is >> legend_line_length;
	panel.LegendTab->_legend_line_length_fld->setValue(legend_line_length);
	
	int legend_just;
	v_is >> legend_just;
	panel.LegendTab->_legend_just_fld->setCurrentIndex(legend_just);
	
	if( qsas_restore_release >= 20306)
	{
		PLINT family, style, weight;
		v_is >> family;
		v_is >> style;
		v_is >> weight;
		switch(family)
		{
			case PL_FCI_SERIF: panel.LegendTab->_legend_font_fld->setSerifFont(); break;
			case PL_FCI_MONO: panel.LegendTab->_legend_font_fld->setMonotypeFont(); break;
			case PL_FCI_SCRIPT: panel.LegendTab->_legend_font_fld->setDecorativeFont(); break;
			default: panel.LegendTab->_legend_font_fld->setSansFont(); break;
		}
		panel.LegendTab->_legend_font_fld->setItalic(style==PL_FCI_ITALIC);
		panel.LegendTab->_legend_font_fld->setBold(weight==PL_FCI_BOLD);
	}
	else
	{
		int legend_font;
		v_is >> legend_font;
		switch(legend_font)
		{
			case 1: // Roman
				panel.LegendTab->_legend_font_fld->setRomanFont();
				break;
				
			case 2: // Italic
				panel.LegendTab->_legend_font_fld->setItalicFont();
				break;
				
			default: // Normal
				panel.LegendTab->_legend_font_fld->setNormalFont();
				break;
		}
	}
	
	int legend_align;
	v_is >> legend_align;
	panel.LegendTab->_legend_align_fld->setCurrentIndex(legend_align);
  
	float legend_loffset;
	v_is >> legend_loffset;
	panel.LegendTab->_legend_loffset_fld->setValue(legend_loffset);
	
	float legend_toffset;
	v_is >> legend_toffset;
	panel.LegendTab->_legend_toffset_fld->setValue(legend_toffset);
	
	float legend_width;
	v_is >> legend_width;
	panel.LegendTab->_legend_width_fld->setValue(legend_width);
	
	float legend_height;
	v_is >> legend_height;
	panel.LegendTab->_legend_height_fld->setValue(legend_height);
	

	// read in the plot items
	int cnt;
	v_is >> cnt;

	//QpPlotItem * after = 0;
	for(int i=0; i<cnt; i++)
	{
		int Type;
		v_is >> Type;
		if(Type == QP_TRACE)
		{
			TraceItem *traceItem = new TraceItem("", &panel, "", (QpItemType)Type);
			v_is >> *traceItem;
// 			after = (QpPlotItem*) traceItem;
		}
		else if(Type == QP_SPECTROGRAM)
		{
			SpectrumItem *spectrumItem = new SpectrumItem("", &panel);
			v_is >> *spectrumItem;

// 			after = (QpPlotItem*) spectrumItem;
		}
		else if(Type == QP_CONST_Y)
		{
			TraceItem *constYItem = new TraceItem("", &panel, "", (QpItemType)Type);
			v_is >> *constYItem;
// 			after = (QpPlotItem*) constYItem;
		}
		else if(Type == QP_CONST_XT)
		{
			TraceItem *constXTItem = new TraceItem("", &panel, "", (QpItemType)Type);
			v_is >> *constXTItem;
// 			after = (QpPlotItem*) constXTItem;
		}
		else if(Type == QP_REG_LINE)
		{
			TraceItem *regLineItem = new TraceItem("", &panel, "", (QpItemType)Type);
			v_is >> *regLineItem;
// 			after = (QpPlotItem*) constXTItem;
		}
		else if(Type == QP_EVENT)
		{
			EventItem *eventItem = new EventItem("", &panel);
			v_is >> *eventItem;
// 			after = (QpPlotItem*) eventItem;
		}
	}

	return v_is;
}

// -----------------------------
 

Vostream& operator<<(Vostream& v_os, PanelItem& panel)
{
	if(QuSlotDebug()) cout << "PanelItem save (<<)" << endl;

	// streams panel object to virtual output stream v_os
	QpuiItem& list_item = panel;
// 	v_os << list_item;
	v_os = save(v_os, list_item);
   
	v_os << ListTypeGroup;
	v_os << panel.I_Type;
	
   	QpAxisTab *axisTab = panel.yAxisTab;
	
	v_os << *axisTab; // do bits common to all axes
	v_os << (int) panel.yAxisTab->_axis_linlog_fld->currentIndex();

	v_os << panel.LegendTab->_show_legend_fld->isChecked();
	v_os << panel.LegendTab->_legend_border_fld->currentIndex();
	v_os << (int)panel.LegendTab->_legend_background_fld->currentIndex();
	v_os << (int)panel.LegendTab->_legend_num_rows_fld->value();
	v_os << (int)panel.LegendTab->_legend_num_cols_fld->value();
	v_os << (float)panel.LegendTab->_legend_char_size_fld->value();
	v_os << (int)panel.LegendTab->_legend_max_char_length_fld->value();
	v_os << (float)panel.LegendTab->_legend_line_length_fld->value();
	v_os << (int)panel.LegendTab->_legend_just_fld->currentIndex();
	PLINT family, style, weight;
	panel.LegendTab->_legend_font_fld->getFontParams(family, style, weight);
	v_os << family;
	v_os << style;
	v_os << weight;
	v_os << (int)panel.LegendTab->_legend_align_fld->currentIndex();
	v_os << (float)panel.LegendTab->_legend_loffset_fld->value();
	v_os << (float)panel.LegendTab->_legend_toffset_fld->value();
	v_os << (float)panel.LegendTab->_legend_width_fld->value();
	v_os << (float)panel.LegendTab->_legend_height_fld->value();
	
	// now output its children
	int cnt = panel.numChildren();
	v_os << cnt;
	
// 	QpPlotItem *plotItem = dynamic_cast<QpPlotItem*>(panel.firstChild());
	PlotItem* plotItem;
// 	while(plotItem)
	for(int i=0; i<cnt; ++i)
	{
		plotItem=dynamic_cast<PlotItem*>(panel.getChildByNumber(i));
		v_os << plotItem->I_Type;
		if(plotItem->I_Type == QP_TRACE)
		{
			TraceItem *traceItem = dynamic_cast<TraceItem*>(plotItem);
			v_os << *traceItem;
		}
		else if(plotItem->I_Type == QP_SPECTROGRAM)
		{
			SpectrumItem *spectrumItem = dynamic_cast<SpectrumItem*>(plotItem);
			v_os << *spectrumItem;
		}
		else if(plotItem->I_Type == QP_CONST_Y)
		{
			TraceItem *constYItem = dynamic_cast<TraceItem*>(plotItem);
			v_os << *constYItem;
		}
		else if(plotItem->I_Type == QP_CONST_XT)
		{
			TraceItem *constXTItem = dynamic_cast<TraceItem*>(plotItem);
			v_os << *constXTItem;
		}
		else if(plotItem->I_Type == QP_REG_LINE)
		{
			TraceItem *regLineItem = dynamic_cast<TraceItem*>(plotItem);
			v_os << *regLineItem;
		}
		else if(plotItem->I_Type == QP_EVENT)
		{
			EventItem *eventItem = dynamic_cast<EventItem*>(plotItem);
			v_os << *eventItem;
		}
		
// 		plotItem = dynamic_cast<QpPlotItem*>(plotItem->nextSibling());
	}

	return v_os;
}


