
// qpui_trace.cc
//
// Plot Interface Implementation - Trace Object
//
// Janet Barnes
//
// July 2002
//
//  Tony Allen
//  Jan 2008, Feb 2013

#include "qpui.h"
#include "qwui.h"
#include "QdTime.h"
#include "qdutil.h"

using namespace QSAS;

extern QPalette QPUI_normal_palette;
extern int qsas_restore_release;

// local wrapper for plpoin() for speed
void Q_plpoin(PLINT n, PLFLT *x, PLFLT *y, PLINT code);


// --------------------------
//     TraceItem CLASS
//     ------------------

PlotItem::PlotItem(const QString& name, QpuiItem* parent, const QString& WLcontext, int position)
{
	if(parent==NULL)
	{
		std::cerr << "PlotItem::PlotItem: creating PlotItem with NULL parent" << std::endl;
		return;
	}
	RootItem* rootParent=dynamic_cast<RootItem*>(parent);
	PageItem* pageParent=dynamic_cast<PageItem*>(parent);
	FrameItem* frameParent=dynamic_cast<FrameItem*>(parent);
	PanelItem* panelParent=dynamic_cast<PanelItem*>(parent);
	if(rootParent==NULL && pageParent==NULL && frameParent==NULL && panelParent==NULL)
	{
		std::cerr << "PlotItem::PlotItem: wrong parent type" << std::endl;
	}
	
	if(name.isEmpty())
	{
		_name=QString("Plot %1").arg(parent->numChildren()+1);
	}
	else
	{
		_name=name;
	}
	parent->addChild(this, position);
	
	
	tabWidget=parent->tabWidget;

	nameLock = false;
	if(WLcontext.isEmpty())
	{
		QdObject_var o=QuGetObjByName(name);
		if(!o.is_nil()) this->WLcontext=name;
		else this->WLcontext="";
	}
	else
	{
		this->WLcontext=WLcontext;
	}
}

PlotItem::PlotItem(PlotItem* plot, QpuiItem* parent, int position)
{
	if(parent==NULL)
	{
		std::cerr << "PlotItem::PlotItem: creating PlotItem with NULL parent" << std::endl;
		return;
	}
	RootItem* rootParent=dynamic_cast<RootItem*>(parent);
	PageItem* pageParent=dynamic_cast<PageItem*>(parent);
	FrameItem* frameParent=dynamic_cast<FrameItem*>(parent);
	PanelItem* panelParent=dynamic_cast<PanelItem*>(parent);
	if(rootParent==NULL && pageParent==NULL && frameParent==NULL && panelParent==NULL)
	{
		std::cerr << "PlotItem::PlotItem: wrong parent type" << std::endl;
	}
	_children.clear();
	setName(plot->_name);
	parent->addChild(this, position);
	
	tabWidget=parent->tabWidget;
	//I_Panel = dynamic_cast<PanelItem*>(_parent);
	nameLock = false;
	WLcontext=plot->WLcontext;
}

PlotItem::~PlotItem()
{

	foreach(PlotItem* i, linkedPlots)
	{
		i->linkedPlots.removeAll(this);
	}
}

bool PlotItem::reparent(QpuiItem* newParent, int position)
{
	if(isParentOf(newParent)) return false;

	RootItem* rootParent=dynamic_cast<RootItem*>(newParent);
	PageItem* pageParent=dynamic_cast<PageItem*>(newParent);
	FrameItem* frameParent=dynamic_cast<FrameItem*>(newParent);
	PanelItem* panelParent=dynamic_cast<PanelItem*>(newParent);
	if(rootParent==NULL && pageParent==NULL && frameParent==NULL && panelParent==NULL)
	{
		bool found=false;
		// find the first PanelItem which is a parent of newParent
		while(!found && newParent->parent()!=NULL)
		{
			newParent=newParent->parent();
			panelParent=dynamic_cast<PanelItem*>(newParent);
			if(panelParent!=NULL) found=true;
		}
		if(!found) return false;
	}

	if(_parent!=NULL)
	{
		_parent->removeChild(this, false);
	}
	if(newParent!=NULL)
	{
		newParent->addChild(this, position);
	}
	
	return true;
}

void PlotItem::setLinks(QList <PlotItem*>& linkList)
{
	linkedPlots = linkList;
}

// ----------------------------------

void PlotItem::removeLink(PlotItem* plot_item)
{
	linkedPlots.removeOne(plot_item);	
}

// ------------------------------------------

QpFrameType PlotItem::frameType()
{
	if(panel()) return panel()->frameType();
    return QP_TS;
}

TraceItem::TraceItem(const QString& name, QpuiItem* parent, const QString& WLcontext, QpItemType plotType, int position):
		PlotItem(name, parent, WLcontext, position)
{
	if(name.isEmpty()) _name=_name+QString(": Trace");
		
	C1_colour = QColor(0,0,0);
	C2_colour = QColor(255,0,0);
	C3_colour = QColor(0,130,0);
	C4_colour = QColor(150,0,255);
	R_colour  = QColor(0,0,0);
	X_colour  = QColor(255,0,0);
	Y_colour  = QColor(0,180,0);
	Z_colour  = QColor(0,0,180);
	T_colour  = QColor(150,0,0);
	P_colour  = QColor(0,130,0);
	L_colour  = QColor(150,0,150);
	userLineColour = false;
	userSymbolColour = false;
	I_Type = plotType;

	DataTab = new QpDataTab(tabWidget, "DataTab", this, plotType);
	DataTab->hide();

	TraceTab = new QpTraceTab(tabWidget, "TraceTab", this);
	TraceTab->hide();

    DataTab->_y_data_fld->setText(WLcontext);
    DataTab->_y_data_fld->reValidate(true);
	
	QColor black = QColor(0, 0, 0);
	if(WLcontext.contains("_C1_") || WLcontext.contains("_dr1_")|| WLcontext.contains("_at1_")){
		TraceTab->setLineColour(C1_colour);
		TraceTab->setSymbolColour(C1_colour);
	}
	else if(WLcontext.contains("_C2_") || WLcontext.contains("_dr2_")|| WLcontext.contains("_at2_")){
		TraceTab->setLineColour(C2_colour);
		TraceTab->setSymbolColour(C2_colour);
	}
	else if(WLcontext.contains("_C3_") || WLcontext.contains("_dr3_")|| WLcontext.contains("_at3_")){
		TraceTab->setLineColour(C3_colour);
		TraceTab->setSymbolColour(C3_colour);
	}
	else if(WLcontext.contains("_C4_") || WLcontext.contains("_dr4_")|| WLcontext.contains("_at4_")){
		TraceTab->setLineColour(C4_colour);
		TraceTab->setSymbolColour(C4_colour);
	}
	else{
		TraceTab->setLineColour(black);
		TraceTab->setSymbolColour(black);
	}


	// set Minimum Size for TabWidget
	tabWidget->adjustSize();
   
	_y_obj_id_getYRange = 0;
	_x_obj_id_getXRange = 0;

}

TraceItem::TraceItem(TraceItem* trace, QpuiItem* parent, int position):
		PlotItem(trace, parent, position)
{
	C1_colour = QColor(0,0,0);
	C2_colour = QColor(255,0,0);
	C3_colour = QColor(0,130,0);
	C4_colour = QColor(150,0,255);
	R_colour  = QColor(0,0,0);
	X_colour  = QColor(255,0,0);
	Y_colour  = QColor(0,180,0);
	Z_colour  = QColor(0,0,180);
	T_colour  = QColor(150,0,0);
	P_colour  = QColor(0,130,0);
	L_colour  = QColor(150,0,150);

	I_Type = trace->I_Type;

	userLineColour = trace->userLineColour;
	userSymbolColour = trace->userLineColour;
	DataTab = new QpDataTab(trace->DataTab, this);
    frame()->changeFrame(trace->frame()->_frameType);
    DataTab->timeAsSec_toggle->setVisible(frameType()==QP_XY);
    DataTab->timeAsSec_toggle->setChecked(trace->DataTab->timeAsSec_toggle->isChecked());
	DataTab->hide();

	TraceTab = new QpTraceTab(trace->TraceTab, this);

	TraceTab->hide();
	
	_line_colour = trace->_line_colour;
	_symbol_colour = trace->_symbol_colour;
	
	// set Minimum Size for TabWidget
	tabWidget->adjustSize();
   
	_y_obj_id_getYRange = 0;
	_x_obj_id_getXRange = 0;
	
	DataTab->_y_data_fld->setText(trace->DataTab->_y_data_fld->text());
	DataTab->_y_data_fld->reValidate(true);
	if(!trace->DataTab->_y_data_fld->text().isEmpty())
	{
		DataTab->_x_data_fld->setText(trace->DataTab->_x_data_fld->text());
		DataTab->_x_data_fld->reValidate(true);
	}
}

TraceItem::~TraceItem()
{
	clear();
	if(DataTab) delete DataTab;
	if(TraceTab) delete TraceTab;
	foreach(PlotItem* i, linkedPlots)
	{
		i->linkedPlots.removeAll(this);
	}
}

// ----------------------------------

void TraceItem::setDefaultOptions()
{
	DataTab->timeAsSec_toggle->setChecked(false);
	DataTab->_show_legend_fld->setChecked(true);
	
	TraceTab->_line_style_fld->setCurrentIndex(QP_SOLID);
	TraceTab->_line_width_fld->setValue(1);
	TraceTab->_connect_style_fld->setCurrentIndex(QPUI_STRAIGHT);
	
	QPalette line_palette = QPalette(QPUI_normal_palette);
	line_palette.setColor(QPalette::Active ,QPalette::ButtonText, Qt::black);
	line_palette.setColor(QPalette::Inactive, QPalette::ButtonText, Qt::black);
	TraceTab->_line_colour_fld->setPalette(line_palette);
	_line_colour = Qt::black;

	TraceTab->_symbol_shape_fld->setCurrentIndex(QPUI_NOSYMBOL);
	TraceTab->_symbol_size_fld->setValue(1.0);
	TraceTab->_symbol_colour_fld->setPalette(line_palette);
	_symbol_colour = Qt::black;
	
	updateAxes();
}

// ------------------------------------

void TraceItem::updateAxes()
{
	if(QuSlotDebug()) cout << "TraceItem::updateAxes " << endl;
	
	frame()->updateAxes();

}

// ------------------------------------

void TraceItem::clearObj_id()
{
   // called when we need to force a range update
	_y_obj_id_getYRange = 0;
	_x_obj_id_getXRange = 0;
}

// -------------------------------------------

void TraceItem::showTabs()
{
	if(QuSlotDebug()) cout << "TraceItem::showTabs '" << this->getName().toLatin1().data() << "'" << endl;
	
	// ensure all tabs are hidden
	int Ntabs = tabWidget->count();
	for(int i = Ntabs-1; i >= 0; i--)
	{
		if(tabWidget->widget(i))
		{
			tabWidget->widget(i)->hide();
			tabWidget->removeTab(i);
		}
	}


	tabWidget->addTab(DataTab, "Data");

	tabWidget->addTab(TraceTab, "Trace");
	// value return is always shown
	tabWidget->addTab(page()->ValueReturnTab, "Value Return");
 	if(this->root()->I_PlotView->I_LastTab == QP_TRACE_TAB)  tabWidget->setCurrentIndex(tabWidget->indexOf(TraceTab));
 	else tabWidget->setCurrentIndex(tabWidget->indexOf(DataTab));
	
}

// -------------------------------------------

void TraceItem::replaceData(QString obj_name)
{
	if(obj_name.isEmpty()) return;
	if(QuSlotDebug()) cout << "TraceItem::replaceData '" <<  obj_name.toLatin1().data() << "'" << endl;

	QString oldCompText = compText( DataTab->_y_data_fld->text() ); 

	if(oldCompText.isEmpty() && !(DataTab->_y_data_fld->text().contains('?')) )
	{
		if(this->frameType() == QP_TS)
		{
			DataTab->_y_data_fld->setDefaultComp(QS_0D, QS_SEQ, obj_name); // set data slot
		}
		else
		{
			DataTab->_y_data_fld->setDefaultComp(QS_1D, QS_NOT_SEQ, obj_name); // set data slot
		}
	}
	else
	{
		// setTextKeepComp can change text for comp or substitute string
		DataTab->_y_data_fld->setTextKeepComp(obj_name);  // set data slot	
	}

	if(!nameLock) this->setName(DataTab->_y_data_fld->text());               // set item name

	this->WLcontext = QString(baseObjectText(DataTab->_y_data_fld->text()).toLatin1().data());
	DataTab->_y_data_fld->reValidate(true);

}

// -------------------------------------------

void TraceItem::replaceXData(QString obj_name)
{
	if(QuSlotDebug()) cout << "TraceItem::replaceXData" << endl;
	if(obj_name.isEmpty()) return;

	DataTab->_x_data_fld->setTextKeepComp(obj_name);  // set data slot
	// setTextKeepComp can change text for comp or substitute string

}

// -------------------------------------------
void TraceItem::replaceTIdata(QString tivl)
{
	if(QuSlotDebug()) cout << "TraceItem::replaceTI with '" << tivl.toLatin1().data() << "'" << endl;
	if(tivl.isEmpty()) return;
	
	DataTab->_mask_data_fld->setText(tivl);  // set TI slot
	DataTab->_mask_data_fld->reValidate();


}

// -------------------------------------------
void TraceItem::replaceTIcomp(QString comp)
{
	if(QuSlotDebug()) cout << "TraceItem::replaceTI with '" << comp.toLatin1().data() << "'" << endl;
	if(comp.isEmpty()) return;
	
	QString newName = baseObjectText(DataTab->_mask_data_fld->text().toLatin1().data());
	newName += comp;
	DataTab->_mask_data_fld->setText(newName);  // set TI slot

	DataTab->_mask_data_fld->reValidate();

}

// -------------------------------------------

void TraceItem::replaceComponent(QString comp)
{
	if(QuSlotDebug()) cout << "TraceItem::replaceComponent with '" << comp.toLatin1().data() << "'" << endl;
	if(comp.isEmpty()) return;
	
	QString newName = baseObjectText(DataTab->_y_data_fld->text().toLatin1().data());
	newName += comp;
	DataTab->_y_data_fld->setText(newName);  // set data slot

	DataTab->_y_data_fld->reValidate();
	this->setName(newName);               // set item name 

}

// -------------------------------------------

void TraceItem::replaceXComponent(QString comp)
{
	if(QuSlotDebug()) cout << "TraceItem::replaceXComponent" << endl;
	if(comp.isEmpty()) return;
	
	QString newName = baseObjectText(DataTab->_x_data_fld->text().toLatin1().data());
	newName += comp;
	DataTab->_x_data_fld->setText(newName);  // set data slot
	
}
// -------------------------------------------

void TraceItem::replace_C_Component(QString comp)
{
	if(QuSlotDebug()) cout << "TraceItem::replace_C_Component" << endl;
	if(comp.isEmpty()) return;
	
	QString newName = baseObjectText(TraceTab->_c_data_fld->text().toLatin1().data());
	newName += comp;
	TraceTab->_c_data_fld->setText(newName);  // set data slot
	TraceTab->_c_data_fld->reValidate(true);
	
}

// -------------------------------------------
void TraceItem::replaceCdata(QString base)
{
	if(QuSlotDebug()) cout << "TraceItem::replaceCdata with '" << base.toLatin1().data() << "'" << endl;
	if(base.isEmpty()) return;
	
	TraceTab->_c_data_fld->setText(base);  
	TraceTab->_c_data_fld->reValidate(true);


}

// -------------------------------------------

void TraceItem::refreshTabs()
{
	if(QuSlotDebug()) cout << "QpTraceItem::refreshTabs" << endl;

	DataTab->_y_data_fld->reValidate();

}


// ------------------------------------------

void TraceItem::changeFrameType(int frameType)
{
	// called when Frame changes type XY <-> TS
	
	if(frameType == QP_TS)
	{
	
		DataTab->_x_label_fld->setVisible(false);
		DataTab->_x_data_fld->setVisible(false);
		DataTab->timeAsSec_toggle->setVisible(false);
        DataTab->_gap_detection_fld->setVisible(true);
        DataTab->_gap_params_container->setVisible(true);

		if(DataTab->I_PlotType == QP_CONST_XT)
		{
			// set to const t trace

			DataTab->_y_label_fld->setText("Time Constant: ");
			DataTab->_y_data_fld->setAcceptable(Qcondition(QS_NARROW, QS_TIME, QS_0D, QS_NOT_SEQ, QS_ANY));
			DataTab->_show_legend_fld->setChecked(false);
		}
		else if(DataTab->I_PlotType == QP_CONST_Y)
		{
			DataTab->_y_label_fld->setText("Y Constant: ");
			DataTab->_y_data_fld->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_0D, QS_NOT_SEQ, QS_ANY));
			DataTab->_show_legend_fld->setChecked(false);
		}
		else if(DataTab->I_PlotType == QP_REG_LINE)
		{
			DataTab->_y_label_fld->setText("Y: ");
			DataTab->_y_data_fld->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_0D, QS_SEQ, QS_ANY));
			DataTab->_show_legend_fld->setChecked(false);
		}
		else
		{
			DataTab->_y_label_fld->setText("Y: ");
			DataTab->_y_data_fld->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_0D, QS_SEQ, QS_TT));
			DataTab->_show_legend_fld->setChecked(true);
		}
	}
	else
	{
		DataTab->timeAsSec_toggle->setVisible(true);
        DataTab->_gap_detection_fld->setVisible(false);
        DataTab->_gap_params_container->setVisible(false);
		
		if(DataTab->I_PlotType == QP_CONST_XT)
		{
			DataTab->_y_label_fld->setText("X Constant: ");
			DataTab->_y_data_fld->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_0D, QS_NOT_SEQ, QS_ANY));

			DataTab->_x_label_fld->hide();
			DataTab->_x_data_fld->hide();
			DataTab->timeAsSec_toggle->hide();
			DataTab->_show_legend_fld->setChecked(false);
		}
		else if(DataTab->I_PlotType == QP_CONST_Y)
		{
			DataTab->_y_label_fld->setText("Y Constant: ");
			DataTab->_y_data_fld->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_0D, QS_NOT_SEQ, QS_ANY));
			
			DataTab->_x_label_fld->hide();
			DataTab->_x_data_fld->hide();
			DataTab->timeAsSec_toggle->hide();
			DataTab->_show_legend_fld->setChecked(false);
		}
		else if(DataTab->I_PlotType == QP_REG_LINE)
		{
			DataTab->_y_label_fld->setText("Y: ");
			DataTab->_y_data_fld->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_0D, QS_SEQ, QS_ANY));
			DataTab->_y_data_fld->addAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_1D, QS_NOT_SEQ, QS_ANY));
			
			DataTab->_x_label_fld->show();
			DataTab->_x_data_fld->show();
			DataTab->_x_data_fld->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_0D, QS_SEQ, QS_ANY));
			DataTab->_y_data_fld->addAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_1D, QS_NOT_SEQ, QS_ANY));
			DataTab->timeAsSec_toggle->show();
			DataTab->_show_legend_fld->setChecked(false);
		}
		else
		{
			DataTab->_y_label_fld->setText("Y: ");
			DataTab->_y_data_fld->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_0D, QS_SEQ, QS_ANY));
			DataTab->_y_data_fld->addAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_1D, QS_NOT_SEQ, QS_ANY));

			DataTab->_x_label_fld->show();
			DataTab->_x_data_fld->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_0D, QS_SEQ, QS_ANY));
			DataTab->_y_data_fld->addAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_1D, QS_NOT_SEQ, QS_ANY));
			DataTab->_x_data_fld->show();
			DataTab->timeAsSec_toggle->show();
			DataTab->_show_legend_fld->setChecked(false);
		}
		

	}
	DataTab->setDataTypeTitle();
	
}

// -------------------------------------------------

void TraceItem::getLabel(const QpAxes& axis, string& label)
{
	// fetch label for axis

	QdObject_var obj;
	string default_label;
	switch(axis)
	{
		case X_AXIS:
			// if time series plot default label is Universal Time
			if(frameType() == QP_TS)
			{
				label = "Universal Time";
				return;
			}
			if(DataTab->timeAsSec_toggle->isChecked())
			{
				label="Seconds";
				return;
			}
			// otherwise fetch x object and check if object needs refreshing
			label = "X Axis";
			obj = _xdo;
			if(obj.is_nil() )
			{
				if( !_ydo.is_nil() && _ydo->xref_exists(DEPEND_1) )  obj = _ydo->get_xref(DEPEND_1);
			}
			break;

		case Y_AXIS:
			// fetch y object and check if object needs refreshing
			label="Y Axis";
			obj = _ydo;
			break;

		default:
			return;
	}

	if(obj.is_nil())  return;
  
	// if object found set default label from LABLAXIS attribute if provided
	if(obj->xref_exists(LABLAXIS))
	{
  	  	QdObject_var xref=obj->get_xref(LABLAXIS);
		if(!xref.is_nil()){
  	  		QdString_var strLab = QdString_var::narrow(xref);
			if(strLab.is_nil()){
			 	QdSMatrix_var strLabMat = QdSMatrix_var::narrow(xref);
				if(!strLabMat.is_nil()){

					valarray<string> ms_data = strLabMat->data();
					default_label = ms_data[0].c_str();
					 for(register int i=1;i<ms_data.size();i++){

						default_label = QuCommonLabl(default_label, ms_data[i].c_str());
					}
				}
				else default_label = "";
			}
			else{
				default_label = strLab->c_str();
			}
		}
	}
	
	// if not use FIELDNAM attribute
	if(default_label.size()==0) default_label = QuGetAttrText(FIELDNAM, obj);
	
	// last resort use object name
	if(default_label.size()==0) default_label = QuGetAttrText(OBJECT_NAME, obj);
	if(default_label.size()!=0) label = default_label;
  
	// attach units
	string units = QuGetAttrText(UNITS, obj);
	if(units.size()!=0 && units!=" ")
	{
		label.append(" (");
		label.append(units);
		label.append(")");
	}
	// and spacecraft id
//	string scid=QuGetAttrText(SC_ID,obj);
//	if(scid.size()!=0 && scid!=" ")
//	{
//		label.append(" SC=");
//		label.append(scid);
//	}
	
	// convert theta and phi
	QString greek = QString(label.c_str());
	greek.replace("theta","#gh");
	greek.replace("phi","#gf");
	label = greek.toLatin1().data();

}


void TraceItem::getLegendText(string& legend_text)
{
	// fetch legend text
	legend_text = " "; // ensure ret string is safe
	
	if(_ydo.is_nil())  return;
  
	if(DataTab->I_PlotType == QP_REG_LINE){

		if(_xdo.is_nil()) return;
		
		if(frameType() == QP_TS){
			linFit_var = QuRegression(_ydo, _xdo);
			DataTab->_legend_text_fld->setToolTip("Slope m, Intercept b (at plot start time), Correlation Coeff r (0 bad, 1 good) ");
		}
		else{
			linFit_var = QuRegression(_ydo, _xdo);
			DataTab->_legend_text_fld->setToolTip("Slope m, Intercept b, Correlation Coeff r (0 bad, 1 good) ");
		}
		
		if(linFit_var.is_nil()) return;
		
		QString params;
		
		QdObject_var xref = linFit_var->get_xref("Slope");
		QdRScalar_var mObj = QdRScalar_var::narrow(xref);
		if( !mObj.is_nil() ) {
			params = QString("m=%1 ").arg(*mObj, 0, 'e', 2);
			if(mObj->xref_exists(UNITS)) {
				QdString_var units = QdString_var::narrow(mObj->get_xref(UNITS));
				if(!units.is_nil())  params += units->c_str();
			}
		}
		xref = linFit_var->get_xref("Intercept");
		QdRScalar_var bObj = QdRScalar_var::narrow(xref);
		if( !bObj.is_nil() ) {
			params += QString(",  b=%1 ").arg(*bObj, 0, 'e', 2);
			if(bObj->xref_exists(UNITS)) {
				QdString_var units = QdString_var::narrow(bObj->get_xref(UNITS));
				if(!units.is_nil())  params += units->c_str();
			}
		}
		
		xref = linFit_var->get_xref("CoeffCorrel");
		QdRScalar_var rObj = QdRScalar_var::narrow(xref);
		if( !rObj.is_nil() ) params += QString(",  r=%1").arg(fabs(*rObj), 0, 'e', 2);

		legend_text = string(params.toLatin1().data());

		return;
	}
	
	DataTab->_legend_text_fld->setToolTip(" ");

	
	string default_label = QuGetAttrText(LABLAXIS, _ydo);
	// if not use FIELDNAM attribute
	if(default_label.size()==0) default_label = QuGetAttrText(FIELDNAM, _ydo);
	// else use object name
	if(default_label.size()==0) default_label = QuGetAttrText(OBJECT_NAME, _ydo);
	if(default_label.size()!=0) legend_text = default_label;
	else legend_text = "";
	
	// attach units
	string units = QuGetAttrText(UNITS, _ydo);
	if(units.size()!=0 && units!=" ")
	{
		legend_text.append(" (");
		legend_text.append(units);
		legend_text.append(")");
	}

	// and spacecraft id
	string scid=QuGetAttrText(SC_ID, _ydo);
	if(scid.size()!=0 && scid!=" ")
	{
		legend_text.append(" SC=");
		legend_text.append(scid);
	}
}

void TraceItem::getFrameName(string& name)
{
	// fetch name of trace object (for display in listview)

	name = DataTab->_y_data_fld->text().toLatin1().data();
	int plotType = I_Type;
	
	if(plotType == QP_CONST_XT && frameType() == QP_TS)
	{
		QmjdTime time_val;
		if(time_val.setFromISOstring(name.c_str()))
			name="T="+string(time_val.getISOstring(0));
		else
			name="Const T Trace";
		return;
	}
	if(plotType == QP_CONST_Y )
	{
		bool okay;
		double d = QString(name.c_str()).toDouble(&okay);
		if(okay)
		{
			name = "Constant=";
			name += name.c_str();
		}
		else
			name="Const Y Trace";
		return;
	}
	if(plotType == QP_CONST_XT && frameType() == QP_XY )
	{
		bool okay;
		double d = QString(name.c_str()).toDouble(&okay);
		if(okay)
		{
			name = "Constant=";
			name += name.c_str();
		}
		else
			name="Const X Trace";
		return;
	}
	if(plotType == QP_TRACE && frameType() == QP_TS)
	{
		if(name=="")  name="TS-Trace";
		return;
	}
	if(plotType == QP_TRACE && frameType() == QP_XY)
	{
		if(name=="") name="XY-Trace";
		return;
	}
	if(plotType == QP_REG_LINE && frameType() == QP_XY)
	{
		if(name=="") name="XY-Linear Fit";
		return;
	}
	if(plotType == QP_REG_LINE && frameType() == QP_TS)
	{
		if(name=="") name="TS-Linear Fit";
		return;
	}
}

string TraceItem::getTitle()
{
	// fetch title for current trace object

	int plotType = I_Type;

	string title ="";
	
	if(linkedPlots.size() != 0)
	{
		QdObject_var dobj = QuGetObjByName( baseObjectText(DataTab->_y_data_fld->text()) );
		if(!dobj.is_nil()) title = QuGetAttrText(FIELDNAM, dobj);
	}
	else
	{
		if(!_ydo.is_nil()) title = QuGetAttrText(FIELDNAM, _ydo);
	}
	
	if(title.size() == 0)  title += DataTab->_y_data_fld->text().toLatin1().data();
	
	if(frameType() == QP_XY && plotType == QP_TRACE)
	{
		if(DataTab->timeAsSec_toggle->isChecked() ) return title;
		
		string x_title;
		if(!_xdo.is_nil())  x_title = QuGetAttrText(FIELDNAM, _xdo);
  
		if(x_title.size() == 0) x_title += DataTab->_x_data_fld->text().toLatin1().data();
		if(x_title.size() != 0 && title.size() != 0)  title += " vs ";
		title += x_title;
	}
	return title;
}

// --------------------------------

int TraceItem::getXmetaIsLog()
{
	if ( QuSlotDebug() )
	{
		printf( " QpTraceItem::getXmetaIsLog \n" );
		fflush(stdout);
	}
	
	int isLog=0;
	// find SCALETYP
		
	if (!_xdo.is_nil() && _xdo->xref_exists(SCALETYP) )
	{
		QdString_var scaleType = QdString_var::narrow(_xdo->get_xref(SCALETYP));
		if(!scaleType.is_nil() )
		{
			if(QuToUpper(scaleType->c_str()) == "LOG")  isLog = 1;
		}
	}
	return isLog;
}

// --------------------------------

void TraceItem::get_dobjs()
{
	if ( QuSlotDebug() )
	{
		printf( " QpTraceItem::get_dobjs \n" );
		fflush(stdout);
	}
	_ydo = DataTab->_y_data_fld->getSlotObject();
	if(_ydo.is_nil()) return;

	
	if(frameType() == QP_TS)
	{
		// if time series fetch timetags
		_xdo = get_timetags(_ydo);
	}
	else
	{
		if(DataTab->timeAsSec_toggle->isChecked())
		{
			QdTimeSeq_var tt = get_timetags(_ydo);
			QdTime_var tOffset = QdTime_var::narrow(DataTab->_x_data_fld->getSlotObject());
			if(tt.is_nil() || tOffset.is_nil()) _xdo = tt;
			else
			{
				QdRScalarSeq_var ss = new QdRScalarSeq(tt->sequence_size());
				for(int i=0; i<tt->sequence_size(); i++)
				{
					ss[i] = tt[i].getDiffSecs(*tOffset);
				}
				_xdo = ss;
			}
		}
		else
		{
			_xdo = DataTab->_x_data_fld->getSlotObject();

			if( _xdo.is_nil() )
			{
				if( _ydo->sequence_size() > 1)
				{
					// use DEPEND_0 for data sequence
					if(_ydo->xref_exists(DEPEND_0) )  _xdo = _ydo->get_xref(DEPEND_0);

				}
				else
				{
					// try Depend_1 for 1D array
					if( _ydo->xref_exists(DEPEND_1) )  _xdo = _ydo->get_xref(DEPEND_1);
				}
			}
		}
	}
	
	// Apply Time Interval Mask if set
	QdObject_var TImaskObj = DataTab->_mask_data_fld->getSlotObject();
	if( !TImaskObj.is_nil()) {
		QdTimeInterval_var TImask = QdTimeInterval_var::narrow(TImaskObj);
		if( !TImask.is_nil()) {
			_ydo = QuGetSubset(_ydo, TImask);
			_xdo = QuGetSubset(_xdo, TImask);
		}
	}

	return;
}

// --------------------------------

void TraceItem::updateColourBar()
{
	if ( QuSlotDebug() )
	{
		printf( " QpTraceItem::updateColourBar() \n" );
		fflush(stdout);
	}
	_cdo = TraceTab->_c_data_fld->getSlotObject();
	if(_cdo.is_nil()) return;

	
	// Apply Time Interval Mask if set
	QdObject_var TImaskObj = DataTab->_mask_data_fld->getSlotObject();
	if( !TImaskObj.is_nil()) {
		QdTimeInterval_var TImask = QdTimeInterval_var::narrow(TImaskObj);
		if( !TImask.is_nil()) {
			_cdo = QuGetSubset(_cdo, TImask);
		}
	}

	// is object joined
	if( QuIsJoined(_cdo, _ydo) == false && !_cdo.is_nil() && !_ydo.is_nil()) {
		QdObject_var joined = QuLinearJoin(_cdo, get_timetags(_ydo));
		if(!joined.is_nil()) _cdo = QdRScalarSeq_var::narrow(joined);
		if( _cdo.is_nil() ) return;
	}

	
	if( TraceTab->c_range_auto->isChecked() ){
		// set min/max
		double min, max;
	
		bool res = QuGetDataRange(_cdo, min, max);
		
		if(res)
		{
			// store values 
			TraceTab->_c_min_fld->setText(QString("%1").arg(min));
			TraceTab->_c_max_fld->setText(QString("%1").arg(max));
		}
	}
	
	// set coour array
	
	// set Label
	if(TraceTab->userColLab == false) {
		string zlabel;
	
		if (_cdo->xref_exists("LABLAXIS")) {
			zlabel = string(QuGetAttrText("LABLAXIS", _cdo));
		}
		else zlabel = string("z");
	
		if (_cdo->xref_exists("UNITS")){
			zlabel += " (";
			zlabel += string(QuGetAttrText("UNITS", _cdo));
			zlabel += ")";
		}

		TraceTab->colLabSlot->setText(zlabel.c_str());
	}

	
	return;
}

// --------------------------------

bool TraceItem::getXmetaRange(double& min, double& max)
{
	if ( QuSlotDebug() )
	{
		printf( " QpTraceItem::getXmetaRange \n" );
		fflush(stdout);
	}

  // gets x scalemin and scalemax values for this trace object
  // returns true if found
  // call with xmin and xmax set to values of range fields
  
	int plotType = I_Type;
	bool rangeFound = false;
	if(plotType == QP_CONST_Y || plotType == QP_CONST_XT)  return false;
	
	if(frameType() == QP_TS)
	{
		// if time series fetch timetags
		QdTimeSeq_var tdo = get_timetags(_ydo);
		// Use 1st and last for speed, getXRange finds true min/max if non-monotonic (slower)
		if(tdo.is_nil() || tdo->sequence_size() == 0) return false;
		min = tdo[0].getEpoch2000();
		max = tdo[tdo->sequence_size()-1].getEpoch2000();
		rangeFound = true;
	}
	else if( frameType() == QP_XY)
	{
		if( !(DataTab->timeAsSec_toggle->isChecked()) )
		{
			if(_xdo.is_nil() ) return false;
			if(_xdo->xref_exists(SCALEMIN) && _xdo->xref_exists(SCALEMAX))
			{
				QdRScalar_var minObj = QdRScalar_var::narrow(_xdo->get_xref(SCALEMIN));
				QdRScalar_var maxObj = QdRScalar_var::narrow(_xdo->get_xref(SCALEMAX));
				if(!minObj.is_nil() && !maxObj.is_nil() )
				{
					min = *minObj;
					max = *maxObj;
					rangeFound =  true;
				}
			}
		}
	}

	if (rangeFound)
	{
		// set current choice for use in y range
		_x_range_min = min;
		_x_range_max = max;
		return rangeFound;
	}
	else	return getXRange(min, max);

}

// --------------------------------

bool TraceItem::getXRange(double& min, double& max)
{
	if ( QuSlotDebug() )
	{
		fflush(stdout);
	}

  // gets x min and max values for this trace object
  // returns true if range calculated
  // call with xmin and xmax set to values of range fields
  
	int plotType = I_Type;

	if(plotType == QP_CONST_Y)  return false;

	if(plotType == QP_CONST_XT && frameType() == QP_XY)
	{
		// In this case ydo is x value
		QdRScalar_var rscalar_var = QdRScalar_var::narrow(_ydo);
		if(!rscalar_var.is_nil())
		{
			min=rscalar_var->scalar();
			max=min;
			return true;
		}
		else return false;
		
	}
	else if(plotType == QP_CONST_XT && frameType() == QP_TS)
	{
		QdTime_var time_val = QdTime_var::narrow(_ydo);
		if(!time_val.is_nil())
		{
			min = time_val->getEpoch2000();
			max = min;
			return true;
		}
		else return false;
	}
		
	if(_xdo.is_nil()) return false;

	int obj_id = _xdo->get_id();

	// check if object id has changed
	if(_x_obj_id_getXRange == obj_id)
	{
		// range has already been calculated just get previously stored values for x min and x max
		min = _x_from;
		max = _x_to;
		return true;
	}
	else
	{
        // if yes, refresh id

        _x_obj_id_getXRange = obj_id;
		bool res;

		// recalculate range for x data object
		res = QuGetDataRange(_xdo, min, max);
		
		if(res)
		{
			// store values 
			_x_from = min;
			_x_to = max;
		}
		return res;
	}
  
}

// --------------------------------

int TraceItem::getYmetaIsLog()
{
	if ( QuSlotDebug() )
	{
		printf( " QpTraceItem::getYmetaIsLog \n" );
		fflush(stdout);
	}
	
	int isLog=0;
	// find SCALETYP
		
	if (!_ydo.is_nil() && _ydo->xref_exists(SCALETYP) )
	{
		QdString_var scaleType = QdString_var::narrow(_ydo->get_xref(SCALETYP));
		if(!scaleType.is_nil() )
		{
			if(QuToUpper(scaleType->c_str()) == "LOG")  isLog = 1;
		}
	}
	return isLog;
}

//--------------------------

bool TraceItem::getYmetaRange(double& ymin, double& ymax)
{
	// gets y min and max values for all y values
	if ( QuSlotDebug() )
	{
		printf( "QpTraceItem::getYmetaRange (yobj is nil? %d)\n" ,  _ydo.is_nil() );
		fflush(stdout);
	}
	
	bool rangeFound = false;

	int plotType = I_Type;

	if(plotType == QP_CONST_XT || plotType == QP_CONST_Y) return false;
	if( _ydo.is_nil() ) return false; // can go no further
	
	if(_ydo->xref_exists(SCALEMIN) && _ydo->xref_exists(SCALEMAX) )
	{
		QdRScalar_var minObj = QdRScalar_var::narrow(_ydo->get_xref(SCALEMIN));
		QdRScalar_var maxObj = QdRScalar_var::narrow(_ydo->get_xref(SCALEMAX));
		
		if( !minObj.is_nil() && !maxObj.is_nil() )
		{
			ymin = *minObj;
			ymax = *maxObj;

			rangeFound = true;
		}
	}

	// choose default spacecraft colour
	if (_ydo->xref_exists(SC_ID) )
	{
		QdString_var sc_id_str = QdString_var::narrow(_ydo->get_xref(SC_ID));
		if(!sc_id_str.is_nil() )
		{
			if(QuToUpper(sc_id_str->c_str()) == "1" || QuToUpper(sc_id_str->c_str()) == "C1")
			{
				if(!userLineColour) TraceTab->setLineColour(C1_colour);
				if(!userSymbolColour) TraceTab->setSymbolColour(C1_colour);
			}
			if(QuToUpper(sc_id_str->c_str()) == "2" || QuToUpper(sc_id_str->c_str()) == "C2")
			{
				if(!userLineColour) TraceTab->setLineColour(C2_colour);
				if(!userSymbolColour) TraceTab->setSymbolColour(C2_colour);
			}
			if(QuToUpper(sc_id_str->c_str()) == "3" || QuToUpper(sc_id_str->c_str()) == "C3")
			{
				if(!userLineColour) TraceTab->setLineColour(C3_colour);
				if(!userSymbolColour) TraceTab->setSymbolColour(C3_colour);
			}
			if(QuToUpper(sc_id_str->c_str()) == "4" || QuToUpper(sc_id_str->c_str()) == "C4")
			{
				if(!userLineColour) TraceTab->setLineColour(C4_colour);
				if(!userSymbolColour) TraceTab->setSymbolColour(C4_colour);
			}
		}
		
	}
	if (rangeFound) return rangeFound;
	else{
		bool autoFound = getYRange(ymin, ymax);
		if (!panel()->yAxisTab->_axis_y_nudge_fld->isChecked()) return autoFound;
		
		// range is not nudged by panel getYmetaRange, so do it here
		// This is not be effective, however, if each trace in a panel 
		// has a small range relative to panel range
		if(panel()->yAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
		{
			double nudgeUp = pow(fabs(ymax / ymin), 0.01);
			double nudgeDown = pow(fabs(ymax / ymin), 0.02);
			ymax *= nudgeUp;
			ymin /= nudgeDown;
		}
		else
		{
			double nudge = fabs(ymax - ymin)*0.02;
			ymax += nudge;
			ymin -= nudge;
		}
		return autoFound;
	}
}


//--------------------------

bool TraceItem::getYRange(double& ymin, double& ymax)
{
	// gets y min and max values for all y values in range xmin - xmax
	// call with ymin and ymax values set to current values of y range fields
	if ( QuSlotDebug() )
	{
		printf( " QpTraceItem::getYRange \n" );
		fflush(stdout);
	}
	
	double xmin, xmax;
	
	if(frameType() == QP_TS)
	{
		xmin=frame()->tAxisTab->_axis_time_fld->getStart().getEpoch2000();
		xmax = frame()->tAxisTab->_axis_time_fld->getEnd().getEpoch2000();
	}
	else
	{
		xmin=frame()->xAxisTab->_axis_xy_fld->getStart();
		xmax = frame()->xAxisTab->_axis_xy_fld->getEnd();
	}
	if(xmin >= xmax) return false;

	bool recalc=false;
		
	int plotType = I_Type;

	if(plotType == QP_CONST_XT ) return false;
	if( _ydo.is_nil() ) return false; // can go no further
	  
	if(plotType == QP_CONST_Y)
	{
		QdRScalar_var rscalar_var = QdRScalar_var::narrow(_ydo);
		if(!rscalar_var.is_nil())
		{
			ymax = rscalar_var->scalar();
			ymin = rscalar_var->scalar();
			return true;
		}
		else return false;
	}
	
	// check if y object has changed
	int obj_id = _ydo->get_id();
	if(  _y_obj_id_getYRange != obj_id)
	{
		// if yes, store id
		recalc = true;
		_y_obj_id_getYRange = obj_id;
	}
 
	if(_xdo.is_nil()) return false;
	
	obj_id = _xdo->get_id();
	  
	// check if x range has changed since last call
	if(_x_range_min != xmin || _x_range_max != xmax)
	{
		_x_range_min = xmin;
		_x_range_max = xmax;
		recalc=true;
	}

	// if nothing has changed fetch range calculated last time
	if( !recalc )
	{
		ymin = _y_from;
		ymax = _y_to;
		return true;
	}
	
	// refresh y range

	// narrow x object to appropriate sequence type and fetch mask of entries in range
	QdRScalarSeq_var rscalarseq_x = QdRScalarSeq_var::narrow(_xdo);
	QdTimeSeq_var timeseq_x = QdTimeSeq_var::narrow(_xdo);
	SequenceMask msk;
	if(!rscalarseq_x.is_nil()) msk = rscalarseq_x->get_mask(QdRScalarSeq::InRange(xmin, xmax));
	if(!timeseq_x.is_nil())
	{
		msk.resize(timeseq_x->size());
		for(register int i=0; i<timeseq_x->size(); i++)
		{
			double t = timeseq_x[i].getEpoch2000();
			msk[i] = (t>=xmin && t<=xmax);
		}
	}
	
	bool res = false;
	
	// narrow y object to sequence of appropriate type and fetch subsequence using mask
	QdRScalarSeq_var ydo = QdRScalarSeq_var::narrow(_ydo);
	if(!ydo.is_nil())
	{
		if(msk.size())
		{
			QdRScalarSeq_var y_var = new QdRScalarSeq;
			ydo->get_subsequence(*y_var, msk);
			ydo = y_var;
		}
		// get range of values in subsequence
		res = QuGetDataRange(ydo, ymin, ymax);
	}
	else
	{
		// or narrow to matrix (but then cannot use msk)
		QdRMatrix_var ydm = QdRMatrix_var::narrow(_ydo);
		if(!ydm.is_nil()) res = QuGetDataRange(ydm, ymin, ymax);
		else
		{
			QdRMatrixSeq_var ydms = QdRMatrixSeq_var::narrow(_ydo);
			if(!ydms.is_nil()) res = QuGetDataRange(ydms, ymin, ymax);
		}
	}
	// store results if min and max found
	if(res)
	{
		_y_from = ymin;
		_y_to = ymax;
	}
	return res;
}


//--------------------------

QpuiReturnStatus TraceItem::plot()
{
	if(QuSlotDebug())
	{
		printf("QpTraceItem::plot \n");
		fflush(stdout);
	}
	int plotType = I_Type;
		
	// plots the Trace Object
	PLFLT *x_data=0;
	PLFLT *y_data=0;
	PLFLT *c_data=0;
	
	RScalarSeq x_diff;
	int n_data=0;
	
	PLINT line_colour_index = getPGPColourIndex(_line_colour);
	PLINT symbol_colour_index = getPGPColourIndex(_symbol_colour);
	
	// Refresh y data object
	if(_ydo.is_nil())
	{
		plmtex("lv",-10.0,0.5,0.0,"Trace Plot ERROR: Null Y Object");
		return QPUI_FAILURE;
	}
	
	if(!DataTab->_y_data_fld->isValid())
	{
		plmtex("lv",-10.0,0.5,0.0,"Trace Plot ERROR: Invalid Y Slot");
		return QPUI_FAILURE;
	}
	
	// find end of panel for text direction
	PLFLT x_end;
	if(frameType() == QP_TS)
	{
		QmjdTime t_plot_start=frame()->tAxisTab->_axis_time_fld->getStart();
		QmjdTime t_plot_end=frame()->tAxisTab->_axis_time_fld->getEnd();

		x_end = (PLFLT)t_plot_end.getDiffSecs(t_plot_start);
	}
	else
	{
		x_end = frame()->xAxisTab->_axis_xy_fld->getEnd();
		if(frame()->xAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
		{
			x_end = SafeLog10(x_end);
		}
	}

	if(plotType == QP_CONST_XT && frameType() == QP_TS)
	{
		QdTimeInterval_var tint = get_timeinterval(_ydo);
		QdTime_var t_plot = QdTime_var::narrow(_ydo);
		QmjdTime t_plot_start=frame()->tAxisTab->_axis_time_fld->getStart();

		if(!tint.is_nil())
		{
			x_data  =new PLFLT[4];
			y_data = new PLFLT[4];
			y_data[0] = y_data[3] = panel()->yAxisTab->_axis_xy_fld->getEnd(); 
			y_data[1] = y_data[2] = panel()->yAxisTab->_axis_xy_fld->getStart();
			if(panel()->yAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
			{
				y_data[0] = SafeLog10(y_data[0]);
				y_data[1] = SafeLog10(y_data[1]);
				y_data[2] = SafeLog10(y_data[2]);
				y_data[3] = SafeLog10(y_data[3]);
			}
			n_data = 4;

			x_data[0] = x_data[1] = (PLFLT)tint->start().getDiffSecs(t_plot_start);
			x_data[2] = x_data[3] = (PLFLT)tint->end().getDiffSecs(t_plot_start);
		}
		else
		{
			QmjdTime time_val;
			if(t_plot.is_nil())
			{
				plmtex("lv",-10.0,0.5,0.0,"Trace Plot ERROR: Invalid constant time");
				return QPUI_FAILURE;
			}
			else time_val=*t_plot;
		 
			PLFLT t_constant = (PLFLT)time_val.getDiffSecs(t_plot_start);
			x_data = new PLFLT[2];
			x_data[0] = x_data[1] = t_constant;
			y_data = new PLFLT[2];
			y_data[0] = panel()->yAxisTab->_axis_xy_fld->getStart();
			y_data[1] = panel()->yAxisTab->_axis_xy_fld->getEnd();
			if(panel()->yAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
			{
				y_data[0] = SafeLog10(y_data[0]);
				y_data[1] = SafeLog10(y_data[1]);
			}
			n_data=2;
		}
//         nominal_spacing=1.5*(double)x_diff[1];
	}
	else if(plotType == QP_CONST_Y)
	{
		PLFLT y_constant;
		QdRScalar_var rscalar_var = QdRScalar_var::narrow(_ydo);
		if(rscalar_var.is_nil())
		{
			QdRScalarSeq_var rscalarseq_var = QdRScalarSeq_var::narrow(_ydo);
			if(!rscalarseq_var.is_nil() && rscalarseq_var->sequence_size() == 1) rscalar_var = new QdRScalar(rscalarseq_var[0]);
		}

		if(rscalar_var.is_nil()) return QPUI_FAILURE;
		y_constant = (PLFLT)rscalar_var->scalar();
		
		n_data=2;
		y_data = new PLFLT[2];
		y_data[0] = y_data[1] = y_constant;
		if(panel()->yAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
		{
			y_data[0] = SafeLog10(y_data[0]);
			y_data[1] = SafeLog10(y_data[1]);
		}
		x_data = new PLFLT[2];
		if(frameType() == QP_TS)
		{
			QmjdTime t_plot_start=frame()->tAxisTab->_axis_time_fld->getStart();

			QmjdTime t_plot_end=frame()->tAxisTab->_axis_time_fld->getEnd();

			x_data[0] = (PLFLT)0.0;
			x_data[1] = (PLFLT)t_plot_end.getDiffSecs(t_plot_start);
		}
		else
		{
			x_data[0] = frame()->xAxisTab->_axis_xy_fld->getStart();
			x_data[1] = frame()->xAxisTab->_axis_xy_fld->getEnd();
			if(frame()->xAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
			{
				x_data[0] = SafeLog10(x_data[0]);
				x_data[1] = SafeLog10(x_data[1]);
			}
		}
	}
	else if(plotType == QP_CONST_XT && frameType() == QP_XY)
	{
		PLFLT x_constant;
		QdRScalar_var rscalar_var = QdRScalar_var::narrow(_ydo);
		if(rscalar_var.is_nil()) return QPUI_FAILURE;
		x_constant = (PLFLT)rscalar_var->scalar();
		
		n_data=2;
		x_data = new PLFLT[2];
		x_data[0] = x_data[1] = x_constant;
		y_data = new PLFLT[2];
		y_data[0] = panel()->yAxisTab->_axis_xy_fld->getStart();
		y_data[1] = panel()->yAxisTab->_axis_xy_fld->getEnd();
		if(frame()->xAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
		{
			x_data[0] = SafeLog10(x_data[0]);
			x_data[1] = SafeLog10(x_data[1]);
		}
		if(panel()->yAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
		{
			y_data[0] = SafeLog10(y_data[0]);
			y_data[1] = SafeLog10(y_data[1]);
		}
	}
	else if(plotType == QP_REG_LINE && frameType() == QP_XY)
	{
		if(linFit_var.is_nil()) linFit_var = QuRegression(_ydo, _xdo);
		if(linFit_var.is_nil()) return QPUI_FAILURE;
		
		QdRScalarSeq_var dep0 = QdRScalarSeq_var::narrow(get_Depend_0(linFit_var));
		QdTimeSeq_var tt = QdTimeSeq_var::narrow(get_Depend_0(linFit_var));

		if(dep0.is_nil()) return QPUI_FAILURE;

		n_data=2;
		x_data = new PLFLT[2];
		x_data[0] = (PLFLT) dep0[0];
		x_data[1] = (PLFLT) dep0[1];
		y_data = new PLFLT[2];
		y_data[0] = (PLFLT) linFit_var[0];
		y_data[1] = (PLFLT) linFit_var[1];
		if(frame()->xAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
		{
			x_data[0] = SafeLog10(x_data[0]);
			x_data[1] = SafeLog10(x_data[1]);
		}
		if(panel()->yAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
		{
			y_data[0] = SafeLog10(y_data[0]);
			y_data[1] = SafeLog10(y_data[1]);
		}
		
		
	}
	if(plotType == QP_REG_LINE && frameType() == QP_TS)
	{
		if(linFit_var.is_nil()) linFit_var = QuRegression(_ydo, _xdo);
		if(linFit_var.is_nil()) return QPUI_FAILURE;
		
		QdTimeSeq_var tt = QdTimeSeq_var::narrow(get_timetags(linFit_var));
		if(tt.is_nil()) return QPUI_FAILURE;

		QmjdTime t_plot_start=frame()->tAxisTab->_axis_time_fld->getStart();

		QmjdTime time_val;
		 
		n_data=2;
		x_data = new PLFLT[2];
		x_data[0] = (PLFLT) tt[0].getDiffSecs(t_plot_start);
		x_data[1] = (PLFLT) tt[1].getDiffSecs(t_plot_start);
			
		y_data = new PLFLT[2];
		y_data[0] = (PLFLT) linFit_var[0];
		y_data[1] = (PLFLT) linFit_var[1];
		
		if(panel()->yAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
		{
			y_data[0] = SafeLog10(y_data[0]);
			y_data[1] = SafeLog10(y_data[1]);
		}
	}
	else if(plotType == QP_TRACE && frameType() == QP_TS)
	{
		QdTimeSeq_var tdo = QdTimeSeq_var::narrow(_xdo);
		if(tdo.is_nil() )
		{
			plmtex("lv",-10.0,0.5,0.0,"Trace Plot ERROR: Null T Object");
			return QPUI_FAILURE;
		}
		// fetch number of elements in each sequence/array
		n_data = QuLength(_ydo);
		int nt_data = tdo->sequence_size();

		if(nt_data != n_data)
		{
			plmtex("lv",-10.0,0.5,0.0,"ERROR: Y and T series of different lengths");
			return QPUI_FAILURE;
		}
		
		// create float c-style arrays for plotting
		x_data = new PLFLT[n_data];
		y_data = new PLFLT[n_data];
		x_diff = RScalarSeq(n_data);

		QmjdTime ts= frame()->tAxisTab->_axis_time_fld->getStart();
		QdTime_var t_plot_start=new QdTime(ts);
		
		// get y object data
		if(!QuSet_Array(n_data, &(y_data[0]), _ydo, t_plot_start))
		{
			plmtex("lv",-10.0,0.5,0.0,"ERROR: Y Object not an Scalar, ScalarSeq or 1D matrix");
			return QPUI_FAILURE;
		}

		QmjdTime half_spacing_t;
//         if(0 == calc_tol_spacing(half_spacing_t, *tdo)) nominal_spacing = half_spacing_t.getEpoch2000();
//         if(0!=calc_tol_spacing(nominal_spacing, *tdo, QSAS_profiles->current()->gapTolerance, 0.01))
//         {
//             std::cout << "nominal" << std::endl;
//             nominal_spacing=DBL_MAX;
//         }
//         std::cout << "nom " << nominal_spacing << std::endl;

		
		// loop through data and set x and y arrays
		double x_diff_sum = 0.;
		for(int i=0; i<n_data; i++)
		{
			// set time in number of seconds from t_midnite
			double s_i = tdo[i].getDiffSecs(ts);
			if(i==0) x_diff[0] = s_i;
			else x_diff[i] = s_i - x_diff_sum;
			x_diff_sum += x_diff[i];
			x_data[i] = (PLFLT)s_i;
		}
		
		// calculate logarithmic values if logs selected
		if(panel()->yAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG) 
		{
			for(int i=0; i<n_data; i++) y_data[i] = (PLFLT) SafeLog10((double)y_data[i]);
		}
	}
	else if(plotType == QP_TRACE && frameType() == QP_XY)
	{
		// Data Series (XY trace)
		
		// fetch number of elements in each sequence/array
		n_data = QuLength(_ydo);
		int nx_data = QuLength(_xdo);

		if(nx_data != n_data)
		{
			plmtex("lv",-10.0,0.5,0.0,"ERROR: Y and X series of different lengths");
			return QPUI_FAILURE;
		}

		// create float c-style arrays for plotting
		x_data = new PLFLT[n_data];
		y_data = new PLFLT[n_data];

		// get y data
		if(!QuSet_Array(n_data, &(y_data[0]), _ydo))
		{
			plmtex("lv",-10.0,0.5,0.0,"ERROR: Y Object not an Scalar, ScalarSeq or 1D matrix");
			return QPUI_FAILURE;
		}
		
		// get x data
		QdRScalarSeq_var rscalarseq_x = QdRScalarSeq_var::narrow(_xdo);
		if(!QuSet_Array(n_data, &(x_data[0]), _xdo))
		{
			plmtex("lv",-10.0,0.5,0.0,"ERROR: X Object not an Scalar, ScalarSeq or 1D Matrix");
			return QPUI_FAILURE;
		}

		// loop through data and set x logs 
		if(frame()->xAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG) 
		{
			for(int i=0; i<n_data; i++) x_data[i] = SafeLog10(x_data[i]);
		
		}
				
		// loop through data and set y logs 
		if(panel()->yAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG) 
		{
			for(int i=0; i<n_data; i++) y_data[i] = SafeLog10(y_data[i]);
		
		}

	} // End if else over trace types

	// Set colour data if available
	if( !_cdo.is_nil() ){

		c_data = new PLFLT[n_data];
		TraceTab->colourTable->getCurrentTable().apply();

		
		bool OKmin, OKmax;
		PLFLT min = (PLFLT) TraceTab->_c_min_fld->text().toFloat(&OKmin);
		PLFLT max = (PLFLT) TraceTab->_c_max_fld->text().toFloat(&OKmax);
		
		PLFLT range = max - min;
		QdRScalarSeq_var obj_c = QdRScalarSeq_var::narrow(_cdo);
		
		if(obj_c.is_nil() || _cdo->sequence_size() != _ydo->sequence_size() ){
			plmtex("lv",-10.0,0.5,0.0,"ERROR: Colour Object not compatible length");
			for(int i=0; i<n_data; i++){
				c_data[i] = 0.5; 
			}
		}
		else{
						
			if(!OKmin || !OKmax || range <= 0.){
				// find range from data as fallback
				min = 1.e10;
				max = -1.e10;
				for(int i=0; i<n_data; i++){
					if(min > obj_c[i]) min = obj_c[i];
					if(max < obj_c[i]) max = obj_c[i];
				}
				range = max - min;
				if (range <= 0.) {
					plmtex("lv",-10.0,0.5,0.0,"ERROR: Colour Object range negative");
					for(int i=0; i<n_data; i++){
						c_data[i] = 0.5; 
					}
				}
			}

			if(TraceTab->c_range_log->isChecked()){
				min = SafeLog10(min);
				max = SafeLog10(max);
				range = max - min;
				for(int i=0; i<n_data; i++){
					c_data[i] = (SafeLog10(obj_c[i]) - min) / range;
					// out of range is supposed to be safe, but is not always
					if(c_data[i] < 0.) c_data[i] = 0.;
					if(c_data[i] > 1.) c_data[i] = 1.;
				}
			}
			else{
				for(int i=0; i<n_data; i++){
					c_data[i] = (obj_c[i] - min) / range;
					// out of range is supposed to be safe, but is not always
					if(c_data[i] < 0.) c_data[i] = 0.;
					if(c_data[i] > 1.) c_data[i] = 1.;
				}
			}
		}
		// set colour bar
		if(TraceTab->_show_colour_bar->isChecked())
		{
			PLFLT x1win_save, x2win_save, y1win_save, y2win_save;
			PLFLT x1vp_save, x2vp_save, y1vp_save, y2vp_save;
			PLFLT x[4], y[4];

			plgvpw(&x1win_save, &x2win_save, &y1win_save, &y2win_save);
			plgvpd(&x1vp_save, &x2vp_save, &y1vp_save, &y2vp_save);
	
			PLFLT chsize_ndc;
		
			chsize_ndc=getCharHeight();
	
			x[0] = x2vp_save + TraceTab->_colour_bar_disp->value() * chsize_ndc*1.5;
			x[2] = x[0] +  TraceTab->_colour_bar_width->value() * chsize_ndc;
			float ht = (y2vp_save - y1vp_save) ;
			y[0] = y1vp_save + ht * TraceTab->_colour_bar_voffset->value();
			y[1] = y[0] + ht * TraceTab->_colour_bar_height->value();
	
			plvpor(x[0],x[2], y[0], y[1]);
		
			plwind(0., 1., min, max);
	
			x[0]=0.;
			x[1]=0.;
			x[2]=1.;
			x[3]=1.;

			PLFLT zstep=(max-min)/200.;
			PLFLT i=0.;
			for(PLFLT col=min;col<max;col+=zstep, i+=1./200.)
			{
				y[0] = col;
				y[3] = col;
				y[1] = col+zstep;
				y[2] = y[1];
				if(i>=1.) i=0.999;
				plcol1(i);
				plfill(4, x, y);
			}
	
			plcol0(1);
			int plot_id=page()->PageTab->_xwin_num_fld->text().toInt();
			page()->setSmooth(plot_id, true);
			string yopt = "bcmtsv";
			if(TraceTab->c_range_log->isChecked()) yopt += "l";
			plschr(0., frame()->DisplayTab->_ticks_char_size_fld->value());
		
			plwid(QtWidth2PlWidth(frame()->DisplayTab->_line_width_fld->value()));

			PLFLT _ticks_length = TraceTab->_colourscale_ticklength->value();
		
			plsmaj(0., _ticks_length);
			plsmin(0., _ticks_length*0.6);

			PLLabelDefaults y_label_opt;
			y_label_opt.exp_label_disp = 2.0 + (PLFLT) TraceTab->_exp_label_offset_y->value();
			y_label_opt.exp_label_pos = 1.02 +(PLFLT) TraceTab->_exp_label_offset_x->value();
			y_label_opt.exp_label_just = 0.0;
			plslabelfunc(NULL, &y_label_opt);
			plbox("bc", 0.0, 0, yopt.c_str(), 0.0, 0);
			plslabelfunc(NULL, NULL);
        
			string label="";
	
			plwid(0);
		
			label = TraceTab->colLabSlot->text().toLatin1().data();
		
			plmtex("l", 1., 0.5, 0.5, SafePGPString(label).c_str());
			plvpor(x1vp_save, x2vp_save, y1vp_save, y2vp_save);
			plwind(x1win_save, x2win_save, y1win_save, y2win_save);
			page()->setSmooth(plot_id, false);
		}

	}
	
	int i=0;
	while (i < n_data)
	{
		int j=i;
		i++;
		if(plotType == QP_TRACE && frameType() == QP_TS)
		{
			// break into sub-sequences on gaps
            if(DataTab->gap_threshold>0)
            {
                while(i < n_data && (double)x_diff[i] <= DataTab->gap_threshold && x_diff[i] > 0. )
			    {
				    i++;  // increment i
			    }
            }
            else
            {
                i = n_data;
            }
		}
		else if( plotType == QP_CONST_XT && frameType() == QP_TS)
		{
			i++;
		}
		else i = n_data; // most trace types ignore gaps
	
		plwid(0); // force reset
		// calculate no. points in this sub-sequence with no gaps
		int n_pts = i-j;
		if(TraceTab->_line_style_fld->currentIndex() != QP_NOLINE)
		{
			plwid(QtWidth2PlWidth(TraceTab->_line_width_fld->value()));
			setDashStyle((QpLineStyle)TraceTab->_line_style_fld->currentIndex(), TraceTab->_line_scale_fld->value());
			plcol0(line_colour_index);
			// draw line between points of sub-sequence of appropriate time
			switch(TraceTab->_connect_style_fld->currentIndex())
			{
				case QPUI_STRAIGHT: default:
					plline(n_pts, x_data+j, y_data+j);
					break;
				case QPUI_HIST:
					plbin(n_pts, x_data+j, y_data+j, PL_BIN_CENTRED|PL_BIN_NOEXPAND);
					break;
				case QPUI_NOCONNECT:
					break;
			}
		}
		
		if(TraceTab->_symbol_shape_fld->currentIndex() == QPUI_NOSYMBOL && n_pts==1)
		{
			// ensure single point is visible!
			plssym(0., TraceTab->_line_width_fld->value());
			plcol0(line_colour_index);
			// plot symbols in sub-sequence
			Q_plpoin(1, x_data+j, y_data+j, (PLINT)QPUI_POINT);
		}
		
		if(TraceTab->_symbol_shape_fld->currentIndex() != QPUI_NOSYMBOL)
		{
			PLINT index=symbolIndexConversion(TraceTab->_symbol_shape_fld->currentIndex());
			if(index==QPUI_CUSTOM_SHAPE)
			{
				plschr(0., TraceTab->_symbol_size_fld->value());
				plcol0(symbol_colour_index);
				
				for(int i=0; i<n_pts; ++i)
				{
					if( !(TraceTab->_symbol_colour_fld->isEnabled())  && TraceTab->_colour_on->isChecked() && c_data) plcol1(c_data[j+i]);
					 if(TraceTab->_size_on->isChecked() && c_data) 
					 plschr(0., TraceTab->_symbol_size_fld->value()*2.*c_data[j+i]);

					plptex(x_data[j+i], y_data[j+i], x_end, 0., 0.5, TraceTab->_custom_symbol_fld->text().toLatin1().data());
				}
			}
			else
			{
				plssym(0., TraceTab->_symbol_size_fld->value());
				plcol0(symbol_colour_index);
				// plot symbols in sub-sequence
				for(int i=0; i<n_pts; ++i)
				{
					if( !(TraceTab->_symbol_colour_fld->isEnabled()) && TraceTab->_colour_on->isChecked() && c_data ) plcol1(c_data[j+i]);
					if(TraceTab->_size_on->isChecked() && c_data ) plssym(0., TraceTab->_symbol_size_fld->value()*2.*c_data[j+i]);

					Q_plpoin(1, x_data+j+i, y_data+j+i, symbolIndexConversion(TraceTab->_symbol_shape_fld->currentIndex()));
				}
			}
		}
	} // repeat until all data exhausted..

	plcol0(0);
	delete x_data;
	delete y_data;
	if(c_data != 0) delete c_data;
	return QPUI_SUCCESS;
}


QpuiReturnStatus TraceItem::plotLegend()
{
    if( !(DataTab->_show_legend_fld->isChecked()) || !(panel()->LegendTab->_show_legend_fld->isChecked()) ) return QPUI_SUCCESS;
 
   
	PLINT line_colour_index = getPGPColourIndex(_line_colour);
	PLINT symbol_colour_index = getPGPColourIndex(_symbol_colour);
    
	PLFLT x0win_save, x1win_save, y0win_save, y1win_save;
	PLFLT x0vp_save, x1vp_save, y0vp_save, y1vp_save;
    
	plgvpw(&x0win_save, &x1win_save, &y0win_save, &y1win_save);
	plgvpd(&x0vp_save, &x1vp_save, &y0vp_save, &y1vp_save);
    
//     QPlot_window->save();
    
	int row_num, col_num, num_cols, num_rows;
    
    row_num = panel()->_legend_entry / panel()->LegendTab->_legend_num_cols_fld->value();
    col_num = panel()->_legend_entry - row_num * panel()->LegendTab->_legend_num_cols_fld->value();
    
//     PLFLT xch, ych; 
    plschr(0., panel()->LegendTab->_legend_char_size_fld->value());
//     plgchr(&xch, &ych);
    
    float legend_width, legend_height;
    legend_width = panel()->LegendTab->_legend_width_fld->value() * (x1vp_save - x0vp_save);
    legend_height = panel()->LegendTab->_legend_height_fld->value() * (y1vp_save - y0vp_save);

    num_cols = panel()->LegendTab->_legend_num_cols_fld->value();
    num_rows = panel()->LegendTab->_legend_num_rows_fld->value();
    float _legend_cell_width, _legend_cell_height;
    _legend_cell_width = legend_width / num_cols;
    _legend_cell_height = legend_height / num_rows;
    
    float x0_vp, x1_vp, y0_vp, y1_vp;
    x0_vp = x0vp_save + panel()->LegendTab->_legend_loffset_fld->value() * (x1vp_save - x0vp_save);
    x0_vp += col_num * _legend_cell_width ;
    x1_vp = x0_vp + _legend_cell_width;
    y1_vp = y1vp_save - panel()->LegendTab->_legend_toffset_fld->value() * (y1vp_save - y0vp_save);
    y1_vp -= row_num * _legend_cell_height;
    y0_vp = y1_vp - _legend_cell_height -0.001; 

    plvpor(x0_vp, x1_vp, y0_vp, y1_vp);
    plwind(0.,1.,0.,1.);
	plscol0(0, 255, 255, 255);
	plscol0(1, 0, 0, 0);
    plcol0(0);
    if((QpuiLegendBackground)panel()->LegendTab->_legend_background_fld->currentIndex() == QPUI_OPAQUE)
    {
		plwid(0);
	    PLFLT x_fill[]={0., 0., 1., 1.};
	    PLFLT y_fill[]={0., 1., 1., 0.};
	    plfill(4, x_fill, y_fill);

    }
    plcol0(1);
    
    float len, marg;
    len = panel()->LegendTab->_legend_line_length_fld->value();
    marg = 0.1*len;
    
	PLFLT x_lgd[2];
	x_lgd[0] = marg;
	x_lgd[1] = marg + len;
	PLFLT y_lgd[2] = {0.,0.};
	float txt_lgd = 0.;
	float ych_cell_height = getCharHeight();/*ych;*/
	if( _legend_cell_height > 1e-4) ych_cell_height /=  _legend_cell_height;
	if( ych_cell_height > 1.) ych_cell_height = 1.;
    
    switch((QpuiLegendJust)panel()->LegendTab->_legend_just_fld->currentIndex())
    {
        case QPUI_VCENTRE: default:
            txt_lgd = 0.5;
            break;
        case QPUI_TOP:
            txt_lgd = 1. - ych_cell_height*1.2;
            break;
        case QPUI_BOTTOM:
		txt_lgd = 0.00 + ych_cell_height*1.2;
            break;
    }

    
    y_lgd[1] = txt_lgd;
    y_lgd[0] = y_lgd[1];
	   
	plschr(0., panel()->LegendTab->_legend_char_size_fld->value());
	setDashStyle((QpLineStyle)TraceTab->_line_style_fld->currentIndex());
	if(TraceTab->_line_style_fld->currentIndex() != QP_NOLINE)
	{
		plwid(QtWidth2PlWidth(TraceTab->_line_width_fld->value()));
		plcol0(line_colour_index);
    }
	else
	{
		plwid(0);
		plcol0(0);
	}
    // for some reason (plp bug?) the line must be plotted for the symbol and text to show!
	plline(2, x_lgd, y_lgd);
    
    if(TraceTab->_symbol_shape_fld->currentIndex() != QPUI_NOSYMBOL)
    {
			PLINT index=symbolIndexConversion(TraceTab->_symbol_shape_fld->currentIndex());

			if(index==QPUI_CUSTOM_SHAPE)
			{
				plschr(0, TraceTab->_symbol_size_fld->value());
				plcol0(symbol_colour_index);
				PLFLT x=0.5*(x_lgd[0]+x_lgd[1]);
				PLFLT y=y_lgd[0];
				plptex(x, y, 1., 0., 0.5, TraceTab->_custom_symbol_fld->text().toLatin1().data());
			}
			else
			{
				plssym(0., TraceTab->_symbol_size_fld->value());
				plcol0(symbol_colour_index);
				PLFLT x=0.5*(x_lgd[0]+x_lgd[1]);
				PLFLT y=y_lgd[0];
				plpoin(1, &x, &y, symbolIndexConversion(TraceTab->_symbol_shape_fld->currentIndex()));
			}

    }
    plcol0(1);
    
    int _text_line_width = 1;

	plschr(0., panel()->LegendTab->_legend_char_size_fld->value());
    PLINT family, style, weight;
#if QT_VERSION >= 0x040300
    panel()->LegendTab->_legend_font_fld->getFontParams(family, style, weight);
    plsfont(family, style, weight);
#else
    family=panel()->LegendTab->_legend_font_fld->getFontIndex();
    plfont(family);
#endif

	
	plwid(QtWidth2PlWidth(_text_line_width));
	string legend_text = DataTab->_legend_text_fld->text().toLatin1().data();

    legend_text = SafePGPString(legend_text);

    QString greek = QString(legend_text.c_str());
    greek.replace("theta","#gh");
    greek.replace("phi","#gf");
    legend_text = greek.toLatin1().data();
    
    if(legend_text.size() > panel()->LegendTab->_legend_max_char_length_fld->value())
	    legend_text = legend_text.substr(0, panel()->LegendTab->_legend_max_char_length_fld->value());

    plptex(x_lgd[1]+marg, txt_lgd, 1., 0., 0., legend_text.c_str());
    panel()->_legend_entry++;
    plscol0(0,255,255,255);
    plscol0(1,0,0,0);
    plcol0(1);

//     QPlot_window->unsave();
	plvpor(x0vp_save, x1vp_save, y0vp_save, y1vp_save);
	plwind(x0win_save, x1win_save, y0win_save, y1win_save);
    
	return QPUI_SUCCESS;
}

QpItemType QpRemapPlotType(int _type, QpFrameType &frameType){

	switch(_type){
		case 0:
			frameType = QP_XY;
			return QP_TRACE;
			break;
		
		case 1: default:
			frameType = QP_TS;
			return QP_TRACE;
			break;
		
		case 2:
			frameType = QP_TS;
			return QP_SPECTROGRAM;
			break;

		case 3:
			frameType = QP_XY;
			return QP_SPECTROGRAM;
			break;

		case 4:
			frameType = QP_TS;
			return QP_CONST_XT;
			break;

		case 5:
			frameType = QP_XY;
			return QP_CONST_Y;
			break;
		
		case 6:
			frameType = QP_XY;
			return QP_CONST_XT;
			break;
		
		case 7:
			frameType = QP_XY;
			return QP_TRACE;
			break;
		
		case 8:
			frameType = QP_TS;
			return QP_TRACE;
			break;
		
		case 9:
		// unknown type flag
			return QP_TRACE;
			break;
		
		case 10:
			frameType = QP_XY;
			return QP_EVENT;
			break;
		
		case 11:
			frameType = QP_TS;
			return QP_EVENT;
			break;
   
	}
   
}

// ----------------------------------

QDomElement TraceItem::toXMLElement(QDomDocument& doc) const
{
    QDomElement xml=doc.createElement("Plot_Item");
    xml.appendChild(createTextXMLNode(doc, "Name", getName()));

    switch(I_Type)
    {
        case QP_CONST_Y:
            xml.setAttribute("Type", "Const_Y");
            break;
        case QP_CONST_XT:
            xml.setAttribute("Type", "Const_XT");
            break;
        case QP_REG_LINE:
            xml.setAttribute("Type", "Reg_Line");
            break;
        case QP_TRACE: default:
            xml.setAttribute("Type", "Trace");
            break;
    }

    xml.appendChild(DataTab->toXMLElement(doc, I_Type));
    
    QString linkText=getLinkText();
    if(!linkText.isEmpty())
    {
        linkText += "S"; // to avoid conflict with existing links on restore
    }
    xml.appendChild(createTextXMLNode(doc, "Link_Text", linkText));

    xml.appendChild(TraceTab->toXMLElement(doc));

    xml.appendChild(createTextXMLNode(doc, "User_Line_Colour", userLineColour));
    xml.appendChild(createTextXMLNode(doc, "Line_Colour_Red", _line_colour.red()));
    xml.appendChild(createTextXMLNode(doc, "Line_Colour_Green", _line_colour.green()));
    xml.appendChild(createTextXMLNode(doc, "Line_Colour_Blue", _line_colour.blue()));

    xml.appendChild(createTextXMLNode(doc, "User_Symbol_Colour", userSymbolColour));
    xml.appendChild(createTextXMLNode(doc, "Symbol_Colour_Red", _symbol_colour.red()));
    xml.appendChild(createTextXMLNode(doc, "Symbol_Colour_Green", _symbol_colour.green()));
    xml.appendChild(createTextXMLNode(doc, "Symbol_Colour_Blue", _symbol_colour.blue()));
    
    return xml;
}

void TraceItem::fromXMLElement(QDomElement& el)
{
    QString s;
	
 	if(QuSlotDebug()) cout <<  "TraceItem restore from XML (>>)" << endl;
   
    if(getXMLNodeText(el, "Name", s))
    {
        setName(s);
    }

    QDomElement dataElement=el.firstChildElement("Data_Tab");
    if(!dataElement.isNull())
    {
        DataTab->fromXMLElement(dataElement, I_Type);
    }

    QDomElement traceElement=el.firstChildElement("Trace_Tab");
    if(!traceElement.isNull())
    {
        TraceTab->fromXMLElement(traceElement);
    }
    
    if(getXMLNodeText(el, "User_Line_Colour", s))
    {
        userLineColour=s.toInt();
    }
    
    if(getXMLNodeText(el, "Link_Text", s))
    {
        setLinkText(s);
    }
    

    int r(0),g(0),b(0);
    if(getXMLNodeText(el, "Line_Colour_Red", s))
    {
        r=s.toInt();
    }
    if(getXMLNodeText(el, "Line_Colour_Green", s))
    {
        g=s.toInt();
    }
    if(getXMLNodeText(el, "Line_Colour_Blue", s))
    {
        b=s.toInt();
    }

    _line_colour=QColor(r,g,b);
    QPalette line_palette = QPalette(QPUI_normal_palette);
    line_palette.setColor(QPalette::Active ,QPalette::ButtonText, _line_colour);
    line_palette.setColor(QPalette::Inactive, QPalette::ButtonText, _line_colour);
    TraceTab->_line_colour_fld->setPalette(line_palette);

    if(getXMLNodeText(el, "User_Symbol_Colour", s))
    {
        userSymbolColour=s.toInt();
    }
	if(getXMLNodeText(el, "Symbol_Colour_Red", s))
    {
        r=s.toInt();
    }
    if(getXMLNodeText(el, "Symbol_Colour_Green", s))
    {
        g=s.toInt();
    }
    if(getXMLNodeText(el, "Symbol_Colour_Blue", s))
    {
        b=s.toInt();
    }
    
    _symbol_colour=QColor(r,g,b);
    line_palette.setColor(QPalette::Active ,QPalette::ButtonText, _symbol_colour);
    line_palette.setColor(QPalette::Inactive, QPalette::ButtonText, _symbol_colour);
    TraceTab->_symbol_colour_fld->setPalette(line_palette);

}

Vistream& operator>>(Vistream& v_is, TraceItem& trace)
{
	if(QuSlotDebug()) cout <<  "TraceItem restore (>>)" << endl;

	// streams from virtual input stream v_is into trace object

	PlotItem& plot_item = trace;
	v_is >> plot_item;
	
	QdObject_var o=QuGetObjByName(trace._name);
	if(!o.is_nil()) trace.WLcontext=trace._name;
	
	string _y_data;
	string _x_data;
	v_is >> _y_data;
	v_is >> _x_data;

	trace.DataTab->_x_data_fld->setText(_x_data.c_str());
	trace.DataTab->_y_data_fld->setText(_y_data.c_str());
	trace.DataTab->old_BaseName = baseObjectText(_y_data.c_str());
	trace.DataTab->old_YComp = compText(_y_data.c_str());
	trace.DataTab->_y_data_fld->reValidate(false); // don't propagate change, just set colour
	
    if( qsas_restore_release >= 20400)
    {
        bool gapDetectionChecked;
        v_is >> gapDetectionChecked;
        trace.DataTab->_gap_detection_fld->setChecked(gapDetectionChecked);
        double gapValue;
        v_is >> gapValue;
        trace.DataTab->_gap_threshold_fld->setValue(gapValue);
        int gapType;
        v_is >> gapType;
        trace.DataTab->_gap_threshold_type_fld->setCurrentIndex(gapType);
        bool useTimeAsSec;
        v_is >> useTimeAsSec;
        trace.DataTab->timeAsSec_toggle->setChecked(useTimeAsSec);
		trace.DataTab->updateGapThreshold();
    }
  
	string linkText;
	v_is >> linkText;
	trace.setLinkText(QString(linkText.c_str()));
	// do the relinking in listview restore after traces etc are there
	
	int _connect_style;
	v_is >> _connect_style;
	trace.TraceTab->_connect_style_fld->setCurrentIndex(_connect_style);
	
	int _line_style;
	v_is >> _line_style;
	trace.TraceTab->_line_style_fld->setCurrentIndex(_line_style);
	
	float _line_width;
	v_is >> _line_width;
	trace.TraceTab->_line_width_fld->setValue(_line_width);

	if( qsas_restore_release >= 20402)
    {
		float _line_scale;
		v_is >> _line_scale;
		trace.TraceTab->_line_scale_fld->setValue(_line_scale);
	}
	
	
	v_is >> trace.userLineColour;

	int r,g,b;
	v_is >> r;
	v_is >> g;
	v_is >> b;
	trace.TraceTab->setLineColour( QColor(r,g,b));
	
	int _symbol_shape;
	v_is >> _symbol_shape;
	trace.TraceTab->_symbol_shape_fld->setCurrentIndex(_symbol_shape);
	if(_symbol_shape==20) // custom
	{
		trace.TraceTab->_custom_symbol_fld->setEnabled(true);
	}
	else
	{
		trace.TraceTab->_custom_symbol_fld->setEnabled(false);
	}
	
	float _symbol_size;
	v_is >> _symbol_size;
	trace.TraceTab->_symbol_size_fld->setValue(_symbol_size);

	if( qsas_restore_release >= 20402)
    {	
		string _custom_symbol;
		v_is >> _custom_symbol;
		trace.TraceTab->_custom_symbol_fld->setText(_custom_symbol.c_str());
	}
	v_is >> trace.userSymbolColour;
	v_is >> r;
	v_is >> g;
	v_is >> b;

	trace.TraceTab->setSymbolColour(QColor(r,g,b));

	bool autoLegend;
	v_is >> autoLegend;
	trace.DataTab->_auto_legend_fld->setChecked(autoLegend);
	
	string _legend_text;
	v_is >> _legend_text;
	trace.DataTab->_legend_text_fld->setText(_legend_text.c_str());
	bool _show_legend;
	v_is >> _show_legend;
	trace.DataTab->_show_legend_fld->setChecked(_show_legend);
	
//	trace.DataTab->_x_data_fld->reValidate(true);
//	trace.DataTab->_y_data_fld->reValidate(true);
	
	return v_is;
}

// ------------------------------------------


Vostream& operator<<(Vostream& v_os, TraceItem& trace)
{
	if(QuSlotDebug()) cout << "QpTraceItem save (<<) '" << trace.getName().toLatin1().data() << "'" << endl;

	// streams trace object to virtual output stream v_os
  
	PlotItem& plot_item = trace;
// 	v_os << plot_item;
	v_os=save(v_os, trace);
	
	v_os << ListTypeEntry;
	v_os << trace.I_Type;
	
	string _y_data = trace.DataTab->_y_data_fld->text().toLatin1().data();
	string _x_data = trace.DataTab->_x_data_fld->text().toLatin1().data();

	v_os << _y_data;
	v_os << _x_data;

    v_os << (bool)trace.DataTab->_gap_detection_fld->isChecked();
    v_os << (double)trace.DataTab->_gap_threshold_fld->value();
    v_os << (int)trace.DataTab->_gap_threshold_type_fld->currentIndex();
    v_os << (bool)trace.DataTab->timeAsSec_toggle->isChecked();

	string linkText;
	if(!trace.getLinkText().isEmpty())
	{
		linkText = string(trace.getLinkText().toLatin1().data());
		linkText += "S"; // to avoid conflict with existing links on restore
	}
	else linkText="";
	v_os << linkText;
	
    int _connect_style = trace.TraceTab->_connect_style_fld->currentIndex();
	v_os << _connect_style;
	
	int _line_style = trace.TraceTab->_line_style_fld->currentIndex();
	v_os << _line_style;
	
	float _line_width = trace.TraceTab->_line_width_fld->value();
	v_os << _line_width;
	
	float _line_scale = trace.TraceTab->_line_scale_fld->value();
	v_os << _line_scale;
	
	v_os << trace.userLineColour;

	v_os << trace._line_colour.red();
	v_os << trace._line_colour.green();
	v_os << trace._line_colour.blue();
	
	int _symbol_shape = trace.TraceTab->_symbol_shape_fld->currentIndex();
	v_os << _symbol_shape;
	
	float _symbol_size = trace.TraceTab->_symbol_size_fld->value();
	v_os << _symbol_size;
	
	string _custom_symbol = trace.TraceTab->_custom_symbol_fld->text().toLatin1().data();
	v_os << _custom_symbol;
	
	v_os << trace.userSymbolColour;

	v_os << trace._symbol_colour.red();
	v_os << trace._symbol_colour.green();
	v_os << trace._symbol_colour.blue();
	
	bool autoLegend = trace.DataTab->_auto_legend_fld->isChecked();
	v_os << autoLegend;


	string _legend_text = trace.DataTab->_legend_text_fld->text().toLatin1().data();
	v_os << _legend_text;

	bool _show_legend = trace.DataTab->_show_legend_fld->isChecked();
	v_os << _show_legend;
	
	return v_os;
}

void Q_plpoin(PLINT n, PLFLT *x, PLFLT *y, PLINT code){

/* interface to plpoin to only plot points that lay within viewport
*  This speeds up plotting for some drivers that apparently render
*  slowly, and so can spend a lot of time rendering points that get
*  clipped by the viewport anyway. Given the overheads in plpoin to
*  look up symbols and fonts, reverting to calling plpoin point by point
*  is probably at least as fast if not faster even in cases where n is 
*  quite small.
*
*  Steve Schwartz; June 2010
*/

	PLFLT xmin, xmax, ymin, ymax, swap, xx, yy;
	int i;

	plgvpw(&xmin, &xmax, &ymin, &ymax);

/* detect if either axis runs backward and swap min & max values */

	if (xmax < xmin){swap = xmax; xmax = xmin; xmin = swap;}
	if (ymax < ymin){swap = ymax; ymax = ymin; ymin = swap;}

	for (i=0; i<n; i++){
		xx = x[i]; yy = y[i];
		if(xx > xmin && xx < xmax && yy> ymin && yy < ymax){
			plpoin(1,&xx,&yy,code);
		}
	}
	return;
}



