//
//  DvObj.h
//
//  Dvos -  Data variable object classes
//
//  Utility calls 
//
//  Tony Allen
//  Nov 2012
//

#include <iostream>

#include "DvObj.h"
#include "DvRecord.h"



// ---------------- Target can be Time Tags or Scalar Sequence -----------------


bool DvObject::isJoinedObj(DvObject &targetObj){

	DvObject_var target = targetObj.getDep0();
	if(target.is_nil()) return true;
	
	return isJoined(*target);
	

	
}

bool DvObject::isJoined(DvObject &target){

	DvObject_var thisDep0 = this->getDep0();
	
	if(thisDep0->get_id() == -1 || thisDep0->is_str()){ // is_str catches badly formed tt
	
		// obj has no Depend_0
		if(seqLen == 1 || seqLen == target.seqLen || target.seqLen == 1) return true; // allow operation without join
		else return false;
	}
	
	if(thisDep0->get_id() == target.get_id()) return true;
	
	if( seqLen == target.seqLen){

		// try fuzzy join
		for(size_t n=0; n<thisDep0->seqSize(); n++) {
			if( fabs(thisDep0->asDouble(n) - target.asDouble(n) ) > DV_CLOSE_ENOUGH) 	return false;
		}
		return true;
	}
	return false;
	
}

bool DvObject::same(DvObject &target){
	
	// e.g. compare time tags or scalar tags directly
	
	if(this->get_id() == target.get_id()) return true;
	
	if( totalSize() == target.totalSize()){
		// try fuzzy join
		for(size_t n=0; n<this->totalSize(); n++) {
			if( fabs(this->asDouble(n) - target.asDouble(n) ) > DV_CLOSE_ENOUGH) 	return false;
		}
		return true;
	}
	return false;
	
}

bool DvObject::isRegular(double &spacing){

	// check if depend_0 is regularly spaced
	double err;
	double Gaptest;
   
	spacing = -1.;
    DvObject_var D0 = getDep0();
	
	if ( D0.is_nil() || D0->seqSize() < 3) return false;

	int n_seq = D0->seqSize();
  
	spacing = D0->asDouble(1) - D0->asDouble(0);
	
	if (spacing < DV_CLOSE_ENOUGH) return false; // too small for fuzzy equality limit
  
	for (int i=2; i<n_seq; i++){
    	Gaptest = D0->asDouble(i) - D0->asDouble(i-1);
    	err = fabs((spacing - Gaptest)/spacing);
    	if (err > DV_CLOSE_ENOUGH ) return false;
	}
	return true;

}

double DvObject::get_spacing(){
    double spacing, next;
	
    if(this->is_nil()) return  0.;
	
	int n_seq = seqSize();

	if( n_seq == 0) return 0.;
	if( n_seq == 1) return this->asDouble(0)* 0.1; // Nominal value

	double spacing_sign = 1.0;
	
  
	next = asDouble(1) - asDouble(0);
	spacing = fabs(next);
	spacing_sign = next / spacing;
	
	for (int i=2; i<n_seq; i++){
    	next = fabs(asDouble(i) - asDouble(i-1));
    	if(next < spacing && next > 1.e-20) {
			spacing = next; // trap duplicate time tags
		}
	}
    return spacing*spacing_sign;
}


DvObject * DvObject::Join(DvObject_var &Target, bool withXrefs){	
	
	// do join
	DvMask Gmsk(1);
	
	if(this->is_dbl() || this->is_int() || this->is_time()) {
		if(withXrefs) return linearJoin(Gmsk, *Target, false);
		else return linearJoinObjOnly(Gmsk, *Target);
	}
	else {
		if(withXrefs) return nnJoin(Gmsk, *Target, false);
		else return nnJoinObjOnly(Gmsk, *Target);
	}

}


DvObject * DvObject::JoinToObj(DvObject_var &dobj, bool withXrefs){	

	// get target DEPEND_0 
	DvObject_var Target = dobj->getDep0();
	
	DvObject *res = Join(Target, withXrefs); // don't return until after Join or Target may go out of scope
	return res;
}

DvObject * DvObject::linearJoinToObj(DvObject &dobj){
	// used by operators, target is DEPEND_0 of dobj
	DvMask Gmsk(1);

	// get target DEPEND_0 
	DvObject_var Target = dobj.getDep0();	
	
	if(Target.is_nil()){
		if (dobj.seqSize() == this->seqSize() || this->seqSize() == 1) return this;
		else return new DvObject();
	}
	
	return linearJoinObjOnly(Gmsk, *Target);
}



DvObject * DvObject::linearJoin(DvMask &Gmsk, DvObject &Target, bool doGaps){	
	
	// do join
	DvObject * res = this->linearJoinObjOnly(Gmsk, Target);
	
	if(res->get_id() == this->get_id()) return this; // unchanged
	
	// attach xrefs from this to result after joining to target
	res->joinXrefs(*this, Target);
	
	if(doGaps && Gmsk.size() == Target.seqSize()){
	
		// do gap mask recursively now as this was not called by MultiJoin.
		res->apply_mask(Gmsk); // recursively applies to xrefs
	}
	
	return res;
}

DvObject * DvObject::linearJoinObjOnly(DvMask &Gmsk, DvObject &Target)
{

	// joins object only, not xrefs
	// joins this onto Target

	// return NN Join if not double

	if( this->is_str() || this->is_event() ) return nnJoinObjOnly(Gmsk, Target); 
	
	if(Target.is_nil()){
		error("[LinearJoin]: no target Depend_0" );
		return new DvObject();
	}

	// get target DEPEND_0 as seq of doubles
	valarray <double> ssTarget = Target.toDouble();	
	
	// test if already joined (or join not needed)
	if( isJoined(Target) ) return this;
	
	DvObject_var gap_opt_ptr = this->get_xref(DV_GAP_OPTION);
	DvString gap_opt;
	if(gap_opt_ptr.is_ok()){
		gap_opt = gap_opt_ptr->str();
	}
	else {
		gap_opt = DV_LINEAR;
	}
	
	// get DEPEND_0 as a sequence of doubles
		
	valarray <double> ssThis = this->getDep0()->toDouble();	

	// invert decreasing tags
	size_t nThis = ssThis.size();
	size_t nTarget = ssTarget.size();

	if(nThis < 2){
		error("[LinearJoin]: Data sequence too short for linear interp");
		return new DvObject();
	}
	
	if(ssThis[0] > ssThis[nThis-1]){
		if(ssTarget[0] > ssTarget[nTarget-1]){
			ssTarget *= -1.;
			ssThis  *= -1.;
		}
		else{
			error("[LinearJoin]: target Depend_0 decreasing, but data Depend_0 increasing" );
			return new DvObject();
		}
	}
	
	if(ssTarget[0] > ssThis[nThis-1] || ssThis[0] > ssTarget[nTarget-1]){
		error("[LinearJoin]: Disjoint data sequences in join");
		return new DvObject();
	}
	

	// gap tolerance in seconds
	DvObject_var gap_ptr = this->get_xref(DV_GAP_BOX_WIDTH);
	double gap;
	if(gap_ptr.is_ok())
		gap = gap_ptr->dbl();
	else{
		gap = Target.get_spacing();
		if(gap <=0) {
			gap = 4.; // default to cluster spin res
			error( "[LinearJoin]: Cannot determine target spacing" );
		}
		gap *=1.5; 
		
	}
	
	// Gmsk has only one element on first call
	// Note gaps are not removed by this method as it 
	// is used to do arithmetic on the fly. Gaps are unly removed by explicit Join operators
	if( Gmsk.size() != nTarget){
		Gmsk.resize(nTarget);
	}
	
	// do join
		
	DvObject * result = new DvObject(this, nTarget, false); // Array dim of this, seq length of target

	size_t i_target=0;
	size_t i_this=1; // ensures interp starts from 1st 2 elements

	
	if(ssTarget[0] < ssThis[0]){
		// data starts after target, get within gap tolerence of data
		while (ssTarget[i_target] + gap < ssThis[0] ) {
			// treat as gap
			if(gap_opt == DV_FILL) {
				(*result)[i_target] = DvNaN;
			}
			else if(gap_opt == DV_LINEAR) {
				double slope = (ssTarget[i_target] - ssThis[i_this]-1) / (ssThis[i_this] - ssThis[i_this-1]);
				(*result)[i_target] = (*this)[i_this];
				(*result)[i_target] -= (*this)[i_this-1] ;
				(*result)[i_target] *= slope;
				(*result)[i_target] += (*this)[i_this-1];
			}
			else if(gap_opt == DV_NN) {
				(*result)[i_target] = (*this)[i_this-1];
			}
			else if(gap_opt == DV_REMOVE) {
				(*result)[i_target] = 0.;
				Gmsk[i_target]=false; // remove (done later)
			}
			else (*result)[i_target] = 0.; // zero fill
			i_target++;
		}
		
		if(ssTarget[i_target] < ssThis[0]){
			// extrapolate (same call as interpolate)
			double slope = (ssTarget[i_target] - ssThis[i_this-1]) / (ssThis[i_this] - ssThis[i_this-1]);
			(*result)[i_target]  = (*this)[i_this];
			(*result)[i_target] -= (*this)[i_this-1];
			(*result)[i_target] *= slope;
			(*result)[i_target] += (*this)[i_this-1];
			i_target++;
		}
		// else fall through to interpolation
		
	}
		
	// ssThis[i_this] starts above ssTarget[i_target] 
	while(i_target < ssTarget.size()){
	
		double target = ssTarget[i_target];
		
		// find data pt above target tag
		while(ssThis[i_this] < target  &&  i_this<ssThis.size()-1 ) i_this++;

		// is this a data gap or off the end of data by more than gap?
		if( ( ssThis[i_this] - ssThis[i_this-1] > gap || target - ssThis[i_this] > gap) ){
		
			// data gap
			
			if(gap_opt == DV_FILL) {
				(*result)[i_target] = DvNaN;
			}
			else if(gap_opt == DV_LINEAR) {
			
				// uses DvRecord operators
				double slope = (target - ssThis[i_this-1]) / (ssThis[i_this] - ssThis[i_this-1]);
				(*result)[i_target]  = (*this)[i_this];
				(*result)[i_target] -= (*this)[i_this-1];
				(*result)[i_target] *= slope;
				(*result)[i_target] += (*this)[i_this-1];
			}
			else if(gap_opt == DV_NN) {
				if( (ssThis[i_this] - target) < (target - ssThis[i_this-1]) ) (*result)[i_target] = (*this)[i_this];
				else (*result)[i_target] = (*this)[i_this-1];
			}
			else if(gap_opt == DV_REMOVE) {
				(*result)[i_target] = 0.; // remove (done later)
				Gmsk[i_target]=false; 
			}
			else (*result)[i_target] = 0.; // zero fill record
			
			
		}
		else{
			// Not a gap, normal interpolation
			
			// if we are off end of data then interpolation algorithm works as extrapolation
		
			double slope = (target - ssThis[i_this-1]) / (ssThis[i_this] - ssThis[i_this-1]);
			(*result)[i_target]  = (*this)[i_this];
			(*result)[i_target] -= (*this)[i_this-1];
			(*result)[i_target] *= slope;
			(*result)[i_target] += (*this)[i_this-1];
		}
			
		i_target++;
	}

	// keep these for future use (xrefs and propagation in operator chains)
		
	result->change_xref(DV_GAP_BOX_WIDTH, gap);
	result->change_xref(DV_JOIN_METHOD, DV_LINEAR); // so double xrefs use same
	result->change_xref(DV_GAP_OPTION, gap_opt);
	
	return result;

}


// ---------- BOXCAR ----------


DvObject * DvObject::boxcarJoin(DvMask &Gmsk, DvObject &Target, bool doGaps){

	// do join
	
	DvObject * res = this->boxcarJoinObjOnly(Gmsk, Target);
	
	if(res->get_id() == this->get_id()) return this; // unchanged
	
	// attach xrefs from this to result after joining to target
	res->joinXrefs(*this, Target); // true means use boxcar
	
	if(doGaps && Gmsk.size() == Target.seqSize()){
	
		// do gap mask recursively now as this was not called by MultiJoin.

		res->apply_mask(Gmsk); // recursively applies to xrefs
	}
	
	return res;

}


DvObject * DvObject::boxcarJoinObjOnly(DvMask &Gmsk, DvObject &Target){


	// joins object only, not xrefs

	// return NN Join if not double
	
	if( !this->is_dbl() ) return nnJoinObjOnly(Gmsk, Target); 

	
	// get target DEPEND_0 as seq of doubles
	valarray <double> ssTarget = Target.toDouble();	
	
	if(ssTarget.size() < 1){
		error("[BoxcarJoin]: no target Depend_0" );
		return new DvObject();
	}

	// Boxcar does not test if already joined since target tt may remain the same

	
	DvObject_var gap_opt_ptr = this->get_xref(DV_GAP_OPTION);
	DvString gap_opt;
	if(gap_opt_ptr.is_ok()){
		gap_opt = gap_opt_ptr->str();
	}
	else {
		gap_opt = DV_FILL;

	}
	
	// get DEPEND_0 as a sequence of doubles
		
	valarray <double> ssThis = this->getDep0()->toDouble();	

	// invert decreasing tags
	size_t nThis = ssThis.size();
	size_t nTarget = ssTarget.size();
	
	if(ssThis[0] > ssThis[nThis-1]){
		if(ssTarget[0] > ssTarget[nTarget-1]){
			ssTarget *= -1.;
			ssThis  *= -1.;
		}
		else{
			error("[BoxcarJoin]: target Depend_0 decreasing, but data Depend_0 increasing" );
			return new DvObject();
		}
	}
	
	if(ssTarget[0] > ssThis[nThis-1] || ssThis[0] > ssTarget[nTarget-1]){
		error("[BoxcarJoin]: Disjoint data sequences in join");
		return new DvObject();
	}
	
	// boxcar in seconds
	DvObject_var width_ptr = this->get_xref(DV_GAP_BOX_WIDTH);
	double width;
	if(width_ptr.is_ok())
		width = width_ptr->dbl();
	else{
		width = Target.get_spacing();
		if(width <=0)  error( "[BoxcarJoin]: Cannot determine target spacing" );
	
		width *=10.; 
	}

	double halfwidth = width*0.5;

	
	// Gmsk has only one element on first call
	// Note, gaps are not removed unitl later
	if( gap_opt == DV_REMOVE  && Gmsk.size() != nTarget){
		Gmsk.resize(nTarget);
	}
	
	// size of boxcar record
	size_t nArray = arraySize();
	
	// do join
		
	DvObject * result = new DvObject(this, nTarget, false); // Array dim of this, seq length of target

	size_t i_target=0;
	size_t i_this=1; // ensures interp starts from 1st 2 elements
	
	if(ssTarget[0] < ssThis[0]){
		// data starts after target, get within gap tolerence of data
		while (ssTarget[i_target] + halfwidth < ssThis[0] ) {
			// treat as gap
			if(gap_opt == DV_FILL) {
				(*result)[i_target] = DvNaN;
			}
			else if(gap_opt == DV_LINEAR) {
				double slope = (ssTarget[i_target] - ssThis[i_this-1]) / (ssThis[i_this] - ssThis[i_this-1]);
				(*result)[i_target] = (*this)[i_this];
				(*result)[i_target] -= (*this)[i_this-1] ;
				(*result)[i_target] *= slope;
				(*result)[i_target] += (*this)[i_this-1];
			}
			else if(gap_opt == DV_NN) {
				(*result)[i_target] = (*this)[i_this-1];
			}
			else if(gap_opt == DV_REMOVE) {
				(*result)[i_target] = 0.;
				Gmsk[i_target]=false; // remove (done later)
			}
			else (*result)[i_target] = 0.; // zero fill
			i_target++;
		}
		
		if(ssTarget[i_target] < ssThis[0]){
			// extrapolate (same call as interpolate)
			double slope = (ssTarget[i_target] - ssThis[i_this-1]) / (ssThis[i_this] - ssThis[i_this-1]);
			(*result)[i_target]  = (*this)[i_this];
			(*result)[i_target] -= (*this)[i_this-1];
			(*result)[i_target] *= slope;
			(*result)[i_target] += (*this)[i_this-1];
			i_target++;
		}
		// else fall through to interpolation
		
	}
		
	// ssThis[i_this] starts above ssTarget[i_target] 
	while(i_target < ssTarget.size()){
	
		double target = ssTarget[i_target];
		
		// find first data pt in boxcar (it will be i_this - 1)
		while(ssThis[i_this-1] < target - halfwidth &&  i_this < nThis-1 ) i_this++;
	
				
		// accumulate boxcar here
		valarray <double> boxcar(0.0, nArray);	
		size_t count= 0;	
		
		size_t n = i_this-1; // first point in boxcar
		while(n < nThis && ssThis[n] < target + halfwidth){
			slice sl(n*nArray, nArray, 1);
			valarray <double> data = this->Ddata[sl];
			for(size_t k=0; k<nArray; k++){
				boxcar[k] += data[k];
			}
			count ++;
			n++;
		}
		if(n > i_this - 1 && count > 0) {
			for(size_t k=0; k<nArray; k++){
				boxcar[k] /= count;
			}
			(*result)[i_target] = boxcar;
		}
		else{		
			// data gap
			// ssThis[i_this-1] will be above target (so use i_this-1 and i_this-2)
			if(gap_opt == DV_FILL) {
				(*result)[i_target] = DvNaN;
			}
			else if(gap_opt == DV_LINEAR) {
				// uses DvRecord operators
				double slope = (target - ssThis[i_this-2]) / (ssThis[i_this-1] - ssThis[i_this-2]);
				(*result)[i_target]  = (*this)[i_this-1];
				(*result)[i_target] -= (*this)[i_this-2];
				(*result)[i_target] *= slope;
				(*result)[i_target] += (*this)[i_this-2];
			}
			else if(gap_opt == DV_NN) {
				if( (ssThis[i_this-1] - target) < (target - ssThis[i_this-2]) ) (*result)[i_target] = ssThis[i_this-1];
				else (*result)[i_target] = ssThis[i_this-2];
			}
			else if(gap_opt == DV_REMOVE) {
				(*result)[i_target] = 0.; // remove (done later)
				Gmsk[i_target]=false; 
			}
			else (*result)[i_target] = 0.; // zero fill record
			
		}

		i_target++;
	}
	
	// this object has no xrefs yet and has not had the gap mask applied	

	// keep these for future use (xrefs and propagation in operator chains)
	result->change_xref(DV_GAP_BOX_WIDTH, width);
	result->change_xref(DV_GAP_OPTION, gap_opt);
	result->change_xref(DV_JOIN_METHOD, DV_BOXCAR); // so double xrefs use same

	return result;
	
}

// ---------- Nearest Neighbour ----------


DvObject * DvObject::nnJoin(DvMask &Gmsk, DvObject &Target, bool doGaps){

	// do join
	
	DvObject * res = this->nnJoinObjOnly(Gmsk, Target);
	
	if(res->get_id() == this->get_id()) return this; // unchanged
	
	// attach xrefs from this to result after joining to target
	res->joinXrefs(*this, Target); // true means use boxcar
	
	if(doGaps && Gmsk.size() == Target.seqSize()){
	
		// do gap mask recursively now as this was not called by MultiJoin.

		res->apply_mask(Gmsk); // recursively applies to xrefs
	}
	
	return res;

}


DvObject * DvObject::nnJoinObjOnly(DvMask &Gmsk, DvObject &Target){

	// Nearest Neighbour
	// joins object only, not xrefs
	// joins this onto Target
	
	// get target DEPEND_0 as seq of doubles
	valarray <double> ssTarget = Target.toDouble();	
	if(ssTarget.size() < 1){
		error("[nnJoin]: no target Depend_0" );
		return new DvObject();
	}

	// test if already joined (or join not needed)
	if( isJoined(Target) ) return this;


	DvObject_var gap_opt_ptr = this->get_xref(DV_GAP_OPTION);
	DvString gap_opt;
	if(gap_opt_ptr.is_ok()){
		gap_opt = gap_opt_ptr->str();
	}
	else {
		gap_opt = DV_FILL;
	}
	
	if(gap_opt == DV_LINEAR && !(this->is_dbl()) ) gap_opt = DV_NN; 
	
	// get DEPEND_0 as a sequence of doubles
		
	valarray <double> ssThis = this->getDep0()->toDouble();	

	// invert decreasing tags
	size_t nThis = ssThis.size();
	size_t nTarget = ssTarget.size();
	
	if(ssThis[0] > ssThis[nThis-1]){
		if(ssTarget[0] > ssTarget[nTarget-1]){
			ssTarget *= -1.;
			ssThis  *= -1.;
		}
		else{
			error("[nnJoin]: target Depend_0 decreasing, but data Depend_0 increasing" );
			return new DvObject();
		}
	}
	
	if(ssTarget[0] > ssThis[nThis-1] || ssThis[0] > ssTarget[nTarget-1]){
		error("[nnJoin]: Disjoint data sequences in join");
		return new DvObject();
	}
	
	// gap tolerance in seconds
	DvObject_var gap_ptr = this->get_xref(DV_GAP_BOX_WIDTH);
	double gap;
	if(gap_ptr.is_ok())
		gap = gap_ptr->dbl();
	else{
		gap = Target.get_spacing();
		if(gap <=0) error( "[LinearJoin]: Cannot determine target spacing" );
	
		gap *=1.5; 
		
	}
	
	// Gmsk has only one element on first call
	// Note, gaps are not removed here
	if( gap_opt == DV_REMOVE  && Gmsk.size() != nTarget){
		Gmsk.resize(nTarget);
	}
	
	
	// do join
		
	DvObject * result = new DvObject(this, nTarget, false); // Array dim of this, seq length of target

	size_t i_target=0;
	size_t i_this=1; // ensures interp starts from 1st 2 elements
	
	if(ssTarget[0] < ssThis[0]){
		// data starts after target, get within gap tolerence of data
		while (ssTarget[i_target] + gap < ssThis[0] ) {
			// treat as gap
			if(gap_opt == DV_FILL) {
				(*result)[i_target] = DvNaN;
			}
			else if(gap_opt == DV_LINEAR ) {
				double slope = (ssTarget[i_target] - ssThis[i_this-1]) / (ssThis[i_this] - ssThis[i_this-1]);
				(*result)[i_target] = (*this)[i_this];
				(*result)[i_target] -= (*this)[i_this-1] ;
				(*result)[i_target] *= slope;
				(*result)[i_target] += (*this)[i_this-1];
			}
			
			else if(gap_opt == DV_REMOVE) {
				(*result)[i_target] = 0.;
				Gmsk[i_target]=false; // remove (done later)
			}
			else if(gap_opt == DV_ZERO_FILL) {
				(*result)[i_target] = 0.; // zero fill
			}
			else  {
				// NN default
				(*result)[i_target] = (*this)[i_this-1];
			}
			i_target++;
		}
		
		if(ssTarget[i_target] < ssThis[0]){
			
			(*result)[i_target]  = (*this)[0];
			i_target++;
		}
		// else fall through to interpolation
		
	}
		
	// ssThis[i_this] starts above ssTarget[i_target] 
	while(i_target < ssTarget.size()){
	
		double target = ssTarget[i_target];
		
		while(ssThis[i_this] < target  &&  i_this<ssThis.size()-1 ) i_this++;
		
		// is this a data gap or off the end of data by more than gap?
		
		if( ( ssThis[i_this] - ssThis[i_this-1] > gap || target - ssThis[i_this] > gap) ){
		
			// data gap
			
			if(gap_opt == DV_FILL) {
				(*result)[i_target] = DvNaN;
			}
			else if(gap_opt == DV_LINEAR ) {
				// uses DvRecord operators
				double slope = (target - ssThis[i_this-1]) / (ssThis[i_this] - ssThis[i_this-1]);
				(*result)[i_target]  = (*this)[i_this];
				(*result)[i_target] -= (*this)[i_this-1];
				(*result)[i_target] *= slope;
				(*result)[i_target] += (*this)[i_this-1];
			}
			else if(gap_opt == DV_NN) {
				if( (ssThis[i_this] - target) < (target - ssThis[i_this-1]) ) (*result)[i_target] = ssThis[i_this];
				else (*result)[i_target] = ssThis[i_this-1];
			}
			else if(gap_opt == DV_REMOVE) {
				(*result)[i_target] = 0.; // remove (done later)
				Gmsk[i_target]=false; 
			}
			else (*result)[i_target] = 0.; // zero fill record
			
			
		}
		else{
			// Not a gap, nn interpolation
			
			if( (ssThis[i_this] - target) < (target - ssThis[i_this-1]) ) (*result)[i_target] = ssThis[i_this];
			else (*result)[i_target] = ssThis[i_this-1];
		}
			
		i_target++;
	}
	
	// this object has no xrefs yet and has not had the gap mask applied	

	// keep these for future use (xrefs and propagation in operator chains)
	result->change_xref(DV_GAP_BOX_WIDTH, gap);
	result->change_xref(DV_GAP_OPTION, gap_opt);
	result->change_xref(DV_JOIN_METHOD, DV_BOXCAR); // so double xrefs use same

	return result;
	
}



// ---------------------

void DvObject::joinXrefs(DvObject &obj, DvObject &Target){

	// xrefs will come from the pre-joined object (obj) (we need their original time tags)
	// and be attached to the joined object
	
	DvMask msk(1); // not used, but needed for method call
	
	DvNode *next = obj.first_xref();
	while(next)
	{
		if(next->name() == DEPEND_0 ) {
			DvObject_var D0(Target);
			this->change_xref(DEPEND_0, D0);
			next = next->next;
			continue;
		}
		  
		if( next->obj()->seqSize() != obj.seqSize() || !next->obj()->is_dbl()) {
			// record varying double type
			this->change_xref(next->name(), next->obj()); ;
			next = next->next;
			continue;
		}
		
		// join rec varying double data xrefs
		DvObject_var xref = next->obj();
		
		DvObject_var D0 = xref->getDep0();
		if( D0.is_nil() ) {
			// sequence length is same as object, so use tt from object
			D0 = obj.getDep0();
			xref = new DvObject(xref); // take copy
			xref->change_xref(DEPEND_0, D0); 
		}
			 			 
		// this is recursive, leave gap mask till last
		bool doGaps = false;

		DvObject_var gapORwidth = this->get_xref(DV_GAP_BOX_WIDTH);
		xref->change_xref(DV_GAP_BOX_WIDTH, gapORwidth);
		
		DvObject_var xref_result;

		DvString join_method = this->getXrefText(DV_JOIN_METHOD);
		
		if(join_method == DV_BOXCAR){
			xref_result = xref->boxcarJoin(msk, Target, doGaps);
		}
		else if(join_method == DV_NN){
			xref_result = xref->nnJoin(msk, Target, doGaps);
		}
		else{
			xref_result = xref->linearJoin(msk, Target, doGaps);
		}
		this->change_xref(next->name(), xref_result);

		next = next->next;
	} 
		
}

// ------------------

DvObject * DvObject::interpAt(DvTime &target){
	size_t iAbove = 0;
	DvObject_var tt = getTimeTags();
	if( tt.is_nil() || tt->not_time() ) return new DvObject();
	
	for (size_t i=0; i<seqSize(); i++) {
			iAbove = i;
			if( tt->time(i) > target ) break;
	}
	if(iAbove == 0) iAbove = 1; // extrapolate
	
	valarray<double> values(arraySize()); 
	double h1 = tt->time(iAbove) - target;
	double h2 = target - tt->time(iAbove-1);
	double h3 = h1 + h2;
	for(size_t k=0; k<arraySize(); k++){
		values[k] = ( this->asDouble(iAbove-1,k)*h1  + this->asDouble(iAbove, k)*h2  ) / h3;
	}

	DvObject *ret = new DvObject(values, this->dims, 1);
	
	ret->copy_xref_from(SI_CONVERSION, *this);
	ret->copy_xref_from(UNITS, *this);
	ret->copy_xref_from(FRAME, *this);
	ret->copy_xref_from(TOFRAME, *this);
	ret->copy_xref_from(DEPEND_1, *this);
	ret->copy_xref_from(DEPEND_2, *this);
	ret->copy_xref_from(DEPEND_3, *this);
	ret->copy_xref_from(DEPEND_4, *this);
	ret->copy_xref_from(DELTA_PLUS, *this);
	ret->copy_xref_from(DELTA_MINUS, *this);
	ret->copy_xref_from(TENSOR_ORDER, *this);
	ret->copy_xref_from(COORDINATE_SYSTEM, *this);
	ret->copy_xref_from(REPRESENTATION, *this);

	return ret;
}

// ----------------------

DvString DvJoinList::MultiJoin( DvObject_var    &target,
								DvJoinList     	&retList)
{
	// join method, gap method and gap/width are attached to objects as xefs
		
	DvMask Gmsk(1); // resized by join if used
	
	DvNode *node = this->first();

    while( node )
    {  
        if (node->obj()->seqSize() == 1)
        {
			// replace single input with sequence of right length and all values the same
			DvObject_var ret = new DvObject( *(node->obj()), target->seqSize());
	     	ret->change_xref(DEPEND_0, target) ; // is automatically joined
			
			retList.append(node->name(), ret);
			node = node->next;
			continue;
        }

	   	bool doGaps = false; // QuMultijoin applies mask after all series done
		
		DvString join_method = node->obj()->getXrefText(DV_JOIN_METHOD);
		
		if(join_method == DV_BOXCAR){
			
			DvObject_var ret = node->obj()->boxcarJoin(Gmsk, *target, doGaps);
			retList.append(node->name(), ret);
		}
		else if(join_method == DV_NN){
			
			DvObject_var ret = node->obj()->nnJoin(Gmsk, *target, doGaps);
			retList.append(node->name(), ret);
		}
		else{
			// DV_LINEAR is default
			DvObject_var ret = node->obj()->linearJoin(Gmsk, *target, doGaps);
			retList.append(node->name(), ret);
		}
		node = node->next;
   }
    
    // Apply gap mask to all objects
	if(Gmsk.size() == target->seqSize() ){
		DvNode *next = retList.first();
		while(next){
			// do gap mask recursively now.
			next->obj()->apply_mask(Gmsk); // recursively applies to xrefs
			next = next->next;
		}
	}
	
	return DvString("Multi Join OK");
}



