//
// StringRep.cc
//
// May 2000
// David Burgess

#include "StringRep.h"
#include "QObject.h"
#include "QdString.h"
#include "QdTime.h"
#include "Matrix.cc"
#include "Numeric.h"

namespace QDOS {

string StringRep::matrix_dims_to_string( const Dim& d ) const
{
  string rstring;
  if(d.rank()==0)
    return rstring;
  vector<int> dims;
  d.get_dims(dims);
  if(dims.size()==0)
      return rstring;
  rstring = get_token(SR_TOKEN_BEGIN_DIMS);
  string separator=get_token(SR_TOKEN_DIM_SEPARATOR);
  for(int i=0;i<dims.size();i++)
  {
     if(i==0) rstring +=  int_to_string( dims[i] );
     else rstring += ( separator + int_to_string( dims[i] ) );
  }
  rstring += get_token(SR_TOKEN_END_DIMS);
  return rstring;
}
// ++++++++++++++++++++++++++++++++++++++++++++++++++ class StringRep
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// -------------------------------------------------------- set_token
void StringRep::set_token( TokenType token_type, const string& tok )
{
  switch( token_type ) {
  case SR_TOKEN_SEPARATOR: token_separator = tok; break;
  case SR_TOKEN_STRING_DELIMITER: token_string_delimiter = tok; break;
  case SR_TOKEN_BEGIN_SEQ: token_begin_seq = tok; break;
  case SR_TOKEN_END_SEQ: token_end_seq = tok; break;
  case SR_TOKEN_BEGIN_DIMS: token_begin_dims = tok; break;
  case SR_TOKEN_END_DIMS: token_end_dims = tok; break;
  case SR_TOKEN_UNKNOWN_DIMS: token_unknown_dims = tok; break;
  case SR_TOKEN_DIM_SEPARATOR: token_dim_separator = tok; break;
  case SR_TOKEN_BEGIN_MATRIX: token_begin_matrix = tok; break;
  case SR_TOKEN_END_MATRIX: token_end_matrix = tok; break;
  case SR_TOKEN_TIME_SEPARATOR: token_time_separator=tok; break;
  case SR_TOKEN_TIME_TRAILING_CHAR: token_time_trailing_char=tok; break;
  default:
    break;
  };
}
// ++++++++++++++++++++++++++++++++++++++++++++++++++ class StringRep
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// -------------------------------------------------------- get_token
string StringRep::get_token( TokenType token_type ) const
{
  string tok;
  switch( token_type ) {
  case SR_TOKEN_SEPARATOR: tok=token_separator; break;
  case SR_TOKEN_STRING_DELIMITER: tok=token_string_delimiter; break;
  case SR_TOKEN_BEGIN_SEQ: tok=token_begin_seq; break;
  case SR_TOKEN_END_SEQ: tok=token_end_seq; break;
  case SR_TOKEN_BEGIN_DIMS: tok=token_begin_dims; break;
  case SR_TOKEN_END_DIMS: tok=token_end_dims; break;
  case SR_TOKEN_UNKNOWN_DIMS: tok=token_unknown_dims; break;
  case SR_TOKEN_DIM_SEPARATOR: tok=token_dim_separator; break;
  case SR_TOKEN_BEGIN_MATRIX: tok=token_begin_matrix; break;
  case SR_TOKEN_END_MATRIX: tok=token_end_matrix; break;
  case SR_TOKEN_TIME_SEPARATOR: tok=token_time_separator; break;
  case SR_TOKEN_TIME_TRAILING_CHAR: tok=token_time_trailing_char; break;
  default:tok="Unknown Token"; break;
  };
  return tok;
}

// ++++++++++++++++++++++++++++++++++++++++++++++++++ class StringRep
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ------------------------------------------------------ object_type

// explicit type_info 
//  including sequence size and matrix dimensions
string StringRep::object_type( QdObject_var obj )
{
  if( obj.is_nil() )
    return "nil_object";
  string s(  obj->object_srep(*this));
  s+=obj->size_srep(*this);
  return s;
}
// ++++++++++++++++++++++++++++++++++++++++++++++++++ class StringRep
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// ---------------------------------------------- function operator()
//
// if object is not a sequence then seqi and seqn are quietly ignored
// if object is a sequence then return string rep for seqn elements
//    starting from element seqi
//    if seqi is out of range (<0 or >=size()) then return empty string
//    if seqn is negative then return ALL elements
//    seqn is quietly truncated if too many elements requested

string StringRep::operator()( QdObject_var obj, int seqi, int seqn )
{
  if( obj.is_nil() )
    return "nil";
  string s;
  if(seqi<0 || seqi>=obj->sequence_size() || seqn==0)
      return obj->size_srep(*this,0)+ token_begin_seq+ token_end_seq;
             
  if(seqn<0) seqn=obj->sequence_size();
  int seqix = seqi + seqn;
  if( seqix > obj->sequence_size() )
  {
    seqix = obj->sequence_size();
    seqn=seqix-seqi;
  }
  s+=obj->size_srep(*this,seqix-seqi);
  if( seqn!=1 )
    s+=token_begin_seq;
 
  for(int i=seqi;i<seqix;i++)
  {
      if(i!=seqi)
          s+=token_separator;
      s+=obj->srep(*this,i);
  }
  if( seqn!=1 ) s+=token_end_seq;
  return s;
}       

// Utility call
// ------------------------------------------

 string upper_case(const string& s)
{
  string res(s);
  int sz=res.size();
  for(int i=0;i<sz;i++)
   res[i]=toupper(res[i]);
  return res;
}

} // end namespace QDOS


